#ifndef __GEOMETRY__
#define __GEOMETRY__

#include <vector>
#include <set>
#include <map>

#include "common.h"


#define EPSILON .0000000000001
//#define EPSILON .000000001


//======================================================================
// Class Geometry
//
// this class implements the useful geometric routines used in the code...
//
// 2007-8-29: Integrated the new Simplex class...
// 2008-1-27: Integrated yet another Simplex class...
// 2008-6-XX: Insphere is better: safer with respect to slivers...
// 2008-12-XX: integrated insphere1toXX predicates
//
//
class Geometry {

public:

   Points * P;
   PLC *C;

   Geometry(PLC * c1);

   bool inSphereStrict(int p, Simplex * s);
   bool inSphereStrict(Point & p, Simplex * s);
   bool inSphereReallyStrict(int p, Simplex * s);
   bool inSphereReallyStrict(Point & p, Simplex * s);
   bool inSphereStrictNoSliver(Point & p, Simplex * s);


   bool IsPoorQuality(Simplex * s);
   bool TwoDIsPoorQuality(Simplex * s);

   Point circumcenter(Simplex * s);

   double circumradius(Simplex * s);

   double ratio(Simplex * s);

// never used!
   double Orient(Simplex * s);

   double Volume(Simplex * s);
   void DihedralAngles(Simplex * s, vector < double >&angles);

   bool AlmostCoplanar(Point & p, int pa, int pb, int pc);

   bool noslivers;

private:

};


#endif              // !defined(__GEOMETRY__)
