#ifndef __INTESTINE_H__
#define __INTESTINE_H__

#include "common.h"
#include <vector>

class Direction {
public:
   double angle;
   Vector direction;
   Mesh *plane;
};

bool operator<(const Direction & a, const Direction & b);

class IntestineRing {
 public:

  int vertex;
  Point center;
  Mesh * segment;
  double radius;

  // circular doubly linked list...
  map<int,int> r2n;
  map<int,int> r2p;


  IntestineRing();

 private:


};


class IntestineCylinder {
 public:

  IntestineRing * ring1;
  IntestineRing * ring2;

  Mesh * segment;
  Simplex * subsegment;

  IntestineCylinder();
  IntestineCylinder(IntestineRing * r1, IntestineRing * r2);

  map<int,set<int> > protectedSegments;

  void AddSeg(int a, int b);
  void RemoveSeg(int a, int b);

 private:

};

class IntestineSphere {
 public:

  int vertex;
  double radius;

  set<int> nbors;

  map<int, set<int> > protectedArcs;

  IntestineSphere();

  void AddArc(int a, int b);
  void RemoveArc(int a, int b);

 private:


};



/*
An intestine point is either in:
1 Sphere
1 Cylinder
1 Ring and 2 Cylinders
OR 
1 Ring, 1 Sphere, 1 Cylinder
*/

class Intestine {
public:


  PLC * C;
  vector<IntestineRing * > rings;
  vector<IntestineCylinder * > cylinders;
  vector<IntestineSphere * > spheres;

  map<int,IntestineRing *> l2r;
  map<int, set<IntestineCylinder *> > l2c;
  map<int,IntestineSphere *> l2s;

  Intestine();

  int BuildIntestine(PLC * C);

  bool IsIntestineEdge(int a, int b);
  bool IsIntestineProtected(int a, int b);
  bool IsIntestineRingEdge(int a, int b);
  bool IsIntestineProtectedEdge(int a, int b);
  bool IsIntestineProtectedArc(int a, int b);
  bool IsIntestineFace(int a, int b, int c);
  bool IsIntestineSphereFace(int a, int b, int c);
  bool IsIntestineCylinderFace(int a, int b, int c);

  IntestineSphere * ContainingIntestineSphere(int a, int b, int c);
  IntestineSphere * ContainingIntestineSphere(int a, int b);
  IntestineCylinder * ContainingIntestineCylinder(int a, int b, int c);
  IntestineCylinder * ContainingIntestineCylinder(int a, int b);
  IntestineRing * ContainingIntestineRing(int a, int b);

  Point IntestineRingSplitPoint(int a, int b);
  Point IntestineArcSplitPoint(int a, int b);
  Point IntestineFaceSplitPoint(int a, int b, int c);


  int SplitIntestine(int a, int b, int c, Mesh * & m);

  bool DoesSphereContain(IntestineSphere * s, int i);
  bool DoesCylinderContain(IntestineCylinder * s, int i);

  void IntestineArcsSegsRings(int i, set<int> &nbs);


private:
  void InitSphere(int iP);
  int ProtectSegment(Mesh * mSeg);
  int ProtectInputPoint(IntestineSphere * s);
  void CopyContainment(int old, int n);


  IntestineRing* GetRing(int i, Mesh * mSeg, vector<Direction> & dirs,
			 int & firstpoint);
  IntestineRing* MidRing(IntestineRing * r1, IntestineRing* r2, Mesh * mSeg, 
			 vector<Direction> & dirs, int & firstpoint);

  double AngleFrom(Vector & v1, Vector & v2, Vector & n);

  Vector RodriguesFormula(Vector u, Vector v, double theta);

  vector<Vector> sd;


};

#endif                          // !defined(__INTESTINE_H__)
