#ifndef __PLC__
#define __PLC__

#include "common.h"

using namespace std;


struct InputPoint {
   Point p;
   int id;
   set < Mesh * >l2m;
};


class PLC {
public:

   string name;

   Points *P;           // pointer to points class
   Geometry *G;         // pointer to geometry routines - make static...

   SimplexPool *SP;

   Mesh *M;             // 3d mesh...

   Configuration *config;
   Stats *stats;

   vector < InputPoint > P0;

   clock_t start;

// constructor...
   PLC();           // plain constructor
   PLC(PLC * C);        // copy constructor...

   void CopyPLC(PLC * C);

   void OpenLogFile(string fname);

   void ReadPLC(string fname, vector < Simplex * >&initialSimplices);

   set < int >pointsToProtect;
   set < Mesh * >segmentsToProtect;
   set < int >lazyProtectionList;
   map <int,double> lazyProtectionDistance;
   map <int, set<int> > lazyProtectionNeighbors;
   map <int, Mesh *> lazyNb2Segment;

   bool IsProtected(int ip);
   bool IsProtected(Mesh * m);

   void MakeProtectionList();
   void FormFaceBoundaries();

   void ReadAndRefine(string fname);

   void Step0();
   void Step1a();
   void Step1b();
   void Step2a();
   void Step2b();
   void Step2c();
   void Step3();
   void Step3Intestine();
   void FixCollarSlivers();
   void StripBBoxTets();
   void OrientPLC();

   void TwoDStep1a();
   void TwoDStep1b();
   void TwoDStep2();

   void CopyMeshOver(PLC * Cbbox);
   bool keepPoint(int i);
   void SetLazyProtectionNeighbors(PLC * COld);

   void PrintRaw(string fname);
   void PrintTetView(string fname);
   void PrintTPLC(string fname);
   void PrintFreeFEM(string fname);

   void PrintOff(string prefix, string suffix);
   void PrintCollarOff(string prefix, string suffix);

   void PrintInfo();

   void InsertInMesh(int i, Mesh * m);
   void InsertInMesh(int i, int hint, Mesh * m);
   void InsertInMesh(int i, Simplex * s, Mesh * m);
   void InsertInMeshAndDequeue(int i, Mesh * m, set < SimplexMesh > &SQ);
   void InsertInMeshAndDequeue(int i, int hint, Mesh * m,
                               set < SimplexMesh > &SQ);
   void InsertInMeshAndParents(int i, Mesh * m);
   void InsertInMeshAndParents(int i, int hint, Mesh * m);

   int GetNextGID();

   map < Mesh *, Mesh * >old2new;

   bool Step3IsEncroached(Simplex * s);

   void SetRadiusAndCC(Simplex * s);

   ofstream *LOGFILE;

   bool logfileisopen;

   void PrintBegin(string s);
   void PrintEnd(string s);

private:

   set < Simplex * >tetsToRemove;
   set < int >pointsToRemove;

   int counter;
   vector < Mesh * >planestoprint;

   Vector NormalInFace(Mesh * segment, Mesh * face);
   Vector DirTowardsSegment(Point p, Mesh * segment);

   int Step1bGroomEdges();
   bool Step1bIsEncroached(Simplex * s);
   bool Step1bDoesEncroach(int iN, Simplex * s);
   void Step1bQueueSegment(Simplex * s, set < SimplexMesh > &SQ);


   void StoreInputNeighbors();

   int HalveAllSegments();
   int HalveNonEndSegments();
   int HalveNewNonEndSegments();

   bool Step2bIsEncroached(Simplex * s, Mesh * m);
   bool Step2bDoesEncroach(int iN, Simplex * s, Mesh * m);
   bool Step2bReject(Point p, Simplex * s, Mesh * m);
   void Step2bQueueSimplex(Simplex * s, Mesh * m,
                           set < SimplexMesh > &SQ);

   int Step3Encroached(Point p, Simplex * s, Mesh * m,
                       set < SimplexMesh > &encroached,
                       set < CollarMesh > &encroachedcollar);
   int Step3EncroachedInt(Point p, Simplex * s, Mesh * m,
                       set < SimplexMesh > &encroached,
                       set < IntestineItem > &encroachedintestine);
   int Step3Encroached(int i, Mesh * m, set < SimplexMesh > &encroached);
   int Step3EncroachedInt(int i, set < SimplexMesh > &encroached,
		       set < IntestineItem > & encroachedintestine);
   void Step3QueueCollarEdge(int i, Simplex * s, Mesh * m,
                             set < CollarMesh > &CQ);
   void Step3EncroachingPoint(Simplex * s, set < int >&enc);

   void GetCenterNormalAndBBox(set < int >&myvert, Point & center,
                               Vector & normal, Vector & b1, Vector & b2,
                               double &bbox);

   void CreateBoundingSquare(double mysize, Point & center, Vector & b1,
                             Vector & b2, Mesh * m);

   void VerifySimplex(Simplex * s);


   bool ContainsBadSliver(set<Simplex * > & star);


   bool IsFace(Simplex * s, int i);

   bool TwoDStep1aIsEncroached(Simplex * s);
   bool TwoDStep2IsEncroached(Simplex * s);
   bool TwoDStep1aDoesEncroach(int p, Simplex * s);
   bool TwoDStep2DoesEncroach(int p, Simplex * s);

   double SpindleLengthSq(int p);

};

#endif              // !defined(__PLC__)
