#ifndef __Point_H__
#define __Point_H__

using namespace std;

/////////////////////////////////////////////////////
//
//  Point Class
//
/////////////////////////////////////////////////////

class Vector;

class Point {
   friend class Vector;

public:
   double xyz[3];

   Point() {
      xyz[0] = 0;
      xyz[1] = 0;
      xyz[2] = 0;
   } Point(int a, int b, int c) {
      xyz[0] = a;
      xyz[1] = b;
      xyz[2] = c;
   }
   Point(double a, double b, double c) {
      xyz[0] = a;
      xyz[1] = b;
      xyz[2] = c;
   }
   ~Point() {
   };

   friend istream & operator>>(istream &, Point &);
   friend ostream & operator<<(ostream &, Point);

   Point & operator =(const Point & p);
   int operator==(Point);
   int operator!=(Point);

   Vector operator-(Point);
   Point operator+(Vector);
   Point operator-(Vector);
   Point & operator+=(Vector);
   Point & operator-=(Vector);

// Scalar Multiplication
   friend Point operator*(int, Point);
   friend Point operator*(double, Point);
   friend Point operator*(Point, int);
   friend Point operator*(Point, double);
// Scalar Division
   friend Point operator/(Point, int);
   friend Point operator/(Point, double);


// Point Addition
   friend Point operator+(Point, Point);

// Point Relations
   friend double d(Point, Point);       // Distance
   friend double d2(Point, Point);      // Distance Squared
};

/////////////////////////////////////////////////////
//
//  Vector Class
//
/////////////////////////////////////////////////////


class Vector : public Point {
public:
   Vector() : Point() {
   };
   Vector(int a, int b, int c) : Point(a, b, c) {
   };
   Vector(double a, double b, double c) : Point(a, b, c) {
   };
   ~Vector() {
   };

   friend ostream & operator<<(ostream &, Vector);

   Vector operator-();

// Scalar Multiplication
   friend Vector operator*(int, Vector);
   friend Vector operator*(double, Vector);
   friend Vector operator*(Vector, int);
   friend Vector operator*(Vector, double);
// Scalar Division
   friend Vector operator/(Vector, int);
   friend Vector operator/(Vector, double);


// Vector Arithmetic Operations
   Vector operator+(Vector);
   Vector operator-(Vector);
   double operator*(Vector);    // dot product
   double operator*(Point);     // dot product
   Vector operator^(Vector);    // cross product

   Vector & operator*=(double);
   Vector & operator/=(double);
   Vector & operator+=(Vector);
   Vector & operator-=(Vector);
   Vector & operator^=(Vector);

   double len() {               // vector length
      return sqrt(xyz[0] *xyz[0] + xyz[1] *xyz[1] + xyz[2] *xyz[2]);
   }
   double len2() {              // vector length squared
      return (xyz[0] *xyz[0] + xyz[1] *xyz[1] + xyz[2] *xyz[2]);
   }

   void normalize();

};

#endif
