#include <stdlib.h>
#include "allh.h"

typedef pair < int, set < int > >I2S;

Boundary::Boundary(Mesh * m1)
{
   C = m1->C;
   M = m1;
}

// arand 9-19-08
// this utility is duplicated in the mesh class?
void Boundary::SegNeighbors(Mesh * m, int i, int &n1, int &n2)
{
   set < int >nbs;
   m->Neighbors(i, nbs);

   if (nbs.size() != 2) {
      cout << "ERROR: SegNeighbors doesn't get 2 neighbors" << endl;
      exit(0);
   }

   set < int >::iterator it;
   it = nbs.begin();
   n1 = *it;
   it++;
   n2 = *it;

}

bool Boundary::Exists(int a, int b)
{

   if (!Exists(a)) {
      return false;
   }
   int b1, b2;
   b1 = next[a];
   b2 = prev[a];
   if (b == b1 || b == b2) {
      return true;
   }
   return false;
}

bool Boundary::Exists(int a)
{
   if (next.find(a) != next.end()) {
      return true;
   }

   return false;
}


// create the boundary...
void Boundary::Build()
{
   int first, current;

   StartBoundary(seglist[0], seglist[1], first, current);


   next[first] = current;
   prev[current] = first;

   for (int i = 1; i < (int)seglist.size(); i++) {
      int mynext = NextPoint(seglist[i], current);
      next[current] = mynext;
      prev[mynext] = current;

      if (mynext != first || i == (int)seglist.size() - 1) {
         current = mynext;
      } else {
         // start a new component...
         StartBoundary(seglist[i + 1], seglist[i + 2], first, current);
         next[first] = current;
         prev[current] = first;
         i++;
      }
   }

   foreach(Mesh * seg, seglist) {
      segments.insert(seg);
   }
   // do not need seglist anymore...
   seglist.clear();


   // check the orientation of the normal of the mesh.
   // If it needs to be corrected, fix it...
   int bboxpoint = -1;
   Mesh::PIterator pIt(M);
   while (pIt.Current() != -1 && bboxpoint == -1) {
      int iP = pIt.Current();
      if (C->P->type[iP] == POINT_BBOX) {
         bboxpoint = iP;
      }
   }

   if (!M->Exterior(M->p2s[bboxpoint])) {
      // flip the normal...
      M->dir = -M->dir;
   }
   //cout << "Exit Build" << endl;

}

void Boundary::StartBoundary(Mesh * seg1, Mesh * seg2, int &first,
                             int &second)
{
   int a, b, c, d;
   seg1->GetSegmentEndPoints(a, b);
   seg2->GetSegmentEndPoints(c, d);

   if (a == c || a == d) {
      first = b;
      second = a;
   } else if (b == c || b == d) {
      first = a;
      second = b;
   } else {
      cout << "ERROR: Start Boundary breaking in the boundary class." <<
      endl;
      cout << "The first two segments do not seem to share an end point"
           << endl;
      exit(0);
   }

}

int Boundary::NextPoint(Mesh * seg, int current)
{
   int a, b;
   seg->GetSegmentEndPoints(a, b);
   if (a == current) {
      return b;
   }
   if (b == current) {
      return a;
   }
   cout <<
   "ERROR: problem finding closed loop for boundary of convex face: "
        << M->name << endl;
   exit(0);

   return -1;
}

void Boundary::InsertPoint(int i, int n1, int n2)
{

   if (next[n1] == n2) {
      next[n1] = i;
      next[i] = n2;
      prev[n2] = i;
      prev[i] = n1;
   } else if (next[n2] == n1) {
      next[n2] = i;
      next[i] = n1;
      prev[n1] = i;
      prev[i] = n2;
   } else {
      cout << "ERROR: Boundary Insertion Failed" << endl;
   }


}


void Boundary::Print()
{

   map < int, int >::iterator it;

   set < int >found;

   // loop through everything to get every component..
   for (it = next.begin(); it != next.end(); it++) {

      if (found.find(it->first) == found.end()) {
         // print this component...
         cout << it->first << endl;;
         int first = it->first;
         int cur = it->first;
         do {

            found.insert(cur);
            cout << C->P->V[cur] << " ";
            cur = next[cur];
         } while (cur != first);
         cout << endl;
      }
   }
}
