#include "allh.h"
#include <algorithm>


typedef pair < int, set < int > >I2S;


bool operator<(const Myangle & a, const Myangle & b)
{
   return a.angle < b.angle;
}


Collar::Collar(Mesh * m1)
{
   C = m1->C;
   M = m1;
}


// this will build the partial collar...
int Collar::BuildCollar()
{

   int pointsadded = 0;

   if (C->logfileisopen && C->config->verbose) {
      (*(C->LOGFILE)) << "Running Collar::BuildCollar() for mesh: " <<
      M->name << endl;
   }

   if (M->dim != 2) {
      cout << "ERROR: attempting to build the collar on a dimension " <<
      M->dim << " mesh." << endl;
      exit(0);
   }
   //cout << "Segments to Protect " << C->segmentsToProtect.size() << endl;

   // build the collar on mesh m
   foreach(Mesh * m, M->children) {
      // all children should be dim 1, so these are all edges...

      //cout << m->name << endl;

      if (C->segmentsToProtect.find(m) != C->segmentsToProtect.end()) {
         pointsadded += CollarEdge(m);
      }
   }

   pointsadded += ConnectEndBalls();

   return pointsadded;
}


int Collar::BuildCollarFull()
{

   int pointsadded = 0;
   //cout << "Collaring " << M->name << endl;

   if (M->dim != 2) {
      cout << "ERROR: attempting to build the collar on a dimension " <<
      M->dim << " mesh." << endl;
      exit(0);
   }
   // build the collar on mesh m
   foreach(Mesh * m, M->children) {
      // all children should be dim 1, so these are all edges...

      pointsadded += CollarEdge(m);
   }

   ConnectEndBalls();

   return pointsadded;

}

int Collar::CollarEdge(Mesh * m)
{

   //cout << "Collar Edge" << endl;

   if (C->logfileisopen && C->config->verbose) {
      (*(C->LOGFILE)) << "Running Collar::CollarEdge for segment: " <<
      m->name << endl;
   }

   int pointsadded = 0;
   // m: edge
   // M: face

   // get the direction of the normal into the face...
   Vector NormalInFace = m->dir ^ M->dir;       // cross the direction of the edge
   NormalInFace.normalize();

   //cout << NormalInFace << " " << m->dir << " " << M->dir << endl;

   // find the first point to collar...

   // check both end points...'
   int end1, end2, in1, in2, n1, n2;
   set < int >nbs1, nbs2;

   set < int >::iterator edgeIt = m->exteriorPoints.begin();
   end1 = *edgeIt;
   edgeIt++;
   end2 = *edgeIt;
   edgeIt++;

   // ERROR Check...
   if (edgeIt != m->exteriorPoints.end()) {
      cout << "ERROR: problem with exteriorPoints in CollarEdge" << endl;
      exit(0);
   }

   m->Neighbors(end1, nbs1);
   m->Neighbors(end2, nbs2);

   if (nbs1.size() != 1 || nbs2.size() != 1) {
      cout << "ERROR: end point has wrong number of neighbors..." <<
      endl;
      exit(0);
   }
   // these are the input points...
   in1 = *(nbs1.begin());
   in2 = *(nbs2.begin());



   int cur, prev, next;

   if ((NormalInFace ^ (C->P->V[in1] - C->P->V[end1])) * M->dir <= 0) {
      prev = in1;

      SegNeighbors(m, in1, n1, n2);

      if (n1 == end1) {
         cur = n2;
      } else {
         cur = n1;
      }

   } else {
      prev = in2;

      SegNeighbors(m, in2, n1, n2);

      if (n1 == end2) {
         cur = n2;
      } else {
         cur = n1;
      }
   }

   SegNeighbors(m, cur, n1, n2);

   if (n1 == prev) {
      next = n2;
   } else {
      next = n1;
   }

   //int nextstart = next;

   int loc;
   int lastloc = -1;

   // find the two triangles containing in1 and cur...
   set < Simplex * >star;
   M->Simplices(prev, star);

   bool pos = false;
   bool neg = false;

   foreach(Simplex * s, star) {
      if (s->contains(cur)) {
         // is this triangle inside?
         if (!M->Exterior(s)) {
            // if so, add a collar in this direction...

            // find the point of s which is not in1 or cur...
            int i3 = -1;
            for (int i = 0; i < s->size(); i++) {
               if (s->v[i] != prev && s->v[i] != cur) {
                  i3 = s->v[i];
               }
            }
            if (i3 == -1) {
               cout <<
               "ERROR: bad vertex index in Collar::CollarEdge()"
                    << endl;
               exit(0);
            }
            //Vector vdir = C->P->V[i3] - C->P->V[in1];
            Vector vdir = C->P->V[i3] - C->P->V[prev];
            vdir.normalize();
            if (vdir * NormalInFace > 0) {
               pos = true;
               //cout << "POS" << endl;
            } else {
               neg = true;
               //cout << "NEG" << endl;
            }

         }

      }
   }

   if (pos) {

      //if (C->logfileisopen && C->config->verbose) {
      //  (*(C->LOGFILE)) << "POS walk" << endl;
      //}


      // walk down the edge adding collar points...
      while (C->P->type[next] != POINT_BBOX) {
         //if (C->logfileisopen && C->config->verbose) {
         //  (*(C->LOGFILE)) << "Begin " << prev << " " << cur << " " << next << endl;
         //}

         Point p = GetCollarPoint(prev, cur, next, NormalInFace);

         loc = C->P->AddPoint(p, cur, POINT_COLLAR);
         pointsadded++;
         //if (C->logfileisopen && C->config->verbose) {
         //  (*(C->LOGFILE)) << "Added " << loc << " " << C->P->V[loc] << endl;
         //}

         C->InsertInMeshAndParents(loc, M);
         //if (C->logfileisopen && C->config->verbose) {
         //  (*(C->LOGFILE)) << "Insertion success" << endl;
         //}

         if (lastloc != -1) {
            col2next[lastloc] = loc;
            col2prev[loc] = lastloc;
         } else {
            col2prev[loc] = -1;
            col2p[loc] = prev;
            p2col[prev].insert(loc);
         }
         col2e[loc] = cur;
         col2m[loc] = m;

         //if (C->logfileisopen && C->config->verbose) {
         //  (*(C->LOGFILE)) << "PRE" << endl;
         //}

         SegNeighbors(m, next, n1, n2);

         //if (C->logfileisopen && C->config->verbose) {
         //  (*(C->LOGFILE)) << "POST" << endl;
         //}


         if (n1 == cur) {
            prev = cur;
            cur = next;
            next = n2;
         } else {
            prev = cur;
            cur = next;
            next = n1;
         }
         lastloc = loc;

      }
      p2col[cur].insert(loc);
      col2p[loc] = cur;
      col2next[loc] = -1;

   }

   // reset the opposite end of the segment...
   //prev = in1;
   //cur = curstart;
   //next = nextstart;
   //lastloc = -1;
   if ((NormalInFace ^ (C->P->V[in1] - C->P->V[end1])) * M->dir >= 0) {
      prev = in1;

      SegNeighbors(m, in1, n1, n2);

      if (n1 == end1) {
         cur = n2;
      } else {
         cur = n1;
      }
   } else {
      prev = in2;

      SegNeighbors(m, in2, n1, n2);

      if (n1 == end2) {
         cur = n2;
      } else {
         cur = n1;
      }
   }

   SegNeighbors(m, cur, n1, n2);

   if (n1 == prev) {
      next = n2;
   } else {
      next = n1;
   }

   lastloc = -1;


   if (neg) {
      //if (C->logfileisopen && C->config->verbose) {
      //  (*(C->LOGFILE)) << "NEG walk" << endl;
      //}

      // walk down the edge adding collar points...
      while (C->P->type[next] != POINT_BBOX) {
         Point p = GetCollarPoint(prev, cur, next, -1.0 * NormalInFace);

         loc = C->P->AddPoint(p, cur, POINT_COLLAR);
         pointsadded++;
         C->InsertInMeshAndParents(loc, M);

         if (lastloc != -1) {
            col2next[lastloc] = loc;
            col2prev[loc] = lastloc;
         } else {
            col2prev[loc] = -1;
            col2p[loc] = prev;
            p2col[prev].insert(loc);
         }
         col2e[loc] = cur;
         col2m[loc] = m;

         SegNeighbors(m, next, n1, n2);

         if (n1 == cur) {
            prev = cur;
            cur = next;
            next = n2;
         } else {
            prev = cur;
            cur = next;
            next = n1;
         }
         lastloc = loc;

      }
      p2col[cur].insert(loc);
      col2p[loc] = cur;
      col2next[loc] = -1;

   }
   //if (C->logfileisopen && C->config->verbose) {
   //  (*(C->LOGFILE)) << "Finished" << endl;
   //}

   return pointsadded;

}


Point Collar::GetCollarPoint(int prev, int cur, int next,
                             Vector NormalInFace)
{

   Point p;

   //Vector NormalInFace;
   //NormalInFace = M->dir^(C->P->V[next] - C->P->V[cur]);
   //NormalInFace.normalize();

   if (C->P->type[prev] == POINT_INPUT) {
      double len = d(C->P->V[next], C->P->V[cur]);

      double dend = d(C->P->V[prev], C->P->V[cur]);

      double myratio = sqrt(1 - len * len / (4.0 * dend * dend));

      p = (1.0 - myratio) * C->P->V[prev] + myratio * C->P->V[cur] +
          len / 2.0 * NormalInFace;


   } else if (C->P->type[next] == POINT_INPUT) {
      double len = d(C->P->V[prev], C->P->V[cur]);
      double dend = d(C->P->V[next], C->P->V[cur]);

      double myratio = sqrt(1 - len * len / (4.0 * dend * dend));

      p = (1.0 - myratio) * C->P->V[next] + myratio * C->P->V[cur] +
          len / 2.0 * NormalInFace;


   } else {
      // usual collar location...
      double d1, d2, dmax;
      d1 = d(C->P->V[prev], C->P->V[cur]);
      d2 = d(C->P->V[next], C->P->V[cur]);
      dmax = ar_max(d1, d2);

      p = C->P->V[cur] + dmax / 2.0 * NormalInFace;

   }
   return p;

}


void Collar::SegNeighbors(Mesh * m, int i, int &n1, int &n2)
{
   set < int >nbs;
   m->Neighbors(i, nbs);

   if (nbs.size() != 2) {
      cout << "ERROR: SegNeighbors doesn't get 2 neighbors" << endl;
      exit(0);
   }

   set < int >::iterator it;
   it = nbs.begin();
   n1 = *it;
   it++;
   n2 = *it;

}

int Collar::ConnectEndBalls()
{

   if (C->logfileisopen && C->config->verbose) {
      (*(C->LOGFILE)) << "Running Collar::ConnectEndBalls() for mesh " <<
      M->name << endl;
   }
   //cout << "Running Collar::ConnectEndBalls() for mesh " << M->name << endl;
   //cout << C->P0.size() << endl;

   int pointsadded = 0;

   // FIXME: this is inefficient... improve...

   Mesh::PIterator pIt(M);
   pIt.Start(M->GetSimplex());

   while (pIt.Current() != -1) {


      //foreach (InputPoint ip, C->P0) {

      //if (ip.l2m.find(M) != ip.l2m.end()) {
      //  int inpoint = C->P->id2loc[ip.id];

      int inpoint = pIt.Current();
      ++pIt;

      if (C->P->type[inpoint] == POINT_INPUT) {

         if (C->IsProtected(inpoint)) {

            //cout << "Protect " << ip.p << " " << ip.id << " " << C->P->V[inpoint] << endl;
            //cout << " dir " << M->dir << endl;

	   bool lazyPoint = false;
	   if (C->lazyProtectionList.find(C->P->ids[inpoint]) != C->lazyProtectionList.end()) {
	     lazyPoint = true;
	   }

	   

	   vector < Myangle > toconnect;
	   if (!lazyPoint) {
	     //  foreach(I2S p2c, p2col) {
	     
	     if (p2col.find(inpoint) != p2col.end()) {
               // get the collar points...
               foreach(int c, p2col[inpoint]) {
		 //cout << "XXX " << c << " " << C->P->V[c] << endl;
		 Myangle tmp;
		 tmp.p = c;
		 tmp.type = 0;
		 toconnect.push_back(tmp);
               }
	     }
	     // get points on non-protected segments...
	     foreach(Mesh * m, C->P->l2m[inpoint]) {
	       if (m->dim == 1 && M->Contains(m)
		   && m->protect == false) {
		 
		 // add another point to the list here...
		 int n1, n2;
		 
		 SegNeighbors(m, inpoint, n1, n2);
		 if (C->P->type[n1] == POINT_BBOX) {
		   n1 = n2;
		 }
		 Myangle tmp;
		 tmp.p = n1;
		 tmp.type = 1;
		 tmp.seg = m;
		 toconnect.push_back(tmp);
		 col2p[n1] = inpoint;
		 col2next[n1] = -1;
		 col2prev[n1] = -1;
	       }
	     }
	   } else {
	     
	     // just get the points in the lazyProtectionList
	     foreach(int nb, C->lazyProtectionNeighbors[C->P->ids[inpoint]]) {
	       
	       if (C->lazyNb2Segment[nb]->IsIn(M)) {
		 Myangle tmp;
		 tmp.p = nb;
		 tmp.type = 1;
		 tmp.seg = C->lazyNb2Segment[nb];
		 toconnect.push_back(tmp);
		 col2p[nb] = inpoint;
		 col2next[nb] = -1;
		 col2prev[nb] = -1;
	       }
	     }
	     
	     
	   }
	   //cout << inpoint << " " << lazyPoint << " " << toconnect.size() << endl;



	    if (toconnect.size() > 0) {
	      // set angles...
	      Vector v1;
	      v1 = C->P->V[toconnect[0].p] - C->P->V[inpoint];
	      v1.normalize();
	      //cout << "Points " << toconnect[0].p  << " " << p2c.first << " " << C->P->V[toconnect[0].p] << " " <<  C->P->V[p2c.first] << endl;
	      for (int ii = 0; ii < (int)toconnect.size(); ii++) {
		Vector v2;
		v2 = C->P->V[toconnect[ii].p] - C->P->V[inpoint];
		v2.normalize();
		//cout << v2 << endl;
		//cout << acos(v1*v2)*180.0/PI << " " << (v1^v2)*M->dir << endl;
		double dotprod = v1 * v2;
		if (dotprod > 1.0) {
		  dotprod = 1.0;
		}
		if (dotprod < -1.0) {
		  dotprod = -1.0;
		}
		
		if ((v1 ^ v2) * M->dir >= 0.0) {
		  toconnect[ii].angle =
		    acos(dotprod) * 180.0 / PI;
		} else {
		  toconnect[ii].angle =
		    360.0 - acos(dotprod) * 180.0 / PI;
		}
	      }
	      sort(toconnect.begin(), toconnect.end());
	    }
	    //cout << "Angles ";
	    //foreach(Myangle ang, toconnect) {
	    //  cout << ang.angle << " " << ang.p << " " << C->P->V[ang.p] << " , ";
	    //} cout << endl;
	    
	    if (toconnect.size() == 0) {
	      
	      Vector v, u, e;
	      foreach(Mesh * m, C->P->l2m[inpoint]) {
		if (m->dim == 1) {
		  set < int >nbors;
		  m->Neighbors(inpoint, nbors);
		  if (nbors.size() != 2) {
		    cout << "Warning: something bad." << endl;
		  }
		  foreach(int i, nbors) {
		    if (C->P->type[i] != POINT_BBOX) {
		      v = C->P->V[i] - C->P->V[inpoint];
		    }
		  }
		  break;
		}
	      }
	      // add one point and use the next case...
	      //cout << v << endl;
	      
	      e.xyz[0] = 1.0;
	      e.xyz[1] = 0.0;
	      e.xyz[2] = 0.0;
	      if (e * M->dir > .9) {
		   e.xyz[0] = 0.0;
		   e.xyz[1] = 1.0;
		   e.xyz[2] = 0.0;
	      }
	      u = e ^ M->dir;
	      u.normalize();
	      
	      //cout << v << u << M->dir << endl;
	      
	      u = v.len() * u;
	      Point pa = C->P->V[inpoint] + u;
	      
	      int loca = C->P->AddPoint(pa, inpoint, POINT_COLLAR);
	      p2col[inpoint].insert(loca);
	      col2p[loca] = inpoint;
	      pointsadded++;
	      InsertPointInMeshes(loca,M,lazyPoint,inpoint);
	      //C->InsertInMeshAndParents(loca, M);

	      
	      Myangle tmp;
	      tmp.p = loca;
		 tmp.type = 2;
		 toconnect.push_back(tmp);
		 // now the one point case is used to build the rest of the ball...
	    }
	    
	    if (toconnect.size() == 1) {
	      // insert three points...
	      
	      int p = toconnect[0].p;
	      Vector v = C->P->V[p] - C->P->V[inpoint];
	      
	      Point pa, pb, pc;
	      int loca, locb, locc;
	      
	      pa = C->P->V[inpoint] - v;
	      Vector v1 = v ^ M->dir;
	      pb = C->P->V[inpoint] + v1;
	      pc = C->P->V[inpoint] - v1;
	      
	      //cout << C->P->V[inpoint] << C->P->V[p] << pa << pb << pc << endl;
	      
	      loca = C->P->AddPoint(pa, inpoint, POINT_COLLAR);
	      p2col[inpoint].insert(loca);
	      col2p[loca] = inpoint;
	      pointsadded++;
	      InsertPointInMeshes(loca,M,lazyPoint,inpoint);
	      //C->InsertInMeshAndParents(loca, M);
		 
	      locb = C->P->AddPoint(pb, inpoint, POINT_COLLAR);
	      p2col[inpoint].insert(locb);
	      col2p[locb] = inpoint;
	      pointsadded++;
	      InsertPointInMeshes(locb,M,lazyPoint,inpoint);
	      //C->InsertInMeshAndParents(locb, M);
	      
	      locc = C->P->AddPoint(pc, inpoint, POINT_COLLAR);
	      p2col[inpoint].insert(locc);
	      col2p[locc] = inpoint;
	      pointsadded++;
	      
	      InsertPointInMeshes(locc,M,lazyPoint,inpoint);
	      //C->InsertInMeshAndParents(locc, M);
	      
	      
	      // connect collar structure...
	      col2prev[p] = locc;
	      col2prev[locc] = loca;
	      col2prev[loca] = locb;
	      col2prev[locb] = p;
	      col2next[p] = locb;
	      col2next[locb] = loca;
	      col2next[loca] = locc;
	      col2next[locc] = p;
	      
	    } else {
		 
	      // loop through and connect when necessary...
	      for (int ii = 0; ii < (int)toconnect.size(); ii++) {
		
		int cur, next, curt, nextt;
		double cura, nexta;
		Mesh *mseg;
		cur = toconnect[ii].p;
		cura = toconnect[ii].angle;
		curt = toconnect[ii].type;
		mseg = toconnect[ii].seg;
		if (ii + 1 < (int)toconnect.size()) {
		  next = toconnect[ii + 1].p;
		  nexta = toconnect[ii + 1].angle;
		  nextt = toconnect[ii + 1].type;
		} else {
		  next = toconnect[0].p;
		  nexta = toconnect[0].angle + 360.0;
		  nextt = toconnect[0].type;
		}
		
		// these are collar points...
		//if (toconnect[ii].type == 0) {
		
		//cout << "Next angle " << cura << " " << nexta << endl;
		
		
		int skipme = 0;
		
		if (curt == 0) {
		  Vector u1, u2;
		  u1 = C->P->V[cur] - C->P->V[inpoint];
		  u1.normalize();
		  u2 = C->P->V[col2e[cur]] - C->P->V[inpoint];
		  u2.normalize();
		  if (M->dir * (u1 ^ u2) >= 0.0) {
		    skipme = 1;
		  }
		}
		
		if (curt == 1
		    && M->bndseg.find(mseg) != M->bndseg.end()) {
		  Vector vv;

		  // old... updated for lazy vertex insertion
		  //M->NormalInFace(cur, inpoint, vv);		  
		  
		  if (C->lazyProtectionList.find(C->P->ids[inpoint])!=C->lazyProtectionList.end() && !M->convex) {
		    int a = M->bnd->next[inpoint];
		    
		    if (!C->P->IsIn(a,C->lazyNb2Segment[cur])) {
		      a = M->bnd->prev[inpoint];
		    }
		    //cout << "Here" << endl;
		    M->NormalInFace(a, inpoint, vv);
		  } else {
		    M->NormalInFace(cur, inpoint, vv);
		  }
		  
		  Vector vtmp =
		    (C->P->V[cur] - C->P->V[inpoint]);
		  
		  vtmp.normalize();
		  
		  Vector vv2 = M->dir ^ vtmp;
		  
		  if (vv * vv2 < 0) {
		    skipme = 1;
		  }
		}
		//cout << "Skipme " << skipme << " " << nexta - cura << endl;
		
		if (skipme == 0) {
		  // I am free to connect...
		  col2prev[cur] = next;
		  col2next[next] = cur;
		  
		  // this is wrong?
		  //if (nextt == 1) {
		  //  col2prev[next] = -1;
		  //}
		  
		  
		  //cout << "Angle " << nexta-cura << endl;
		  
		  if (nexta - cura > 270) {          // make this 270...
		    
		    // for 270 degrees or more... insert 3 points...
		    int loc = SplitCollarReverse(cur, next);
		    pointsadded++;
		    InsertPointInMeshes(loc,M,lazyPoint,inpoint);
		    //C->InsertInMeshAndParents(loc, M);
		    
		    int loc2 = SplitCollar(cur, loc);
		    pointsadded++;
		    InsertPointInMeshes(loc2,M,lazyPoint,inpoint);
		    //C->InsertInMeshAndParents(loc2, M);
		    
		    int loc3 = SplitCollar(loc, next);
		    pointsadded++;
		    InsertPointInMeshes(loc3,M,lazyPoint,inpoint);
		    //C->InsertInMeshAndParents(loc3, M);
		  } else if (nexta - cura > 180) {
		    
		    // for 270 degrees or more... insert 3 points...
		    int loc2;
		    int loc = SplitCollar3(cur, next, loc2);
		    pointsadded++;
		    InsertPointInMeshes(loc,M,lazyPoint,inpoint);
		    //C->InsertInMeshAndParents(loc, M);
		    pointsadded++;
		    InsertPointInMeshes(loc2,M,lazyPoint,inpoint);
		    //C->InsertInMeshAndParents(loc2, M);
		    
		    
		  } else if (nexta - cura > 90) {
		    int loc = SplitCollar(cur, next);
		    //cout << "Cur " << cur << " " << C->P->V[cur] << " Next " << next << " " << C->P->V[next] << endl;
		    //cout << "Midpoint " << loc << " " << C->P->V[loc] << endl;
		    pointsadded++;
		    InsertPointInMeshes(loc,M,lazyPoint,inpoint);
		    //C->InsertInMeshAndParents(loc, M);
		  }
		}
                  // else do nothing... this side does not need to be connected..
	      }
            }
	    
            //int tmppp = toconnect[0].p;
            //int first = tmppp;
            //cout << "Collar Loop" << endl;
            //do {
            //  cout << tmppp << ":" << C->P->V[tmppp] << " " << col2next[tmppp] << " " << col2prev[tmppp] << endl; ;
            //  tmppp = col2next[tmppp];
            //} while (tmppp != first and tmppp != -1);

         }
      }
   }

   return pointsadded;

}


int Collar::SplitCollar(int a, int b)
{
   // split the collar segment between a and b...
   // add the point to the point set and return the location...

   Point p;
   int loc;

   // decide if this is an end arc or a line
   if (col2p.find(a) != col2p.end() && col2p.find(b) != col2p.end()
       && col2p[a] == col2p[b]) {
      // split on the arc...

      int p0 = col2p[a];

      double dist = d(C->P->V[a], C->P->V[p0]);

      Point ptmp = (C->P->V[a] + C->P->V[b]) / 2;
      Vector vv = ptmp - C->P->V[p0];


      if (vv.len() < .01 * dist) {
         vv = (-1.0 * ((C->P->V[a] - C->P->V[p0]) ^ M->dir) +
               ((C->P->V[b] - C->P->V[p0]) ^ M->dir));
         //cout << "XXX " << M->dir << " " << C->P->V[a]-C->P->V[p0] << " "  << vv << endl;
         //cout << (M->dir^(C->P->V[a]-C->P->V[p0])) << endl;
      }

      vv.normalize();

      p = C->P->V[p0] + vv * dist;

      loc = C->P->AddPoint(p, p0, POINT_COLLAR);

      p2col[p0].insert(loc);
      col2p[loc] = p0;

   } else {
      if (C->P->type[a] == POINT_SEGMENT
          || C->P->type[b] == POINT_SEGMENT) {
         cout << "ERROR: segment points should only lie on collar arcs."
              << endl;
         exit(0);
      }
      // split at the midpoint...
      p = (C->P->V[a] + C->P->V[b]) / 2.0;

      loc = C->P->AddPoint(p, b, POINT_COLLAR);
   }
   //stitch loc into the collar maps...
   if (col2next[a] == b) {
      col2next[a] = loc;
      col2next[loc] = b;
      col2prev[loc] = a;
      col2prev[b] = loc;
   } else if (col2prev[a] == b) {
      col2next[b] = loc;
      col2next[loc] = a;
      col2prev[loc] = b;
      col2prev[a] = loc;
   } else {
      cout << "ERROR: bad collar split" << endl;
      exit(0);
   }

   return loc;                  // return for insertion into the mesh...
}

int Collar::SplitCollarReverse(int a, int b)
{
   // split the collar segment between a and b...
   // add the point to the point set and return the location...

   //cout << "Split CollarReverse " << a << " " << b << endl;

   Point p;
   int loc;

   // decide if this is an end arc or a line
   if (col2p.find(a) != col2p.end() && col2p.find(b) != col2p.end()
       && col2p[a] == col2p[b]) {
      // split on the arc...

      int p0 = col2p[a];

      double dist = d(C->P->V[a], C->P->V[p0]);

      Point ptmp = (C->P->V[a] + C->P->V[b]) / 2;
      Vector vv = ptmp - C->P->V[p0];

      if (vv.len() < .01 * dist) {
         vv = (-1.0 * ((C->P->V[a] - C->P->V[p0]) ^ M->dir) +
               ((C->P->V[b] - C->P->V[p0]) ^ M->dir));
         //cout << vv << endl;
      }

      vv.normalize();

      p = C->P->V[p0] - vv * dist;      // NOTE: sign change here...

      loc = C->P->AddPoint(p, p0, POINT_COLLAR);

      p2col[p0].insert(loc);
      col2p[loc] = p0;

   } else {
      cout << a << " " << b << " " << col2p[a] << " " << col2p[b] << " "
           << C->P->type[a] << " " << C->P->type[b] << endl;
      cout << "ERROR: bad data in SplitCollarReverse" << endl;
      exit(0);
   }
   //stitch loc into the collar maps...
   //if (col2next[a] == b) {
   //  col2next[a] = loc;
   //  col2next[loc] = b;
   //  col2prev[loc] = a;
   //  col2prev[b] = loc;
   //} else if (col2prev[a] == b) {

   // weird bug above...
   if (col2prev[a] == b) {
      col2next[b] = loc;
      col2next[loc] = a;
      col2prev[loc] = b;
      col2prev[a] = loc;
   } else {
      cout << "ERROR: bad collar split" << endl;
      exit(0);
   }

   return loc;                  // return for insertion into the mesh...
}

int Collar::SplitCollar3(int a, int b, int &loc2)
{
   //cout << "Split Collar3 " << a << " " << b << endl;

   Point p, p2;
   int loc;

   // decide if this is an end arc or a line
   if (col2p.find(a) != col2p.end() && col2p.find(b) != col2p.end()
       && col2p[a] == col2p[b]) {
      // split on the arc...

      int p0 = col2p[a];

      Vector va = C->P->V[a] - C->P->V[p0];
      va.normalize();
      Vector vb = C->P->V[b] - C->P->V[p0];
      vb.normalize();

      Vector v1 =.5 * ((M->dir ^ (va)) + ((vb) ^ M->dir));// arand - WARNING sign? 
      v1.normalize();

      Vector v21 = .5 * (v1 + va);
      v21.normalize();
      Vector v22 = .5 * (v1 + vb);
      v22.normalize();

      Vector v31 = .5 * (v21 + v1);
      v31.normalize();
      Vector v32 = .5 * (v22 + v1);
      v32.normalize();

      Vector v41 = .5 * (v31 + v21);
      v41.normalize();
      Vector v42 = .5 * (v32 + v22);
      v42.normalize();


      double dist = d(C->P->V[a], C->P->V[p0]);

      p = C->P->V[p0] + v41 * dist;
      p2 = C->P->V[p0] + v42 * dist;

      loc = C->P->AddPoint(p, p0, POINT_COLLAR);
      loc2 = C->P->AddPoint(p2, p0, POINT_COLLAR);

      p2col[p0].insert(loc);
      col2p[loc] = p0;
      p2col[p0].insert(loc2);
      col2p[loc2] = p0;


   } else {
      cout << "ERROR: bad data in SplitCollarReverse" << endl;
      exit(0);
   }
   //stitch loc into the collar maps...
   //if (col2next[a] == b) {
   //  col2next[a] = loc;
   //  col2next[loc] = loc2;
   //  col2next[loc2] = b;
   //  col2prev[loc] = a;
   //  col2prev[loc2] = loc;
   //  col2prev[b] = loc2;
   //} else if (col2prev[a] == b) {

   // weird bug above...
   if (col2prev[a] == b) {
      col2next[b] = loc2;
      col2next[loc2] = loc;
      col2next[loc] = a;
      col2prev[loc2] = b;
      col2prev[loc] = loc2;
      col2prev[a] = loc;
   } else {
      cout << "ERROR: bad collar split" << endl;
      exit(0);
   }

   return loc;                  // return for insertion into the mesh...
}





bool Collar::Exists(int a, int b)
{

   if (!Exists(a)) {
      return false;
   }
   int b1, b2;
   b1 = col2next[a];
   b2 = col2prev[a];
   if (b == b1 || b == b2) {
      return true;
   }
   return false;
}

bool Collar::Exists(int a)
{
   if (col2next.find(a) != col2next.end()) {
      return true;
   }
   if (C->P->type[a] == POINT_COLLAR) {
      cout << "ERROR: Something is wrong with the collar" << endl;
   }

   return false;
}


int Collar::CollarInputPoint(int iP)
{
   int pointsadded = 0;


   //cout << "CIP " << iP << " " << C->P->V[iP] << " " << C->P->ids[iP] << endl;
   //cout << M->name << endl;

   vector < Mesh * >segs;
   foreach(Mesh * m, C->P->l2m[iP]) {
      if (M->Contains(m)) {
         segs.push_back(m);
      }
   }

   if (segs.size() != 2) {
      cout << "ERROR connecting collar balls" << endl;
      cout << segs.size() << endl;
      return 0;
      exit(0);
   }

   set < int >nbs1;
   set < int >nbs2;

   segs[0]->Neighbors(iP, nbs1);
   segs[1]->Neighbors(iP, nbs2);

   if (nbs1.size() != 2 || nbs2.size() != 2) {
      cout << "Error in CollarInputPoint" << endl;
      exit(0);
   }
   int n1 = -1, n2 = -1;

   foreach(int in, nbs1) {
      if (C->P->type[in] != POINT_BBOX) {
         n1 = in;
      }
   }

   foreach(int in, nbs2) {
      if (C->P->type[in] != POINT_BBOX) {
         n2 = in;
      }
   }

   if (n1 == -1 || n2 == -1) {
      cout << "Error in CollarInputPoint" << endl;
      exit(0);
   }

   col2p[n1] = iP;
   col2p[n2] = iP;

   Vector v1, v2;

   v1 = C->P->V[n2] - C->P->V[n1];
   v2 = C->P->V[iP] - C->P->V[n1];

   if ((v1 ^ v2) * M->dir < 0) {
      col2next[n1] = n2;
      col2prev[n1] = -1;
      col2next[n2] = -1;
      col2prev[n2] = n1;
   } else {
      col2next[n1] = -1;
      col2prev[n1] = n2;
      col2next[n2] = n1;
      col2prev[n2] = -1;
   }

   v1 = C->P->V[n1] - C->P->V[iP];
   v2 = C->P->V[n2] - C->P->V[iP];
   if (v1 * v2 < 0) {
      int loc = SplitCollar(n1, n2);
      pointsadded++;
      C->InsertInMeshAndParents(loc, M);
   }

   return pointsadded;
}


void Collar::Print()
{

   cout << "---" << endl;
   cout << "Collar For " << M->name << endl;
   map < int, int >::iterator mIt;
   for (mIt = col2next.begin(); mIt != col2next.end(); mIt++) {
      cout << mIt->first << " " << mIt->second << " " << C->P->V[mIt->
                                                                 first]
           << endl;
   }

   cout << "---" << endl;
}


void Collar::InsertPointInMeshes(int loc, Mesh * m, bool lazyPoint, int inpoint) {
  if (lazyPoint) {

    C->lazyProtectionNeighbors[C->P->ids[inpoint]].insert(loc);
    C->lazyNb2Segment[loc] = m;

  } else {
    // add the point now...
    C->InsertInMeshAndParents(loc, m);
  }
  
}
