#include "allh.h"
#include <iostream>
#include <fstream>

#include "iniparser.h"

Configuration::Configuration(PLC * C1)
{

   C = C1;

   ini = NULL;

   epsilon = .000000001;
   infinity = 1000000000.0;

   boundingboxratio = 8.0;
   maxradiusedgeratio = 2.5;
   maxsliverangle = -1.0;

   maxoutputradius = -1.0;
   hsq = -1.0;

   step2asplitfactor = 2;
   facetoprint = -1;
   
   maxiterations = -1;

   exactinsphere = true;
   inexactsliverinsphere = false;
   protectallsegments = false;
   lazyvertexprotection = true;

   printtplc = true;
   printraw = false;
   printtetview = false;
   printoff = true;
   printgmv = false;
   printfreefem = false;
   printintermediate = false;
   printmonochrome = false;
   colorcollar = false;
   useintestine = false;


   verbose = false;

   removeboundingbox = true;

}

void Configuration::ReadConfiguration(string fname)
{

   ini = iniparser_load(fname.c_str());

   if (ini == NULL) {
      if (C->logfileisopen) {
         (*(C->LOGFILE)) << "Config file " << fname << " not found.   Defaults used."
                         << endl;
      }

   } else {
      (*(C->LOGFILE)) << "Config file: " << fname
                      <<" loaded successfully."  << endl;

   // double options
   epsilon = iniparser_getdouble(ini,"dir3:epsilon",epsilon);
   infinity = iniparser_getdouble(ini,"dir3:infinity",infinity);
   maxradiusedgeratio = iniparser_getdouble(ini,"dir3:maxradiusedgeratio", maxradiusedgeratio);
   maxsliverangle = iniparser_getdouble(ini,"dir3:maxsliverangle", maxsliverangle);
   maxoutputradius = iniparser_getdouble(ini,"dir3:maxoutputradius", maxoutputradius);
   boundingboxratio = iniparser_getdouble(ini,"dir3:boundingboxratio",boundingboxratio);
   hsq = maxoutputradius*maxoutputradius;

   // integer options
   step2asplitfactor = iniparser_getint(ini,"dir3:step2asplitfactor",step2asplitfactor);
   maxiterations = iniparser_getint(ini,"dir3:maxiterations",maxiterations);
   facetoprint = iniparser_getint(ini,"dir3:facetoprint",facetoprint);
   if (facetoprint != -1) {
     planestoprint.insert(facetoprint);
   }

   // boolean options
   exactinsphere = iniparser_getboolean(ini,"dir3:exactinsphere",exactinsphere);
   inexactsliverinsphere = iniparser_getboolean(ini,"dir3:inexactsliverinsphere",inexactsliverinsphere);
   protectallsegments = iniparser_getboolean(ini,"dir3:protectallsegments",protectallsegments);
   lazyvertexprotection = iniparser_getboolean(ini,"dir3:lazyvertexprotection",lazyvertexprotection);

   printtplc = iniparser_getboolean(ini,"dir3:printtplc",printtplc);
   printraw = iniparser_getboolean(ini,"dir3:printraw",printraw);
   printtetview = iniparser_getboolean(ini,"dir3:printtetview",printtetview);
   printoff = iniparser_getboolean(ini,"dir3:printoff",printoff);
   printgmv = iniparser_getboolean(ini,"dir3:printgmv",printgmv);
   printfreefem = iniparser_getboolean(ini,"dir3:printfreefem",printfreefem);
   printintermediate = iniparser_getboolean(ini,"dir3:printintermediate",printintermediate);
   printmonochrome = iniparser_getboolean(ini,"dir3:printmonochrome",printmonochrome);
   colorcollar = iniparser_getboolean(ini,"dir3:colorcollar",colorcollar);

   verbose = iniparser_getboolean(ini,"dir3:verbose",verbose);

   removeboundingbox = iniparser_getboolean(ini,"dir3:removeboundingbox",removeboundingbox);

   useintestine = iniparser_getboolean(ini,"dir3:useintestine",useintestine);


   }

   LogConfiguration();

   C->PrintEnd("Reading Configuration File.");

}

void Configuration::LogConfiguration()
{
   if (C->logfileisopen) {
      (*(C->LOGFILE)) << "Configuration:" << endl;
      (*(C->LOGFILE)) << "  Epsilon = " << epsilon << endl;
      (*(C->LOGFILE)) << "  Infinity = " << infinity << endl;
      (*(C->LOGFILE)) << "  BoundingBoxRatio = " << boundingboxratio << endl;
      (*(C->LOGFILE)) << "  MaxRadiusEdgeRatio = " << maxradiusedgeratio << endl;
      (*(C->LOGFILE)) << "  MaxSliverAngle = " << maxsliverangle << endl;
      (*(C->LOGFILE)) << "  MaxOutputRadius = " << maxoutputradius << endl;
      (*(C->LOGFILE)) << "  Step2aSplitFactor = " << step2asplitfactor << endl;

      (*(C->LOGFILE)) << "  MaxIterations = " << maxiterations << endl;
      (*(C->LOGFILE)) << "  ExactInsphere = " << exactinsphere << endl;
      (*(C->LOGFILE)) << "  InexactSliverInsphere = " << inexactsliverinsphere << endl;
      (*(C->LOGFILE)) << "  ProtectAllSegments = " << protectallsegments<< endl;
      (*(C->LOGFILE)) << "  LazyVertexProtection = " << lazyvertexprotection<< endl;

      (*(C->LOGFILE)) << "  PrintTPLC = " << printtplc << endl;
      (*(C->LOGFILE)) << "  PrintOFF = " << printoff << endl;
      (*(C->LOGFILE)) << "  PrintRaw = " << printraw << endl;
      (*(C->LOGFILE)) << "  PrintTetView = " << printtetview << endl;
      (*(C->LOGFILE)) << "  PrintGMV = " << printgmv << endl;
      (*(C->LOGFILE)) << "  PrintFreeFEM = " << printfreefem << endl;
      (*(C->LOGFILE)) << "  Printintermediate = " << printintermediate << endl;
      (*(C->LOGFILE)) << "  FaceToPrint = " << facetoprint << endl;

      (*(C->LOGFILE)) << "  PrintMonochrome = " << printmonochrome << endl;
      (*(C->LOGFILE)) << "  ColorCollar = " << colorcollar << endl;
      (*(C->LOGFILE)) << "  UseIntestine = " << useintestine << endl;
      (*(C->LOGFILE)) << "  Verbose = " << verbose << endl;
      (*(C->LOGFILE)) << "  RemoveBoundingBox = " << removeboundingbox << endl;
   }
}
