#include <algorithm>

#include "allh.h"

#include "insphere.h"

Geometry::Geometry(PLC * c1)
{
   C = c1;
   P = C->P;

   noslivers = false;

   insphere_init();
}

double Geometry::ratio(Simplex * s)
{
   double cr, se, tmp;

   cr = d(circumcenter(s), P->V[s->v[0]]);

   se = d(P->V[s->v[0]], P->V[s->v[1]]);

   for (int i = 2; i < 4; i++) {
      tmp = d(P->V[s->v[0]], P->V[s->v[i]]);
      if (tmp < se)
         se = tmp;
   }

   for (int i = 2; i < 4; i++) {
      tmp = d(P->V[s->v[1]], P->V[s->v[i]]);
      if (tmp < se)
         se = tmp;
   }

   if (s->size() == 4) {
      tmp = d(P->V[s->v[2]], P->V[s->v[3]]);
      if (tmp < se)
         se = tmp;
   }

   return se / cr;
}

bool Geometry::inSphereReallyStrict(int p, Simplex * s)
{
   if (d2(s->cc, P->V[p]) < s->rsq - C->config->epsilon) {
      return true;
   }
   return false;

}

bool Geometry::inSphereReallyStrict(Point & p, Simplex * s)
{
   if (d2(s->cc, p) < s->rsq - C->config->epsilon) {
      return true;
   }
   return false;
}


bool Geometry::inSphereStrict(int p, Simplex * s)
{
   return inSphereStrict(P->V[p], s);
}


bool Geometry::inSphereStrict(Point & p, Simplex * s)
{
  
  // DEBUG code...
  //if (s->rsq == -1.0) {
  //    cout << "ERROR: In function inSphere, simplex does not seem to be set up correctly." << endl;
  //    cout << p << " " << s << endl;
  
  //    cout << s->v[0] << " " << s->v[1] << " " << s->v[2] << " " << s->v[3] << " " << endl;
  
  //    exit(0);
  //}
  
  if (C->config->exactinsphere) {
    
    if (s->size() == 4) {
      if (insphere3to3(P->V[s->v[0]].xyz[0], P->V[s->v[0]].xyz[1],
		       P->V[s->v[0]].xyz[2], P->V[s->v[1]].xyz[0],
		       P->V[s->v[1]].xyz[1], P->V[s->v[1]].xyz[2],
		       P->V[s->v[2]].xyz[0], P->V[s->v[2]].xyz[1],
		       P->V[s->v[2]].xyz[2], P->V[s->v[3]].xyz[0],
		       P->V[s->v[3]].xyz[1], P->V[s->v[3]].xyz[2], p.xyz[0],
		       p.xyz[1], p.xyz[2]) < 0) {
	return true;
      }
      
      return false;
    } else if (s->size() == 3) {
      if (insphere2to3(P->V[s->v[0]].xyz[0], P->V[s->v[0]].xyz[1],
		       P->V[s->v[0]].xyz[2], P->V[s->v[1]].xyz[0],
		       P->V[s->v[1]].xyz[1], P->V[s->v[1]].xyz[2],
		       P->V[s->v[2]].xyz[0], P->V[s->v[2]].xyz[1],
		       P->V[s->v[2]].xyz[2], p.xyz[0], p.xyz[1], p.xyz[2]) < 0) {
	return true;
      }
      return false;
      
      
    } else if (s->size() == 2) {
      

      if (insphere1to3(P->V[s->v[0]].xyz[0], P->V[s->v[0]].xyz[1],
		       P->V[s->v[0]].xyz[2], P->V[s->v[1]].xyz[0],
		       P->V[s->v[1]].xyz[1], P->V[s->v[1]].xyz[2], p.xyz[0],
		       p.xyz[1], p.xyz[2]) < 0) {
	return true;
      }
      
      return false;
    }
    
  } else {
    // inexact version... duplicated in inSphereReallyStrict
    if (d2(s->cc, p) < s->rsq - C->config->epsilon) {
      return true;
    }
  }
  return false;  
}


bool Geometry::AlmostCoplanar(Point & p, int pa, int pb, int pc) {

   Vector v1,v2,v3,n;
   int tmp;
   if (pa < pb) {
      tmp = pa;
      pa = pb;
      pb = tmp;
   }
   if (pb < pc) {
      tmp = pb;
      pb = pc;
      pc = tmp;
   }
   if (pa < pb) {
      tmp = pa;
      pa = pb;
      pb = tmp;
   }


   v1 = P->V[pb] - P->V[pa];
   v2 = P->V[pc] - P->V[pa];
   v3 = p - P->V[pa];

   v1.normalize();
   v2.normalize();
   v3.normalize();

   n  = v1^v2;
   n.normalize();

   if (fabs(n*v3) < C->config->epsilon) {

      return true;

   }
   return false;
}

bool Geometry::inSphereStrictNoSliver(Point & p, Simplex * s)
{
   if (!C->config->inexactsliverinsphere || s->size() != 4) {
      return inSphereStrict(p,s);
   }

   if (inSphereStrict(p,s)) {


	   // is p nearly coplanar with any faces of s?
	   // if so, only return true if it encoraches the
	   // neighboring tet also...
	   int a,b,c,d,tmp;
	   a = s->v[1];
	   b = s->v[2];
	   c = s->v[3];
	   d = s->v[0];
	   
	   for (int i=0; i<4; i++) {
	       // is the point nearly coplanar?
	       if (AlmostCoplanar(p,a,b,c)) {
		   //noslivers = false;
		   if (noslivers && s->n[i] != NULL && !inSphereStrict(p,s->n[i])) {
		       
		       double distmin,distmax,disttmp;;
		       distmin = d2(p,P->V[a]);
		       distmax = distmin;

		       disttmp = d2(p,P->V[b]);
		       if( disttmp < distmin) {
			   distmin = disttmp;
		       }
		       if( disttmp > distmax) {
			   distmax = disttmp;
		       }

		       disttmp = d2(p,P->V[c]);
		       if( disttmp < distmin) {
			   distmin = disttmp;
		       }
		       if( disttmp > distmax) {
			   distmax = disttmp;
		       }

		       disttmp = d2(P->V[b],P->V[c]);
		       if( disttmp < distmin) {
			   distmin = disttmp;
		       }
		       if( disttmp > distmax) {
			   distmax = disttmp;
		       }

		       disttmp = d2(P->V[b],P->V[a]);
		       if( disttmp < distmin) {
			   distmin = disttmp;
		       }
		       if( disttmp > distmax) {
			   distmax = disttmp;
		       }


		       disttmp = d2(P->V[a],P->V[c]);
		       if( disttmp < distmin) {
			   distmin = disttmp;
		       }
		       if( disttmp > distmax) {
			   distmax = disttmp;
		       }

		       if (distmin/distmax > .01*.01) {

			   cout << distmin << " " << distmax << " " << distmin/distmax << endl;
			   
			   cout << "Sliver " << s << p << " " << ratio(s) << endl;
			   cout << i << " " << a << " " << b << " " << c << " " << d << endl;
			   cout << P->V[a] << " " << P->V[b] << " " << P->V[c] << endl;
			   C->P->print(s);
			   C->P->print(s->n[i]);
			   
			   // fail since I don't encroach the adjacent tet...
			   return false;
		       }
		   }
	       }
	  
	       tmp = a;
	       a = b;
	       b = c;
	       c = d;
	       d = tmp;
	       
	   }


       return true;
   }
   
   return false;
}


double Geometry::circumradius(Simplex * s)
{
   return d(P->V[s->v[0]], circumcenter(s));
}

void Geometry::DihedralAngles(Simplex * s, vector < double >&angles)
{
   // s must be a tet...

   if (s->size() < 4) {
      cout << "ERROR: a tet not given to DihedralAngles" << endl;
      cout << s->size() << endl;
   }

   Point p0, p1, p2, p3;
   Vector n1, n2;
   p0 = P->V[s->v[0]];
   p1 = P->V[s->v[1]];
   p2 = P->V[s->v[2]];
   p3 = P->V[s->v[3]];

   // 01 dihedral
   n1 = (p1 - p0) ^ (p2 - p0);
   n1.normalize();
   n2 = (p1 - p0) ^ (p3 - p0);
   n2.normalize();
   double val = n1 * n2;
   if (val > 1.0 || val < -1.0) {
      angles.push_back(0.0);
   } else {
      angles.push_back(acos(val));
   }

   // 02 dihedral
   n1 = (p2 - p0) ^ (p1 - p0);
   n1.normalize();
   n2 = (p2 - p0) ^ (p3 - p0);
   n2.normalize();
   val = n1 * n2;
   if (val > 1.0 || val < -1.0) {
      angles.push_back(0.0);
   } else {
      angles.push_back(acos(val));
   }

   // 03 dihedral
   n1 = (p3 - p0) ^ (p1 - p0);
   n1.normalize();
   n2 = (p3 - p0) ^ (p2 - p0);
   n2.normalize();
   val = n1 * n2;
   if (val > 1.0 || val < -1.0) {
      angles.push_back(0.0);
   } else {
      angles.push_back(acos(val));
   }

   // 12 dihedral
   n1 = (p2 - p1) ^ (p0 - p1);
   n1.normalize();
   n2 = (p2 - p1) ^ (p3 - p1);
   n2.normalize();
   val = n1 * n2;
   if (val > 1.0 || val < -1.0) {
      angles.push_back(0.0);
   } else {
      angles.push_back(acos(val));
   }

   // 13 dihedral
   n1 = (p3 - p1) ^ (p0 - p1);
   n1.normalize();
   n2 = (p3 - p1) ^ (p2 - p1);
   n2.normalize();
   val = n1 * n2;
   if (val > 1.0 || val < -1.0) {
      angles.push_back(0.0);
   } else {
      angles.push_back(acos(val));
   }

   // 23 dihedral
   n1 = (p3 - p2) ^ (p0 - p2);
   n1.normalize();
   n2 = (p3 - p2) ^ (p1 - p2);
   n2.normalize();
   val = n1 * n2;
   if (val > 1.0 || val < -1.0) {
      angles.push_back(0.0);
   } else {
      angles.push_back(acos(val));
   }


   sort(angles.begin(), angles.end());

   // convert to degrees...
   for (int i = 0; i < (int)angles.size(); i++) {
      angles[i] = angles[i] *180.0 / PI;
   }

}

bool Geometry::TwoDIsPoorQuality(Simplex * s)
{

   if (ratio(s) < 1.0 / C->config->maxradiusedgeratio) {
      return true;
   }

   //double size = C->M->bbox / C->config->bbox_factor + C->config->epsilon;

   if (C->config->hsq > 0.0) {

      if (s->rsq > C->config->hsq) {

         for (int i = 0; i < s->size(); i++) {

            if (P->V[s->v[i]].xyz[0] < C->M->upperbound[0]+ C->config->epsilon
                && P->V[s->v[i]].xyz[0]  > C->M->lowerbound[0] - C->config->epsilon
                && P->V[s->v[i]].xyz[1]  < C->M->upperbound[1] + C->config->epsilon
                && P->V[s->v[i]].xyz[1]  > C->M->lowerbound[1] - C->config->epsilon) {
               return true;
            } else {

            }
         }
         return false;
      }
   }

   return false;
}

bool Geometry::IsPoorQuality(Simplex * s)
{

   if (ratio(s) < 1.0 / C->config->maxradiusedgeratio) {
      return true;
   }

   //double size = C->M->bbox / C->config->bbox_factor + C->config->epsilon;

   if (C->config->hsq > 0.0) {

      if (s->rsq > C->config->hsq) {

         for (int i = 0; i < s->size(); i++) {

	   
            if (P->V[s->v[i]].xyz[0] < C->M->upperbound[0]+ C->config->epsilon
                && P->V[s->v[i]].xyz[0]  > C->M->lowerbound[0] - C->config->epsilon
                && P->V[s->v[i]].xyz[1]  < C->M->upperbound[1] + C->config->epsilon
                && P->V[s->v[i]].xyz[1]  > C->M->lowerbound[1] - C->config->epsilon
                && P->V[s->v[i]].xyz[2]  < C->M->upperbound[2] + C->config->epsilon
                && P->V[s->v[i]].xyz[2]  > C->M->lowerbound[2] - C->config->epsilon) {
	   
	     /*
	   double val0 = C->M->upperbound[0] - C->M->lowerbound[0];
	   double val1 = C->M->upperbound[1] - C->M->lowerbound[1];
	   double val2 = C->M->upperbound[2] - C->M->lowerbound[2];
	   if (P->V[s->v[i]].xyz[0] < C->M->upperbound[0]+ .2*val0
	       && P->V[s->v[i]].xyz[0]  > C->M->lowerbound[0] - .2*val0
	       && P->V[s->v[i]].xyz[1]  < C->M->upperbound[1] + .2*val1
	       && P->V[s->v[i]].xyz[1]  > C->M->lowerbound[1] - .2*val1
	       && P->V[s->v[i]].xyz[2]  < C->M->upperbound[2] + .2*val2
                && P->V[s->v[i]].xyz[2]  > C->M->lowerbound[2] - .2*val2) {

	     //if (P->type[s->v[i]] == POINT_RING || 
	     // P->type[s->v[i]] == POINT_SPHERE || 
	     // P->type[s->v[i]] == POINT_CYLINDER) {
	     */
               return true;
	       //}
            } else {
            }
         }
         return false;
      }
   }

   if (C->config->maxsliverangle > 0) {
   
     vector<double> angles;
     
     DihedralAngles(s,angles);
     
     if (angles[0] <= C->config->maxsliverangle) {
       return true;
     }
   }
   return false;
}





double Geometry::Orient(Simplex * s)
{
   if (s->size() == 4) {
      double c11, c12, c13, c21, c22, c23, c31, c32, c33;
      double d1, d2, d3;
      c11 = P->V[s->v[0]].xyz[0] - P->V[s->v[3]].xyz[0];
      c12 = P->V[s->v[0]].xyz[1] - P->V[s->v[3]].xyz[1];
      c13 = P->V[s->v[0]].xyz[2] - P->V[s->v[3]].xyz[2];
      c21 = P->V[s->v[1]].xyz[0] - P->V[s->v[3]].xyz[0];
      c22 = P->V[s->v[1]].xyz[1] - P->V[s->v[3]].xyz[1];
      c23 = P->V[s->v[1]].xyz[2] - P->V[s->v[3]].xyz[2];
      c31 = P->V[s->v[2]].xyz[0] - P->V[s->v[3]].xyz[0];
      c32 = P->V[s->v[2]].xyz[1] - P->V[s->v[3]].xyz[1];
      c33 = P->V[s->v[2]].xyz[2] - P->V[s->v[3]].xyz[2];
      d1 = c22 * c33 - c23 * c32;
      d2 = c21 * c33 - c31 * c23;
      d3 = c21 * c32 - c22 * c31;
      return c11 * d1 - c12 * d2 + c13 * d3;
   }
   return 0;
}

// faster version of circumcenter
// this is more robust for slivers than before...
Point Geometry::circumcenter(Simplex * s)
{
   int size = s->size();

   if (size == 1) {
      return P->V[s->v[0]];
   } else if (size == 2) {
      return (P->V[s->v[0]] + P->V[s->v[1]]) / 2.0;
   } else if (size == 3) {

      double rb, rc, rd;
      double bx, by, bz;
      double cx, cy, cz;
      double dx, dy, dz;

      bx = P->V[s->v[1]].xyz[0] - P->V[s->v[0]].xyz[0];
      by = P->V[s->v[1]].xyz[1] - P->V[s->v[0]].xyz[1];
      bz = P->V[s->v[1]].xyz[2] - P->V[s->v[0]].xyz[2];
      cx = P->V[s->v[2]].xyz[0] - P->V[s->v[0]].xyz[0];
      cy = P->V[s->v[2]].xyz[1] - P->V[s->v[0]].xyz[1];
      cz = P->V[s->v[2]].xyz[2] - P->V[s->v[0]].xyz[2];

      rb = .5 * (P->V[s->v[1]] - P->V[s->v[0]]) * (P->V[s->v[1]] +
                                                   P->V[s->v[0]]);
      rc = .5 * (P->V[s->v[2]] - P->V[s->v[0]]) * (P->V[s->v[2]] +
                                                   P->V[s->v[0]]);

      Vector n =
         (P->V[s->v[1]] - P->V[s->v[0]]) ^ (P->V[s->v[2]] -
                                            P->V[s->v[0]]);
      n.normalize();
      dx = n.xyz[0];
      dy = n.xyz[1];
      dz = n.xyz[2];

      rd = n * P->V[s->v[0]];

      double den =
         bx * cy * dz + by * cz * dx + bz * cx * dy - dx * cy * bz -
         dy * cz * bx - dz * cx * by;


      if (den < EPSILON && den > -EPSILON) {
	//cout << "WARNING: Colinear points in a simplex" << endl;
	double d1 = d(P->V[s->v[1]],P->V[s->v[0]]);
	double d2 = d(P->V[s->v[2]],P->V[s->v[0]]);
	if (d1 > d2) {
	  return (P->V[s->v[1]] + P->V[s->v[0]]) / 2.0;
	}
	return (P->V[s->v[2]] + P->V[s->v[0]]) / 2.0;
      }

      double numx, numy, numz;
      numx =
         rb * cy * dz + by * cz * rd + bz * rc * dy - rd * cy * bz -
         dy * cz * rb - dz * rc * by;
      numy =
         bx * rc * dz + rb * cz * dx + bz * cx * rd - dx * rc * bz -
         rd * cz * bx - dz * cx * rb;
      numz =
         bx * cy * rd + by * rc * dx + rb * cx * dy - dx * cy * rb -
         dy * rc * bx - rd * cx * by;

      Point p(numx / den, numy / den, numz / den);
      return p;



   } else if (size == 4) {

      double rb, rc, rd;
      double bx, by, bz;
      double cx, cy, cz;
      double dx, dy, dz;

      bx = P->V[s->v[1]].xyz[0] - P->V[s->v[0]].xyz[0];
      by = P->V[s->v[1]].xyz[1] - P->V[s->v[0]].xyz[1];
      bz = P->V[s->v[1]].xyz[2] - P->V[s->v[0]].xyz[2];
      cx = P->V[s->v[2]].xyz[0] - P->V[s->v[0]].xyz[0];
      cy = P->V[s->v[2]].xyz[1] - P->V[s->v[0]].xyz[1];
      cz = P->V[s->v[2]].xyz[2] - P->V[s->v[0]].xyz[2];
      dx = P->V[s->v[3]].xyz[0] - P->V[s->v[0]].xyz[0];
      dy = P->V[s->v[3]].xyz[1] - P->V[s->v[0]].xyz[1];
      dz = P->V[s->v[3]].xyz[2] - P->V[s->v[0]].xyz[2];

      rb = .5 * (P->V[s->v[1]] - P->V[s->v[0]]) * (P->V[s->v[1]] +
                                                   P->V[s->v[0]]);
      rc = .5 * (P->V[s->v[2]] - P->V[s->v[0]]) * (P->V[s->v[2]] +
                                                   P->V[s->v[0]]);
      rd = .5 * (P->V[s->v[3]] - P->V[s->v[0]]) * (P->V[s->v[3]] +
                                                   P->V[s->v[0]]);

      double den =
         bx * cy * dz + by * cz * dx + bz * cx * dy - dx * cy * bz -
         dy * cz * bx - dz * cx * by;


      // FIXME: try different values here
      if (den < EPSILON && den > -EPSILON) {
         // sliver case...

         Vector n =
            (P->V[s->v[3]] - P->V[s->v[0]]) ^ (P->V[s->v[2]] -
                                               P->V[s->v[0]]);
         n.normalize();
         dx = n.xyz[0];
         dy = n.xyz[1];
         dz = n.xyz[2];

         rd = n * P->V[s->v[0]];

         den =
            bx * cy * dz + by * cz * dx + bz * cx * dy - dx * cy * bz -
            dy * cz * bx - dz * cx * by;
         if (den < EPSILON && den > -EPSILON) {
	   //cout << "WARNING: 3 colinear points in a tet." << endl;
	   //cout << s << endl;
	   //P->print(s);
	   //exit(0);
	   double d1 = d(P->V[s->v[1]],P->V[s->v[0]]);
	   double d2 = d(P->V[s->v[2]],P->V[s->v[0]]);
	   double d3 = d(P->V[s->v[3]],P->V[s->v[0]]);

	   if (d1 > d2 && d1 > d3) {
	     return (P->V[s->v[1]] + P->V[s->v[0]]) / 2.0;
	   } else if (d2 > d2) {
	     return (P->V[s->v[2]] + P->V[s->v[0]]) / 2.0;	     
	   }
	   return (P->V[s->v[3]] + P->V[s->v[0]]) / 2.0;
         }

      }

      double numx, numy, numz;

      numx =
         rb * cy * dz + by * cz * rd + bz * rc * dy - rd * cy * bz -
         dy * cz * rb - dz * rc * by;
      numy =
         bx * rc * dz + rb * cz * dx + bz * cx * rd - dx * rc * bz -
         rd * cz * bx - dz * cx * rb;
      numz =
         bx * cy * rd + by * rc * dx + rb * cx * dy - dx * cy * rb -
         dy * rc * bx - rd * cx * by;

      Point p(numx / den, numy / den, numz / den);
      return p;

   } else {
      cout << "ERROR: Circumcenter called with a simplex of size " <<
      size << endl;
   }
   Point p;
   return p;

}

double Geometry::Volume(Simplex * s) {

   if (s->size() == 2) {
      return d(P->V[s->v[0]],P->V[s->v[1]]);
   } else if (s->size() == 3) {

   } else if (s->size() == 4) {
      Vector a,b,c;
      a = P->V[s->v[0]]-P->V[s->v[3]];
      b = P->V[s->v[1]]-P->V[s->v[3]];
      c = P->V[s->v[2]]-P->V[s->v[3]];

      double val = (a*(b^c))/6.0;
      if (val < 0.0) {
         return -1.0*val;
      }
      return val;

   }

   return 0.0;
}
