/*********************************************
**
** Expansions limited to 10000 elements only in
** order to save memory.  This may influence the
** correctness of the results.  To avoid this,
** further analysis of the program is necessary
** in order to recycle unused space.
**
***********************************************/


#include "insphere.h"

#include "predicates1.cpp"

REAL errA1to3, errB1to3, errC1to3, errCX1to3;
REAL errA2to3, errB2to3, errC2to3, errCX2to3;
REAL errA3to3, errB3to3, errC3to3, errCX3to3;

void insphere_init()
{
   exactinit();
   insphere1to3_init();
   insphere2to3_init();
   insphere3to3_init();
}


void insphere1to3_init()
{
   double E = epsilon;

#ifndef CYGWIN
   int round_up = 6770, round_near = 4722;
   _FPU_SETCW(round_up);
#endif
   errA1to3 = (4.0 + 19.0 * E) * E;
   errB1to3 = (2.0 + 14.0 * E) * E;
   errC1to3 = (9.0 + 67.0 * E) * E * E;
   errCX1to3 = (2.0 + 7.0 * E) * E;
#ifndef CYGWIN
   _FPU_SETCW(round_near);
#endif
}

int insphere1to3(REAL ax, REAL ay, REAL az, REAL bx, REAL by, REAL bz,
                 REAL px, REAL py, REAL pz)
{
   INEXACT REAL c, abig, ahi, alo, bhi, blo, err1, err2, err3, _i, bvirt,
                avirt, bround, around, _j, _0;
   REAL yCX, yCE, yBX, yBE, yAE, wzC, wz, wyC, wy, wxC, wx, vzC, vz, vyC,
        vy, vxC, vx, t11, t10, t9, t8, t7, t6, t5, t4, t3, t2, t,
        insphereP5, insphereP4, insphereC5, insphereC4, insphereC3,
        insphereC2, insphereC, insphere5, insphere4, insphere3, insphere2,
        insphere;
   REAL t23[2], t22[2], t21[2], t19[2], t18[2], t17[2], t15[2], t14[2],
        t13[2], insphereB3[2], insphereB2[2], insphereB[2];
   REAL t24[24], t20[4], t16[4], t12[4], insphereD5[18], insphereD4[12],
        insphereD3[6], insphereD2[6], insphereD[6], insphereB5[6],
        insphereB4[4];
   int t24_len, t20_len, t16_len, t12_len, insphereD5_len, insphereD4_len,
       insphereD3_len, insphereD2_len, insphereD_len, insphereB5_len,
       insphereB4_len;

   vx = (REAL) (px - ax);
   vy = (REAL) (py - ay);
   vz = (REAL) (pz - az);
   wx = (REAL) (px - bx);
   wy = (REAL) (py - by);
   wz = (REAL) (pz - bz);
   insphere = (REAL) (vx * wx);
   insphere2 = (REAL) (vy * wy);
   insphere3 = (REAL) (vz * wz);
   insphere4 = (REAL) (insphere2 + insphere3);
   t = Absolute(insphere2);
   t2 = Absolute(insphere3);
   insphereP4 = (REAL) (t + t2);
   insphere5 = (REAL) (insphere + insphere4);
   t3 = Absolute(insphere);
   insphereP5 = (REAL) (t3 + insphereP4);
   yAE = (REAL) (errA1to3 * insphereP5);

   if (insphere5 > yAE)
      return 1;
   else {
      if (-insphere5 > yAE)
         return -1;
      else {
         Two_Product(vx, wx, insphereB[1], insphereB[0]);
         Two_Product(vy, wy, insphereB2[1], insphereB2[0]);
         Two_Product(vz, wz, insphereB3[1], insphereB3[0]);
         Two_Two_Sum(insphereB2[1], insphereB2[0], insphereB3[1],
                     insphereB3[0], insphereB4[3], insphereB4[2],
                     insphereB4[1], insphereB4[0]);
         insphereB4_len = 4;
         insphereB5_len =
            fast_expansion_sum_zeroelim(2, insphereB, insphereB4_len,
                                        insphereB4, insphereB5);
         yBX = estimate(insphereB5_len, insphereB5);
         yBE = (REAL) (errB1to3 * insphereP5);
         if (yBX > yBE)
            return 1;
         else {
            if (-yBX > yBE)
               return -1;
            else {
               Two_Diff_Tail(px, ax, vx, vxC);
               Two_Diff_Tail(py, ay, vy, vyC);
               Two_Diff_Tail(pz, az, vz, vzC);
               Two_Diff_Tail(px, bx, wx, wxC);
               Two_Diff_Tail(py, by, wy, wyC);
               Two_Diff_Tail(pz, bz, wz, wzC);
               t4 = (REAL) (vx * wxC);
               t5 = (REAL) (vxC * wx);
               insphereC = (REAL) (t4 + t5);
               t6 = (REAL) (vy * wyC);
               t7 = (REAL) (vyC * wy);
               insphereC2 = (REAL) (t6 + t7);
               t8 = (REAL) (vz * wzC);
               t9 = (REAL) (vzC * wz);
               insphereC3 = (REAL) (t8 + t9);
               insphereC4 = (REAL) (insphereC2 + insphereC3);
               insphereC5 = (REAL) (insphereC + insphereC4);
               yCX = (REAL) (yBX + insphereC5);
               t10 = (REAL) (errCX1to3 * yBX);
               t11 = (REAL) (errC1to3 * insphereP5);
               yCE = (REAL) (t10 + t11);
               if (yCX > yCE)
                  return 1;
               else {
                  if (-yCX > yCE)
                     return -1;
                  else {
                     Two_Product(vx, wxC, t13[1], t13[0]);
                     Two_Product(vxC, wx, t14[1], t14[0]);
                     Two_Two_Sum(t13[1], t13[0], t14[1], t14[0],
                                 t12[3], t12[2], t12[1], t12[0]);
                     t12_len = 4;
                     Two_Product(vxC, wxC, t15[1], t15[0]);
                     insphereD_len =
                        fast_expansion_sum_zeroelim(2, t15,
                                                    t12_len, t12,
                                                    insphereD);
                     Two_Product(vy, wyC, t17[1], t17[0]);
                     Two_Product(vyC, wy, t18[1], t18[0]);
                     Two_Two_Sum(t17[1], t17[0], t18[1], t18[0],
                                 t16[3], t16[2], t16[1], t16[0]);
                     t16_len = 4;
                     Two_Product(vyC, wyC, t19[1], t19[0]);
                     insphereD2_len =
                        fast_expansion_sum_zeroelim(2, t19,
                                                    t16_len, t16,
                                                    insphereD2);
                     Two_Product(vz, wzC, t21[1], t21[0]);
                     Two_Product(vzC, wz, t22[1], t22[0]);
                     Two_Two_Sum(t21[1], t21[0], t22[1], t22[0],
                                 t20[3], t20[2], t20[1], t20[0]);
                     t20_len = 4;
                     Two_Product(vzC, wzC, t23[1], t23[0]);
                     insphereD3_len =
                        fast_expansion_sum_zeroelim(2, t23,
                                                    t20_len, t20,
                                                    insphereD3);
                     insphereD4_len =
                        fast_expansion_sum_zeroelim(insphereD3_len,
                                                    insphereD3,
                                                    insphereD2_len,
                                                    insphereD2,
                                                    insphereD4);
                     insphereD5_len =
                        fast_expansion_sum_zeroelim(insphereD4_len,
                                                    insphereD4,
                                                    insphereD_len,
                                                    insphereD,
                                                    insphereD5);
                     t24_len =
                        fast_expansion_sum_zeroelim(insphereD5_len,
                                                    insphereD5,
                                                    insphereB5_len,
                                                    insphereB5,
                                                    t24);
                     if (t24[t24_len - 1] > 0.0)
                        return 1;
                     else if (t24[t24_len - 1] < 0.0)
                        return -1;
                     else
                        return 0;
                  }
               };
            }
         };
      }
   };

}



void insphere2to3_init()
{

   double E = epsilon;
#ifndef CYGWIN
   int round_up = 6770, round_near = 4722;
   _FPU_SETCW(round_up);
#endif
   errA2to3 = (32.0 + 593.0 * E) * E;
   errB2to3 = (10.0 + 150.0 * E) * E;
   errC2to3 = (307.0 + 8955.0 * E) * E * E;
   errCX2to3 = (2.0 + 7.0 * E) * E;
#ifndef CYGWIN
   _FPU_SETCW(round_near);
#endif
}

int insphere2to3(REAL ax, REAL ay, REAL az, REAL bx, REAL by, REAL bz,
                 REAL cx, REAL cy, REAL cz, REAL px, REAL py, REAL pz)
{
   INEXACT REAL c, abig, ahi, alo, bhi, blo, err1, err2, err3, _i, bvirt,
                avirt, bround, around, _j, _0;
   REAL yMinorP3, yMinorP2, yMinorP, yMinorC3, yMinorC2, yMinorC, yMinor3,
        yMinor2, yMinor, yCX, yCE, yBX, yBE, yAE, xMinorP3, xMinorP2,
        xMinorP, xMinorC3, xMinorC2, xMinorC, xMinor3, xMinor2, xMinor,
        wwC5, wwC4, wwC3, wwC2, wwC, ww5, ww4, ww3, ww2, ww, wvP10, wvP9,
        wvP5, wvP4, wvC10, wvC9, wvC8, wvC7, wvC6, wvC5, wvC4, wvC3, wvC2,
        wvC, wv10, wv9, wv8, wv7, wv6, wv5, wv4, wv3, wv2, wv, vP15, vP14,
        vC15, vC14, vC13, vC12, vC11, vC10, vC9, vC8, vC7, vC6, vC5, vC4,
        vC3, vC2, vC, v15, v14, v13, v12, v11, v10, v9, v8, v7, v6, v5, v4,
        v3, v2, v, t58, t57, t56, t55, t54, t53, t52, t51, t50, t49, t48,
        t47, t46, t45, t44, t43, t42, t41, t40, t39, t38, t37, t36, t35,
        t34, t33, t32, t31, t30, t29, t28, t27, t26, t25, t24, t23, t22,
        t21, t20, t19, t18, t17, t16, t15, t14, t13, t12, t11, t10, t9, t8,
        t7, t6, t5, t4, t3, t2, t, przC, prz, pryC, pry, prxC, prx,
        numerP5, numerP4, numerP3, numerP2, numerP, numerC5, numerC4,
        numerC3, numerC2, numerC, numer5, numer4, numer3, numer2, numer,
        insphereP, insphereC, insphere, denomP3, denomP2, denomC3, denomC2,
        denomC, denom3, denom2, denom, crzC, crz, cryC, cry, crxC, crx,
        brzC, brz, bryC, bry, brxC, brx;
   REAL wwB3[2], wwB2[2], wwB[2], wvB8[2], wvB7[2], wvB6[2], wvB3[2],
        wvB2[2], wvB[2], vB13[2], vB12[2], vB11[2], vB8[2], vB7[2], vB6[2],
        vB3[2], vB2[2], vB[2], t121[2], t120[2], t119[2], t117[2], t116[2],
        t115[2], t113[2], t112[2], t111[2], t109[2], t108[2], t107[2],
        t105[2], t104[2], t103[2], t101[2], t100[2], t99[2], t97[2],
        t96[2], t95[2], t93[2], t92[2], t91[2], t89[2], t88[2], t87[2],
        t85[2], t84[2], t83[2], t82[2], t81[2], t80[2], t79[2], t78[2],
        t77[2], t76[2], t75[2], t74[2], t73[2], t72[2], t71[2], t70[2],
        t69[2], t68[2], t67[2], t66[2], t65[2], t64[2], t63[2], t62[2],
        t61[2], t60[2], t59[2];
   REAL yMinorD3[1872], yMinorD2[1080], yMinorD[792], yMinorB3[144],
        yMinorB2[72], yMinorB[72], xMinorD3[1872], xMinorD2[792],
        xMinorD[1080], xMinorB3[144], xMinorB2[72], xMinorB[72], wwD5[12],
        wwD4[8], wwD3[4], wwD2[4], wwD[4], wwB5[6], wwB4[4], wvD10[18],
        wvD9[12], wvD8[6], wvD7[6], wvD6[6], wvD5[18], wvD4[12], wvD3[6],
        wvD2[6], wvD[6], wvB10[6], wvB9[4], wvB5[6], wvB4[4], vD15[18],
        vD14[12], vD13[6], vD12[6], vD11[6], vD10[12], vD9[8], vD8[4],
        vD7[4], vD6[4], vD5[12], vD4[8], vD3[4], vD2[4], vD[4], vB15[6],
        vB14[4], vB10[6], vB9[4], vB5[6], vB4[4], t162[10000], t161[10000],
        t160[10000], t159[10000], t158[10000], t157[10000], t156[3456],
        t155[10000], t154[10000], t153[10000], t152[3456], t151[10000],
        t150[10000], t149[10000], t148[3456], t147[10000], t146[10000],
        t145[648], t144[216], t143[216], t142[432], t141[288], t140[144],
        t139[144], t138[288], t137[648], t136[216], t135[216], t134[432],
        t133[432], t132[144], t131[216], t130[360], t129[432], t128[216],
        t127[144], t126[360], t125[648], t124[216], t123[216], t122[432],
        t118[4], t114[4], t110[4], t106[4], t102[4], t98[4], t94[4],
        t90[4], t86[4], numerD5[10000], numerD4[10000], numerD3[10000],
        numerD2[10000], numerD[10000], numerB5[5184], numerB4[1728],
        numerB3[3456], numerB2[1728], numerB[1728], insphereD[10000],
        insphereB[10000], denomD3[1656], denomD2[1080], denomD[576],
        denomB3[144], denomB2[72], denomB[72];
   int yMinorD3_len, yMinorD2_len, yMinorD_len, yMinorB3_len,
       yMinorB2_len, yMinorB_len, xMinorD3_len, xMinorD2_len, xMinorD_len,
       xMinorB3_len, xMinorB2_len, xMinorB_len, wwD5_len, wwD4_len,
       wwD3_len, wwD2_len, wwD_len, wwB5_len, wwB4_len, wvD10_len,
       wvD9_len, wvD8_len, wvD7_len, wvD6_len, wvD5_len, wvD4_len,
       wvD3_len, wvD2_len, wvD_len, wvB10_len, wvB9_len, wvB5_len,
       wvB4_len, vD15_len, vD14_len, vD13_len, vD12_len, vD11_len,
       vD10_len, vD9_len, vD8_len, vD7_len, vD6_len, vD5_len, vD4_len,
       vD3_len, vD2_len, vD_len, vB15_len, vB14_len, vB10_len, vB9_len,
       vB5_len, vB4_len, t162_len, t161_len, t160_len, t159_len, t158_len,
       t157_len, t156_len, t155_len, t154_len, t153_len, t152_len,
       t151_len, t150_len, t149_len, t148_len, t147_len, t146_len,
       t145_len, t144_len, t143_len, t142_len, t141_len, t140_len,
       t139_len, t138_len, t137_len, t136_len, t135_len, t134_len,
       t133_len, t132_len, t131_len, t130_len, t129_len, t128_len,
       t127_len, t126_len, t125_len, t124_len, t123_len, t122_len,
       t118_len, t114_len, t110_len, t106_len, t102_len, t98_len, t94_len,
       t90_len, t86_len, numerD5_len, numerD4_len, numerD3_len,
       numerD2_len, numerD_len, numerB5_len, numerB4_len, numerB3_len,
       numerB2_len, numerB_len, insphereD_len, insphereB_len, denomD3_len,
       denomD2_len, denomD_len, denomB3_len, denomB2_len, denomB_len;

   brx = (REAL) (bx - ax);
   bry = (REAL) (by - ay);
   brz = (REAL) (bz - az);
   crx = (REAL) (cx - ax);
   cry = (REAL) (cy - ay);
   crz = (REAL) (cz - az);
   prx = (REAL) (px - ax);
   pry = (REAL) (py - ay);
   prz = (REAL) (pz - az);
   v = (REAL) (brx * brx);
   v2 = (REAL) (bry * bry);
   v3 = (REAL) (brz * brz);
   v4 = (REAL) (v2 + v3);
   v5 = (REAL) (v + v4);
   v6 = (REAL) (crx * crx);
   v7 = (REAL) (cry * cry);
   v8 = (REAL) (crz * crz);
   v9 = (REAL) (v7 + v8);
   v10 = (REAL) (v6 + v9);
   ww = (REAL) (prx * prx);
   ww2 = (REAL) (pry * pry);
   ww3 = (REAL) (prz * prz);
   ww4 = (REAL) (ww2 + ww3);
   ww5 = (REAL) (ww + ww4);
   v11 = (REAL) (brx * crx);
   v12 = (REAL) (bry * cry);
   v13 = (REAL) (brz * crz);
   v14 = (REAL) (v12 + v13);
   t = Absolute(v12);
   t2 = Absolute(v13);
   vP14 = (REAL) (t + t2);
   v15 = (REAL) (v11 + v14);
   t3 = Absolute(v11);
   vP15 = (REAL) (t3 + vP14);
   wv = (REAL) (brx * prx);
   wv2 = (REAL) (bry * pry);
   wv3 = (REAL) (brz * prz);
   wv4 = (REAL) (wv2 + wv3);
   t4 = Absolute(wv2);
   t5 = Absolute(wv3);
   wvP4 = (REAL) (t4 + t5);
   wv5 = (REAL) (wv + wv4);
   t6 = Absolute(wv);
   wvP5 = (REAL) (t6 + wvP4);
   wv6 = (REAL) (prx * crx);
   wv7 = (REAL) (pry * cry);
   wv8 = (REAL) (prz * crz);
   wv9 = (REAL) (wv7 + wv8);
   t7 = Absolute(wv7);
   t8 = Absolute(wv8);
   wvP9 = (REAL) (t7 + t8);
   wv10 = (REAL) (wv6 + wv9);
   t9 = Absolute(wv6);
   wvP10 = (REAL) (t9 + wvP9);
   xMinor = (REAL) (v15 * wv10);
   xMinorP = (REAL) (vP15 * wvP10);
   xMinor2 = (REAL) (wv5 * v10);
   xMinorP2 = (REAL) (wvP5 * v10);
   xMinor3 = (REAL) (xMinor - xMinor2);
   xMinorP3 = (REAL) (xMinorP + xMinorP2);
   yMinor = (REAL) (v5 * wv10);
   yMinorP = (REAL) (v5 * wvP10);
   yMinor2 = (REAL) (wv5 * v15);
   yMinorP2 = (REAL) (wvP5 * vP15);
   yMinor3 = (REAL) (yMinor - yMinor2);
   yMinorP3 = (REAL) (yMinorP + yMinorP2);
   denom = (REAL) (v5 * v10);
   denom2 = (REAL) (v15 * v15);
   denomP2 = (REAL) (vP15 * vP15);
   denom3 = (REAL) (denom - denom2);
   denomP3 = (REAL) (denom + denomP2);
   numer = (REAL) (v5 * xMinor3);
   numerP = (REAL) (v5 * xMinorP3);
   numer2 = (REAL) (v10 * yMinor3);
   numerP2 = (REAL) (v10 * yMinorP3);
   numer3 = (REAL) (numer - numer2);
   numerP3 = (REAL) (numerP + numerP2);
   numer4 = (REAL) (ww5 * denom3);
   numerP4 = (REAL) (ww5 * denomP3);
   numer5 = (REAL) (numer3 + numer4);
   numerP5 = (REAL) (numerP3 + numerP4);
   insphere = (REAL) (numer5 * denom3);
   insphereP = (REAL) (numerP5 * denomP3);
   yAE = (REAL) (errA2to3 * insphereP);

   if (insphere > yAE)
      return 1;
   else {
      if (-insphere > yAE)
         return -1;
      else {
         Square(brx, vB[1], vB[0]);
         Square(bry, vB2[1], vB2[0]);
         Square(brz, vB3[1], vB3[0]);
         Two_Two_Sum(vB2[1], vB2[0], vB3[1], vB3[0], vB4[3], vB4[2],
                     vB4[1], vB4[0]);
         vB4_len = 4;
         vB5_len =
            fast_expansion_sum_zeroelim(2, vB, vB4_len, vB4, vB5);
         Square(crx, vB6[1], vB6[0]);
         Square(cry, vB7[1], vB7[0]);
         Square(crz, vB8[1], vB8[0]);
         Two_Two_Sum(vB7[1], vB7[0], vB8[1], vB8[0], vB9[3], vB9[2],
                     vB9[1], vB9[0]);
         vB9_len = 4;
         vB10_len =
            fast_expansion_sum_zeroelim(2, vB6, vB9_len, vB9, vB10);
         Square(prx, wwB[1], wwB[0]);
         Square(pry, wwB2[1], wwB2[0]);
         Square(prz, wwB3[1], wwB3[0]);
         Two_Two_Sum(wwB2[1], wwB2[0], wwB3[1], wwB3[0], wwB4[3],
                     wwB4[2], wwB4[1], wwB4[0]);
         wwB4_len = 4;
         wwB5_len =
            fast_expansion_sum_zeroelim(2, wwB, wwB4_len, wwB4, wwB5);
         Two_Product(brx, crx, vB11[1], vB11[0]);
         Two_Product(bry, cry, vB12[1], vB12[0]);
         Two_Product(brz, crz, vB13[1], vB13[0]);
         Two_Two_Sum(vB12[1], vB12[0], vB13[1], vB13[0], vB14[3],
                     vB14[2], vB14[1], vB14[0]);
         vB14_len = 4;
         vB15_len =
            fast_expansion_sum_zeroelim(2, vB11, vB14_len, vB14, vB15);
         Two_Product(brx, prx, wvB[1], wvB[0]);
         Two_Product(bry, pry, wvB2[1], wvB2[0]);
         Two_Product(brz, prz, wvB3[1], wvB3[0]);
         Two_Two_Sum(wvB2[1], wvB2[0], wvB3[1], wvB3[0], wvB4[3],
                     wvB4[2], wvB4[1], wvB4[0]);
         wvB4_len = 4;
         wvB5_len =
            fast_expansion_sum_zeroelim(2, wvB, wvB4_len, wvB4, wvB5);
         Two_Product(prx, crx, wvB6[1], wvB6[0]);
         Two_Product(pry, cry, wvB7[1], wvB7[0]);
         Two_Product(prz, crz, wvB8[1], wvB8[0]);
         Two_Two_Sum(wvB7[1], wvB7[0], wvB8[1], wvB8[0], wvB9[3],
                     wvB9[2], wvB9[1], wvB9[0]);
         wvB9_len = 4;
         wvB10_len =
            fast_expansion_sum_zeroelim(2, wvB6, wvB9_len, wvB9,
                                        wvB10);
         xMinorB_len =
            mult_expansion_zeroelim(vB15_len, vB15, wvB10_len, wvB10,
                                    xMinorB);
         xMinorB2_len =
            mult_expansion_zeroelim(wvB5_len, wvB5, vB10_len, vB10,
                                    xMinorB2);
         xMinorB3_len =
            fast_expansion_diff_zeroelim(xMinorB_len, xMinorB,
                                         xMinorB2_len, xMinorB2,
                                         xMinorB3);
         yMinorB_len =
            mult_expansion_zeroelim(vB5_len, vB5, wvB10_len, wvB10,
                                    yMinorB);
         yMinorB2_len =
            mult_expansion_zeroelim(wvB5_len, wvB5, vB15_len, vB15,
                                    yMinorB2);
         yMinorB3_len =
            fast_expansion_diff_zeroelim(yMinorB_len, yMinorB,
                                         yMinorB2_len, yMinorB2,
                                         yMinorB3);
         denomB_len =
            mult_expansion_zeroelim(vB5_len, vB5, vB10_len, vB10,
                                    denomB);
         denomB2_len =
            mult_expansion_zeroelim(vB15_len, vB15, vB15_len, vB15,
                                    denomB2);
         denomB3_len =
            fast_expansion_diff_zeroelim(denomB_len, denomB,
                                         denomB2_len, denomB2,
                                         denomB3);
         numerB_len =
            mult_expansion_zeroelim(vB5_len, vB5, xMinorB3_len,
                                    xMinorB3, numerB);
         numerB2_len =
            mult_expansion_zeroelim(vB10_len, vB10, yMinorB3_len,
                                    yMinorB3, numerB2);
         numerB3_len =
            fast_expansion_diff_zeroelim(numerB_len, numerB,
                                         numerB2_len, numerB2,
                                         numerB3);
         numerB4_len =
            mult_expansion_zeroelim(wwB5_len, wwB5, denomB3_len,
                                    denomB3, numerB4);
         numerB5_len =
            fast_expansion_sum_zeroelim(numerB4_len, numerB4,
                                        numerB3_len, numerB3, numerB5);
         insphereB_len =
            mult_expansion_zeroelim(numerB5_len, numerB5, denomB3_len,
                                    denomB3, insphereB);
         yBX = estimate(insphereB_len, insphereB);
         yBE = (REAL) (errB2to3 * insphereP);
         if (yBX > yBE)
            return 1;
         else {
            if (-yBX > yBE)
               return -1;
            else {
               Two_Diff_Tail(bx, ax, brx, brxC);
               Two_Diff_Tail(by, ay, bry, bryC);
               Two_Diff_Tail(bz, az, brz, brzC);
               Two_Diff_Tail(cx, ax, crx, crxC);
               Two_Diff_Tail(cy, ay, cry, cryC);
               Two_Diff_Tail(cz, az, crz, crzC);
               Two_Diff_Tail(px, ax, prx, prxC);
               Two_Diff_Tail(py, ay, pry, pryC);
               Two_Diff_Tail(pz, az, prz, przC);
               t10 = (REAL) (brx * brxC);
               vC = (REAL) (2.0 * t10);
               t11 = (REAL) (bry * bryC);
               vC2 = (REAL) (2.0 * t11);
               t12 = (REAL) (brz * brzC);
               vC3 = (REAL) (2.0 * t12);
               vC4 = (REAL) (vC2 + vC3);
               vC5 = (REAL) (vC + vC4);
               t13 = (REAL) (crx * crxC);
               vC6 = (REAL) (2.0 * t13);
               t14 = (REAL) (cry * cryC);
               vC7 = (REAL) (2.0 * t14);
               t15 = (REAL) (crz * crzC);
               vC8 = (REAL) (2.0 * t15);
               vC9 = (REAL) (vC7 + vC8);
               vC10 = (REAL) (vC6 + vC9);
               t16 = (REAL) (prx * prxC);
               wwC = (REAL) (2.0 * t16);
               t17 = (REAL) (pry * pryC);
               wwC2 = (REAL) (2.0 * t17);
               t18 = (REAL) (prz * przC);
               wwC3 = (REAL) (2.0 * t18);
               wwC4 = (REAL) (wwC2 + wwC3);
               wwC5 = (REAL) (wwC + wwC4);
               t19 = (REAL) (brx * crxC);
               t20 = (REAL) (brxC * crx);
               vC11 = (REAL) (t19 + t20);
               t21 = (REAL) (bry * cryC);
               t22 = (REAL) (bryC * cry);
               vC12 = (REAL) (t21 + t22);
               t23 = (REAL) (brz * crzC);
               t24 = (REAL) (brzC * crz);
               vC13 = (REAL) (t23 + t24);
               vC14 = (REAL) (vC12 + vC13);
               vC15 = (REAL) (vC11 + vC14);
               t25 = (REAL) (brx * prxC);
               t26 = (REAL) (brxC * prx);
               wvC = (REAL) (t25 + t26);
               t27 = (REAL) (bry * pryC);
               t28 = (REAL) (bryC * pry);
               wvC2 = (REAL) (t27 + t28);
               t29 = (REAL) (brz * przC);
               t30 = (REAL) (brzC * prz);
               wvC3 = (REAL) (t29 + t30);
               wvC4 = (REAL) (wvC2 + wvC3);
               wvC5 = (REAL) (wvC + wvC4);
               t31 = (REAL) (prx * crxC);
               t32 = (REAL) (prxC * crx);
               wvC6 = (REAL) (t31 + t32);
               t33 = (REAL) (pry * cryC);
               t34 = (REAL) (pryC * cry);
               wvC7 = (REAL) (t33 + t34);
               t35 = (REAL) (prz * crzC);
               t36 = (REAL) (przC * crz);
               wvC8 = (REAL) (t35 + t36);
               wvC9 = (REAL) (wvC7 + wvC8);
               wvC10 = (REAL) (wvC6 + wvC9);
               t37 = (REAL) (v15 * wvC10);
               t38 = (REAL) (vC15 * wv10);
               xMinorC = (REAL) (t37 + t38);
               t39 = (REAL) (wv5 * vC10);
               t40 = (REAL) (wvC5 * v10);
               xMinorC2 = (REAL) (t39 + t40);
               xMinorC3 = (REAL) (xMinorC - xMinorC2);
               t41 = (REAL) (v5 * wvC10);
               t42 = (REAL) (vC5 * wv10);
               yMinorC = (REAL) (t41 + t42);
               t43 = (REAL) (wv5 * vC15);
               t44 = (REAL) (wvC5 * v15);
               yMinorC2 = (REAL) (t43 + t44);
               yMinorC3 = (REAL) (yMinorC - yMinorC2);
               t45 = (REAL) (v5 * vC10);
               t46 = (REAL) (vC5 * v10);
               denomC = (REAL) (t45 + t46);
               t47 = (REAL) (v15 * vC15);
               t48 = (REAL) (vC15 * v15);
               denomC2 = (REAL) (t47 + t48);
               denomC3 = (REAL) (denomC - denomC2);
               t49 = (REAL) (v5 * xMinorC3);
               t50 = (REAL) (vC5 * xMinor3);
               numerC = (REAL) (t49 + t50);
               t51 = (REAL) (v10 * yMinorC3);
               t52 = (REAL) (vC10 * yMinor3);
               numerC2 = (REAL) (t51 + t52);
               numerC3 = (REAL) (numerC - numerC2);
               t53 = (REAL) (ww5 * denomC3);
               t54 = (REAL) (wwC5 * denom3);
               numerC4 = (REAL) (t53 + t54);
               numerC5 = (REAL) (numerC3 + numerC4);
               t55 = (REAL) (numer5 * denomC3);
               t56 = (REAL) (numerC5 * denom3);
               insphereC = (REAL) (t55 + t56);
               yCX = (REAL) (yBX + insphereC);
               t57 = (REAL) (errCX2to3 * yBX);
               t58 = (REAL) (errC2to3 * insphereP);
               yCE = (REAL) (t57 + t58);
               if (yCX > yCE)
                  return 1;
               else {
                  if (-yCX > yCE)
                     return -1;
                  else {
                     Two_Product(brx, brxC, t60[1], t60[0]);
                     t59[1] = 2.0 * t60[1];
                     t59[0] = 2.0 * t60[0];
                     Square(brxC, t61[1], t61[0]);
                     Two_Two_Sum(t59[1], t59[0], t61[1], t61[0],
                                 vD[3], vD[2], vD[1], vD[0]);
                     vD_len = 4;
                     Two_Product(bry, bryC, t63[1], t63[0]);
                     t62[1] = 2.0 * t63[1];
                     t62[0] = 2.0 * t63[0];
                     Square(bryC, t64[1], t64[0]);
                     Two_Two_Sum(t62[1], t62[0], t64[1], t64[0],
                                 vD2[3], vD2[2], vD2[1], vD2[0]);
                     vD2_len = 4;
                     Two_Product(brz, brzC, t66[1], t66[0]);
                     t65[1] = 2.0 * t66[1];
                     t65[0] = 2.0 * t66[0];
                     Square(brzC, t67[1], t67[0]);
                     Two_Two_Sum(t65[1], t65[0], t67[1], t67[0],
                                 vD3[3], vD3[2], vD3[1], vD3[0]);
                     vD3_len = 4;
                     vD4_len =
                        fast_expansion_sum_zeroelim(vD3_len, vD3,
                                                    vD2_len, vD2,
                                                    vD4);
                     vD5_len =
                        fast_expansion_sum_zeroelim(vD4_len, vD4,
                                                    vD_len, vD,
                                                    vD5);
                     Two_Product(crx, crxC, t69[1], t69[0]);
                     t68[1] = 2.0 * t69[1];
                     t68[0] = 2.0 * t69[0];
                     Square(crxC, t70[1], t70[0]);
                     Two_Two_Sum(t68[1], t68[0], t70[1], t70[0],
                                 vD6[3], vD6[2], vD6[1], vD6[0]);
                     vD6_len = 4;
                     Two_Product(cry, cryC, t72[1], t72[0]);
                     t71[1] = 2.0 * t72[1];
                     t71[0] = 2.0 * t72[0];
                     Square(cryC, t73[1], t73[0]);
                     Two_Two_Sum(t71[1], t71[0], t73[1], t73[0],
                                 vD7[3], vD7[2], vD7[1], vD7[0]);
                     vD7_len = 4;
                     Two_Product(crz, crzC, t75[1], t75[0]);
                     t74[1] = 2.0 * t75[1];
                     t74[0] = 2.0 * t75[0];
                     Square(crzC, t76[1], t76[0]);
                     Two_Two_Sum(t74[1], t74[0], t76[1], t76[0],
                                 vD8[3], vD8[2], vD8[1], vD8[0]);
                     vD8_len = 4;
                     vD9_len =
                        fast_expansion_sum_zeroelim(vD8_len, vD8,
                                                    vD7_len, vD7,
                                                    vD9);
                     vD10_len =
                        fast_expansion_sum_zeroelim(vD9_len, vD9,
                                                    vD6_len, vD6,
                                                    vD10);
                     Two_Product(prx, prxC, t78[1], t78[0]);
                     t77[1] = 2.0 * t78[1];
                     t77[0] = 2.0 * t78[0];
                     Square(prxC, t79[1], t79[0]);
                     Two_Two_Sum(t77[1], t77[0], t79[1], t79[0],
                                 wwD[3], wwD[2], wwD[1], wwD[0]);
                     wwD_len = 4;
                     Two_Product(pry, pryC, t81[1], t81[0]);
                     t80[1] = 2.0 * t81[1];
                     t80[0] = 2.0 * t81[0];
                     Square(pryC, t82[1], t82[0]);
                     Two_Two_Sum(t80[1], t80[0], t82[1], t82[0],
                                 wwD2[3], wwD2[2], wwD2[1],
                                 wwD2[0]);
                     wwD2_len = 4;
                     Two_Product(prz, przC, t84[1], t84[0]);
                     t83[1] = 2.0 * t84[1];
                     t83[0] = 2.0 * t84[0];
                     Square(przC, t85[1], t85[0]);
                     Two_Two_Sum(t83[1], t83[0], t85[1], t85[0],
                                 wwD3[3], wwD3[2], wwD3[1],
                                 wwD3[0]);
                     wwD3_len = 4;
                     wwD4_len =
                        fast_expansion_sum_zeroelim(wwD3_len, wwD3,
                                                    wwD2_len, wwD2,
                                                    wwD4);
                     wwD5_len =
                        fast_expansion_sum_zeroelim(wwD4_len, wwD4,
                                                    wwD_len, wwD,
                                                    wwD5);
                     Two_Product(brx, crxC, t87[1], t87[0]);
                     Two_Product(brxC, crx, t88[1], t88[0]);
                     Two_Two_Sum(t87[1], t87[0], t88[1], t88[0],
                                 t86[3], t86[2], t86[1], t86[0]);
                     t86_len = 4;
                     Two_Product(brxC, crxC, t89[1], t89[0]);
                     vD11_len =
                        fast_expansion_sum_zeroelim(2, t89,
                                                    t86_len, t86,
                                                    vD11);
                     Two_Product(bry, cryC, t91[1], t91[0]);
                     Two_Product(bryC, cry, t92[1], t92[0]);
                     Two_Two_Sum(t91[1], t91[0], t92[1], t92[0],
                                 t90[3], t90[2], t90[1], t90[0]);
                     t90_len = 4;
                     Two_Product(bryC, cryC, t93[1], t93[0]);
                     vD12_len =
                        fast_expansion_sum_zeroelim(2, t93,
                                                    t90_len, t90,
                                                    vD12);
                     Two_Product(brz, crzC, t95[1], t95[0]);
                     Two_Product(brzC, crz, t96[1], t96[0]);
                     Two_Two_Sum(t95[1], t95[0], t96[1], t96[0],
                                 t94[3], t94[2], t94[1], t94[0]);
                     t94_len = 4;
                     Two_Product(brzC, crzC, t97[1], t97[0]);
                     vD13_len =
                        fast_expansion_sum_zeroelim(2, t97,
                                                    t94_len, t94,
                                                    vD13);
                     vD14_len =
                        fast_expansion_sum_zeroelim(vD13_len, vD13,
                                                    vD12_len, vD12,
                                                    vD14);
                     vD15_len =
                        fast_expansion_sum_zeroelim(vD14_len, vD14,
                                                    vD11_len, vD11,
                                                    vD15);
                     Two_Product(brx, prxC, t99[1], t99[0]);
                     Two_Product(brxC, prx, t100[1], t100[0]);
                     Two_Two_Sum(t99[1], t99[0], t100[1], t100[0],
                                 t98[3], t98[2], t98[1], t98[0]);
                     t98_len = 4;
                     Two_Product(brxC, prxC, t101[1], t101[0]);
                     wvD_len =
                        fast_expansion_sum_zeroelim(2, t101,
                                                    t98_len, t98,
                                                    wvD);
                     Two_Product(bry, pryC, t103[1], t103[0]);
                     Two_Product(bryC, pry, t104[1], t104[0]);
                     Two_Two_Sum(t103[1], t103[0], t104[1], t104[0],
                                 t102[3], t102[2], t102[1],
                                 t102[0]);
                     t102_len = 4;
                     Two_Product(bryC, pryC, t105[1], t105[0]);
                     wvD2_len =
                        fast_expansion_sum_zeroelim(2, t105,
                                                    t102_len, t102,
                                                    wvD2);
                     Two_Product(brz, przC, t107[1], t107[0]);
                     Two_Product(brzC, prz, t108[1], t108[0]);
                     Two_Two_Sum(t107[1], t107[0], t108[1], t108[0],
                                 t106[3], t106[2], t106[1],
                                 t106[0]);
                     t106_len = 4;
                     Two_Product(brzC, przC, t109[1], t109[0]);
                     wvD3_len =
                        fast_expansion_sum_zeroelim(2, t109,
                                                    t106_len, t106,
                                                    wvD3);
                     wvD4_len =
                        fast_expansion_sum_zeroelim(wvD3_len, wvD3,
                                                    wvD2_len, wvD2,
                                                    wvD4);
                     wvD5_len =
                        fast_expansion_sum_zeroelim(wvD4_len, wvD4,
                                                    wvD_len, wvD,
                                                    wvD5);
                     Two_Product(prx, crxC, t111[1], t111[0]);
                     Two_Product(prxC, crx, t112[1], t112[0]);
                     Two_Two_Sum(t111[1], t111[0], t112[1], t112[0],
                                 t110[3], t110[2], t110[1],
                                 t110[0]);
                     t110_len = 4;
                     Two_Product(prxC, crxC, t113[1], t113[0]);
                     wvD6_len =
                        fast_expansion_sum_zeroelim(2, t113,
                                                    t110_len, t110,
                                                    wvD6);
                     Two_Product(pry, cryC, t115[1], t115[0]);
                     Two_Product(pryC, cry, t116[1], t116[0]);
                     Two_Two_Sum(t115[1], t115[0], t116[1], t116[0],
                                 t114[3], t114[2], t114[1],
                                 t114[0]);
                     t114_len = 4;
                     Two_Product(pryC, cryC, t117[1], t117[0]);
                     wvD7_len =
                        fast_expansion_sum_zeroelim(2, t117,
                                                    t114_len, t114,
                                                    wvD7);
                     Two_Product(prz, crzC, t119[1], t119[0]);
                     Two_Product(przC, crz, t120[1], t120[0]);
                     Two_Two_Sum(t119[1], t119[0], t120[1], t120[0],
                                 t118[3], t118[2], t118[1],
                                 t118[0]);
                     t118_len = 4;
                     Two_Product(przC, crzC, t121[1], t121[0]);
                     wvD8_len =
                        fast_expansion_sum_zeroelim(2, t121,
                                                    t118_len, t118,
                                                    wvD8);
                     wvD9_len =
                        fast_expansion_sum_zeroelim(wvD8_len, wvD8,
                                                    wvD7_len, wvD7,
                                                    wvD9);
                     wvD10_len =
                        fast_expansion_sum_zeroelim(wvD9_len, wvD9,
                                                    wvD6_len, wvD6,
                                                    wvD10);
                     t123_len =
                        mult_expansion_zeroelim(vB15_len, vB15,
                                                wvD10_len, wvD10,
                                                t123);
                     t124_len =
                        mult_expansion_zeroelim(vD15_len, vD15,
                                                wvB10_len, wvB10,
                                                t124);
                     t122_len =
                        fast_expansion_sum_zeroelim(t124_len, t124,
                                                    t123_len, t123,
                                                    t122);
                     t125_len =
                        mult_expansion_zeroelim(vD15_len, vD15,
                                                wvD10_len, wvD10,
                                                t125);
                     xMinorD_len =
                        fast_expansion_sum_zeroelim(t125_len, t125,
                                                    t122_len, t122,
                                                    xMinorD);
                     t127_len =
                        mult_expansion_zeroelim(wvB5_len, wvB5,
                                                vD10_len, vD10,
                                                t127);
                     t128_len =
                        mult_expansion_zeroelim(wvD5_len, wvD5,
                                                vB10_len, vB10,
                                                t128);
                     t126_len =
                        fast_expansion_sum_zeroelim(t128_len, t128,
                                                    t127_len, t127,
                                                    t126);
                     t129_len =
                        mult_expansion_zeroelim(wvD5_len, wvD5,
                                                vD10_len, vD10,
                                                t129);
                     xMinorD2_len =
                        fast_expansion_sum_zeroelim(t129_len, t129,
                                                    t126_len, t126,
                                                    xMinorD2);
                     xMinorD3_len =
                        fast_expansion_diff_zeroelim(xMinorD_len,
                                                     xMinorD,
                                                     xMinorD2_len,
                                                     xMinorD2,
                                                     xMinorD3);
                     t131_len =
                        mult_expansion_zeroelim(vB5_len, vB5,
                                                wvD10_len, wvD10,
                                                t131);
                     t132_len =
                        mult_expansion_zeroelim(vD5_len, vD5,
                                                wvB10_len, wvB10,
                                                t132);
                     t130_len =
                        fast_expansion_sum_zeroelim(t132_len, t132,
                                                    t131_len, t131,
                                                    t130);
                     t133_len =
                        mult_expansion_zeroelim(vD5_len, vD5,
                                                wvD10_len, wvD10,
                                                t133);
                     yMinorD_len =
                        fast_expansion_sum_zeroelim(t133_len, t133,
                                                    t130_len, t130,
                                                    yMinorD);
                     t135_len =
                        mult_expansion_zeroelim(wvB5_len, wvB5,
                                                vD15_len, vD15,
                                                t135);
                     t136_len =
                        mult_expansion_zeroelim(wvD5_len, wvD5,
                                                vB15_len, vB15,
                                                t136);
                     t134_len =
                        fast_expansion_sum_zeroelim(t136_len, t136,
                                                    t135_len, t135,
                                                    t134);
                     t137_len =
                        mult_expansion_zeroelim(wvD5_len, wvD5,
                                                vD15_len, vD15,
                                                t137);
                     yMinorD2_len =
                        fast_expansion_sum_zeroelim(t137_len, t137,
                                                    t134_len, t134,
                                                    yMinorD2);
                     yMinorD3_len =
                        fast_expansion_diff_zeroelim(yMinorD_len,
                                                     yMinorD,
                                                     yMinorD2_len,
                                                     yMinorD2,
                                                     yMinorD3);
                     t139_len =
                        mult_expansion_zeroelim(vB5_len, vB5,
                                                vD10_len, vD10,
                                                t139);
                     t140_len =
                        mult_expansion_zeroelim(vD5_len, vD5,
                                                vB10_len, vB10,
                                                t140);
                     t138_len =
                        fast_expansion_sum_zeroelim(t140_len, t140,
                                                    t139_len, t139,
                                                    t138);
                     t141_len =
                        mult_expansion_zeroelim(vD5_len, vD5,
                                                vD10_len, vD10,
                                                t141);
                     denomD_len =
                        fast_expansion_sum_zeroelim(t141_len, t141,
                                                    t138_len, t138,
                                                    denomD);
                     t143_len =
                        mult_expansion_zeroelim(vB15_len, vB15,
                                                vD15_len, vD15,
                                                t143);
                     t144_len =
                        mult_expansion_zeroelim(vD15_len, vD15,
                                                vB15_len, vB15,
                                                t144);
                     t142_len =
                        fast_expansion_sum_zeroelim(t144_len, t144,
                                                    t143_len, t143,
                                                    t142);
                     t145_len =
                        mult_expansion_zeroelim(vD15_len, vD15,
                                                vD15_len, vD15,
                                                t145);
                     denomD2_len =
                        fast_expansion_sum_zeroelim(t145_len, t145,
                                                    t142_len, t142,
                                                    denomD2);
                     denomD3_len =
                        fast_expansion_diff_zeroelim(denomD_len,
                                                     denomD,
                                                     denomD2_len,
                                                     denomD2,
                                                     denomD3);
                     t147_len =
                        mult_expansion_zeroelim(vB5_len, vB5,
                                                xMinorD3_len,
                                                xMinorD3, t147);
                     t148_len =
                        mult_expansion_zeroelim(vD5_len, vD5,
                                                xMinorB3_len,
                                                xMinorB3, t148);
                     t146_len =
                        fast_expansion_sum_zeroelim(t148_len, t148,
                                                    t147_len, t147,
                                                    t146);
                     t149_len =
                        mult_expansion_zeroelim(vD5_len, vD5,
                                                xMinorD3_len,
                                                xMinorD3, t149);
                     numerD_len =
                        fast_expansion_sum_zeroelim(t149_len, t149,
                                                    t146_len, t146,
                                                    numerD);
                     t151_len =
                        mult_expansion_zeroelim(vB10_len, vB10,
                                                yMinorD3_len,
                                                yMinorD3, t151);
                     t152_len =
                        mult_expansion_zeroelim(vD10_len, vD10,
                                                yMinorB3_len,
                                                yMinorB3, t152);
                     t150_len =
                        fast_expansion_sum_zeroelim(t152_len, t152,
                                                    t151_len, t151,
                                                    t150);
                     t153_len =
                        mult_expansion_zeroelim(vD10_len, vD10,
                                                yMinorD3_len,
                                                yMinorD3, t153);
                     numerD2_len =
                        fast_expansion_sum_zeroelim(t153_len, t153,
                                                    t150_len, t150,
                                                    numerD2);
                     numerD3_len =
                        fast_expansion_diff_zeroelim(numerD_len,
                                                     numerD,
                                                     numerD2_len,
                                                     numerD2,
                                                     numerD3);
                     t155_len =
                        mult_expansion_zeroelim(wwB5_len, wwB5,
                                                denomD3_len,
                                                denomD3, t155);
                     t156_len =
                        mult_expansion_zeroelim(wwD5_len, wwD5,
                                                denomB3_len,
                                                denomB3, t156);
                     t154_len =
                        fast_expansion_sum_zeroelim(t156_len, t156,
                                                    t155_len, t155,
                                                    t154);
                     t157_len =
                        mult_expansion_zeroelim(wwD5_len, wwD5,
                                                denomD3_len,
                                                denomD3, t157);
                     numerD4_len =
                        fast_expansion_sum_zeroelim(t157_len, t157,
                                                    t154_len, t154,
                                                    numerD4);
                     numerD5_len =
                        fast_expansion_sum_zeroelim(numerD4_len,
                                                    numerD4,
                                                    numerD3_len,
                                                    numerD3,
                                                    numerD5);
                     t159_len =
                        mult_expansion_zeroelim(numerB5_len,
                                                numerB5,
                                                denomD3_len,
                                                denomD3, t159);
                     t160_len =
                        mult_expansion_zeroelim(numerD5_len,
                                                numerD5,
                                                denomB3_len,
                                                denomB3, t160);
                     t158_len =
                        fast_expansion_sum_zeroelim(t160_len, t160,
                                                    t159_len, t159,
                                                    t158);
                     t161_len =
                        mult_expansion_zeroelim(numerD5_len,
                                                numerD5,
                                                denomD3_len,
                                                denomD3, t161);
                     insphereD_len =
                        fast_expansion_sum_zeroelim(t161_len, t161,
                                                    t158_len, t158,
                                                    insphereD);
                     t162_len =
                        fast_expansion_sum_zeroelim(insphereD_len,
                                                    insphereD,
                                                    insphereB_len,
                                                    insphereB,
                                                    t162);
                     if (t162[t162_len - 1] > 0.0)
                        return 1;
                     else if (t162[t162_len - 1] < 0.0)
                        return -1;
                     else
                        return 0;
                  }
               };
            }
         };
      }
   };

}


void insphere3to3_init()
{

   double E = epsilon;
#ifndef CYGWIN
   int round_up = 6770, round_near = 4722;
   _FPU_SETCW(round_up);
#endif
   errA3to3 = (24.0 + 349.0 * E) * E;
   errB3to3 = (8.0 + 111.0 * E) * E;
   errC3to3 = (188.0 + 4565.0 * E) * E * E;
   errCX3to3 = (2.0 + 7.0 * E) * E;
#ifndef CYGWIN
   _FPU_SETCW(round_near);
#endif
}

int insphere3to3(REAL ax, REAL ay, REAL az, REAL bx, REAL by, REAL bz,
                 REAL cx, REAL cy, REAL cz, REAL dx, REAL dy, REAL dz,
                 REAL px, REAL py, REAL pz)
{
   INEXACT REAL c, abig, ahi, alo, bhi, blo, err1, err2, err3, _i, bvirt,
                avirt, bround, around, _j, _0;
   REAL zMinorP5, zMinorP4, zMinorP3, zMinorP2, zMinorP, zMinorC5,
        zMinorC4, zMinorC3, zMinorC2, zMinorC, zMinor5, zMinor4, zMinor3,
        zMinor2, zMinor, yMinorP5, yMinorP4, yMinorP3, yMinorP2, yMinorP,
        yMinorC5, yMinorC4, yMinorC3, yMinorC2, yMinorC, yMinor5, yMinor4,
        yMinor3, yMinor2, yMinor, yCX, yCE, yBX, yBE, yAE, xMinorP5,
        xMinorP4, xMinorP3, xMinorP2, xMinorP, xMinorC5, xMinorC4,
        xMinorC3, xMinorC2, xMinorC, xMinor5, xMinor4, xMinor3, xMinor2,
        xMinor, t108, t107, t106, t105, t104, t103, t102, t101, t100, t99,
        t98, t97, t96, t95, t94, t93, t92, t91, t90, t89, t88, t87, t86,
        t85, t84, t83, t82, t81, t80, t79, t78, t77, t76, t75, t74, t73,
        t72, t71, t70, t69, t68, t67, t66, t65, t64, t63, t62, t61, t60,
        t59, t58, t57, t56, t55, t54, t53, t52, t51, t50, t49, t48, t47,
        t46, t45, t44, t43, t42, t41, t40, t39, t38, t37, t36, t35, t34,
        t33, t32, t31, t30, t29, t28, t27, t26, t25, t24, t23, t22, t21,
        t20, t19, t18, t17, t16, t15, t14, t13, t12, t11, t10, t9, t8, t7,
        t6, t5, t4, t3, t2, t, przC, prz, pryC, pry, prxC, prx, prSqrdC5,
        prSqrdC4, prSqrdC3, prSqrdC2, prSqrdC, prSqrd5, prSqrd4, prSqrd3,
        prSqrd2, prSqrd, numerP7, numerP6, numerP5, numerP4, numerP3,
        numerP2, numerP, numerC7, numerC6, numerC5, numerC4, numerC3,
        numerC2, numerC, numer7, numer6, numer5, numer4, numer3, numer2,
        numer, insphereP, insphereC, insphere, dzminP3, dzminC3, dzminC2,
        dzminC, dzmin3, dzmin2, dzmin, dyminP3, dyminC3, dyminC2, dyminC,
        dymin3, dymin2, dymin, dxminP3, dxminC3, dxminC2, dxminC, dxmin3,
        dxmin2, dxmin, drzC, drz, dryC, dry, drxC, drx, drSqrdC5, drSqrdC4,
        drSqrdC3, drSqrdC2, drSqrdC, drSqrd5, drSqrd4, drSqrd3, drSqrd2,
        drSqrd, denomP5, denomP4, denomP3, denomP2, denomP, denomC5,
        denomC4, denomC3, denomC2, denomC, denom5, denom4, denom3, denom2,
        denom, czminP3, czminC3, czminC2, czminC, czmin3, czmin2, czmin,
        cyminP3, cyminC3, cyminC2, cyminC, cymin3, cymin2, cymin, cxminP3,
        cxminC3, cxminC2, cxminC, cxmin3, cxmin2, cxmin, crzC, crz, cryC,
        cry, crxC, crx, crSqrdC5, crSqrdC4, crSqrdC3, crSqrdC2, crSqrdC,
        crSqrd5, crSqrd4, crSqrd3, crSqrd2, crSqrd, bzminP3, bzminC3,
        bzminC2, bzminC, bzmin3, bzmin2, bzmin, byminP3, byminC3, byminC2,
        byminC, bymin3, bymin2, bymin, bxminP3, bxminC3, bxminC2, bxminC,
        bxmin3, bxmin2, bxmin, brzC, brz, bryC, bry, brxC, brx, brSqrdC5,
        brSqrdC4, brSqrdC3, brSqrdC2, brSqrdC, brSqrd5, brSqrd4, brSqrd3,
        brSqrd2, brSqrd;
   REAL t216[2], t215[2], t214[2], t212[2], t211[2], t210[2], t208[2],
        t207[2], t206[2], t204[2], t203[2], t202[2], t200[2], t199[2],
        t198[2], t196[2], t195[2], t194[2], t192[2], t191[2], t190[2],
        t188[2], t187[2], t186[2], t184[2], t183[2], t182[2], t180[2],
        t179[2], t178[2], t176[2], t175[2], t174[2], t172[2], t171[2],
        t170[2], t168[2], t167[2], t166[2], t164[2], t163[2], t162[2],
        t160[2], t159[2], t158[2], t156[2], t155[2], t154[2], t152[2],
        t151[2], t150[2], t148[2], t147[2], t146[2], t144[2], t143[2],
        t142[2], t141[2], t140[2], t139[2], t138[2], t137[2], t136[2],
        t135[2], t134[2], t133[2], t132[2], t131[2], t130[2], t129[2],
        t128[2], t127[2], t126[2], t125[2], t124[2], t123[2], t122[2],
        t121[2], t120[2], t119[2], t118[2], t117[2], t116[2], t115[2],
        t114[2], t113[2], t112[2], t111[2], t110[2], t109[2], prSqrdB3[2],
        prSqrdB2[2], prSqrdB[2], dzminB2[2], dzminB[2], dyminB2[2],
        dyminB[2], dxminB2[2], dxminB[2], drSqrdB3[2], drSqrdB2[2],
        drSqrdB[2], czminB2[2], czminB[2], cyminB2[2], cyminB[2],
        cxminB2[2], cxminB[2], crSqrdB3[2], crSqrdB2[2], crSqrdB[2],
        bzminB2[2], bzminB[2], byminB2[2], byminB[2], bxminB2[2],
        bxminB[2], brSqrdB3[2], brSqrdB2[2], brSqrdB[2];
   REAL zMinorD5[1584], zMinorD4[528], zMinorD3[1056], zMinorD2[528],
        zMinorD[528], zMinorB5[144], zMinorB4[48], zMinorB3[96],
        zMinorB2[48], zMinorB[48], yMinorD5[1584], yMinorD4[528],
        yMinorD3[1056], yMinorD2[528], yMinorD[528], yMinorB5[144],
        yMinorB4[48], yMinorB3[96], yMinorB2[48], yMinorB[48],
        xMinorD5[1584], xMinorD4[528], xMinorD3[1056], xMinorD2[528],
        xMinorD[528], xMinorB5[144], xMinorB4[48], xMinorB3[96],
        xMinorB2[48], xMinorB[48], t285[10000], t284[10000], t283[10000],
        t282[10000], t281[10000], t280[3168], t279[288], t278[3168],
        t277[3456], t276[4032], t275[576], t274[2016], t273[2592],
        t272[3168], t271[288], t270[3168], t269[3456], t268[3168],
        t267[288], t266[3168], t265[3456], t264[24], t263[8], t262[24],
        t261[32], t260[24], t259[8], t258[24], t257[32], t256[24], t255[8],
        t254[24], t253[32], t252[288], t251[144], t250[96], t249[240],
        t248[288], t247[144], t246[96], t245[240], t244[288], t243[144],
        t242[96], t241[240], t240[288], t239[144], t238[96], t237[240],
        t236[288], t235[144], t234[96], t233[240], t232[288], t231[144],
        t230[96], t229[240], t228[288], t227[144], t226[96], t225[240],
        t224[288], t223[144], t222[96], t221[240], t220[288], t219[144],
        t218[96], t217[240], t213[4], t209[4], t205[4], t201[4], t197[4],
        t193[4], t189[4], t185[4], t181[4], t177[4], t173[4], t169[4],
        t165[4], t161[4], t157[4], t153[4], t149[4], t145[4], prSqrdD5[12],
        prSqrdD4[8], prSqrdD3[4], prSqrdD2[4], prSqrdD[4], prSqrdB5[6],
        prSqrdB4[4], numerD7[10000], numerD6[10000], numerD5[6624],
        numerD4[6624], numerD3[10000], numerD2[6624], numerD[6624],
        numerB7[1152], numerB6[576], numerB5[288], numerB4[288],
        numerB3[576], numerB2[288], numerB[288], insphereD[10000],
        insphereB[10000], dzminD3[12], dzminD2[6], dzminD[6], dzminB3[4],
        dyminD3[12], dyminD2[6], dyminD[6], dyminB3[4], dxminD3[12],
        dxminD2[6], dxminD[6], dxminB3[4], drSqrdD5[12], drSqrdD4[8],
        drSqrdD3[4], drSqrdD2[4], drSqrdD[4], drSqrdB5[6], drSqrdB4[4],
        denomD5[168], denomD4[56], denomD3[112], denomD2[56], denomD[56],
        denomB5[24], denomB4[8], denomB3[16], denomB2[8], denomB[8],
        czminD3[12], czminD2[6], czminD[6], czminB3[4], cyminD3[12],
        cyminD2[6], cyminD[6], cyminB3[4], cxminD3[12], cxminD2[6],
        cxminD[6], cxminB3[4], crSqrdD5[12], crSqrdD4[8], crSqrdD3[4],
        crSqrdD2[4], crSqrdD[4], crSqrdB5[6], crSqrdB4[4], bzminD3[12],
        bzminD2[6], bzminD[6], bzminB3[4], byminD3[12], byminD2[6],
        byminD[6], byminB3[4], bxminD3[12], bxminD2[6], bxminD[6],
        bxminB3[4], brSqrdD5[12], brSqrdD4[8], brSqrdD3[4], brSqrdD2[4],
        brSqrdD[4], brSqrdB5[6], brSqrdB4[4];
   int zMinorD5_len, zMinorD4_len, zMinorD3_len, zMinorD2_len,
       zMinorD_len, zMinorB5_len, zMinorB4_len, zMinorB3_len,
       zMinorB2_len, zMinorB_len, yMinorD5_len, yMinorD4_len,
       yMinorD3_len, yMinorD2_len, yMinorD_len, yMinorB5_len,
       yMinorB4_len, yMinorB3_len, yMinorB2_len, yMinorB_len,
       xMinorD5_len, xMinorD4_len, xMinorD3_len, xMinorD2_len,
       xMinorD_len, xMinorB5_len, xMinorB4_len, xMinorB3_len,
       xMinorB2_len, xMinorB_len, t285_len, t284_len, t283_len, t282_len,
       t281_len, t280_len, t279_len, t278_len, t277_len, t276_len,
       t275_len, t274_len, t273_len, t272_len, t271_len, t270_len,
       t269_len, t268_len, t267_len, t266_len, t265_len, t264_len,
       t263_len, t262_len, t261_len, t260_len, t259_len, t258_len,
       t257_len, t256_len, t255_len, t254_len, t253_len, t252_len,
       t251_len, t250_len, t249_len, t248_len, t247_len, t246_len,
       t245_len, t244_len, t243_len, t242_len, t241_len, t240_len,
       t239_len, t238_len, t237_len, t236_len, t235_len, t234_len,
       t233_len, t232_len, t231_len, t230_len, t229_len, t228_len,
       t227_len, t226_len, t225_len, t224_len, t223_len, t222_len,
       t221_len, t220_len, t219_len, t218_len, t217_len, t213_len,
       t209_len, t205_len, t201_len, t197_len, t193_len, t189_len,
       t185_len, t181_len, t177_len, t173_len, t169_len, t165_len,
       t161_len, t157_len, t153_len, t149_len, t145_len, prSqrdD5_len,
       prSqrdD4_len, prSqrdD3_len, prSqrdD2_len, prSqrdD_len,
       prSqrdB5_len, prSqrdB4_len, numerD7_len, numerD6_len, numerD5_len,
       numerD4_len, numerD3_len, numerD2_len, numerD_len, numerB7_len,
       numerB6_len, numerB5_len, numerB4_len, numerB3_len, numerB2_len,
       numerB_len, insphereD_len, insphereB_len, dzminD3_len, dzminD2_len,
       dzminD_len, dzminB3_len, dyminD3_len, dyminD2_len, dyminD_len,
       dyminB3_len, dxminD3_len, dxminD2_len, dxminD_len, dxminB3_len,
       drSqrdD5_len, drSqrdD4_len, drSqrdD3_len, drSqrdD2_len,
       drSqrdD_len, drSqrdB5_len, drSqrdB4_len, denomD5_len, denomD4_len,
       denomD3_len, denomD2_len, denomD_len, denomB5_len, denomB4_len,
       denomB3_len, denomB2_len, denomB_len, czminD3_len, czminD2_len,
       czminD_len, czminB3_len, cyminD3_len, cyminD2_len, cyminD_len,
       cyminB3_len, cxminD3_len, cxminD2_len, cxminD_len, cxminB3_len,
       crSqrdD5_len, crSqrdD4_len, crSqrdD3_len, crSqrdD2_len,
       crSqrdD_len, crSqrdB5_len, crSqrdB4_len, bzminD3_len, bzminD2_len,
       bzminD_len, bzminB3_len, byminD3_len, byminD2_len, byminD_len,
       byminB3_len, bxminD3_len, bxminD2_len, bxminD_len, bxminB3_len,
       brSqrdD5_len, brSqrdD4_len, brSqrdD3_len, brSqrdD2_len,
       brSqrdD_len, brSqrdB5_len, brSqrdB4_len;

   brx = (REAL) (bx - ax);
   bry = (REAL) (by - ay);
   brz = (REAL) (bz - az);
   crx = (REAL) (cx - ax);
   cry = (REAL) (cy - ay);
   crz = (REAL) (cz - az);
   drx = (REAL) (dx - ax);
   dry = (REAL) (dy - ay);
   drz = (REAL) (dz - az);
   prx = (REAL) (px - ax);
   pry = (REAL) (py - ay);
   prz = (REAL) (pz - az);
   brSqrd = (REAL) (brx * brx);
   brSqrd2 = (REAL) (bry * bry);
   brSqrd3 = (REAL) (brz * brz);
   brSqrd4 = (REAL) (brSqrd2 + brSqrd3);
   brSqrd5 = (REAL) (brSqrd + brSqrd4);
   crSqrd = (REAL) (crx * crx);
   crSqrd2 = (REAL) (cry * cry);
   crSqrd3 = (REAL) (crz * crz);
   crSqrd4 = (REAL) (crSqrd2 + crSqrd3);
   crSqrd5 = (REAL) (crSqrd + crSqrd4);
   drSqrd = (REAL) (drx * drx);
   drSqrd2 = (REAL) (dry * dry);
   drSqrd3 = (REAL) (drz * drz);
   drSqrd4 = (REAL) (drSqrd2 + drSqrd3);
   drSqrd5 = (REAL) (drSqrd + drSqrd4);
   prSqrd = (REAL) (prx * prx);
   prSqrd2 = (REAL) (pry * pry);
   prSqrd3 = (REAL) (prz * prz);
   prSqrd4 = (REAL) (prSqrd2 + prSqrd3);
   prSqrd5 = (REAL) (prSqrd + prSqrd4);
   bxmin = (REAL) (cry * drz);
   bxmin2 = (REAL) (dry * crz);
   bxmin3 = (REAL) (bxmin - bxmin2);
   t = Absolute(bxmin);
   t2 = Absolute(bxmin2);
   bxminP3 = (REAL) (t + t2);
   bymin = (REAL) (crx * drz);
   bymin2 = (REAL) (drx * crz);
   bymin3 = (REAL) (bymin - bymin2);
   t3 = Absolute(bymin);
   t4 = Absolute(bymin2);
   byminP3 = (REAL) (t3 + t4);
   bzmin = (REAL) (crx * dry);
   bzmin2 = (REAL) (drx * cry);
   bzmin3 = (REAL) (bzmin - bzmin2);
   t5 = Absolute(bzmin);
   t6 = Absolute(bzmin2);
   bzminP3 = (REAL) (t5 + t6);
   cxmin = (REAL) (bry * drz);
   cxmin2 = (REAL) (dry * brz);
   cxmin3 = (REAL) (cxmin - cxmin2);
   t7 = Absolute(cxmin);
   t8 = Absolute(cxmin2);
   cxminP3 = (REAL) (t7 + t8);
   cymin = (REAL) (brx * drz);
   cymin2 = (REAL) (drx * brz);
   cymin3 = (REAL) (cymin - cymin2);
   t9 = Absolute(cymin);
   t10 = Absolute(cymin2);
   cyminP3 = (REAL) (t9 + t10);
   czmin = (REAL) (brx * dry);
   czmin2 = (REAL) (drx * bry);
   czmin3 = (REAL) (czmin - czmin2);
   t11 = Absolute(czmin);
   t12 = Absolute(czmin2);
   czminP3 = (REAL) (t11 + t12);
   dxmin = (REAL) (bry * crz);
   dxmin2 = (REAL) (cry * brz);
   dxmin3 = (REAL) (dxmin - dxmin2);
   t13 = Absolute(dxmin);
   t14 = Absolute(dxmin2);
   dxminP3 = (REAL) (t13 + t14);
   dymin = (REAL) (brx * crz);
   dymin2 = (REAL) (crx * brz);
   dymin3 = (REAL) (dymin - dymin2);
   t15 = Absolute(dymin);
   t16 = Absolute(dymin2);
   dyminP3 = (REAL) (t15 + t16);
   dzmin = (REAL) (brx * cry);
   dzmin2 = (REAL) (crx * bry);
   dzmin3 = (REAL) (dzmin - dzmin2);
   t17 = Absolute(dzmin);
   t18 = Absolute(dzmin2);
   dzminP3 = (REAL) (t17 + t18);
   xMinor = (REAL) (bxmin3 * brSqrd5);
   xMinorP = (REAL) (bxminP3 * brSqrd5);
   xMinor2 = (REAL) (cxmin3 * crSqrd5);
   xMinorP2 = (REAL) (cxminP3 * crSqrd5);
   xMinor3 = (REAL) (xMinor - xMinor2);
   xMinorP3 = (REAL) (xMinorP + xMinorP2);
   xMinor4 = (REAL) (dxmin3 * drSqrd5);
   xMinorP4 = (REAL) (dxminP3 * drSqrd5);
   xMinor5 = (REAL) (xMinor3 + xMinor4);
   xMinorP5 = (REAL) (xMinorP3 + xMinorP4);
   yMinor = (REAL) (bymin3 * brSqrd5);
   yMinorP = (REAL) (byminP3 * brSqrd5);
   yMinor2 = (REAL) (cymin3 * crSqrd5);
   yMinorP2 = (REAL) (cyminP3 * crSqrd5);
   yMinor3 = (REAL) (yMinor - yMinor2);
   yMinorP3 = (REAL) (yMinorP + yMinorP2);
   yMinor4 = (REAL) (dymin3 * drSqrd5);
   yMinorP4 = (REAL) (dyminP3 * drSqrd5);
   yMinor5 = (REAL) (yMinor3 + yMinor4);
   yMinorP5 = (REAL) (yMinorP3 + yMinorP4);
   zMinor = (REAL) (bzmin3 * brSqrd5);
   zMinorP = (REAL) (bzminP3 * brSqrd5);
   zMinor2 = (REAL) (czmin3 * crSqrd5);
   zMinorP2 = (REAL) (czminP3 * crSqrd5);
   zMinor3 = (REAL) (zMinor - zMinor2);
   zMinorP3 = (REAL) (zMinorP + zMinorP2);
   zMinor4 = (REAL) (dzmin3 * drSqrd5);
   zMinorP4 = (REAL) (dzminP3 * drSqrd5);
   zMinor5 = (REAL) (zMinor3 + zMinor4);
   zMinorP5 = (REAL) (zMinorP3 + zMinorP4);
   denom = (REAL) (brx * bxmin3);
   t19 = Absolute(brx);
   denomP = (REAL) (t19 * bxminP3);
   denom2 = (REAL) (crx * cxmin3);
   t20 = Absolute(crx);
   denomP2 = (REAL) (t20 * cxminP3);
   denom3 = (REAL) (denom - denom2);
   denomP3 = (REAL) (denomP + denomP2);
   denom4 = (REAL) (drx * dxmin3);
   t21 = Absolute(drx);
   denomP4 = (REAL) (t21 * dxminP3);
   denom5 = (REAL) (denom3 + denom4);
   denomP5 = (REAL) (denomP3 + denomP4);
   numer = (REAL) (pry * yMinor5);
   t22 = Absolute(pry);
   numerP = (REAL) (t22 * yMinorP5);
   numer2 = (REAL) (prx * xMinor5);
   t23 = Absolute(prx);
   numerP2 = (REAL) (t23 * xMinorP5);
   numer3 = (REAL) (numer - numer2);
   numerP3 = (REAL) (numerP + numerP2);
   numer4 = (REAL) (prSqrd5 * denom5);
   numerP4 = (REAL) (prSqrd5 * denomP5);
   numer5 = (REAL) (prz * zMinor5);
   t24 = Absolute(prz);
   numerP5 = (REAL) (t24 * zMinorP5);
   numer6 = (REAL) (numer4 - numer5);
   numerP6 = (REAL) (numerP4 + numerP5);
   numer7 = (REAL) (numer3 + numer6);
   numerP7 = (REAL) (numerP3 + numerP6);
   insphere = (REAL) (numer7 * denom5);
   insphereP = (REAL) (numerP7 * denomP5);
   yAE = (REAL) (errA3to3 * insphereP);

   if (insphere > yAE)
      return 1;
   else {
      if (-insphere > yAE)
         return -1;
      else {
         Square(brx, brSqrdB[1], brSqrdB[0]);
         Square(bry, brSqrdB2[1], brSqrdB2[0]);
         Square(brz, brSqrdB3[1], brSqrdB3[0]);
         Two_Two_Sum(brSqrdB2[1], brSqrdB2[0], brSqrdB3[1], brSqrdB3[0],
                     brSqrdB4[3], brSqrdB4[2], brSqrdB4[1],
                     brSqrdB4[0]);
         brSqrdB4_len = 4;
         brSqrdB5_len =
            fast_expansion_sum_zeroelim(2, brSqrdB, brSqrdB4_len,
                                        brSqrdB4, brSqrdB5);
         Square(crx, crSqrdB[1], crSqrdB[0]);
         Square(cry, crSqrdB2[1], crSqrdB2[0]);
         Square(crz, crSqrdB3[1], crSqrdB3[0]);
         Two_Two_Sum(crSqrdB2[1], crSqrdB2[0], crSqrdB3[1], crSqrdB3[0],
                     crSqrdB4[3], crSqrdB4[2], crSqrdB4[1],
                     crSqrdB4[0]);
         crSqrdB4_len = 4;
         crSqrdB5_len =
            fast_expansion_sum_zeroelim(2, crSqrdB, crSqrdB4_len,
                                        crSqrdB4, crSqrdB5);
         Square(drx, drSqrdB[1], drSqrdB[0]);
         Square(dry, drSqrdB2[1], drSqrdB2[0]);
         Square(drz, drSqrdB3[1], drSqrdB3[0]);
         Two_Two_Sum(drSqrdB2[1], drSqrdB2[0], drSqrdB3[1], drSqrdB3[0],
                     drSqrdB4[3], drSqrdB4[2], drSqrdB4[1],
                     drSqrdB4[0]);
         drSqrdB4_len = 4;
         drSqrdB5_len =
            fast_expansion_sum_zeroelim(2, drSqrdB, drSqrdB4_len,
                                        drSqrdB4, drSqrdB5);
         Square(prx, prSqrdB[1], prSqrdB[0]);
         Square(pry, prSqrdB2[1], prSqrdB2[0]);
         Square(prz, prSqrdB3[1], prSqrdB3[0]);
         Two_Two_Sum(prSqrdB2[1], prSqrdB2[0], prSqrdB3[1], prSqrdB3[0],
                     prSqrdB4[3], prSqrdB4[2], prSqrdB4[1],
                     prSqrdB4[0]);
         prSqrdB4_len = 4;
         prSqrdB5_len =
            fast_expansion_sum_zeroelim(2, prSqrdB, prSqrdB4_len,
                                        prSqrdB4, prSqrdB5);
         Two_Product(cry, drz, bxminB[1], bxminB[0]);
         Two_Product(dry, crz, bxminB2[1], bxminB2[0]);
         Two_Two_Diff(bxminB[1], bxminB[0], bxminB2[1], bxminB2[0],
                      bxminB3[3], bxminB3[2], bxminB3[1], bxminB3[0]);
         bxminB3_len = 4;
         Two_Product(crx, drz, byminB[1], byminB[0]);
         Two_Product(drx, crz, byminB2[1], byminB2[0]);
         Two_Two_Diff(byminB[1], byminB[0], byminB2[1], byminB2[0],
                      byminB3[3], byminB3[2], byminB3[1], byminB3[0]);
         byminB3_len = 4;
         Two_Product(crx, dry, bzminB[1], bzminB[0]);
         Two_Product(drx, cry, bzminB2[1], bzminB2[0]);
         Two_Two_Diff(bzminB[1], bzminB[0], bzminB2[1], bzminB2[0],
                      bzminB3[3], bzminB3[2], bzminB3[1], bzminB3[0]);
         bzminB3_len = 4;
         Two_Product(bry, drz, cxminB[1], cxminB[0]);
         Two_Product(dry, brz, cxminB2[1], cxminB2[0]);
         Two_Two_Diff(cxminB[1], cxminB[0], cxminB2[1], cxminB2[0],
                      cxminB3[3], cxminB3[2], cxminB3[1], cxminB3[0]);
         cxminB3_len = 4;
         Two_Product(brx, drz, cyminB[1], cyminB[0]);
         Two_Product(drx, brz, cyminB2[1], cyminB2[0]);
         Two_Two_Diff(cyminB[1], cyminB[0], cyminB2[1], cyminB2[0],
                      cyminB3[3], cyminB3[2], cyminB3[1], cyminB3[0]);
         cyminB3_len = 4;
         Two_Product(brx, dry, czminB[1], czminB[0]);
         Two_Product(drx, bry, czminB2[1], czminB2[0]);
         Two_Two_Diff(czminB[1], czminB[0], czminB2[1], czminB2[0],
                      czminB3[3], czminB3[2], czminB3[1], czminB3[0]);
         czminB3_len = 4;
         Two_Product(bry, crz, dxminB[1], dxminB[0]);
         Two_Product(cry, brz, dxminB2[1], dxminB2[0]);
         Two_Two_Diff(dxminB[1], dxminB[0], dxminB2[1], dxminB2[0],
                      dxminB3[3], dxminB3[2], dxminB3[1], dxminB3[0]);
         dxminB3_len = 4;
         Two_Product(brx, crz, dyminB[1], dyminB[0]);
         Two_Product(crx, brz, dyminB2[1], dyminB2[0]);
         Two_Two_Diff(dyminB[1], dyminB[0], dyminB2[1], dyminB2[0],
                      dyminB3[3], dyminB3[2], dyminB3[1], dyminB3[0]);
         dyminB3_len = 4;
         Two_Product(brx, cry, dzminB[1], dzminB[0]);
         Two_Product(crx, bry, dzminB2[1], dzminB2[0]);
         Two_Two_Diff(dzminB[1], dzminB[0], dzminB2[1], dzminB2[0],
                      dzminB3[3], dzminB3[2], dzminB3[1], dzminB3[0]);
         dzminB3_len = 4;
         xMinorB_len =
            mult_expansion_zeroelim(bxminB3_len, bxminB3, brSqrdB5_len,
                                    brSqrdB5, xMinorB);
         xMinorB2_len =
            mult_expansion_zeroelim(cxminB3_len, cxminB3, crSqrdB5_len,
                                    crSqrdB5, xMinorB2);
         xMinorB3_len =
            fast_expansion_diff_zeroelim(xMinorB_len, xMinorB,
                                         xMinorB2_len, xMinorB2,
                                         xMinorB3);
         xMinorB4_len =
            mult_expansion_zeroelim(dxminB3_len, dxminB3, drSqrdB5_len,
                                    drSqrdB5, xMinorB4);
         xMinorB5_len =
            fast_expansion_sum_zeroelim(xMinorB4_len, xMinorB4,
                                        xMinorB3_len, xMinorB3,
                                        xMinorB5);
         yMinorB_len =
            mult_expansion_zeroelim(byminB3_len, byminB3, brSqrdB5_len,
                                    brSqrdB5, yMinorB);
         yMinorB2_len =
            mult_expansion_zeroelim(cyminB3_len, cyminB3, crSqrdB5_len,
                                    crSqrdB5, yMinorB2);
         yMinorB3_len =
            fast_expansion_diff_zeroelim(yMinorB_len, yMinorB,
                                         yMinorB2_len, yMinorB2,
                                         yMinorB3);
         yMinorB4_len =
            mult_expansion_zeroelim(dyminB3_len, dyminB3, drSqrdB5_len,
                                    drSqrdB5, yMinorB4);
         yMinorB5_len =
            fast_expansion_sum_zeroelim(yMinorB4_len, yMinorB4,
                                        yMinorB3_len, yMinorB3,
                                        yMinorB5);
         zMinorB_len =
            mult_expansion_zeroelim(bzminB3_len, bzminB3, brSqrdB5_len,
                                    brSqrdB5, zMinorB);
         zMinorB2_len =
            mult_expansion_zeroelim(czminB3_len, czminB3, crSqrdB5_len,
                                    crSqrdB5, zMinorB2);
         zMinorB3_len =
            fast_expansion_diff_zeroelim(zMinorB_len, zMinorB,
                                         zMinorB2_len, zMinorB2,
                                         zMinorB3);
         zMinorB4_len =
            mult_expansion_zeroelim(dzminB3_len, dzminB3, drSqrdB5_len,
                                    drSqrdB5, zMinorB4);
         zMinorB5_len =
            fast_expansion_sum_zeroelim(zMinorB4_len, zMinorB4,
                                        zMinorB3_len, zMinorB3,
                                        zMinorB5);
         denomB_len =
            scale_expansion_zeroelim(bxminB3_len, bxminB3, brx,
                                     denomB);
         denomB2_len =
            scale_expansion_zeroelim(cxminB3_len, cxminB3, crx,
                                     denomB2);
         denomB3_len =
            fast_expansion_diff_zeroelim(denomB_len, denomB,
                                         denomB2_len, denomB2,
                                         denomB3);
         denomB4_len =
            scale_expansion_zeroelim(dxminB3_len, dxminB3, drx,
                                     denomB4);
         denomB5_len =
            fast_expansion_sum_zeroelim(denomB4_len, denomB4,
                                        denomB3_len, denomB3, denomB5);
         numerB_len =
            scale_expansion_zeroelim(yMinorB5_len, yMinorB5, pry,
                                     numerB);
         numerB2_len =
            scale_expansion_zeroelim(xMinorB5_len, xMinorB5, prx,
                                     numerB2);
         numerB3_len =
            fast_expansion_diff_zeroelim(numerB_len, numerB,
                                         numerB2_len, numerB2,
                                         numerB3);
         numerB4_len =
            mult_expansion_zeroelim(prSqrdB5_len, prSqrdB5,
                                    denomB5_len, denomB5, numerB4);
         numerB5_len =
            scale_expansion_zeroelim(zMinorB5_len, zMinorB5, prz,
                                     numerB5);
         numerB6_len =
            fast_expansion_diff_zeroelim(numerB4_len, numerB4,
                                         numerB5_len, numerB5,
                                         numerB6);
         numerB7_len =
            fast_expansion_sum_zeroelim(numerB6_len, numerB6,
                                        numerB3_len, numerB3, numerB7);
         insphereB_len =
            mult_expansion_zeroelim(numerB7_len, numerB7, denomB5_len,
                                    denomB5, insphereB);
         yBX = estimate(insphereB_len, insphereB);
         yBE = (REAL) (errB3to3 * insphereP);
         if (yBX > yBE)
            return 1;
         else {
            if (-yBX > yBE)
               return -1;
            else {
               Two_Diff_Tail(bx, ax, brx, brxC);
               Two_Diff_Tail(by, ay, bry, bryC);
               Two_Diff_Tail(bz, az, brz, brzC);
               Two_Diff_Tail(cx, ax, crx, crxC);
               Two_Diff_Tail(cy, ay, cry, cryC);
               Two_Diff_Tail(cz, az, crz, crzC);
               Two_Diff_Tail(dx, ax, drx, drxC);
               Two_Diff_Tail(dy, ay, dry, dryC);
               Two_Diff_Tail(dz, az, drz, drzC);
               Two_Diff_Tail(px, ax, prx, prxC);
               Two_Diff_Tail(py, ay, pry, pryC);
               Two_Diff_Tail(pz, az, prz, przC);
               t25 = (REAL) (brx * brxC);
               brSqrdC = (REAL) (2.0 * t25);
               t26 = (REAL) (bry * bryC);
               brSqrdC2 = (REAL) (2.0 * t26);
               t27 = (REAL) (brz * brzC);
               brSqrdC3 = (REAL) (2.0 * t27);
               brSqrdC4 = (REAL) (brSqrdC2 + brSqrdC3);
               brSqrdC5 = (REAL) (brSqrdC + brSqrdC4);
               t28 = (REAL) (crx * crxC);
               crSqrdC = (REAL) (2.0 * t28);
               t29 = (REAL) (cry * cryC);
               crSqrdC2 = (REAL) (2.0 * t29);
               t30 = (REAL) (crz * crzC);
               crSqrdC3 = (REAL) (2.0 * t30);
               crSqrdC4 = (REAL) (crSqrdC2 + crSqrdC3);
               crSqrdC5 = (REAL) (crSqrdC + crSqrdC4);
               t31 = (REAL) (drx * drxC);
               drSqrdC = (REAL) (2.0 * t31);
               t32 = (REAL) (dry * dryC);
               drSqrdC2 = (REAL) (2.0 * t32);
               t33 = (REAL) (drz * drzC);
               drSqrdC3 = (REAL) (2.0 * t33);
               drSqrdC4 = (REAL) (drSqrdC2 + drSqrdC3);
               drSqrdC5 = (REAL) (drSqrdC + drSqrdC4);
               t34 = (REAL) (prx * prxC);
               prSqrdC = (REAL) (2.0 * t34);
               t35 = (REAL) (pry * pryC);
               prSqrdC2 = (REAL) (2.0 * t35);
               t36 = (REAL) (prz * przC);
               prSqrdC3 = (REAL) (2.0 * t36);
               prSqrdC4 = (REAL) (prSqrdC2 + prSqrdC3);
               prSqrdC5 = (REAL) (prSqrdC + prSqrdC4);
               t37 = (REAL) (cry * drzC);
               t38 = (REAL) (cryC * drz);
               bxminC = (REAL) (t37 + t38);
               t39 = (REAL) (dry * crzC);
               t40 = (REAL) (dryC * crz);
               bxminC2 = (REAL) (t39 + t40);
               bxminC3 = (REAL) (bxminC - bxminC2);
               t41 = (REAL) (crx * drzC);
               t42 = (REAL) (crxC * drz);
               byminC = (REAL) (t41 + t42);
               t43 = (REAL) (drx * crzC);
               t44 = (REAL) (drxC * crz);
               byminC2 = (REAL) (t43 + t44);
               byminC3 = (REAL) (byminC - byminC2);
               t45 = (REAL) (crx * dryC);
               t46 = (REAL) (crxC * dry);
               bzminC = (REAL) (t45 + t46);
               t47 = (REAL) (drx * cryC);
               t48 = (REAL) (drxC * cry);
               bzminC2 = (REAL) (t47 + t48);
               bzminC3 = (REAL) (bzminC - bzminC2);
               t49 = (REAL) (bry * drzC);
               t50 = (REAL) (bryC * drz);
               cxminC = (REAL) (t49 + t50);
               t51 = (REAL) (dry * brzC);
               t52 = (REAL) (dryC * brz);
               cxminC2 = (REAL) (t51 + t52);
               cxminC3 = (REAL) (cxminC - cxminC2);
               t53 = (REAL) (brx * drzC);
               t54 = (REAL) (brxC * drz);
               cyminC = (REAL) (t53 + t54);
               t55 = (REAL) (drx * brzC);
               t56 = (REAL) (drxC * brz);
               cyminC2 = (REAL) (t55 + t56);
               cyminC3 = (REAL) (cyminC - cyminC2);
               t57 = (REAL) (brx * dryC);
               t58 = (REAL) (brxC * dry);
               czminC = (REAL) (t57 + t58);
               t59 = (REAL) (drx * bryC);
               t60 = (REAL) (drxC * bry);
               czminC2 = (REAL) (t59 + t60);
               czminC3 = (REAL) (czminC - czminC2);
               t61 = (REAL) (bry * crzC);
               t62 = (REAL) (bryC * crz);
               dxminC = (REAL) (t61 + t62);
               t63 = (REAL) (cry * brzC);
               t64 = (REAL) (cryC * brz);
               dxminC2 = (REAL) (t63 + t64);
               dxminC3 = (REAL) (dxminC - dxminC2);
               t65 = (REAL) (brx * crzC);
               t66 = (REAL) (brxC * crz);
               dyminC = (REAL) (t65 + t66);
               t67 = (REAL) (crx * brzC);
               t68 = (REAL) (crxC * brz);
               dyminC2 = (REAL) (t67 + t68);
               dyminC3 = (REAL) (dyminC - dyminC2);
               t69 = (REAL) (brx * cryC);
               t70 = (REAL) (brxC * cry);
               dzminC = (REAL) (t69 + t70);
               t71 = (REAL) (crx * bryC);
               t72 = (REAL) (crxC * bry);
               dzminC2 = (REAL) (t71 + t72);
               dzminC3 = (REAL) (dzminC - dzminC2);
               t73 = (REAL) (bxmin3 * brSqrdC5);
               t74 = (REAL) (bxminC3 * brSqrd5);
               xMinorC = (REAL) (t73 + t74);
               t75 = (REAL) (cxmin3 * crSqrdC5);
               t76 = (REAL) (cxminC3 * crSqrd5);
               xMinorC2 = (REAL) (t75 + t76);
               xMinorC3 = (REAL) (xMinorC - xMinorC2);
               t77 = (REAL) (dxmin3 * drSqrdC5);
               t78 = (REAL) (dxminC3 * drSqrd5);
               xMinorC4 = (REAL) (t77 + t78);
               xMinorC5 = (REAL) (xMinorC3 + xMinorC4);
               t79 = (REAL) (bymin3 * brSqrdC5);
               t80 = (REAL) (byminC3 * brSqrd5);
               yMinorC = (REAL) (t79 + t80);
               t81 = (REAL) (cymin3 * crSqrdC5);
               t82 = (REAL) (cyminC3 * crSqrd5);
               yMinorC2 = (REAL) (t81 + t82);
               yMinorC3 = (REAL) (yMinorC - yMinorC2);
               t83 = (REAL) (dymin3 * drSqrdC5);
               t84 = (REAL) (dyminC3 * drSqrd5);
               yMinorC4 = (REAL) (t83 + t84);
               yMinorC5 = (REAL) (yMinorC3 + yMinorC4);
               t85 = (REAL) (bzmin3 * brSqrdC5);
               t86 = (REAL) (bzminC3 * brSqrd5);
               zMinorC = (REAL) (t85 + t86);
               t87 = (REAL) (czmin3 * crSqrdC5);
               t88 = (REAL) (czminC3 * crSqrd5);
               zMinorC2 = (REAL) (t87 + t88);
               zMinorC3 = (REAL) (zMinorC - zMinorC2);
               t89 = (REAL) (dzmin3 * drSqrdC5);
               t90 = (REAL) (dzminC3 * drSqrd5);
               zMinorC4 = (REAL) (t89 + t90);
               zMinorC5 = (REAL) (zMinorC3 + zMinorC4);
               t91 = (REAL) (brx * bxminC3);
               t92 = (REAL) (brxC * bxmin3);
               denomC = (REAL) (t91 + t92);
               t93 = (REAL) (crx * cxminC3);
               t94 = (REAL) (crxC * cxmin3);
               denomC2 = (REAL) (t93 + t94);
               denomC3 = (REAL) (denomC - denomC2);
               t95 = (REAL) (drx * dxminC3);
               t96 = (REAL) (drxC * dxmin3);
               denomC4 = (REAL) (t95 + t96);
               denomC5 = (REAL) (denomC3 + denomC4);
               t97 = (REAL) (pry * yMinorC5);
               t98 = (REAL) (pryC * yMinor5);
               numerC = (REAL) (t97 + t98);
               t99 = (REAL) (prx * xMinorC5);
               t100 = (REAL) (prxC * xMinor5);
               numerC2 = (REAL) (t99 + t100);
               numerC3 = (REAL) (numerC - numerC2);
               t101 = (REAL) (prSqrd5 * denomC5);
               t102 = (REAL) (prSqrdC5 * denom5);
               numerC4 = (REAL) (t101 + t102);
               t103 = (REAL) (prz * zMinorC5);
               t104 = (REAL) (przC * zMinor5);
               numerC5 = (REAL) (t103 + t104);
               numerC6 = (REAL) (numerC4 - numerC5);
               numerC7 = (REAL) (numerC3 + numerC6);
               t105 = (REAL) (numer7 * denomC5);
               t106 = (REAL) (numerC7 * denom5);
               insphereC = (REAL) (t105 + t106);
               yCX = (REAL) (yBX + insphereC);
               t107 = (REAL) (errCX3to3 * yBX);
               t108 = (REAL) (errC3to3 * insphereP);
               yCE = (REAL) (t107 + t108);
               if (yCX > yCE)
                  return 1;
               else {
                  if (-yCX > yCE)
                     return -1;
                  else {
                     Two_Product(brx, brxC, t110[1], t110[0]);
                     t109[1] = 2.0 * t110[1];
                     t109[0] = 2.0 * t110[0];
                     Square(brxC, t111[1], t111[0]);
                     Two_Two_Sum(t109[1], t109[0], t111[1], t111[0],
                                 brSqrdD[3], brSqrdD[2], brSqrdD[1],
                                 brSqrdD[0]);
                     brSqrdD_len = 4;
                     Two_Product(bry, bryC, t113[1], t113[0]);
                     t112[1] = 2.0 * t113[1];
                     t112[0] = 2.0 * t113[0];
                     Square(bryC, t114[1], t114[0]);
                     Two_Two_Sum(t112[1], t112[0], t114[1], t114[0],
                                 brSqrdD2[3], brSqrdD2[2],
                                 brSqrdD2[1], brSqrdD2[0]);
                     brSqrdD2_len = 4;
                     Two_Product(brz, brzC, t116[1], t116[0]);
                     t115[1] = 2.0 * t116[1];
                     t115[0] = 2.0 * t116[0];
                     Square(brzC, t117[1], t117[0]);
                     Two_Two_Sum(t115[1], t115[0], t117[1], t117[0],
                                 brSqrdD3[3], brSqrdD3[2],
                                 brSqrdD3[1], brSqrdD3[0]);
                     brSqrdD3_len = 4;
                     brSqrdD4_len =
                        fast_expansion_sum_zeroelim(brSqrdD3_len,
                                                    brSqrdD3,
                                                    brSqrdD2_len,
                                                    brSqrdD2,
                                                    brSqrdD4);
                     brSqrdD5_len =
                        fast_expansion_sum_zeroelim(brSqrdD4_len,
                                                    brSqrdD4,
                                                    brSqrdD_len,
                                                    brSqrdD,
                                                    brSqrdD5);
                     Two_Product(crx, crxC, t119[1], t119[0]);
                     t118[1] = 2.0 * t119[1];
                     t118[0] = 2.0 * t119[0];
                     Square(crxC, t120[1], t120[0]);
                     Two_Two_Sum(t118[1], t118[0], t120[1], t120[0],
                                 crSqrdD[3], crSqrdD[2], crSqrdD[1],
                                 crSqrdD[0]);
                     crSqrdD_len = 4;
                     Two_Product(cry, cryC, t122[1], t122[0]);
                     t121[1] = 2.0 * t122[1];
                     t121[0] = 2.0 * t122[0];
                     Square(cryC, t123[1], t123[0]);
                     Two_Two_Sum(t121[1], t121[0], t123[1], t123[0],
                                 crSqrdD2[3], crSqrdD2[2],
                                 crSqrdD2[1], crSqrdD2[0]);
                     crSqrdD2_len = 4;
                     Two_Product(crz, crzC, t125[1], t125[0]);
                     t124[1] = 2.0 * t125[1];
                     t124[0] = 2.0 * t125[0];
                     Square(crzC, t126[1], t126[0]);
                     Two_Two_Sum(t124[1], t124[0], t126[1], t126[0],
                                 crSqrdD3[3], crSqrdD3[2],
                                 crSqrdD3[1], crSqrdD3[0]);
                     crSqrdD3_len = 4;
                     crSqrdD4_len =
                        fast_expansion_sum_zeroelim(crSqrdD3_len,
                                                    crSqrdD3,
                                                    crSqrdD2_len,
                                                    crSqrdD2,
                                                    crSqrdD4);
                     crSqrdD5_len =
                        fast_expansion_sum_zeroelim(crSqrdD4_len,
                                                    crSqrdD4,
                                                    crSqrdD_len,
                                                    crSqrdD,
                                                    crSqrdD5);
                     Two_Product(drx, drxC, t128[1], t128[0]);
                     t127[1] = 2.0 * t128[1];
                     t127[0] = 2.0 * t128[0];
                     Square(drxC, t129[1], t129[0]);
                     Two_Two_Sum(t127[1], t127[0], t129[1], t129[0],
                                 drSqrdD[3], drSqrdD[2], drSqrdD[1],
                                 drSqrdD[0]);
                     drSqrdD_len = 4;
                     Two_Product(dry, dryC, t131[1], t131[0]);
                     t130[1] = 2.0 * t131[1];
                     t130[0] = 2.0 * t131[0];
                     Square(dryC, t132[1], t132[0]);
                     Two_Two_Sum(t130[1], t130[0], t132[1], t132[0],
                                 drSqrdD2[3], drSqrdD2[2],
                                 drSqrdD2[1], drSqrdD2[0]);
                     drSqrdD2_len = 4;
                     Two_Product(drz, drzC, t134[1], t134[0]);
                     t133[1] = 2.0 * t134[1];
                     t133[0] = 2.0 * t134[0];
                     Square(drzC, t135[1], t135[0]);
                     Two_Two_Sum(t133[1], t133[0], t135[1], t135[0],
                                 drSqrdD3[3], drSqrdD3[2],
                                 drSqrdD3[1], drSqrdD3[0]);
                     drSqrdD3_len = 4;
                     drSqrdD4_len =
                        fast_expansion_sum_zeroelim(drSqrdD3_len,
                                                    drSqrdD3,
                                                    drSqrdD2_len,
                                                    drSqrdD2,
                                                    drSqrdD4);
                     drSqrdD5_len =
                        fast_expansion_sum_zeroelim(drSqrdD4_len,
                                                    drSqrdD4,
                                                    drSqrdD_len,
                                                    drSqrdD,
                                                    drSqrdD5);
                     Two_Product(prx, prxC, t137[1], t137[0]);
                     t136[1] = 2.0 * t137[1];
                     t136[0] = 2.0 * t137[0];
                     Square(prxC, t138[1], t138[0]);
                     Two_Two_Sum(t136[1], t136[0], t138[1], t138[0],
                                 prSqrdD[3], prSqrdD[2], prSqrdD[1],
                                 prSqrdD[0]);
                     prSqrdD_len = 4;
                     Two_Product(pry, pryC, t140[1], t140[0]);
                     t139[1] = 2.0 * t140[1];
                     t139[0] = 2.0 * t140[0];
                     Square(pryC, t141[1], t141[0]);
                     Two_Two_Sum(t139[1], t139[0], t141[1], t141[0],
                                 prSqrdD2[3], prSqrdD2[2],
                                 prSqrdD2[1], prSqrdD2[0]);
                     prSqrdD2_len = 4;
                     Two_Product(prz, przC, t143[1], t143[0]);
                     t142[1] = 2.0 * t143[1];
                     t142[0] = 2.0 * t143[0];
                     Square(przC, t144[1], t144[0]);
                     Two_Two_Sum(t142[1], t142[0], t144[1], t144[0],
                                 prSqrdD3[3], prSqrdD3[2],
                                 prSqrdD3[1], prSqrdD3[0]);
                     prSqrdD3_len = 4;
                     prSqrdD4_len =
                        fast_expansion_sum_zeroelim(prSqrdD3_len,
                                                    prSqrdD3,
                                                    prSqrdD2_len,
                                                    prSqrdD2,
                                                    prSqrdD4);
                     prSqrdD5_len =
                        fast_expansion_sum_zeroelim(prSqrdD4_len,
                                                    prSqrdD4,
                                                    prSqrdD_len,
                                                    prSqrdD,
                                                    prSqrdD5);
                     Two_Product(cry, drzC, t146[1], t146[0]);
                     Two_Product(cryC, drz, t147[1], t147[0]);
                     Two_Two_Sum(t146[1], t146[0], t147[1], t147[0],
                                 t145[3], t145[2], t145[1],
                                 t145[0]);
                     t145_len = 4;
                     Two_Product(cryC, drzC, t148[1], t148[0]);
                     bxminD_len =
                        fast_expansion_sum_zeroelim(2, t148,
                                                    t145_len, t145,
                                                    bxminD);
                     Two_Product(dry, crzC, t150[1], t150[0]);
                     Two_Product(dryC, crz, t151[1], t151[0]);
                     Two_Two_Sum(t150[1], t150[0], t151[1], t151[0],
                                 t149[3], t149[2], t149[1],
                                 t149[0]);
                     t149_len = 4;
                     Two_Product(dryC, crzC, t152[1], t152[0]);
                     bxminD2_len =
                        fast_expansion_sum_zeroelim(2, t152,
                                                    t149_len, t149,
                                                    bxminD2);
                     bxminD3_len =
                        fast_expansion_diff_zeroelim(bxminD_len,
                                                     bxminD,
                                                     bxminD2_len,
                                                     bxminD2,
                                                     bxminD3);
                     Two_Product(crx, drzC, t154[1], t154[0]);
                     Two_Product(crxC, drz, t155[1], t155[0]);
                     Two_Two_Sum(t154[1], t154[0], t155[1], t155[0],
                                 t153[3], t153[2], t153[1],
                                 t153[0]);
                     t153_len = 4;
                     Two_Product(crxC, drzC, t156[1], t156[0]);
                     byminD_len =
                        fast_expansion_sum_zeroelim(2, t156,
                                                    t153_len, t153,
                                                    byminD);
                     Two_Product(drx, crzC, t158[1], t158[0]);
                     Two_Product(drxC, crz, t159[1], t159[0]);
                     Two_Two_Sum(t158[1], t158[0], t159[1], t159[0],
                                 t157[3], t157[2], t157[1],
                                 t157[0]);
                     t157_len = 4;
                     Two_Product(drxC, crzC, t160[1], t160[0]);
                     byminD2_len =
                        fast_expansion_sum_zeroelim(2, t160,
                                                    t157_len, t157,
                                                    byminD2);
                     byminD3_len =
                        fast_expansion_diff_zeroelim(byminD_len,
                                                     byminD,
                                                     byminD2_len,
                                                     byminD2,
                                                     byminD3);
                     Two_Product(crx, dryC, t162[1], t162[0]);
                     Two_Product(crxC, dry, t163[1], t163[0]);
                     Two_Two_Sum(t162[1], t162[0], t163[1], t163[0],
                                 t161[3], t161[2], t161[1],
                                 t161[0]);
                     t161_len = 4;
                     Two_Product(crxC, dryC, t164[1], t164[0]);
                     bzminD_len =
                        fast_expansion_sum_zeroelim(2, t164,
                                                    t161_len, t161,
                                                    bzminD);
                     Two_Product(drx, cryC, t166[1], t166[0]);
                     Two_Product(drxC, cry, t167[1], t167[0]);
                     Two_Two_Sum(t166[1], t166[0], t167[1], t167[0],
                                 t165[3], t165[2], t165[1],
                                 t165[0]);
                     t165_len = 4;
                     Two_Product(drxC, cryC, t168[1], t168[0]);
                     bzminD2_len =
                        fast_expansion_sum_zeroelim(2, t168,
                                                    t165_len, t165,
                                                    bzminD2);
                     bzminD3_len =
                        fast_expansion_diff_zeroelim(bzminD_len,
                                                     bzminD,
                                                     bzminD2_len,
                                                     bzminD2,
                                                     bzminD3);
                     Two_Product(bry, drzC, t170[1], t170[0]);
                     Two_Product(bryC, drz, t171[1], t171[0]);
                     Two_Two_Sum(t170[1], t170[0], t171[1], t171[0],
                                 t169[3], t169[2], t169[1],
                                 t169[0]);
                     t169_len = 4;
                     Two_Product(bryC, drzC, t172[1], t172[0]);
                     cxminD_len =
                        fast_expansion_sum_zeroelim(2, t172,
                                                    t169_len, t169,
                                                    cxminD);
                     Two_Product(dry, brzC, t174[1], t174[0]);
                     Two_Product(dryC, brz, t175[1], t175[0]);
                     Two_Two_Sum(t174[1], t174[0], t175[1], t175[0],
                                 t173[3], t173[2], t173[1],
                                 t173[0]);
                     t173_len = 4;
                     Two_Product(dryC, brzC, t176[1], t176[0]);
                     cxminD2_len =
                        fast_expansion_sum_zeroelim(2, t176,
                                                    t173_len, t173,
                                                    cxminD2);
                     cxminD3_len =
                        fast_expansion_diff_zeroelim(cxminD_len,
                                                     cxminD,
                                                     cxminD2_len,
                                                     cxminD2,
                                                     cxminD3);
                     Two_Product(brx, drzC, t178[1], t178[0]);
                     Two_Product(brxC, drz, t179[1], t179[0]);
                     Two_Two_Sum(t178[1], t178[0], t179[1], t179[0],
                                 t177[3], t177[2], t177[1],
                                 t177[0]);
                     t177_len = 4;
                     Two_Product(brxC, drzC, t180[1], t180[0]);
                     cyminD_len =
                        fast_expansion_sum_zeroelim(2, t180,
                                                    t177_len, t177,
                                                    cyminD);
                     Two_Product(drx, brzC, t182[1], t182[0]);
                     Two_Product(drxC, brz, t183[1], t183[0]);
                     Two_Two_Sum(t182[1], t182[0], t183[1], t183[0],
                                 t181[3], t181[2], t181[1],
                                 t181[0]);
                     t181_len = 4;
                     Two_Product(drxC, brzC, t184[1], t184[0]);
                     cyminD2_len =
                        fast_expansion_sum_zeroelim(2, t184,
                                                    t181_len, t181,
                                                    cyminD2);
                     cyminD3_len =
                        fast_expansion_diff_zeroelim(cyminD_len,
                                                     cyminD,
                                                     cyminD2_len,
                                                     cyminD2,
                                                     cyminD3);
                     Two_Product(brx, dryC, t186[1], t186[0]);
                     Two_Product(brxC, dry, t187[1], t187[0]);
                     Two_Two_Sum(t186[1], t186[0], t187[1], t187[0],
                                 t185[3], t185[2], t185[1],
                                 t185[0]);
                     t185_len = 4;
                     Two_Product(brxC, dryC, t188[1], t188[0]);
                     czminD_len =
                        fast_expansion_sum_zeroelim(2, t188,
                                                    t185_len, t185,
                                                    czminD);
                     Two_Product(drx, bryC, t190[1], t190[0]);
                     Two_Product(drxC, bry, t191[1], t191[0]);
                     Two_Two_Sum(t190[1], t190[0], t191[1], t191[0],
                                 t189[3], t189[2], t189[1],
                                 t189[0]);
                     t189_len = 4;
                     Two_Product(drxC, bryC, t192[1], t192[0]);
                     czminD2_len =
                        fast_expansion_sum_zeroelim(2, t192,
                                                    t189_len, t189,
                                                    czminD2);
                     czminD3_len =
                        fast_expansion_diff_zeroelim(czminD_len,
                                                     czminD,
                                                     czminD2_len,
                                                     czminD2,
                                                     czminD3);
                     Two_Product(bry, crzC, t194[1], t194[0]);
                     Two_Product(bryC, crz, t195[1], t195[0]);
                     Two_Two_Sum(t194[1], t194[0], t195[1], t195[0],
                                 t193[3], t193[2], t193[1],
                                 t193[0]);
                     t193_len = 4;
                     Two_Product(bryC, crzC, t196[1], t196[0]);
                     dxminD_len =
                        fast_expansion_sum_zeroelim(2, t196,
                                                    t193_len, t193,
                                                    dxminD);
                     Two_Product(cry, brzC, t198[1], t198[0]);
                     Two_Product(cryC, brz, t199[1], t199[0]);
                     Two_Two_Sum(t198[1], t198[0], t199[1], t199[0],
                                 t197[3], t197[2], t197[1],
                                 t197[0]);
                     t197_len = 4;
                     Two_Product(cryC, brzC, t200[1], t200[0]);
                     dxminD2_len =
                        fast_expansion_sum_zeroelim(2, t200,
                                                    t197_len, t197,
                                                    dxminD2);
                     dxminD3_len =
                        fast_expansion_diff_zeroelim(dxminD_len,
                                                     dxminD,
                                                     dxminD2_len,
                                                     dxminD2,
                                                     dxminD3);
                     Two_Product(brx, crzC, t202[1], t202[0]);
                     Two_Product(brxC, crz, t203[1], t203[0]);
                     Two_Two_Sum(t202[1], t202[0], t203[1], t203[0],
                                 t201[3], t201[2], t201[1],
                                 t201[0]);
                     t201_len = 4;
                     Two_Product(brxC, crzC, t204[1], t204[0]);
                     dyminD_len =
                        fast_expansion_sum_zeroelim(2, t204,
                                                    t201_len, t201,
                                                    dyminD);
                     Two_Product(crx, brzC, t206[1], t206[0]);
                     Two_Product(crxC, brz, t207[1], t207[0]);
                     Two_Two_Sum(t206[1], t206[0], t207[1], t207[0],
                                 t205[3], t205[2], t205[1],
                                 t205[0]);
                     t205_len = 4;
                     Two_Product(crxC, brzC, t208[1], t208[0]);
                     dyminD2_len =
                        fast_expansion_sum_zeroelim(2, t208,
                                                    t205_len, t205,
                                                    dyminD2);
                     dyminD3_len =
                        fast_expansion_diff_zeroelim(dyminD_len,
                                                     dyminD,
                                                     dyminD2_len,
                                                     dyminD2,
                                                     dyminD3);
                     Two_Product(brx, cryC, t210[1], t210[0]);
                     Two_Product(brxC, cry, t211[1], t211[0]);
                     Two_Two_Sum(t210[1], t210[0], t211[1], t211[0],
                                 t209[3], t209[2], t209[1],
                                 t209[0]);
                     t209_len = 4;
                     Two_Product(brxC, cryC, t212[1], t212[0]);
                     dzminD_len =
                        fast_expansion_sum_zeroelim(2, t212,
                                                    t209_len, t209,
                                                    dzminD);
                     Two_Product(crx, bryC, t214[1], t214[0]);
                     Two_Product(crxC, bry, t215[1], t215[0]);
                     Two_Two_Sum(t214[1], t214[0], t215[1], t215[0],
                                 t213[3], t213[2], t213[1],
                                 t213[0]);
                     t213_len = 4;
                     Two_Product(crxC, bryC, t216[1], t216[0]);
                     dzminD2_len =
                        fast_expansion_sum_zeroelim(2, t216,
                                                    t213_len, t213,
                                                    dzminD2);
                     dzminD3_len =
                        fast_expansion_diff_zeroelim(dzminD_len,
                                                     dzminD,
                                                     dzminD2_len,
                                                     dzminD2,
                                                     dzminD3);
                     t218_len =
                        mult_expansion_zeroelim(bxminB3_len,
                                                bxminB3,
                                                brSqrdD5_len,
                                                brSqrdD5, t218);
                     t219_len =
                        mult_expansion_zeroelim(bxminD3_len,
                                                bxminD3,
                                                brSqrdB5_len,
                                                brSqrdB5, t219);
                     t217_len =
                        fast_expansion_sum_zeroelim(t219_len, t219,
                                                    t218_len, t218,
                                                    t217);
                     t220_len =
                        mult_expansion_zeroelim(bxminD3_len,
                                                bxminD3,
                                                brSqrdD5_len,
                                                brSqrdD5, t220);
                     xMinorD_len =
                        fast_expansion_sum_zeroelim(t220_len, t220,
                                                    t217_len, t217,
                                                    xMinorD);
                     t222_len =
                        mult_expansion_zeroelim(cxminB3_len,
                                                cxminB3,
                                                crSqrdD5_len,
                                                crSqrdD5, t222);
                     t223_len =
                        mult_expansion_zeroelim(cxminD3_len,
                                                cxminD3,
                                                crSqrdB5_len,
                                                crSqrdB5, t223);
                     t221_len =
                        fast_expansion_sum_zeroelim(t223_len, t223,
                                                    t222_len, t222,
                                                    t221);
                     t224_len =
                        mult_expansion_zeroelim(cxminD3_len,
                                                cxminD3,
                                                crSqrdD5_len,
                                                crSqrdD5, t224);
                     xMinorD2_len =
                        fast_expansion_sum_zeroelim(t224_len, t224,
                                                    t221_len, t221,
                                                    xMinorD2);
                     xMinorD3_len =
                        fast_expansion_diff_zeroelim(xMinorD_len,
                                                     xMinorD,
                                                     xMinorD2_len,
                                                     xMinorD2,
                                                     xMinorD3);
                     t226_len =
                        mult_expansion_zeroelim(dxminB3_len,
                                                dxminB3,
                                                drSqrdD5_len,
                                                drSqrdD5, t226);
                     t227_len =
                        mult_expansion_zeroelim(dxminD3_len,
                                                dxminD3,
                                                drSqrdB5_len,
                                                drSqrdB5, t227);
                     t225_len =
                        fast_expansion_sum_zeroelim(t227_len, t227,
                                                    t226_len, t226,
                                                    t225);
                     t228_len =
                        mult_expansion_zeroelim(dxminD3_len,
                                                dxminD3,
                                                drSqrdD5_len,
                                                drSqrdD5, t228);
                     xMinorD4_len =
                        fast_expansion_sum_zeroelim(t228_len, t228,
                                                    t225_len, t225,
                                                    xMinorD4);
                     xMinorD5_len =
                        fast_expansion_sum_zeroelim(xMinorD4_len,
                                                    xMinorD4,
                                                    xMinorD3_len,
                                                    xMinorD3,
                                                    xMinorD5);
                     t230_len =
                        mult_expansion_zeroelim(byminB3_len,
                                                byminB3,
                                                brSqrdD5_len,
                                                brSqrdD5, t230);
                     t231_len =
                        mult_expansion_zeroelim(byminD3_len,
                                                byminD3,
                                                brSqrdB5_len,
                                                brSqrdB5, t231);
                     t229_len =
                        fast_expansion_sum_zeroelim(t231_len, t231,
                                                    t230_len, t230,
                                                    t229);
                     t232_len =
                        mult_expansion_zeroelim(byminD3_len,
                                                byminD3,
                                                brSqrdD5_len,
                                                brSqrdD5, t232);
                     yMinorD_len =
                        fast_expansion_sum_zeroelim(t232_len, t232,
                                                    t229_len, t229,
                                                    yMinorD);
                     t234_len =
                        mult_expansion_zeroelim(cyminB3_len,
                                                cyminB3,
                                                crSqrdD5_len,
                                                crSqrdD5, t234);
                     t235_len =
                        mult_expansion_zeroelim(cyminD3_len,
                                                cyminD3,
                                                crSqrdB5_len,
                                                crSqrdB5, t235);
                     t233_len =
                        fast_expansion_sum_zeroelim(t235_len, t235,
                                                    t234_len, t234,
                                                    t233);
                     t236_len =
                        mult_expansion_zeroelim(cyminD3_len,
                                                cyminD3,
                                                crSqrdD5_len,
                                                crSqrdD5, t236);
                     yMinorD2_len =
                        fast_expansion_sum_zeroelim(t236_len, t236,
                                                    t233_len, t233,
                                                    yMinorD2);
                     yMinorD3_len =
                        fast_expansion_diff_zeroelim(yMinorD_len,
                                                     yMinorD,
                                                     yMinorD2_len,
                                                     yMinorD2,
                                                     yMinorD3);
                     t238_len =
                        mult_expansion_zeroelim(dyminB3_len,
                                                dyminB3,
                                                drSqrdD5_len,
                                                drSqrdD5, t238);
                     t239_len =
                        mult_expansion_zeroelim(dyminD3_len,
                                                dyminD3,
                                                drSqrdB5_len,
                                                drSqrdB5, t239);
                     t237_len =
                        fast_expansion_sum_zeroelim(t239_len, t239,
                                                    t238_len, t238,
                                                    t237);
                     t240_len =
                        mult_expansion_zeroelim(dyminD3_len,
                                                dyminD3,
                                                drSqrdD5_len,
                                                drSqrdD5, t240);
                     yMinorD4_len =
                        fast_expansion_sum_zeroelim(t240_len, t240,
                                                    t237_len, t237,
                                                    yMinorD4);
                     yMinorD5_len =
                        fast_expansion_sum_zeroelim(yMinorD4_len,
                                                    yMinorD4,
                                                    yMinorD3_len,
                                                    yMinorD3,
                                                    yMinorD5);
                     t242_len =
                        mult_expansion_zeroelim(bzminB3_len,
                                                bzminB3,
                                                brSqrdD5_len,
                                                brSqrdD5, t242);
                     t243_len =
                        mult_expansion_zeroelim(bzminD3_len,
                                                bzminD3,
                                                brSqrdB5_len,
                                                brSqrdB5, t243);
                     t241_len =
                        fast_expansion_sum_zeroelim(t243_len, t243,
                                                    t242_len, t242,
                                                    t241);
                     t244_len =
                        mult_expansion_zeroelim(bzminD3_len,
                                                bzminD3,
                                                brSqrdD5_len,
                                                brSqrdD5, t244);
                     zMinorD_len =
                        fast_expansion_sum_zeroelim(t244_len, t244,
                                                    t241_len, t241,
                                                    zMinorD);
                     t246_len =
                        mult_expansion_zeroelim(czminB3_len,
                                                czminB3,
                                                crSqrdD5_len,
                                                crSqrdD5, t246);
                     t247_len =
                        mult_expansion_zeroelim(czminD3_len,
                                                czminD3,
                                                crSqrdB5_len,
                                                crSqrdB5, t247);
                     t245_len =
                        fast_expansion_sum_zeroelim(t247_len, t247,
                                                    t246_len, t246,
                                                    t245);
                     t248_len =
                        mult_expansion_zeroelim(czminD3_len,
                                                czminD3,
                                                crSqrdD5_len,
                                                crSqrdD5, t248);
                     zMinorD2_len =
                        fast_expansion_sum_zeroelim(t248_len, t248,
                                                    t245_len, t245,
                                                    zMinorD2);
                     zMinorD3_len =
                        fast_expansion_diff_zeroelim(zMinorD_len,
                                                     zMinorD,
                                                     zMinorD2_len,
                                                     zMinorD2,
                                                     zMinorD3);
                     t250_len =
                        mult_expansion_zeroelim(dzminB3_len,
                                                dzminB3,
                                                drSqrdD5_len,
                                                drSqrdD5, t250);
                     t251_len =
                        mult_expansion_zeroelim(dzminD3_len,
                                                dzminD3,
                                                drSqrdB5_len,
                                                drSqrdB5, t251);
                     t249_len =
                        fast_expansion_sum_zeroelim(t251_len, t251,
                                                    t250_len, t250,
                                                    t249);
                     t252_len =
                        mult_expansion_zeroelim(dzminD3_len,
                                                dzminD3,
                                                drSqrdD5_len,
                                                drSqrdD5, t252);
                     zMinorD4_len =
                        fast_expansion_sum_zeroelim(t252_len, t252,
                                                    t249_len, t249,
                                                    zMinorD4);
                     zMinorD5_len =
                        fast_expansion_sum_zeroelim(zMinorD4_len,
                                                    zMinorD4,
                                                    zMinorD3_len,
                                                    zMinorD3,
                                                    zMinorD5);
                     t254_len =
                        scale_expansion_zeroelim(bxminD3_len,
                                                 bxminD3, brx,
                                                 t254);
                     t255_len =
                        scale_expansion_zeroelim(bxminB3_len,
                                                 bxminB3, brxC,
                                                 t255);
                     t253_len =
                        fast_expansion_sum_zeroelim(t255_len, t255,
                                                    t254_len, t254,
                                                    t253);
                     t256_len =
                        scale_expansion_zeroelim(bxminD3_len,
                                                 bxminD3, brxC,
                                                 t256);
                     denomD_len =
                        fast_expansion_sum_zeroelim(t256_len, t256,
                                                    t253_len, t253,
                                                    denomD);
                     t258_len =
                        scale_expansion_zeroelim(cxminD3_len,
                                                 cxminD3, crx,
                                                 t258);
                     t259_len =
                        scale_expansion_zeroelim(cxminB3_len,
                                                 cxminB3, crxC,
                                                 t259);
                     t257_len =
                        fast_expansion_sum_zeroelim(t259_len, t259,
                                                    t258_len, t258,
                                                    t257);
                     t260_len =
                        scale_expansion_zeroelim(cxminD3_len,
                                                 cxminD3, crxC,
                                                 t260);
                     denomD2_len =
                        fast_expansion_sum_zeroelim(t260_len, t260,
                                                    t257_len, t257,
                                                    denomD2);
                     denomD3_len =
                        fast_expansion_diff_zeroelim(denomD_len,
                                                     denomD,
                                                     denomD2_len,
                                                     denomD2,
                                                     denomD3);
                     t262_len =
                        scale_expansion_zeroelim(dxminD3_len,
                                                 dxminD3, drx,
                                                 t262);
                     t263_len =
                        scale_expansion_zeroelim(dxminB3_len,
                                                 dxminB3, drxC,
                                                 t263);
                     t261_len =
                        fast_expansion_sum_zeroelim(t263_len, t263,
                                                    t262_len, t262,
                                                    t261);
                     t264_len =
                        scale_expansion_zeroelim(dxminD3_len,
                                                 dxminD3, drxC,
                                                 t264);
                     denomD4_len =
                        fast_expansion_sum_zeroelim(t264_len, t264,
                                                    t261_len, t261,
                                                    denomD4);
                     denomD5_len =
                        fast_expansion_sum_zeroelim(denomD4_len,
                                                    denomD4,
                                                    denomD3_len,
                                                    denomD3,
                                                    denomD5);
                     t266_len =
                        scale_expansion_zeroelim(yMinorD5_len,
                                                 yMinorD5, pry,
                                                 t266);
                     t267_len =
                        scale_expansion_zeroelim(yMinorB5_len,
                                                 yMinorB5, pryC,
                                                 t267);
                     t265_len =
                        fast_expansion_sum_zeroelim(t267_len, t267,
                                                    t266_len, t266,
                                                    t265);
                     t268_len =
                        scale_expansion_zeroelim(yMinorD5_len,
                                                 yMinorD5, pryC,
                                                 t268);
                     numerD_len =
                        fast_expansion_sum_zeroelim(t268_len, t268,
                                                    t265_len, t265,
                                                    numerD);
                     t270_len =
                        scale_expansion_zeroelim(xMinorD5_len,
                                                 xMinorD5, prx,
                                                 t270);
                     t271_len =
                        scale_expansion_zeroelim(xMinorB5_len,
                                                 xMinorB5, prxC,
                                                 t271);
                     t269_len =
                        fast_expansion_sum_zeroelim(t271_len, t271,
                                                    t270_len, t270,
                                                    t269);
                     t272_len =
                        scale_expansion_zeroelim(xMinorD5_len,
                                                 xMinorD5, prxC,
                                                 t272);
                     numerD2_len =
                        fast_expansion_sum_zeroelim(t272_len, t272,
                                                    t269_len, t269,
                                                    numerD2);
                     numerD3_len =
                        fast_expansion_diff_zeroelim(numerD_len,
                                                     numerD,
                                                     numerD2_len,
                                                     numerD2,
                                                     numerD3);
                     t274_len =
                        mult_expansion_zeroelim(prSqrdB5_len,
                                                prSqrdB5,
                                                denomD5_len,
                                                denomD5, t274);
                     t275_len =
                        mult_expansion_zeroelim(prSqrdD5_len,
                                                prSqrdD5,
                                                denomB5_len,
                                                denomB5, t275);
                     t273_len =
                        fast_expansion_sum_zeroelim(t275_len, t275,
                                                    t274_len, t274,
                                                    t273);
                     t276_len =
                        mult_expansion_zeroelim(prSqrdD5_len,
                                                prSqrdD5,
                                                denomD5_len,
                                                denomD5, t276);
                     numerD4_len =
                        fast_expansion_sum_zeroelim(t276_len, t276,
                                                    t273_len, t273,
                                                    numerD4);
                     t278_len =
                        scale_expansion_zeroelim(zMinorD5_len,
                                                 zMinorD5, prz,
                                                 t278);
                     t279_len =
                        scale_expansion_zeroelim(zMinorB5_len,
                                                 zMinorB5, przC,
                                                 t279);
                     t277_len =
                        fast_expansion_sum_zeroelim(t279_len, t279,
                                                    t278_len, t278,
                                                    t277);
                     t280_len =
                        scale_expansion_zeroelim(zMinorD5_len,
                                                 zMinorD5, przC,
                                                 t280);
                     numerD5_len =
                        fast_expansion_sum_zeroelim(t280_len, t280,
                                                    t277_len, t277,
                                                    numerD5);
                     numerD6_len =
                        fast_expansion_diff_zeroelim(numerD4_len,
                                                     numerD4,
                                                     numerD5_len,
                                                     numerD5,
                                                     numerD6);
                     numerD7_len =
                        fast_expansion_sum_zeroelim(numerD6_len,
                                                    numerD6,
                                                    numerD3_len,
                                                    numerD3,
                                                    numerD7);
                     t282_len =
                        mult_expansion_zeroelim(numerB7_len,
                                                numerB7,
                                                denomD5_len,
                                                denomD5, t282);
                     t283_len =
                        mult_expansion_zeroelim(numerD7_len,
                                                numerD7,
                                                denomB5_len,
                                                denomB5, t283);
                     t281_len =
                        fast_expansion_sum_zeroelim(t283_len, t283,
                                                    t282_len, t282,
                                                    t281);
                     t284_len =
                        mult_expansion_zeroelim(numerD7_len,
                                                numerD7,
                                                denomD5_len,
                                                denomD5, t284);
                     insphereD_len =
                        fast_expansion_sum_zeroelim(t284_len, t284,
                                                    t281_len, t281,
                                                    insphereD);
                     t285_len =
                        fast_expansion_sum_zeroelim(insphereD_len,
                                                    insphereD,
                                                    insphereB_len,
                                                    insphereB,
                                                    t285);
                     if (t285[t285_len - 1] > 0.0)
                        return 1;
                     else if (t285[t285_len - 1] < 0.0)
                        return -1;
                     else
                        return 0;
                  }
               };
            }
         };
      }
   };

}
