#include "allh.h"
#include <algorithm>


bool operator<(const Direction & a, const Direction & b)
{
   return a.angle < b.angle;
}

IntestineRing::IntestineRing() {
  vertex = -1;
  segment = NULL;
}


IntestineCylinder::IntestineCylinder() {
  ring1 = NULL;
  ring2 = NULL;
  segment = NULL;
  subsegment = NULL;
}


IntestineCylinder::IntestineCylinder(IntestineRing * r1, IntestineRing * r2) {
  IntestineCylinder();
  ring1 = r1;
  ring2 = r2;

  segment = r1->segment;
}

void IntestineCylinder::AddSeg(int a, int b) {
  protectedSegments[a].insert(b);
  protectedSegments[b].insert(a);
}

void IntestineCylinder::RemoveSeg(int a, int b) {
  protectedSegments[a].erase(protectedSegments[a].find(b));
  protectedSegments[b].erase(protectedSegments[b].find(a));
}



IntestineSphere::IntestineSphere() {
  vertex = -1;
}

void IntestineSphere::AddArc(int a, int b) {
  protectedArcs[a].insert(b);
  protectedArcs[b].insert(a);
}

void IntestineSphere::RemoveArc(int a, int b) {
  protectedArcs[a].erase(protectedArcs[a].find(b));
  protectedArcs[b].erase(protectedArcs[b].find(a));
}


Intestine::Intestine() {
  
  int a,b,c;
  for (int i=0; i<27; i++) {
    a = (i/9)-1;
    b = ((i/3)%3)-1;
    c = (i%3) - 1;
    if (a != 0 || b != 0 || c != 0) {
      Vector myv(a,b,c);
      myv.normalize();
      sd.push_back(myv);
    }
  }

}

int Intestine::BuildIntestine(PLC * C) {
  int pointsAdded = 0;
  
  this->C = C;

  foreach(int id, C->pointsToProtect) {
    InitSphere(C->P->id2loc[id]);
  }


  // loop through the segments to be protected
  // and protect each one...
  foreach(Mesh * mSeg, C->segmentsToProtect) {
    pointsAdded += ProtectSegment(mSeg);
  }
  

  // loop through the input points and form protecting spheres...
  foreach(IntestineSphere * s, spheres) {
    ProtectInputPoint(s);
  }



  return pointsAdded;

}

double Intestine::AngleFrom(Vector & v1, Vector & v2, Vector & n) {
    double dotprod = v1*v2;
    if (dotprod > 1.0) {
      dotprod = 1.0;
    }
    if (dotprod < -1.0) {
      dotprod = -1.0;
    }    
    if ((v1 ^ v2) * n >= 0.0) {    
      return acos(dotprod) * 180.0 / PI;
    } else {
      return 360.0 - acos(dotprod) * 180.0 / PI;
    }
    return -1.0;
}

// http://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
Vector Intestine::RodriguesFormula(Vector u, Vector v, double theta) {
  theta = theta*PI/180.0;

  return cos(theta)*v + sin(theta)*(u^v) + (u*v)*u*(1-cos(theta));
}

int Intestine::ProtectSegment(Mesh * mSeg) {

  //cout << "Protecting Segment " << mSeg->name << endl;


  int pointsAdded = 0;

  vector<Direction> dirs;

  foreach (Mesh * mPlane, mSeg->parents) {
    if (mPlane->dim == 2) {
      
      
      if (mPlane->bndseg.find(mSeg) == mPlane->bndseg.end()) {
	Vector normal;

	normal = mSeg->dir^mPlane->dir;

	Direction mydir;
	mydir.plane = mPlane;
	mydir.direction = normal;
	dirs.push_back(mydir);

	Direction mydir2;
	mydir2.plane = mPlane;
	mydir2.direction = -1.0*normal;
	dirs.push_back(mydir2);
	


      } else {
	Vector normal;
	mPlane->NormalInFace(mSeg,normal);
	Direction mydir;
	mydir.plane = mPlane;
	mydir.direction = normal;
	
	dirs.push_back(mydir);
      }



    }
  }

  Vector zerodegrees = dirs[0].direction;
  // set the angle
  for (int iD=0; iD<(int)dirs.size(); iD++) {
    dirs[iD].angle = AngleFrom(zerodegrees,dirs[iD].direction,mSeg->dir);
  }

  sort(dirs.begin(), dirs.end());

  // insert dummy directions to ensure the intestine surrounds
  // the segment...
  int nDir = (int)dirs.size();

  for (int iD=1; iD<=nDir; iD++) {
    double diff = 0.0;

    int a,b;
    if (iD < nDir) {
      diff = dirs[iD].angle - dirs[iD-1].angle;
      a = iD - 1;
      b = iD;
    } else {
      diff = 360.0 + dirs[0].angle - dirs[iD-1].angle;
      a = nDir - 1;
      b = 0;
    }
    
    //cout << "Diff " << diff << " " << dirs[iD].angle << endl;
    if (diff > 270) {
      //insert 3 extra directions...

      Direction d1, d2, d3;

      d1.plane = NULL;
      d2.plane = NULL;
      d3.plane = NULL;

      d1.direction = RodriguesFormula(mSeg->dir,dirs[a].direction,diff/4.0);
      d2.direction = RodriguesFormula(mSeg->dir,dirs[a].direction,2.0*diff/4.0);
      d3.direction = RodriguesFormula(mSeg->dir,dirs[a].direction,3.0*diff/4.0);
      
      d1.angle = AngleFrom(zerodegrees,d1.direction,mSeg->dir);
      d2.angle = AngleFrom(zerodegrees,d2.direction,mSeg->dir);
      d3.angle = AngleFrom(zerodegrees,d3.direction,mSeg->dir);

      dirs.push_back(d1);
      dirs.push_back(d2);
      dirs.push_back(d3);

    } else if (diff > 180) {
      //insert 2 extra directions

      Direction d1, d2;

      d1.plane = NULL;
      d2.plane = NULL;

      d1.direction = RodriguesFormula(mSeg->dir,dirs[a].direction,diff/3.0);
      d2.direction = RodriguesFormula(mSeg->dir,dirs[a].direction,2.0*diff/3.0);

      d1.angle = AngleFrom(zerodegrees,d1.direction,mSeg->dir);
      d2.angle = AngleFrom(zerodegrees,d2.direction,mSeg->dir);

      dirs.push_back(d1);
      dirs.push_back(d2);
      
    } else if (diff > 90) {
      Direction d1;

      d1.plane = NULL;

      d1.direction = RodriguesFormula(mSeg->dir,dirs[a].direction,diff/2.0);
      d1.angle = AngleFrom(zerodegrees,d1.direction,mSeg->dir);

      dirs.push_back(d1);
      
    }
  }

  sort(dirs.begin(), dirs.end());


  // insert intestine points into each face (and dummy direction)
  // and form rings and cylinders...
   // check both end points...'
   int end1,in1,n1,n2;
   set < int >nbs1;

   // get a bounding box point for the segment
   set < int >::iterator edgeIt = mSeg->exteriorPoints.begin();
   end1 = *edgeIt;
   edgeIt++;

   mSeg->Neighbors(end1, nbs1);

   if (nbs1.size() != 1) {
      cout << "ERROR: end point has wrong number of neighbors..." <<
      endl;
      exit(0);
   }
   // these are the input points...
   in1 = *(nbs1.begin());

   int cur, prev, next;

   mSeg->SegNeighbors(in1, n1, n2);   

   prev = in1;
   if (n1 == end1) {
     cur = n2; 
   } else {
     cur = n1;
   }

   IntestineRing * prevRing = NULL;

   int ringfirst=-1, prevfirst=-1, midfirst = -1;

   while (C->P->type[cur] != POINT_INPUT) {

     mSeg->SegNeighbors(cur, n1, n2);
     
     if (n1 == prev) {
       next = n2;
     } else {
       next = n1;
     }



     IntestineRing * newRing = GetRing(cur,mSeg,dirs,ringfirst);
     pointsAdded += newRing->r2n.size();

     rings.push_back(newRing);
     

     if (prevRing != NULL) {

       IntestineRing * midRing = MidRing(prevRing,newRing, mSeg, dirs, midfirst);

       // form the cylinder...

       IntestineCylinder * newCyl = new IntestineCylinder(prevRing, newRing);
       cylinders.push_back(newCyl);

       map<int,int>::iterator mIt;
       for(mIt = prevRing->r2n.begin(); mIt!= prevRing->r2n.end(); mIt++) {
	 l2c[mIt->first].insert(newCyl);
       }
       for(mIt = newRing->r2n.begin(); mIt!= newRing->r2n.end(); mIt++) {
	 l2c[mIt->first].insert(newCyl);
       }
       for(mIt = midRing->r2n.begin(); mIt!= midRing->r2n.end(); mIt++) {
	 l2c[mIt->first].insert(newCyl);
       }

       int a = ringfirst;
       int b = prevfirst;
       int c = midfirst;

       do {
	 newCyl->AddSeg(a,c);
	 newCyl->AddSeg(c,b);
	 a = newRing->r2n[a];
	 b = prevRing->r2n[b];
	 c = midRing->r2n[c];
       } while (a != ringfirst);
       

     }

     prevRing = newRing;
     prev = cur;
     cur = next;
     prevfirst = ringfirst;
   }

  return pointsAdded;
}



IntestineRing * Intestine::GetRing(int i, Mesh * mSeg, vector<Direction> & dirs, 
				   int & firstpoint) {

  IntestineRing * myring = new IntestineRing();

  myring->segment = mSeg;

  int n1,n2;

  mSeg->SegNeighbors(i,n1,n2);

  bool endcase = false;
  if (C->P->type[n1] == POINT_INPUT) {
    endcase = true;
  } else if (C->P->type[n2] == POINT_INPUT) {
    endcase = true;
    int tmp = n1;
    n1 = n2;
    n2 = tmp;
  }
  if (endcase) {

      double len = d(C->P->V[n2], C->P->V[i]);

      double dend = d(C->P->V[n1], C->P->V[i]);

      double myratio = sqrt(1 - len * len / (4.0 * dend * dend));

      myring->center = (1-myratio)*C->P->V[n1] + (myratio)*C->P->V[i];
      myring->radius = len/2.0;
  } else {
    myring->center = C->P->V[i];

    myring->radius = max(d(C->P->V[n1], C->P->V[i]),d(C->P->V[n2], C->P->V[i]))/2.0;

  }
  
  // insert the points...

  int first = -1,last = -1,prev = -1;;
 

  foreach(Direction mydir, dirs) {

    Point p = myring->center + myring->radius * mydir.direction;
    
    int loc = C->P->AddPoint(p,i, POINT_RING);
    if (mydir.plane != NULL) {
      C->InsertInMeshAndParents(loc, mydir.plane);    
    } else {
      // point only lives in 3D
      C->InsertInMeshAndParents(loc, C->M);
    }
    if (first == -1) {
      first = loc;
    } else {
      myring->r2n[prev] = loc;
      myring->r2p[loc] = prev;
      last = loc;
    }
    prev = loc;

    l2r[loc] = myring;

    if (endcase) {
      l2s[loc] = l2s[i];
      l2s[loc]->nbors.insert(loc);
    }

  }

  myring->r2n[last] = first;
  myring->r2p[first] = last;  

  firstpoint = first;

  return myring;
}


IntestineRing * Intestine::MidRing(IntestineRing * r1, IntestineRing* r2, Mesh * mSeg, 
					vector<Direction> & dirs, int & firstpoint) {

  IntestineRing * myring = new IntestineRing();

  myring->segment = mSeg;
  
  myring->center = (r1->center + r2->center)/2;
  myring->radius = (r1->radius + r2->radius)/2;
  
  myring->vertex = -1;

  // insert the points...

  int first = -1,last = -1,prev = -1;;
 

  foreach(Direction mydir, dirs) {

    Point p = myring->center + myring->radius * mydir.direction;
    
    int loc = C->P->AddPoint(p,r1->vertex, POINT_CYLINDER);
    if (mydir.plane != NULL) {
      C->InsertInMeshAndParents(loc, mydir.plane);    
    } else {
      // point only lives in 3D
      C->InsertInMeshAndParents(loc, C->M);
    }
    if (first == -1) {
      first = loc;
    } else {
      myring->r2n[prev] = loc;
      myring->r2p[loc] = prev;
      last = loc;
    }
    prev = loc;

  }

  myring->r2n[last] = first;
  myring->r2p[first] = last;  

  firstpoint = first;

  return myring;
}


void Intestine::InitSphere(int iP) {
  
  IntestineSphere * newSphere = new IntestineSphere();

  newSphere->vertex = iP;

  foreach(Mesh * m, C->P->l2m[iP]) {

    if (m->dim == 1) {
      // get the neighbor and add to the sphere...
      int n1,n2;

      m->SegNeighbors(iP,n1,n2);
      if (C->P->type[n1] != POINT_BBOX) {
	newSphere->nbors.insert(n1);
	l2s[n1] = newSphere;
      } else if (C->P->type[n2] != POINT_BBOX) {
	newSphere->nbors.insert(n2);
	l2s[n2] = newSphere;
      } else {
	cout <<"ERROR: Intestine::InitSphere -- no neighbor on segment" << endl;
      }

    }
  }
  
  if (newSphere->nbors.size() == 0) {
    cout << "ERROR: no points on initial IntestineSphere." << endl;
  } else {
    newSphere->radius = d(C->P->V[iP],C->P->V[*(newSphere->nbors.begin())]);
  }

  //cout << newSphere->radius << " " << newSphere->nbors.size() << endl;

  spheres.push_back(newSphere);

}

//
// Add points to the sphere to prevent any large gaps...
//
int Intestine::ProtectInputPoint(IntestineSphere * s) {
  int pointsAdded = 0;

  foreach (Vector thisDir, sd) {
    // should I insert a point in this direction?
    
    bool pointNeeded = true;
    foreach(int iN, s->nbors) {
      Vector otherDir = C->P->V[iN] - C->P->V[s->vertex];
      otherDir.normalize();

      if (otherDir*thisDir > sqrt(.5)) {
	pointNeeded = false;
	break;
      }

    }

    if (pointNeeded) {
      // insert the point...
      Point p = C->P->V[s->vertex] + s->radius*thisDir;
      int loc = C->P->AddPoint(p,s->vertex, POINT_SPHERE);
      // point only lives in 3D
      C->InsertInMeshAndParents(loc, C->M);
      s->nbors.insert(loc);

      l2s[loc] = s;

      pointsAdded++;
      //cout << "Sphere Insert " << s->vertex << " " << p << " " << loc << s->radius << endl;
          
    }
  }

  // FIXME: set up protectedArcs!!!!
  // NOT GENERAL : only handles simple case
  foreach(int iN, s->nbors) {
    Mesh * mLow = C->P->LowestDimensionalContainingMesh(iN);

    foreach(Mesh * m, C->P->l2m[iN]) {
      if (m->dim == 2) { // only deal with planes...
	if ( (mLow->dim == 1 && 
	      C->segmentsToProtect.find(mLow) == C->segmentsToProtect.end()) ||
	     mLow->dim == 2) {

	  // find the matching point...
	    foreach(int jN, s->nbors) {

	      if (iN != jN) {
		Mesh * mLow1 = C->P->LowestDimensionalContainingMesh(jN);
		if (m->Exists(jN) && 
		    (mLow1->dim == 2 || 
		     C->segmentsToProtect.find(mLow1) == C->segmentsToProtect.end() ) ) {
		  s->protectedArcs[iN].insert(jN);
		  //cout << s->vertex << " :: " << iN <<" " << jN << endl;
		}
	      }

	    }
	  
	}
      }
    }
  }


  return pointsAdded;
}


bool Intestine::IsIntestineEdge(int a, int b) {
  if (C->P->type[a] != POINT_RING && C->P->type[a] != POINT_CYLINDER 
      && C->P->type[a] != POINT_SPHERE && C->P->type[a] != POINT_SEGMENT) {
    return false;
  }
  if (C->P->type[b] != POINT_RING && C->P->type[b] != POINT_CYLINDER 
      && C->P->type[b] != POINT_SPHERE && C->P->type[b] != POINT_SEGMENT) {
    return false;
  }

  if (l2r.find(a) != l2r.end() && l2r.find(b) != l2r.end()) {
    if (l2r[a] == l2r[b]) {
      return true;
    }
  }
  if (l2s.find(a) != l2s.end() && l2s.find(b) != l2s.end()) {
    if (l2s[a] == l2s[b]) {
      return true;
    }
  }
  if (l2c.find(a) != l2c.end() && l2c.find(b) != l2c.end()) {
  
    foreach (IntestineCylinder * c, l2c[a]) {
      if (l2c[b].find(c) != l2c[b].end()) {
	return true;
      }
    }
  }

  return false;
}


bool Intestine::IsIntestineRingEdge(int a, int b) {
  if (C->P->type[a] != POINT_RING) {
    return false;
  }
  if (C->P->type[b] != POINT_RING) {
    return false;
  }

  if (l2r.find(a) != l2r.end() && l2r.find(b) != l2r.end()) {
    if (l2r[a] == l2r[b]) {
      if (l2r[a]->r2n[a] == b || l2r[a]->r2p[a] == b) {
	return true;
      }
    }
  }
  return false;
}

bool Intestine::IsIntestineProtectedEdge(int a, int b) {
  // find a common cylinder...
  if (l2c.find(a) != l2c.end() && l2c.find(b) != l2c.end()) {
  
    foreach (IntestineCylinder * c, l2c[a]) {
      if (l2c[b].find(c) != l2c[b].end()) {
	if (c->protectedSegments.find(a) != c->protectedSegments.end()) {
	  if (c->protectedSegments[a].find(b) != c->protectedSegments[a].end()) {
	    return true;
	  }
	}
      }
    }
  }
  return false;
}

bool Intestine::IsIntestineProtectedArc(int a, int b) {
  if (l2s.find(a) != l2s.end() && l2s.find(b) != l2s.end()) {
    if (l2s[a] == l2s[b]) {
      if (l2s[a]->protectedArcs.find(a) != l2s[a]->protectedArcs.end()) {
	if (l2s[a]->protectedArcs[a].find(b) != l2s[a]->protectedArcs[a].end()) {
	  return true;
	}
      }
    }
  }
  return false;
}


bool Intestine::IsIntestineSphereFace(int a, int b, int c) {
  if (l2s.find(a) != l2s.end()) {

    if (l2s.find(b) != l2s.end() && l2s.find(c) != l2s.end()) {
      if (l2s[a] == l2s[b] && l2s[a] == l2s[c]) {
	return true;
      }
    }
  }
  return false;
}

bool Intestine::IsIntestineCylinderFace(int a, int b, int c) {

  if (l2c.find(a) != l2c.end()) {
    if (l2c.find(b) != l2c.end() && l2c.find(c) != l2c.end()) {
      foreach(IntestineCylinder * cyl, l2c[a]) {
	if (l2c[b].find(cyl) != l2c[b].end() && 
	    l2c[c].find(cyl) != l2c[c].end()) {
	  return true;
	}
      }
    }
  }

  return false;
}

bool Intestine::IsIntestineFace(int a, int b, int c) {
  return IsIntestineCylinderFace(a,b,c) || IsIntestineSphereFace(a,b,c);
}

//
// Get the midpoint of the arc between a and b...
//
Point Intestine::IntestineRingSplitPoint(int a, int b) {

  Point p;
  
  IntestineRing * thering = ContainingIntestineRing(a,b);

  if (thering != NULL) {

    Vector v = (.5*(C->P->V[a]+C->P->V[b])) - thering->center;
    v.normalize();
    p = thering->center + thering->radius*v;
  } else {
    cout << "WARNING: IntestineRingSplitPoint Fails." << endl;
  }

  return p;
}


Point Intestine::IntestineFaceSplitPoint(int a, int b, int c) {
  Point p;

  // is this a sphere face?
  IntestineSphere * thesphere = ContainingIntestineSphere(a,b,c);
  if (thesphere != NULL) {
    //project the circumcenter to the surface of the sphere...
    Simplex s(a,b,c);

    p = C->G->circumcenter(&s);
    
    Vector v = p - C->P->V[thesphere->vertex];
    v.normalize();

    p = C->P->V[thesphere->vertex] + thesphere->radius*v;

  } else {
    IntestineCylinder * thecyl = ContainingIntestineCylinder(a,b,c);
    if (thecyl != NULL) {
      //cout << a << " " << b << " " << c << endl;
      //cout << "Split Cyl " << endl;
	

      Simplex s(a,b,c);
      p = C->G->circumcenter(&s);
      
      Vector v1 = p - thecyl->ring1->center;
      Vector v2 = thecyl->ring2->center - thecyl->ring1->center;

      v2.normalize();
      Point p1 = (v1*v2)*v2 + thecyl->ring1->center;
      Vector v3 = p - p1;
      v3.normalize();

      double ratio = d(p1,thecyl->ring1->center)/d(thecyl->ring1->center,thecyl->ring2->center);

      p = p1  + (ratio*thecyl->ring2->radius + (1-ratio)*thecyl->ring1->radius)*v3;

      //cout << "Cyl Split Point " << p << endl;

    } else {
      cout << "WARNING: IntestineFaceSplitPoint failing" << endl;
    }
  }
  
  return p;
}


Point Intestine::IntestineArcSplitPoint(int a, int b) {
  Point p;

  IntestineSphere * thesphere = ContainingIntestineSphere(a,b);
  if (thesphere != NULL) {
    //project the circumcenter to the surface of the sphere...

    p = .5*(C->P->V[a]+C->P->V[b]);
    
    Vector v = p - C->P->V[thesphere->vertex];
    v.normalize();

    p = C->P->V[thesphere->vertex] + thesphere->radius*v;
  } 

  //cout << "Arc Split Point " << p << endl;

  return p;
}

IntestineSphere * Intestine::ContainingIntestineSphere(int a, int b) {
    if (l2s.find(a) != l2s.end()) {

    if (l2s.find(b) != l2s.end()) {
      if (l2s[a] == l2s[b]) {
	return l2s[a];
      }
    }
  }

  return NULL;
}

IntestineSphere * Intestine::ContainingIntestineSphere(int a, int b, int c) {
    if (l2s.find(a) != l2s.end()) {

    if (l2s.find(b) != l2s.end() && l2s.find(c) != l2s.end()) {
      if (l2s[a] == l2s[b] && l2s[a] == l2s[c]) {
	return l2s[a];
      }
    }
  }

  return NULL;
}
  
IntestineCylinder * Intestine::ContainingIntestineCylinder(int a, int b, int c) {
  if (l2c.find(a) != l2c.end()) {
    if (l2c.find(b) != l2c.end() && l2c.find(c) != l2c.end()) {
      foreach(IntestineCylinder * cyl, l2c[a]) {
	if (l2c[b].find(cyl) != l2c[b].end() && 
	    l2c[c].find(cyl) != l2c[c].end()) {
	  return cyl;
	}
      }
    }
  }

  return NULL;
}

IntestineCylinder * Intestine::ContainingIntestineCylinder(int a, int b) {


  if (l2c.find(a) != l2c.end()) {
  //if (C->P->type[a] == POINT_RING || C->P->type[a] == POINT_CYLINDER) {
    if (l2c.find(b) != l2c.end()) {
      //if (C->P->type[b] == POINT_RING || C->P->type[b] == POINT_CYLINDER) {
      if (l2c[a].size() == 0 || l2c[b].size() == 0) {
	return NULL;
      }
      foreach(IntestineCylinder * cyl, l2c[a]) {
	if (l2c[b].find(cyl) != l2c[b].end() ) {
	  return cyl;
	}
      }
    }
  }
  return NULL;
}



IntestineRing * Intestine::ContainingIntestineRing(int a, int b) {
  if (l2r.find(a) != l2r.end() && l2r.find(b) != l2r.end()) {
    if (l2r[a] == l2r[b]) {
      return l2r[a];
    }
  }  
  return NULL;
}

int Intestine::SplitIntestine(int a, int b, int c, Mesh * &m) {
  int loc = -1;

  //cout << "SplitIntestine " << a << " " << b <<" " << c  << endl;

  if ( c == -1) {

    if (C->M->intestine->IsIntestineRingEdge(a,b)) {
      IntestineRing * ir = ContainingIntestineRing(a,b);
      
      Point p = IntestineRingSplitPoint(a,b);      
      loc = C->P->AddPoint(p, a, POINT_RING);
      m = C->M;

      if (ir->r2n[a] == b) {
	ir->r2n[a] = loc;
	ir->r2n[loc] = b;
	ir->r2p[b] = loc;
	ir->r2p[loc] = a;	
      } else {
	ir->r2p[a] = loc;
	ir->r2p[loc] = b;
	ir->r2n[b] = loc;
	ir->r2n[loc] = a;	
      }

      CopyContainment(a,loc);

    } else if (C->M->intestine->IsIntestineProtectedEdge(a,b)) {

      IntestineCylinder * ic = ContainingIntestineCylinder(a,b);
      
      Point p = .5*(C->P->V[a] + C->P->V[b]); 
      loc = C->P->AddPoint(p, a, POINT_CYLINDER);

      // find lowest dimensional mesh...
      
      m = C->P->LowestDimensionalContainingMesh(a);

      ic->RemoveSeg(a,b);
      ic->AddSeg(a,loc);
      ic->AddSeg(b,loc);

      l2c[loc].insert(ic);

      
    } else if (C->M->intestine->IsIntestineProtectedArc(a,b)) {
      IntestineSphere * is = ContainingIntestineSphere(a,b);
      
      Point p = IntestineArcSplitPoint(a,b);
      loc = C->P->AddPoint(p, a, POINT_SPHERE);

      // find lowest dimensional mesh containing both a and b...
      //m = C->P->LowestDimensionalContainingMesh(a);
      m = C->M;
      int dim = 3;
      foreach( Mesh * m1, C->P->l2m[a]) {
	if (m1->dim < dim && m1->Exists(b)) {
	  m = m1;
	  dim = m1->dim;
	}
      }

      is->RemoveArc(a,b);
      is->AddArc(a,loc);
      is->AddArc(b,loc);

      l2s[loc] = is;

    }

  } else {

    if (IsIntestineFace(a,b,c)) {

      m = C->M;

      Point p = IntestineFaceSplitPoint(a,b,c);

      IntestineSphere * is = ContainingIntestineSphere(a,b,c);
      
      if (is != NULL) {

	loc = C->P->AddPoint(p, a, POINT_SPHERE);

	l2s[loc] = is;
      } else {
	IntestineCylinder * ic = ContainingIntestineCylinder(a,b,c);
	if (ic != NULL) {

	  loc = C->P->AddPoint(p, a, POINT_CYLINDER);

	  l2c[loc].insert(ic);
	}
      }
    }
    
  }
  if (m != NULL) {
      //cout << m->name << endl;
  }

  return loc;
}


void Intestine::CopyContainment(int old, int n) {
  if (l2r.find(old) != l2r.end()) {
    l2r[n] = l2r[old];
  }

  if (l2s.find(old) != l2s.end()) {
    l2s[n] = l2s[old];
  }

  if (l2c.find(old) != l2c.end()) {
    foreach (IntestineCylinder * c, l2c[old]) {
      l2c[n].insert(c);
    }
  }


}


bool Intestine::IsIntestineProtected(int a, int b) {
  return IsIntestineProtectedArc(a,b) || IsIntestineRingEdge(a,b) || IsIntestineProtectedEdge(a,b);
}

void Intestine::IntestineArcsSegsRings(int i, set<int> & nbs) {
  if (l2r.find(i) != l2r.end()) {
    nbs.insert(l2r[i]->r2n[i]);
    nbs.insert(l2r[i]->r2p[i]);
  }

  if (l2s.find(i) != l2s.end()) {
    if (l2s[i]->protectedArcs.find(i) != l2s[i]->protectedArcs.end()) {
      foreach (int j, l2s[i]->protectedArcs[i]) {
	nbs.insert(j);
      }
    }
  }  


  if (l2c.find(i) != l2c.end()) {
    foreach(IntestineCylinder * c, l2c[i]) {
      if (c->protectedSegments.find(i) != c->protectedSegments.end()) {
	foreach( int j, c->protectedSegments[i]) {
	  nbs.insert(j);
	}
      }
    }
  }
}
