#include <iostream>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include <string>
#include <list>

#include "allh.h"

typedef pair < int, Simplex * >I2S;
typedef pair < int, int >duple;

Mesh::Mesh(PLC * complex)
{
   C = complex;
   std::ostringstream oss1;
   gid = C->GetNextGID();
   oss1 << "Unlabeled-Mesh-" << gid;
   name = oss1.str();
   col = NULL;
   intestine = NULL;
   bnd = NULL;
   np = 0;
   ns = 0;
   convex = true;
   protect = false;
}


// this builds a box around the mesh
// and creates its initial triangulation
void Mesh::Initialize(vector < Simplex * >&initialSimplices)
{

   // FIXME: add dim = 2 and dim = 1 initialization routine which
   //      using dir and base and some size info to make a box...
   if (dim == 2) {
      istringstream INIT;
      string initial =
         "8 1 1 0 1 -1 1 -1 0 -1 -1 0 -1 1 -1 1 0 6 0 1 7 1 2 3 3 4 5 5 6 7 7 1 3 3 5 7";
      INIT.str(initial);
      INIT >> np;

      for (int i = 0; i < np; i++) {
         double d1, d2;

         INIT >> d1 >> d2;
         Point tmp(d1, d2, 0.0);
         tmp = tmp * bbox + base;
         int myl = C->P->AddPoint(tmp, POINT_BBOX);
         C->P->l2m[myl].insert(this);
      }
      INIT >> ns;

      for (int i = 0; i < ns; i++) {
         int a, b, c;
         INIT >> a >> b >> c;

         Simplex *s = C->SP->NewSimplex(a, b, c);
         SetRadiusAndCC(s);
         initialSimplices.push_back(s);
         Setp2s(s);
      }
      // link the tets together with an n^2 search...
      for (int i = 0; i < ns; i++) {
         Simplex *n1, *n2, *n0, *n3;
         n0 = NULL;
         n1 = NULL;
         n2 = NULL;
         n3 = NULL;

         for (int j = 0; j < ns; j++) {

            if (j != i) {
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[1])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[2]) ) {
                  n0 = initialSimplices[j];

               }
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[0])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[2]) ) {
                  n1 = initialSimplices[j];
               }
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[0])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[1])) {
                  n2 = initialSimplices[j];
               }
            }
         }
         // order important here...
         initialSimplices[i]->link(n0, n1, n2, n3);
      }
      if (C->logfileisopen) {
         (*(C->LOGFILE)) << "Bounding box created with " << np <<
         " vertices and " << ns << " tets." << endl;
      }
   } else if (dim == 3) {
      istringstream INIT;
      // hardcoding of the initial delaunay triangulation
      string initial =
         "26 -1 -1 -1 0 -1 -1 1 -1 -1 -1 0 -1 0 0 -1 1 0 -1 -1 1 -1 0 1 -1 1 1 -1 -1 -1 0 0 -1 0 1 -1 0 -1 0 0 1 0 0 -1 1 0 0 1 0 1 1 0 -1 -1 1 0 -1 1 1 -1 1 -1 0 1 0 0 1 1 0 1 -1 1 1 0 1 1 1 1 1 44 1 3 12 0 1 3 4 12 1 10 4 12 1 9 10 12 1 9 12 0 7 3 4 12 7 3 12 6 7 15 4 12 7 14 15 12 7 14 12 6 16 7 8 4 16 7 15 4 16 13 15 4 16 5 8 4 16 5 13 4 11 1 4 2 11 1 10 4 11 13 10 4 11 5 4 2 11 5 13 4 21 10 4 12 21 15 4 12 21 13 15 4 21 13 10 4 20 14 15 23 20 14 15 12 20 24 21 15 20 21 15 12 20 24 15 23 18 9 12 17 18 9 10 12 18 21 10 12 18 20 21 12 18 20 12 17 22 11 19 10 22 11 13 10 22 21 13 10 22 18 19 10 22 18 21 10 22 16 15 25 22 16 13 15 22 21 13 15 22 24 15 25 22 24 21 15";
      INIT.str(initial);

      INIT >> np;

      int start = C->P->V.size();

      for (int i = 0; i < np; i++) {
         double d1, d2, d3;


         INIT >> d1 >> d2 >> d3;
         Point tmp(d1, d2, d3);
         tmp = tmp * bbox + base;
         int myl = C->P->AddPoint(tmp, POINT_BBOX);
         C->P->l2m[myl].insert(this);
      }
      INIT >> ns;

      for (int i = 0; i < ns; i++) {
         int a, b, c, d;
         INIT >> a >> b >> c >> d;

         Simplex *s = C->SP->NewSimplex(a+start, b+start, c+start, d+start);
         SetRadiusAndCC(s);
         initialSimplices.push_back(s);
         Setp2s(s);
      }

      // link the tets together with an n^2 search...
      for (int i = 0; i < ns; i++) {
         Simplex *n1, *n2, *n3, *n0;
         n0 = NULL;
         n1 = NULL;
         n2 = NULL;
         n3 = NULL;

         for (int j = 0; j < ns; j++) {

            if (j != i) {
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[1])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[2])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[3])) {
                  n0 = initialSimplices[j];

               }
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[0])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[2])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[3])) {
                  n1 = initialSimplices[j];
               }
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[0])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[1])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[3])) {
                  n2 = initialSimplices[j];

               }
               if (initialSimplices[j]->
                   contains(initialSimplices[i]->v[0])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[1])
                   && initialSimplices[j]->
                   contains(initialSimplices[i]->v[2])) {
                  n3 = initialSimplices[j];

               }
            }
         }
         // order important here...
         initialSimplices[i]->link(n0, n1, n2, n3);
      }

      if (C->logfileisopen) {
         (*(C->LOGFILE)) << "Bounding box created with " << np <<
         " vertices and " << ns << " tets." << endl;
      }
   } else {
      cout << "ERROR: Initialize called on a mesh of dimension " << dim
           << endl;
      exit(0);
   }
}

void Mesh::Setp2s(Simplex * s)
{
   for (int i = 0; i < s->size(); i++) {
      p2s[s->v[i]] = s;
      if (s->v[i] == -1) {
         cout << "ERROR: bad simplex in p2s" << endl;
         cout << s << endl;
         C->P->print(s);
         exit(0);
      }
   }
}

Simplex *Mesh::Locate(int i)
{
   return Locate(C->P->V[i], C->P->parent[i]);
}

Simplex *Mesh::Locate(int i, int hint)
{
   return Locate(C->P->V[i], hint);
}

Simplex *Mesh::Locate(int i, Simplex * s)
{
   return Locate(C->P->V[i], s);
}

Simplex *Mesh::Locate(Point p)
{
   return Locate(p, p2s.begin()->first);
}


Simplex *Mesh::Locate(Point p, Simplex * s)
{

   return Locate(p, s->v[0]);

}

// new locate, if this works, need to optimize a little for efficiency...
Simplex *Mesh::Locate(Point p, int hint)
{

   if (hint == -1) {
      hint = p2s.begin()->first;
   }

   int numSteps = 0;

   int curLoc = hint;

   int finished = 0;
   while (finished == 0) {

      finished = 1;
      set < int >nbs;
      double mydist = d2(p, C->P->V[curLoc]);
      Neighbors(curLoc, nbs);
      foreach(int iNb, nbs) {
         double dtmp = d2(p, C->P->V[iNb]);
         if (dtmp < mydist - C->config->epsilon) {
	   //if (dtmp < mydist - C->config->epsilon*C->config->epsilon) {
            finished = 0;
            curLoc = iNb;
            mydist = dtmp;
         }
      }
      numSteps++;
   }

   C->stats->AddLocate(numSteps);

   set < Simplex * >star;

   Simplices(curLoc, star);

   foreach(Simplex * ss, star) {
       //if (C->G->inSphereStrictNoSliver(p, ss)) {
     //if (C->G->inSphereStrict(p, ss) && C->G->inSphereReallyStrict(p, ss) ) {
      if (C->G->inSphereStrict(p, ss) ) {
         return ss;
      }
   }

   // beyond this point should never occur...

   cout << "WARNING: Locate Failed" << endl;

   cout << p << " " << hint << " " << C->P->V[hint] << " " << name << endl;
   cout << curLoc << " " << C->P->V[curLoc] 
	<< " " << C->P->type[curLoc] << endl;
   cout << "Dist " << d2(p, C->P->V[curLoc]) << endl;
   cout << "N Points " << C->P->V.size() << endl;
   cout << p2s[curLoc] << " " << (void*)p2s[curLoc] << endl;


   cout << "Cavity " << endl;

   foreach(Simplex * s, star) {
      cout << s << " " << C->G->inSphereStrict(p, s) << endl;
      cout << C->P->V[s->v[0]] << " " << C->P->V[s->v[1]] << " " << C->
      P->V[s->v[2]] << " " << C->P->V[s->v[3]] << endl;
      cout << s->n[0] << " " << s->n[1] << " " << s->n[2] << " " << s->
      n[3] << endl;
      cout << s->cc << endl;
   }
   cout << "done " << endl;

   set < int >nbs;
   Neighbors(curLoc, nbs);
   foreach(int iNb, nbs) {
      cout << "Dist " << iNb << " " << d2(p, C->P->V[iNb]) << endl;

      set < Simplex * >star;

      Simplices(iNb, star);

      foreach(Simplex * ss, star) {
         if (C->G->inSphereStrict(p, ss)) {
            cout << "WARNING: Locate Issue..." << endl;
            return ss;
            //cout << "I am in something here..." << endl;
         }
         //cout << ss << " " << C->G->InSphereVal(p,ss)  << endl;
      }

   }
   return NULL;
}



void Mesh::InsertPoint(int i)
{

    /*
    if (C->config->inexactsliverinsphere ||
	C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = true;
    }
    */
   // cout << "Insert 0" << endl;
   InsertPoint(i, Locate(i));
}

void Mesh::InsertPoint(int i, set < Simplex * >&oldcavity,
                       set < Simplex * >&newcavity)
{
    /*
    if (C->config->inexactsliverinsphere ||
	C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = true;
    }
    */

   // cout << "Insert 0" << endl;
   InsertPoint(i, Locate(i), oldcavity, newcavity);
}

void Mesh::InsertPoint(int i, int hint)
{
    /*
    if (C->config->inexactsliverinsphere ||
	C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = true;
    }
    */

    //if (C->P->type[i] == POINT_COLLAR) {
    //C->G->noslivers = true;
    //}
   // cout << "Insert 0" << endl;
   InsertPoint(i, Locate(i, hint));
}

void Mesh::InsertPoint(int i, int hint, set < Simplex * >&oldcavity,
                       set < Simplex * >&newcavity)
{
    /*
    if (C->config->inexactsliverinsphere ||
	C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = true;
    }
    */

    //if (C->P->type[i] == POINT_COLLAR) {
//	C->G->noslivers = true;
    //}
   // cout << "Insert 0" << endl;
   InsertPoint(i, Locate(i, hint), oldcavity, newcavity);
}



void Mesh::InsertPoint(int i, Simplex * s)
{
    /*
    if (C->config->inexactsliverinsphere ||
	C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = true;
    }
    */
    //if (C->P->type[i] == POINT_COLLAR) {
    //C->G->noslivers = true;
    //}

   // cout << "Insert 0" << endl;
   set < Simplex * >oldcavity;
   set < Simplex * >newcavity;
   InsertPoint(i, s, oldcavity, newcavity);
}

// InsertPoint
//
// Add point i to the mesh.  Returns both the new and old simplices.
// Does not free the old memory.  Does update p2s.
//
void Mesh::InsertPoint(int i, Simplex * s, set < Simplex * >&oldcavity,
                       set < Simplex * >&newcavity)
{
    /*
    if (C->config->inexactsliverinsphere ||
	C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = true;
    }
    */
    //if (i == 42927) {
    //cout << name << endl;
    //}

   InsertPointNoP2S(i, s, oldcavity, newcavity);

   C->P->l2m[i].insert(this);

   //cout << "Insert " << i << " " << C->P->V[i] << " " << C->P->type[i] << " "<< name << endl;

   foreach(Simplex * s1, newcavity) {
      Setp2s(s1);
   }
   int jj = i;

   foreach(Simplex * s2, oldcavity) {
       for (int i=0; i<s2->size(); i++) {
	   if (p2s[s2->v[i]] == s2) {
	     cout << endl << "p2s now broken " << s2 << " " << s2->v[i] << endl;
	     cout << name << " " << jj << " " << s->v[i] << endl;
	       cout << C->P->V[jj] << " " << C->P->V[s->v[i]] << endl;
	       cout << C->P->type[jj] << " " << C->P->type[s->v[i]] << endl;
	       cout << d2(C->P->V[jj],C->P->V[s->v[i]]) << endl;
	       cout << "New Cavity" << endl;
 	       foreach(Simplex * s3, newcavity) {
 		   cout << s3 << endl;
	       }
	       cout << "End New Cavity" << endl;

	       cout << "Old Cavity" << endl;
 	       foreach(Simplex * s4, oldcavity) {
 		   cout << s4 << endl;
	       }
	       cout << "End Old Cavity" << endl;
	       exit(0);
	   }
       }
   }

    if (C->P->type[i] == POINT_COLLAR) {
	C->G->noslivers = false;
    }
    if (C->P->type[i] == POINT_COLLAR ||
	C->P->type[i] == POINT_FACE ||
	C->P->type[i] == POINT_3D) {
	C->G->noslivers = false;
    }



}


// InsertPointNoP2S
//
// Add point i to the mesh.  Returns both the new and old simplices.
// Does not free the old memory or update p2s.
// Note: not updating p2s needed for current remove operation...
//
void Mesh::InsertPointNoP2S(int i, Simplex * s,
                            set < Simplex * >&oldcavity,
                            set < Simplex * >&newcavity)
{

   // remove this? 2-18-08
   if (dim == 3) {
      // make sure I am inside the bounding box.  If not
      // print some error messages...
      if (C->P->V[i].xyz[0] > base.xyz[0] + bbox + EPSILON ||
          C->P->V[i].xyz[1] > base.xyz[1] + bbox + EPSILON ||
          C->P->V[i].xyz[2] > base.xyz[2] + bbox + EPSILON ||
          C->P->V[i].xyz[0] < base.xyz[0] -bbox - EPSILON ||
          C->P->V[i].xyz[1] < base.xyz[1] -bbox - EPSILON ||
          C->P->V[i].xyz[2] < base.xyz[2] -bbox - EPSILON) {

         cout << "ERROR: Attempting to insert outside bbox:" << C->P->
         V[i] << ", bbox size " << bbox << endl;
         cout << "Point id " << i << endl;
         cout << C->P->V[i].xyz[1] - (bbox + EPSILON) << endl;
         cout << "Mesh name " << name << endl;
         cout << s << endl;
         C->P->print(s);

         vector < double >angles;
         C->G->DihedralAngles(s, angles);


         exit(0);
      }
   }
   //cout << "Insert " << i << " " << C->P->V[i] << " into " << name << endl;

   //cout << "Insert 1" << endl;

   //cout << "Cavity "; printMe(s);


   set < Face > bndry;

   Cavity(i, s, oldcavity, bndry);


   //cout << "Insert 3" << endl;
   Stitch(i, bndry, newcavity);

   //cout << "Insert 4" << endl;

   //update the number of points info...
   np++;
   ns += newcavity.size() - oldcavity.size();


   if (bnd != NULL && convex == false) {

      // is this a boundary point?
      if (C->P->type[i] == POINT_SEGMENT) {
         Mesh *seg = C->P->LowestDimensionalContainingMesh(i);

         if (bnd->segments.find(seg) != bnd->segments.end()) {
            // this is a new boundary point...

            set < int >nbs;
            seg->Neighbors(i, nbs);

            if (nbs.size() != 2) {
               cout <<
               "ERROR in InsertPointNoP2S handling nonconvex boundary."
                    << endl;
               exit(0);
            }
            int n1, n2;
            set < int >::iterator it = nbs.begin();
            n1 = *it;
            it++;
            n2 = *it;


	    //cout << name << " " << i << " " << n1 << " " << n2 << endl;

            //cout << "BND " << i << " : " << n1 << " " << n2 << endl;
            //cout << C->P->V[i] << C->P->V[n1] << C->P->V[n2] << endl;
            bnd->InsertPoint(i, n1, n2);
         }
      }


   }

}


// grab the cavity and it's boundary...
void Mesh::Cavity(int p, Simplex * t, set < Simplex * >&cavity,
                  set < Face > &bndry)
{
   return Cavity(C->P->V[p], t, cavity, bndry);
}

// Cavity
//
// This function returns the cavity and the bounary of the cavity associated with
// point p (which is not necessarily in the mesh yet...
//
//
void Mesh::Cavity(Point & p, Simplex * t, set < Simplex * >&cavity,
                  set < Face > &bndry)
{

   //if (t->size() != dim + 1) {
   //    cout << "ERROR: bad simplex dimention. " << t->size() << " " << dim << " " << name << endl;
   //}

   if (!C->G->inSphereStrict(p,t)) {
       //if (!C->G->inSphereStrictNoSliver(p,t)) {
      cout << "ERROR: inSphereStrict failing in Cavity()" << endl;
      cout << p << " " << t << endl;
      C->P->print(t);
   }


   vector < Simplex * >tocheck;

   tocheck.push_back(t);

   cavity.insert(t);

   //cout << "Cavity 1" << endl;
   while (tocheck.size() > 0) {

      Simplex *myt = tocheck[tocheck.size() - 1];
      tocheck.pop_back();

      //if (myt->size() != dim + 1) {
      //  cout << "ERROR: Cavity simplex size problem." << endl;
      //}


      for (int i = 0; i < dim + 1; i++) {
         if (myt->n[i] == NULL) {

            // error check?
            /*
               for (int k=0; k<dim+1; k++) {


               if (i != k && C->P->type[myt->v[k]] != POINT_BBOX) {
               cout << "ERROR: cavity issue." << endl;
               cout << myt << " " << k << endl;
               for (int j=0; j<=dim; j++) {
               cout << ":: " << j << " " << myt->v[j] << " " << C->P->type[myt->v[j]] << " " << C->P->V[myt->v[j]] << endl;
               }
               cout << C->G->circumradius(myt) << " " <<  C->G->circumcenter(myt) << endl;

               cout << name << " " << myt->f << endl;

               for (int ii=0; ii<dim+1; ii++) {
               cout << myt->n[i] << endl;
               }

               cout << "-------------" << endl;
               foreach (Simplex * ss, cavity) {
               cout << ss << endl;
               }

               exit(0);
               }

               }
             */

            // create a dummy simplex representing the boundary...

            int a, b, c;
            a = myt->v[0];
            b = myt->v[1];
            c = myt->v[2];
            if (i == 0) {
               a = myt->v[dim];
            } else if (i == 1) {
               b = myt->v[dim];
            } else if (i == 2) {
               c = myt->v[dim];
            }
            //Simplex * newt = new Simplex(a,b,c,-1);
            Simplex *newt = C->SP->NewSimplex(a, b, c, -1);

            Face pr;
            pr.s = newt;
            pr.v = -1;
            // no need to link this simplex to anything...
            bndry.insert(pr);

         } else {
	     //if (C->G->inSphereStrictNoSliver(p, myt->n[i])) {
	     if (C->G->inSphereStrict(p, myt->n[i])) {
               if (cavity.find(myt->n[i]) == cavity.end()) {
                  // add neighbor to the cavity and queue to check...
                  cavity.insert(myt->n[i]);
                  tocheck.push_back(myt->n[i]);
               }
            } else {
               // this is on the boundary
               Face pr;
               pr.s = myt->n[i];
               for (int k = 0; k < myt->size(); k++) {
                  if (myt->n[i]->n[k] == myt) {
                     pr.v = k;
                  }
               }
               bndry.insert(pr);

            }
         }
      }
   }

   bool quit = false;
   foreach(Face ff, bndry) {
       if (ff.v != -1) {
	   if (cavity.find(ff.s) != cavity.end()) {

	       cout << "Bad face on the boundary" << endl;
	       quit = true;
	       cout << ff.v << " " << ff.s << endl;
	       cout << "Cavity " << endl;
	       foreach (Simplex * ss, cavity) {
		   cout << ss << endl;
	       }
	   }
	   
       }
   }

   if (quit) exit(0);



   //cout << "Cavity 2" << endl;

   //cout << "BOUNDARY" << endl;
   //set<Simplex>::iterator iter;
   //for (iter=bndry.begin(); iter !=bndry.end(); iter++) {
   //  printMe(*iter);
   //}

   //cout << endl << "Cavity " << p << " " << name << " " << t << " " << t->f << " " << cavity.size() << endl;
   //foreach(Simplex * ss, cavity) {
   //  cout << ss << " ";
   //} cout << endl;

   //cout << "CAVITY" << endl;
   //for (iter=cavity.begin(); iter !=cavity.end(); iter++) {
   //  printMe(*iter);
   //}



   //cout << cavity.size() << endl;
   //cout << bndry.size() << endl;
   return;


}
void Mesh::Cavity(Point & p, set < int >&nbs)
{
   Cavity(p, Locate(p), nbs);
}

void Mesh::Cavity(Point & p, Simplex * s, set < int >&nbs)
{

   set < Simplex * >cavity;
   set < Face > bndry;
   Cavity(p, Locate(p, s->v[0]), cavity, bndry);

   foreach(Simplex * s1, cavity) {
      for (int i = 0; i < s1->size(); i++) {
         nbs.insert(s1->v[i]);
      }
   }

}

// Stitch
//
// This functions connects point number i to the boundary faces
// and returns the new simplices in newcavity.
//
// Important: does not update p2s!
//
void Mesh::Stitch(int i, set < Face > &bndry, set < Simplex * >&newcavity)
{

   Simplex *toHook = NULL;

   //if (newcavity.size() != 0) {
   //    cout << "Warning... Stitch called with a nonempty newcavity..." << endl;
   //}

   // clear the cavity and refil with the new simplices containing i...
   //newcavity.clear();

   if (dim == 3) {
      map < duple, Simplex * >toBeStitched;
      // make each simplex and hook it to the outside of the cavity...
      foreach(Face pr, bndry) {
         Simplex *news;

         if (pr.v == -1) {
            // turn the dummy simplex into the simplex to insert...
            pr.s->v[dim] = i;
            pr.s->n[dim] = NULL; // the opposite face is outside the mesh...
            news = pr.s;
            newcavity.insert(pr.s);
         } else {
            int a, b, c;
            a = pr.s->v[0];
            b = pr.s->v[1];
            c = pr.s->v[2];
            if (pr.v == 0) {
               a = pr.s->v[dim];
            } else if (pr.v == 1) {
               b = pr.s->v[dim];
            } else if (pr.v == 2) {
               c = pr.s->v[dim];
            }
            news = C->SP->NewSimplex(a, b, c, i);
            news->n[dim] = pr.s;
            newcavity.insert(news);
            pr.s->n[pr.v] = news;
	    //if (a == -1 || b == -1 || c == -1) {
	    //cout << pr.s << endl;
	    //	cout << news << endl;
	    //	cout << a << " " << b << " " << c << " " << i << endl;
	    //}


         }
         SetRadiusAndCC(news);

         for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
               if (i != j && news->v[i] < news->v[j]) {
                  duple tmpd;
                  tmpd.first = news->v[i];
                  tmpd.second = news->v[j];
                  map < duple, Simplex * >::iterator mIt =
                     toBeStitched.find(tmpd);
                  if (mIt == toBeStitched.end()) {
                     toBeStitched[tmpd] = news;
                  } else {
                     int k = -1, k1 = -1; // k will be set to be the missing value from i and j...
                     if (i != 2 && j != 2) {
                        k = 2;
                     } else if (i != 1 && j != 1) {
                        k = 1;
                     } else {
                        k = 0;
                     }
                     toHook = mIt->second;
                     if ((toHook->v[0] == news->v[i]
                          && toHook->v[1] == news->v[j])
                         || (toHook->v[0] == news->v[j]
                             && toHook->v[1] == news->v[i])) {
                        k1 = 2;
                     } else if ((toHook->v[0] == news->v[i]
                                 && toHook->v[2] == news->v[j])
                                || (toHook->v[0] == news->v[j]
                                    && toHook->v[2] == news->v[i])) {
                        k1 = 1;
			//} else if ((toHook->v[1] == news->v[i]
                        //         && toHook->v[2] == news->v[j])
                        //        || (toHook->v[1] == news->v[j]
                        //            && toHook->v[2] == news->v[i])) {
			// k1 = 0;
			//} else {
			// cout << "Problem. " << endl;
			//}
		     } else {
			 k1 = 0;
		     }

                     toHook->n[k1] = news;
                     news->n[k] = toHook;
                     toBeStitched.erase(mIt);
                  }
               }
            }
         }
      }
      if (toBeStitched.size() != 0) {
	  cout << "ERROR: problem stitching all tets together." << endl;
	  map < duple, Simplex * >::iterator mIt = toBeStitched.begin();
	  while (mIt != toBeStitched.end()) {

	      cout << mIt->first.first <<" " << mIt->first.second << " " << mIt->second << endl;

	      mIt++;
	  }
	  cout << "ToBeStitched " << toBeStitched.size() << endl;
	  cout << "New simplices " << endl;
	  foreach(Simplex * ss, newcavity) {
	      cout << ss << " ";
	      if (ss->n[dim]!=NULL) {
		  cout << ss->n[dim] << endl;
	      }cout << endl;
	      
	  }

	  exit(0);

      }

      foreach (Simplex *mys, newcavity) {
	  for (int i=0; i<4; i++) {
	      if (mys->n[i] != NULL && mys->n[i]->size() != 4) {
		  cout << "Bad nbor simplex: " << mys << " " << mys->n[i] << endl;
	      }
	      if (mys->v[i] == -1) {
		  cout << "Bad tet " << mys << endl;
	      }
	  }
      }

   } else if (dim == 2) {
      map < int, Simplex * >toBeStitched2d;
      // make each simplex and hook it to the outside of the cavity...
      foreach(Face pr, bndry) {
         Simplex *news;

         if (pr.v == -1) {
            // turn the dummy simplex into the simplex to insert...
            pr.s->v[dim] = i;
            pr.s->n[dim] = NULL; // the opposite face is outside the mesh...
            news = pr.s;
            newcavity.insert(pr.s);
         } else {
            int a, b;
            a = pr.s->v[0];
            b = pr.s->v[1];
            if (pr.v == 0) {
               a = pr.s->v[dim];
            } else if (pr.v == 1) {
               b = pr.s->v[dim];
            }

            news = C->SP->NewSimplex(a, b, i);

            news->n[dim] = pr.s;
            newcavity.insert(news);
            pr.s->n[pr.v] = news;
         }
         SetRadiusAndCC(news);

         for (int i = 0; i < 2; i++) {
            map < int, Simplex * >::iterator mIt =
               toBeStitched2d.find(news->v[i]);
            if (mIt == toBeStitched2d.end()) {
               toBeStitched2d[news->v[i]] = news;
            } else {
               int k = -1, k1 = -1; // k will be set to be the missing value from i and j...
               if (i != 1) {
                  k = 1;
               } else {
                  k = 0;
               }
               toHook = mIt->second;
               if (toHook->v[0] == news->v[i]) {
                  k1 = 1;
               } else {
                  k1 = 0;
               }
               toHook->n[k1] = news;
               news->n[k] = toHook;
               toBeStitched2d.erase(mIt);
            }
         }
      }
   } else {
      // dim = 1

      foreach(Face pr, bndry) {
         Simplex *news;

         if (pr.v == -1) {
            // turn the dummy simplex into the simplex to insert...
            pr.s->v[dim] = i;
            pr.s->n[dim] = NULL; // the opposite face is outside the mesh...
            news = pr.s;
            newcavity.insert(pr.s);
         } else {
            int a;

            if (pr.v == 0) {
               a = pr.s->v[dim];
            } else {
               a = pr.s->v[0];
            }

            news = C->SP->NewSimplex(a, i);

            news->n[dim] = pr.s;
            newcavity.insert(news);
            pr.s->n[pr.v] = news;
         }
         SetRadiusAndCC(news);
         if (toHook == NULL) {
            toHook = news;
         } else {
            toHook->n[0] = news;
            news->n[0] = toHook;
         }
      }
   }

   //cout << "Stitched " << bndry.size() << " tets" << endl;
   return;

}

bool Mesh::IsIn(Mesh * m)
{
   foreach(Mesh * m1, parents) {
      if (m1 == m) {
         return true;
      }
   }
   return false;
}


bool Mesh::Contains(Mesh * m)
{
   foreach(Mesh * m1, children) {
      if (m1 == m) {
         return true;
      }
   }
   return false;

}



void Mesh::Neighbors(int p, set < int >&nbs)
{
   set < Simplex * >checked;
   vector < Simplex * >tocheck;

   if (p2s.find(p) == p2s.end()) {
      return;
   }

   tocheck.push_back(p2s[p]);

   checked.insert(p2s[p]);

   while (tocheck.size() > 0) {

      Simplex *myt = tocheck[tocheck.size() - 1];
      tocheck.pop_back();

      for (int i = 0; i < myt->size(); i++) {
         if (myt->v[i] != p) {
            nbs.insert(myt->v[i]);
         }
         if (myt->n[i] != NULL && myt->n[i]->contains(p)
             && checked.find(myt->n[i]) == checked.end()) {
            tocheck.push_back(myt->n[i]);
            checked.insert(myt->n[i]);
         }
      }
   }

}

void Mesh::SegNeighbors(int i, int &n1, int &n2)
{

  if (dim != 1) {
    cout << "ERROR: SegNeighbors called on mesh of dim " << dim << endl;
      exit(0);
  }

   set < int >nbs;
   Neighbors(i, nbs);

   if (nbs.size() != 2) {
      cout << "ERROR: SegNeighbors doesn't get 2 neighbors" << endl;
      exit(0);
   }

   set < int >::iterator it;
   it = nbs.begin();
   n1 = *it;
   it++;
   n2 = *it;

}


void Mesh::Simplices(int p, set < Simplex * >&star)
{

   vector < Simplex * >tocheck;

   if (p2s.find(p) == p2s.end()) {
      return;
   }

   tocheck.push_back(p2s[p]);

   star.insert(p2s[p]);

   while (tocheck.size() > 0) {

      Simplex *myt = tocheck[tocheck.size() - 1];
      tocheck.pop_back();

      for (int i = 0; i < myt->size(); i++) {
         if (myt->n[i] != NULL && myt->n[i]->contains(p)
             && star.find(myt->n[i]) == star.end()) {
            tocheck.push_back(myt->n[i]);
            star.insert(myt->n[i]);
         }
      }
   }

}


bool Mesh::Exists(int i)
{
   if (p2s.find(i) != p2s.end()) {
      return true;
   }
   return false;
}

Simplex *Mesh::GetSimplex()
{
   if (p2s.size() > 0) {
      return p2s.begin()->second;
   } else {
      cout << "ERROR: GetSimplex Called on mesh with p2s empty" << endl;
      return NULL;
   }
}

Simplex *Mesh::GetSimplex(int i)
{
   if (p2s.find(i) == p2s.end()) {
      return NULL;
   }
   return p2s[i];
}

int Mesh::GetPoint()
{
   if (p2s.size() > 0) {
      return p2s.begin()->first;
   } else {
      cout << "ERROR: GetPoint Called on mesh with p2s empty" << endl;

      return -1;
   }
}

bool Mesh::Exterior(int p)
{

   if (C->P->type[p] == POINT_BBOX) {
      return true;
   }
   return false;
}


bool Mesh::Exterior(Simplex * s)
{

   if (dim == 3) {
      return false;
   }

   if (bnd == NULL || bnd->segments.size() == 0) {
      // convex case or if the boundary of a non convex face
      // has not yet been set up...
      for (int i = 0; i < s->size(); i++) {
         if (C->P->type[s->v[i]] == POINT_BBOX) {
            return true;
         }
      }
      return false;
   } else {
      // non convex case...
      // reusing code from the collar...
      if (IsBoundary(s->v[0])) {
         return IsBndHelp(s->v[0], s->v[1], s->v[2]);
      }
      if (IsBoundary(s->v[1])) {
         return IsBndHelp(s->v[1], s->v[2], s->v[0]);
      }
      if (IsBoundary(s->v[2])) {
         return IsBndHelp(s->v[2], s->v[0], s->v[1]);
      }

      return false;

   }
   return false;

}



bool Mesh::IsBoundary(int p)
{

   // return true if p is a boundary point of a nonconvex face...
   if (dim != 2) {
      cout << "WARNING: IsBoundary called on a mesh of dim " << dim <<
      endl;
      return false;
   }
   if (convex) {
      cout << "WARNING: IsBoundary called on a convex face." << endl;
      return false;
   }

   if (bnd == NULL) {
      return false;
   }

   if (bnd->segments.size() == 0) {
      return false;
   }
   // FIXME: an alternative implementation will use the bnd class to see if the point is
   //        there but it shoud work either way
   if (C->P->type[p] == POINT_SEGMENT || C->P->type[p] == POINT_INPUT) {
      foreach(Mesh * m, C->P->l2m[p]) {
         // is this a boundary segment...
         if (m->dim == 1
             && bnd->segments.find(m) != bnd->segments.end()) {
            return true;
         }
      }
   }
   return false;
}


// FIXME: this is not general!
bool Mesh::IsIntestine(Simplex * s) {

    
    if (s->size() == 3) {
	// find two tets containing s

	// if exactly one of these belongs to the intestine
	// then return true
	set<Simplex * > star;
	Simplices(s->v[0],star);
	
	vector<Simplex *> good;

	foreach(Simplex * s1, star) {
	    if (s1->contains(s->v[1]) && 
		s1->contains(s->v[2]) ) {
		good.push_back(s1);
	    }
	}

	if (good.size() != 2) {
	    return false;
	}

	if (IsIntestine(good[0]) != IsIntestine(good[1])) {
	    return true;
	}
	return false;
	}

  for (int i=0; i<s->size(); i++) {
    if ((C->P->type[s->v[i]] == POINT_INPUT && C->IsProtected(s->v[i])) || 
	(C->P->type[s->v[i]] == POINT_SEGMENT && 
	 C->IsProtected(C->P->LowestDimensionalContainingMesh(s->v[i])))) {
      return true;
    }
  }

  for (int i=0; i<s->size(); i++) {
    if (C->P->type[s->v[i]] == POINT_INPUT || 
	C->P->type[s->v[i]] == POINT_SEGMENT ||
	C->P->type[s->v[i]] == POINT_3D || 
	C->P->type[s->v[i]] == POINT_FACE ||
	C->P->type[s->v[i]] == POINT_SPHERE ||
	C->P->type[s->v[i]] == POINT_BBOX) {
      // note: all sphere points are connected to an input point...
      // so the previous rule works...
      return false;
    }
  }

  if (s->size() < 4) {
    return false;
  }
  // FIXME: for faces this may not be general
  // should work for tets?
  IntestineCylinder * cyl;
  int k = 3;
  int a=0,b=1,c=2;
  cyl = C->M->intestine->ContainingIntestineCylinder(s->v[0], s->v[1],s->v[2]);
  if (cyl == NULL) {
    cyl = C->M->intestine->ContainingIntestineCylinder(s->v[0], s->v[1],s->v[3]);
    k = 2;
    c = 3;
  }
  if (cyl == NULL) {
    cyl = C->M->intestine->ContainingIntestineCylinder(s->v[0], s->v[2],s->v[3]);
    k = 1;
    b = 2;
  }
  if (cyl == NULL) {
    cyl = C->M->intestine->ContainingIntestineCylinder(s->v[1], s->v[2],s->v[3]);
    k = 0;
    a = 1;
  }
  if (cyl == NULL) {
    return false;
  }
  if (C->M->intestine->l2c[s->v[k]].find(cyl) != C->M->intestine->l2c[s->v[k]].end()) {
    return true;
  }
  
  Vector vb = C->P->V[s->v[b]] - C->P->V[s->v[a]];
  Vector vc = C->P->V[s->v[c]] - C->P->V[s->v[a]];
  Vector vk = C->P->V[s->v[k]] - C->P->V[s->v[a]];
  Vector vi = cyl->ring1->center - C->P->V[s->v[a]];
  Vector n = vb^vc;
  n.normalize();
  vi.normalize();
  vk.normalize();
  
  if ((vi*n)*(vk*n) > 0) {
    cout << "TRUE" << endl;
    //exit(0);
    return true;
  }
  
  return false;
}


bool Mesh::IsCollar(Simplex * s)
{
   if (dim == 2) {
      if (col == NULL) {
         return false;
      }

      if (C->P->type[s->v[0]] == POINT_FACE ||
          C->P->type[s->v[0]] == POINT_3D ||
          C->P->type[s->v[0]] == POINT_BBOX ||
          C->P->type[s->v[1]] == POINT_FACE ||
          C->P->type[s->v[1]] == POINT_3D ||
          C->P->type[s->v[1]] == POINT_BBOX ||
          C->P->type[s->v[2]] == POINT_FACE ||
          C->P->type[s->v[2]] == POINT_3D ||
          C->P->type[s->v[2]] == POINT_BBOX) {
         return false;
      }

      //int cp = -1;
      //int pa=-1,pb=-1;
      if (col->Exists(s->v[0])) {
         return IsCollarHelp(s->v[0], s->v[1], s->v[2]);
      }
      if (col->Exists(s->v[1])) {
         return IsCollarHelp(s->v[1], s->v[2], s->v[0]);
      }
      if (col->Exists(s->v[2])) {
         return IsCollarHelp(s->v[2], s->v[0], s->v[1]);
      }

      return false;

   } else if (dim == 1) {

      if (C->P->type[s->v[0]] == POINT_FACE ||
          C->P->type[s->v[0]] == POINT_3D ||
          C->P->type[s->v[0]] == POINT_BBOX ||
          C->P->type[s->v[1]] == POINT_FACE ||
          C->P->type[s->v[1]] == POINT_3D ||
          C->P->type[s->v[1]] == POINT_BBOX) {
         return false;
      }
      // ask if this segment is protected...
      if (C->segmentsToProtect.find(this) != C->segmentsToProtect.end()) {
         return true;
      }
      int ip = -1;

      if (C->P->type[s->v[0]] == POINT_INPUT) {
         ip = s->v[0];
      } else if (C->P->type[s->v[1]] == POINT_INPUT) {
         ip = s->v[1];
      }

      if (ip >= 0) {

         if (C->pointsToProtect.find(C->P->ids[ip]) !=
             C->pointsToProtect.end()) {
	   if (C->lazyProtectionList.find(C->P->ids[ip]) ==
             C->lazyProtectionList.end()) {

	     return true;
	   }
         }
         // check if the input point is part of a protected segment...
         foreach(Mesh * mm, C->P->l2m[ip]) {
            if (C->segmentsToProtect.find(mm) !=
                C->segmentsToProtect.end()) {
               return true;
            }
         }

         return false;

      }

      return false;

   }

   return false;

}

// only pass cp to this function if
// cp is a point in the collar!
//
// FIXME: this has some issues in non-convex faces...
// works now?
bool Mesh::IsCollarHelp(int cp, int pa, int pb)
{

   int cnext, cprev;
   cnext = col->col2next[cp];
   cprev = col->col2prev[cp];

   Vector vnext, vprev, va, vb;

   if (pb == cnext || pa == cprev) {
      // swap pa and pb
      int tmp = pa;
      pa = pb;
      pb = tmp;
   }

   if (cnext == -1) {
      // fix here...
      vprev = C->P->V[cprev] - C->P->V[cp];
      va = C->P->V[pa] - C->P->V[cp];
      if (va * vprev < 0.0) {
         return false;
      }
      if ((va ^ vprev) * dir > 0.0) {
         return false;
      }
      return true;

   } else if (cprev == -1) {
      vnext = C->P->V[cnext] - C->P->V[cp];
      vb = C->P->V[pb] - C->P->V[cp];
      if (vb * vnext < 0.0) {
         return false;
      }
      if ((vnext ^ vb) * dir > 0.0) {
         return false;
      }
      return true;
   } else {

      vnext = C->P->V[cnext] - C->P->V[cp];
      vprev = C->P->V[cprev] - C->P->V[cp];
      va = C->P->V[pa] - C->P->V[cp];
      vb = C->P->V[pb] - C->P->V[cp];

      if (pa == cnext) {
         if ((vnext ^ vb) * dir > 0.0) {
            return false;
         }
         return true;
      }
      if (pb == cprev) {
         if ((va ^ vprev) * dir > 0.0) {
            return false;
         }
         return true;
      }

      double s1, s2;
      s1 = (vnext ^ va) * dir;
      s2 = (va ^ vprev) * dir;
      if (s1 > 0.0 && s2 > 0.0) {
         return false;
      }
      if (s1 < 0.0 && s2 < 0.0) {
         return true;
      }

      if (va * vnext > 0.0) {
         if (s1 > 0.0) {
            return false;
         }
         return true;
      }
      if (va * vprev > 0.0) {
         if (s2 > 0.0) {
            return false;
         }
         return true;
      }
      return false;

      cout << "CP " <<  cp << endl 
	   << "Next/Prev " << cnext << " " << cprev << endl 
	   << "Vectors " << vnext << " "  << vprev << endl 
	   << "A/B " << pa << " " << pb << endl 
	   << "Vector " << va << " " << vb << endl;

      cout << C->P->type[cp] << " " << C->P->type[pa] << " " << C->P->
      type[pb] << " " << C->P->type[cnext] << " " << C->P->
      type[cprev] << endl;
      cout << C->P->LowestDimensionalContainingMesh(cp)->name << endl;
      cout << C->P->LowestDimensionalContainingMesh(cnext)->name << endl;
      cout << C->P->LowestDimensionalContainingMesh(cprev)->name << endl;
      cout << C->P->LowestDimensionalContainingMesh(pa)->name << endl;
      cout << C->P->LowestDimensionalContainingMesh(pb)->name << endl;

      cout << "ERROR: the collar is confusing... " << endl;
      //exit(0);

   }
   return false;

}


// only pass cp to this function if
// cp is a point in the collar!
// this function is horribly confusing but
// gets the job done...
//
// is there a better geometric predicate to use?
bool Mesh::IsBndHelp(int cp, int pa, int pb)
{

   int cnext, cprev;
   cnext = bnd->next[cp];
   cprev = bnd->prev[cp];

   Vector vnext, vprev, va, vb;

   if (pb == cnext || pa == cprev) {
      // swap pa and pb
      int tmp = pa;
      pa = pb;
      pb = tmp;
   }

   vnext = C->P->V[cnext] - C->P->V[cp];
   vprev = C->P->V[cprev] - C->P->V[cp];

   va = C->P->V[pa] - C->P->V[cp];
   vb = C->P->V[pb] - C->P->V[cp];

   if (pa == cnext) {
      if ((vnext ^ vb) * dir > 0.0) {
         return false;
      }
      return true;
   }
   if (pb == cprev) {
      if ((va ^ vprev) * dir > 0.0) {
         return false;
      }
      return true;
   }



   double s1, s2;
   s1 = (vnext ^ va) * dir;
   s2 = (va ^ vprev) * dir;
   if (s1 >= 0.0 && s2 >= 0.0) {
      return false;
   }
   if (s1 <= 0.0 && s2 <= 0.0) {
      return true;
   }

   // consider
   // vnext^vprev*dir
   // this tells me if I am inside or outside here...
   if ((vnext ^ vprev) * dir > 0) {
      return true;
   }
   return false;
}


void Mesh::GetSegmentEndPoints(int &a, int &b)
{

   if (dim != 1) {
      cout << "ERROR: calling a segment function on a dimension " << dim
           << " mesh." << endl;
      cout << name << endl;
      exit(0);
   }

   int afound = 0;
   int bfound = 0;

   PIterator pIt(this);
   pIt.Start(GetSimplex());
   while (pIt.Current() != -1) {
      int iP = pIt.Current();

      if (C->P->type[iP] == POINT_INPUT) {
         if (afound == 0) {
            a = iP;
            afound = 1;
         } else if (bfound == 0) {
            b = iP;
            bfound = 1;
         } else {
            cout <<
            "ERROR: 3 input points in a segment.  This is not supported."
                 << endl;

            cout << name << " " << a << " " << b << endl;
            cout << iP;
            exit(0);
         }
      }
      ++pIt;
   }
}


bool Mesh::IsEndSegment(Simplex * s)
{
   if (dim != 1) {
      cout << "Warning: IsEndSegment called on a dimension " << dim <<
      " mesh." << endl;
      return false;
   }

   if (C->P->type[s->v[0]] == POINT_INPUT
       || C->P->type[s->v[1]] == POINT_INPUT) {
      return true;
   }

   return false;
}


void Mesh::PrintOff(string fname)
{

   //cout << "Printing Mesh " << name << " to OFF file " << fname << endl;

   ofstream fout;
   fout.open((fname).c_str());


   map < int, int >l2i;
   map < int, int >i2l;
   // first, loop over the mesh and get all the vertices...
   int vertexCount = 0;
   PIterator pIt(this);
   pIt.Start(GetSimplex());
   while (pIt.Current() != -1) {
      int iP = pIt.Current();

      if (C->P->type[iP] != POINT_BBOX) {
         //have I seen this point?
         if (l2i.find(iP) == l2i.end()) {
            l2i[iP] = vertexCount;
            i2l[vertexCount] = iP;
            vertexCount++;
            //cout << iP << " " << vertexCount << endl;
         }
      }
      ++pIt;
   }

   //cout << "Vertex Count "<< vertexCount << endl;

   // loop over the simplices

   int simplexCount = 0;

   Iterator it;
   it.Start(GetSimplex());
   while (it.Current() != NULL) {

      // count simplices to use...
      //if (!Exterior(it.Current()) && !C->Step3IsEncroached(it.Current())) {
      if (!Exterior(it.Current())) {
         simplexCount++;
      }
      ++it;

   }

   //cout << "Simplex Count" << simplexCount << endl;

   fout << "OFF" << endl;
   fout << vertexCount << " " << 2*simplexCount << " 0" << endl;


   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[0] << " " << C->P->V[i2l[i]].
      xyz[1] << " " << C->P->V[i2l[i]].xyz[2] << endl;

   }

   // reloop simplices and print...
   it.reset();
   it.Start(GetSimplex());
   while (it.Current() != NULL) {


      Simplex *s = it.Current();
      // count simplices to use...
      //if (!Exterior(s) && !C->Step3IsEncroached(it.Current())) {
      if (!Exterior(s)) {
         fout << s->size();
         for (int iv = 0; iv < s->size(); iv++) {
            fout << " " << l2i[s->v[iv]];
         }
         fout << endl;

	 // double print...
         fout << s->size();
         for (int iv = s->size()-1; iv >= 0; iv--) {
            fout << " " << l2i[s->v[iv]];
         }
         fout << endl;


      }
      ++it;
   }
}

// Mesh mSeg must be a boundary segment
// of the face...
void Mesh::NormalInFace(Mesh * mSeg, Vector & n)
{
   if (mSeg->dim != 1 || dim != 2) {
      cout << "Error: bad call of NormalInFace" << endl;
      exit(0);
   }
  Mesh::Iterator sIt;
  sIt.Start(mSeg->GetSimplex());
  while (sIt.Current() != NULL) {
    Simplex * s = sIt.Current();
    
    if (C->P->type[s->v[0]] != POINT_BBOX &&
	C->P->type[s->v[1]] != POINT_BBOX){
  
      NormalInFace(s->v[0],s->v[1],n);
      return;
    }

    ++sIt;
  }

}

// the segment from a to b must be a boundary segment of the mesh
void Mesh::NormalInFace(int a, int b, Vector & n)
{

   if (dim != 2) {
      cout << "Error: bad call of NormalInFace" << endl;
      exit(0);
   }
   // check if a and b are boundary points?

   if (convex) {

      Vector v1 = C->P->V[b] - C->P->V[a];
      Vector v2 = base - C->P->V[a]; // base is in the interior of the convex face...

      v1.normalize();
      v2.normalize();

      n = v2 - (v1 * v2) * v1;
      n.normalize();

   } else {

      // WARNING: this may be backwards...

      if (!bnd->Exists(a,b)) {
         cout << "ERROR: bad segment passed to NormalInFace" << endl;
	 cout << a << " " << b << endl;
	 cout << C->P->type[a] << " " << C->P->type[b] << endl;
	 cout << bnd->next[a] << " : " << bnd->prev[a] 
	      << " --- " << bnd->next[b] << " : " << bnd->prev[b] << endl;

         exit(0);
      }
      Vector v1;
      if (bnd->next[a] == b) {
	v1 = C->P->V[b] - C->P->V[a];
      } else {
	v1 = C->P->V[a] - C->P->V[b];
      }
      v1.normalize();
      n = dir ^ v1;
   }
}



void Mesh::PrintCollarOff(string fname)
{

   //cout << "Printing Mesh " << name << " to OFF file " << fname << endl;

   ofstream fout;
   fout.open((fname).c_str());


   map < int, int >l2i;
   map < int, int >i2l;
   // first, loop over the mesh and get all the vertices...
   int vertexCount = 0;
   PIterator pIt(this);
   pIt.Start(GetSimplex());
   while (pIt.Current() != -1) {
      int iP = pIt.Current();

      if (C->P->type[iP] != POINT_BBOX) {
         //have I seen this point?
         if (l2i.find(iP) == l2i.end()) {
            l2i[iP] = vertexCount;
            i2l[vertexCount] = iP;
            vertexCount++;
            //cout << iP << " " << vertexCount << endl;
         }
      }
      ++pIt;
   }

   // loop over the simplices

   int simplexCount = 0;

   int counta = 0;
   Iterator it;
   it.Start(GetSimplex());
   while (it.Current() != NULL) {

      // count simplices to use...
      if (!Exterior(it.Current()) && IsCollar(it.Current())) {
         simplexCount++;
      } else if (!Exterior(it.Current())) {
         counta++;
      }
      ++it;
   }
   //cout << name << " : Non-collar vs. Collar :  " << counta << " " << simplexCount << endl;

   fout << "OFF" << endl;
   fout << vertexCount << " " << simplexCount << " 0" << endl;

   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[0] << " " << C->P->V[i2l[i]].
      xyz[1] << " " << C->P->V[i2l[i]].xyz[2] << endl;

   }

   // reloop simplices and print...
   it.reset();
   it.Start(GetSimplex());
   while (it.Current() != NULL) {


      Simplex *s = it.Current();

      if (!Exterior(s) && IsCollar(s)) {
         fout << s->size();
         for (int iv = 0; iv < s->size(); iv++) {
            fout << " " << l2i[s->v[iv]];
         }
         fout << endl;

      }
      ++it;

   }
}

void Mesh::PrintSurfaceOff(string fname)
{


   if (dim != 3) {
      cout << "ERROR: wrong dimensional mesh to build surface mesh." <<
      endl;
      exit(0);
   }

   ofstream fout;
   fout.open((fname).c_str());


   map < int, int >l2i;
   map < int, int >i2l;
   // first, loop over the mesh and get all the vertices...
   int vertexCount = 0;
   PIterator pIt(this);
   pIt.Start(GetSimplex());
   while (pIt.Current() != -1) {
      int iP = pIt.Current();

      if (C->P->type[iP] != POINT_BBOX && C->P->type[iP] != POINT_3D) {
         //have I seen this point?
         if (l2i.find(iP) == l2i.end()) {
            l2i[iP] = vertexCount;
            i2l[vertexCount] = iP;
            vertexCount++;
            //cout << iP << " " << vertexCount << endl;
         }
      }
      ++pIt;
   }


   // change below to only print surface simplices...

   int simplexCount = 0;


   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      if (m->dim == 2 || m->parents.size() == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               simplexCount++;
            }
            ++it;
         }
      }
   }

   fout << "OFF" << endl;
   fout << vertexCount << " " << simplexCount << " 0" << endl;


   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[0] << " " << C->P->V[i2l[i]].
      xyz[1] << " " << C->P->V[i2l[i]].xyz[2] << endl;
   }

   // reloop simplices and print...

   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      if (m->dim == 2 || m->parents.size() == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            Simplex *s = it.Current();
            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               fout << s->size();
               for (int iv = 0; iv < s->size(); iv++) {
                  fout << " " << l2i[s->v[iv]];
               }
               fout << endl;
            }
            ++it;
         }
      }
   }
}

void Mesh::PrintSurfaceGMV(string fname)
{


   if (dim != 3) {
      cout << "ERROR: wrong dimensional mesh to build surface mesh." <<
      endl;
      exit(0);
   }

   ofstream fout;
   fout.open((fname).c_str());


   map < int, int >l2i;
   map < int, int >i2l;
   // first, loop over the mesh and get all the vertices...
   int vertexCount = 0;
   PIterator pIt(this);
   pIt.Start(GetSimplex());
   while (pIt.Current() != -1) {
      int iP = pIt.Current();

      if (C->P->type[iP] != POINT_BBOX && C->P->type[iP] != POINT_3D) {
         //have I seen this point?
         if (l2i.find(iP) == l2i.end()) {
            l2i[iP] = vertexCount + 1;
            i2l[vertexCount] = iP;
            vertexCount++;
            //cout << iP << " " << vertexCount << endl;
         }
      }
      ++pIt;
   }


   // change below to only print surface simplices...

   int simplexCount = 0;


   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      if (m->dim == 2 || m->parents.size() == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               simplexCount++;
            }
            ++it;
         }
      }
   }

   fout << "gmvinput ascii" << endl;
   fout << "nodes " << vertexCount << endl;

   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[0] << " ";
   } fout << endl;
   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[1] << " ";
   } fout << endl;
   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[2] << " ";
   } fout << endl;



   // reloop simplices and print...

   fout << "cells " << simplexCount << endl;

   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      if (m->dim == 2 || m->parents.size() == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            Simplex *s = it.Current();
            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               if (s->size() == 2) {
                  fout << "line ";
               } else if (s->size() == 3) {
                  fout << "tri ";
               } else if (s->size() == 4) {
                  fout << "tet ";
               }
               fout << s->size();
               for (int iv = 0; iv < s->size(); iv++) {
                  fout << " " << l2i[s->v[iv]];
               }
               fout << endl;
            }
            ++it;
         }
      }
   }

   fout << "endgmv" << endl;
}



void Mesh::PrintSurfaceCOff(string fname)
{

    bool printIntestine = true;
    bool printIntEdges = false;
  
  // prints each face in a different color...
  
  
  if (dim != 3) {
    cout << "ERROR: wrong dimensional mesh to build surface mesh." 
	 << endl;
    exit(0);
  }
  
  ofstream fout;
  fout.open((fname).c_str());
  
  
  map < int, int >l2i;
  map < int, int >i2l;
  // first, loop over the mesh and get all the vertices...
  int vertexCount = 0;
  PIterator pIt(this);
  pIt.Start(GetSimplex());
  while (pIt.Current() != -1) {
      int iP = pIt.Current();

      if (C->P->type[iP] != POINT_BBOX && C->P->type[iP] != POINT_3D) {
         //have I seen this point?
         if (l2i.find(iP) == l2i.end()) {
            l2i[iP] = vertexCount;
            i2l[vertexCount] = iP;
            vertexCount++;
            //cout << iP << " " << vertexCount << endl;
         }
      }
      ++pIt;
   }


   // change below to only print surface simplices...

   int simplexCount = 0;


   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      //if (m->dim == 2 || m->parents.size() == 1) {
      if (m->dim == 2 || m->dim == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               simplexCount++;
            }
            ++it;
         }
      }
   }

   if (printIntestine && intestine != NULL) {


     if (printIntEdges) {
       PIterator pIt2(this);
       pIt2.Start(GetSimplex());
       while (pIt2.Current() != -1) {
	 int iP = pIt2.Current();
	 
	 if (C->P->type[iP] == POINT_RING || 
	     C->P->type[iP] == POINT_CYLINDER || 
	     C->P->type[iP] == POINT_SPHERE || 
	     C->P->type[iP] == POINT_SEGMENT) {
	   set<int> nbors;
	   Neighbors(iP,nbors);
	   foreach(int iN, nbors) {
	     if (iN > iP && intestine->IsIntestineEdge(iP,iN)) {
	       simplexCount++;
	     }
	   }
	   
	 }
	 ++pIt2;
       }
     }

     Iterator it;
     it.Start(GetSimplex());
     while (it.Current() != NULL) {
       
       vector<Simplex> tmp;
       Simplex * s = it.Current();
       s->Faces(tmp);

       foreach(Simplex s1, tmp) {
	 if (intestine->IsIntestineFace(s1.v[0],s1.v[1],s1.v[2])) {
	   simplexCount++;
	 }
       }

       ++it;
     }     


   }

   // seems that I should be using the keyword COFF
   // but most viewer programs don't like that
   fout << "OFF" << endl;
   fout << vertexCount << " " << 2*simplexCount << " 0" << endl;


   for (int i = 0; i < vertexCount; i++) {
      fout << C->P->V[i2l[i]].xyz[0] << " " << C->P->V[i2l[i]].
      xyz[1] << " " << C->P->V[i2l[i]].xyz[2] << endl;
   }

   // reloop simplices and print...

   double r, g, b;

   r = 0.0;
   g = 0.0;
   b = 0.0;

   double bright = .75;
   r = bright;
   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      //if (m->dim == 2 || m->parents.size() == 1) {
      if (m->dim == 2 || m->dim == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            Simplex *s = it.Current();
            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               fout << s->size();
               for (int iv = 0; iv < s->size(); iv++) {
                  fout << " " << l2i[s->v[iv]];
               }
               // print the color here...
               if (m->dim == 2) {
		 if ( (C->config->colorcollar) 
                      && m->IsCollar(it.Current())) {
                     fout << " " << 1.0 << " " << 1.0 << " " << 1.0
                          << " 1.0" << endl;
                  } else {
		   if (C->config->printmonochrome) {
		    fout << " " << bright << " " << bright << " " << bright <<
                     " 1.0" << endl;
		   } else {
		    fout << " " << r << " " << g << " " << b <<
                     " 1.0" << endl;
                  }
		 }
               } else {
                  fout << " 0.0 0.0 0.0 1.0" << endl;
               }

	       
	       // double printing for meshlab...
               fout << s->size();
               for (int iv = s->size()-1; iv >= 0; iv--) {
                  fout << " " << l2i[s->v[iv]];
               }
               // print the color here...
               if (m->dim == 2) {
		 if ( (C->config->colorcollar) 
                      && m->IsCollar(it.Current())) {
                     fout << " " << 1.0 << " " << 1.0 << " " << 1.0
                          << " 1.0" << endl;
                  } else {
		   if (C->config->printmonochrome) {
		    fout << " " << bright << " " << bright << " " << bright <<
                     " 1.0" << endl;
		   } else {
		    fout << " " << r << " " << g << " " << b <<
                     " 1.0" << endl;
                  }
		 }
               } else {
                  fout << " 0.0 0.0 0.0 1.0" << endl;
               }


            }
            ++it;
         }

         if (m->dim == 2) {
            //increment the color...
            if (r == 0.0) {
               r = bright;
            } else {
               r = 0.0;
               if (g == 0.0) {
                  g = bright;
               } else {
                  g = 0.0;
                  if (b == 0.0) {
                     b = bright;
                  } else {
                     b = 0.0;
                     r = bright;
                  }
               }
            }
         }
      }
   }

   // print the rings... OLD
   //if (printIntestine && intestine != NULL) {
   //  foreach(IntestineRing * ring, intestine->rings) {
   //    map<int,int>::iterator mIt;
   //   for (mIt = ring->r2n.begin(); mIt != ring->r2n.end(); mIt++) {
   // fout << "2 " << l2i[mIt->first] << " " << l2i[mIt->second] << " 0.0 0.0 0.0 1.0" << endl;
   //    }
   //  }
   //}

   if (intestine != NULL) {
     
     if (printIntEdges) {

       PIterator pIt3(this);
       pIt3.Start(GetSimplex());
       while (pIt3.Current() != -1) {
	 int iP = pIt3.Current();
	 
	 if (C->P->type[iP] == POINT_RING || 
	     C->P->type[iP] == POINT_CYLINDER || 
	     C->P->type[iP] == POINT_SPHERE || 
	     C->P->type[iP] == POINT_SEGMENT) {
	   set<int> nbors;
	   Neighbors(iP,nbors);
	   foreach(int iN, nbors) {
	     if (iN > iP && intestine->IsIntestineEdge(iP,iN)) {
	       fout << "2 " << l2i[iP] << " " << l2i[iN] << " 0.0 0.0 0.0 1.0" << endl;
	       fout << "2 " << l2i[iN] << " " << l2i[iP] << " 0.0 0.0 0.0 1.0" << endl;
	     }
	   }
	   
	 }
	 ++pIt3;
	 
       }
     }
       
     Iterator it;
     it.Start(GetSimplex());
     while (it.Current() != NULL) {
       
       vector<Simplex> tmp;
       Simplex * s = it.Current();
       s->Faces(tmp);

       foreach(Simplex s1, tmp) {
	 if (intestine->IsIntestineFace(s1.v[0],s1.v[1],s1.v[2])) {
	   fout << "3 " << l2i[s1.v[0]] << " " << l2i[s1.v[1]] << " " << l2i[s1.v[2]] << " 1.0 1.0 1.0 0.5" << endl;
	   fout << "3 " << l2i[s1.v[2]] << " " << l2i[s1.v[1]] << " " << l2i[s1.v[0]] << " 1.0 1.0 1.0 0.5" << endl;
	   
	 }
       }

       ++it;
     }     


   }



}


void Mesh::PrintEPS(string fn)
{
   ofstream psout;
   psout.open(fn.c_str());

   double xmin = base.xyz[0]-bbox/C->config->boundingboxratio,
          ymin = base.xyz[1]-bbox/C->config->boundingboxratio,
          xmax = base.xyz[0]+bbox/C->config->boundingboxratio,
          ymax = base.xyz[1]+bbox/C->config->boundingboxratio;

   // Scale and translate so that
   // lower LH corner is at (1,1) inches = (72,72) points
   // upper RH corner scaled to fit on page with 1 inch margin

   //double res = 800*C->config->boundingboxratio;
   double res = 550;

   double sx = res / (xmax - xmin);
   // arand... square below...
   double sy = res / (ymax - ymin);

   double scale = fmin(sx,sy);


   double center = res/2.0;
   double width = res/2.0;
   //double width = res/(C->config->boundingboxratio)/2.0;
   //double height = .75*width;
   double height = width;

   psout << "%!" << endl
         << "%%Title: tri2ps" << endl
         << "%%BoundingBox: " << (int)(center-width)-10 << " " << (int)center-10 << " "
         << (int)( center+width)+10 << " "
         << (int)( center+height)+10 << endl
   //<< "%%BoundingBox: " << (int)(center-width)-2 << " " << (int)center-2 << " "
   //<< (int)( center+width)+2 << " "
   //<< (int)( center+height)+2 << endl
         << "/setc { aload pop setrgbcolor } bind def" << endl
         << "/col1 { [1 0 0] setc } def" << endl
         << "/col2 { [0 0 1] setc } def" << endl
         << "/col0 { 0 setgray } def" << endl
         << "%%Page: 1 1" << endl
         << "%%BeginPageSetup" << endl
         << "%%EndPageSetup" << endl
         << "2.0 setlinewidth" << endl;

   // foreach edge...

   Mesh::Iterator sIt;
   sIt.Start(GetSimplex());
   while (sIt.Current() != NULL) {

      Simplex * s = sIt.Current();
      if (s->size() != 3) {
         cout << "Simplex size error." << endl;
      }

      //if (1) {
      if (C->P->type[s->v[0]] != POINT_BBOX && C->P->type[s->v[1]] != POINT_BBOX && C->P->type[s->v[2]] != POINT_BBOX) {
         Point p0,p1,p2;
         p0 = C->P->V[s->v[0]];
         p1 = C->P->V[s->v[1]];
         p2 = C->P->V[s->v[2]];

         psout << "newpath "
               << (int) (scale*(p0.xyz[0] - xmin))+10 << " "
               << (int) (scale*(p0.xyz[1] - ymin))+10 << " moveto "
               << (int) (scale*(p1.xyz[0] - xmin))+10 << " "
               << (int) (scale*(p1.xyz[1] - ymin))+10 << " lineto ";
         //if (seg2d[s->v[0]].find(s->v[1])!= seg2d[s->v[0]].end() ||
         //  seg2d[s->v[1]].find(s->v[0])!= seg2d[s->v[1]].end()) {
         //  //psout << "col1 stroke" << endl;
         //  psout << "col0 stroke" << endl;
         //} else {
         psout << "col0 stroke" << endl;
         //}

         psout << "newpath "
               << (int) (scale*(p0.xyz[0] - xmin))+10 << " "
               << (int) (scale*(p0.xyz[1] - ymin))+10 << " moveto "
               << (int) (scale*(p2.xyz[0] - xmin))+10 << " "
               << (int) (scale*(p2.xyz[1] - ymin))+10 << " lineto ";
         //if (seg2d[s->v[2]].find(s->v[0])!= seg2d[s->v[2]].end() ||
         //  seg2d[s->v[0]].find(s->v[2])!= seg2d[s->v[0]].end()) {
         //  //psout << "col1 stroke" << endl;
         //  psout << "col0 stroke" << endl;
         //} else {
         psout << "col0 stroke" << endl;
         //}

         psout << "newpath "
               << (int) (scale*(p2.xyz[0] - xmin))+10 << " "
               << (int) (scale*(p2.xyz[1] - ymin))+10 << " moveto "
               << (int) (scale*(p1.xyz[0] - xmin))+10 << " "
               << (int) (scale*(p1.xyz[1] - ymin))+10 << " lineto ";

         //if (seg2d[s->v[2]].find(s->v[1])!= seg2d[s->v[2]].end() ||
         //  seg2d[s->v[1]].find(s->v[2])!= seg2d[s->v[1]].end()) {
         //  //psout << "col1 stroke" << endl;
         //  psout << "col0 stroke" << endl;
         //} else {
         psout << "col0 stroke" << endl;
         //}
      }
      ++sIt;
   }

   for (int i=0; i<(int)C->P->V.size(); i++) {

      if (C->P->type[i] == POINT_INPUT) {
         psout << (int) (scale*(C->P->V[i].xyz[0] - xmin))+10 << " " << (int) (scale*(C->P->V[i].xyz[1] - ymin))+10
               << " 5 0 360 arc closepath 0.0 setgray fill stroke" << endl;
      }
      if (C->P->type[i] == POINT_SEGMENT) {
         psout << (int) (scale*(C->P->V[i].xyz[0] - xmin))+10 << " " << (int) (scale*(C->P->V[i].xyz[1] - ymin))+10
               << " 3 0 360 arc closepath col1 fill stroke" << endl;
      }
      if (C->P->type[i] == POINT_FACE) {
         psout << (int) (scale*(C->P->V[i].xyz[0] - xmin))+10 << " " << (int) (scale*(C->P->V[i].xyz[1] - ymin))+10
               << " 3 0 360 arc closepath col2 fill stroke" << endl;
      }

   }

   psout << "showpage" << endl
         << "%%PageTrailer" << endl
         << "%%EndPage" << endl;

   psout.close();
}

void Mesh::PrintSurfaceEPS(string fn) {
   ofstream psout;
   psout.open(fn.c_str());

   double xmin = base.xyz[0]-bbox/C->config->boundingboxratio,
          ymin = base.xyz[1]-bbox/C->config->boundingboxratio,
          xmax = base.xyz[0]+bbox/C->config->boundingboxratio,
          ymax = base.xyz[1]+bbox/C->config->boundingboxratio;

   // Scale and translate so that
   // lower LH corner is at (1,1) inches = (72,72) points
   // upper RH corner scaled to fit on page with 1 inch margin

   //double res = 800*C->config->boundingboxratio;
   double res = 550;

   double sx = res / (xmax - xmin);
   // arand... square below...
   double sy = res / (ymax - ymin);

   double scale = fmin(sx,sy);


   double center = res/2.0;
   double width = res/2.0;
   //double width = res/(C->config->boundingboxratio)/2.0;
   //double height = .75*width;
   double height = width;

   psout << "%!" << endl
         << "%%Title: tri2ps" << endl
         << "%%BoundingBox: " << (int)(center-width)-10 << " " << (int)center-10 << " "
         << (int)( center+width)+10 << " "
         << (int)( center+height)+10 << endl
   //<< "%%BoundingBox: " << (int)(center-width)-2 << " " << (int)center-2 << " "
   //<< (int)( center+width)+2 << " "
   //<< (int)( center+height)+2 << endl
         << "/setc { aload pop setrgbcolor } bind def" << endl
         << "/col1 { [1 0 0] setc } def" << endl
         << "/col2 { [0 0 1] setc } def" << endl
         << "/col0 { 0 setgray } def" << endl
         << "%%Page: 1 1" << endl
         << "%%BeginPageSetup" << endl
         << "%%EndPageSetup" << endl
         << "2.0 setlinewidth" << endl;

   // loop through all child meshes
   foreach(Mesh * m, children) {
      // deal with the surfaces and the free edges...
      if (m->dim == 1) {
         //count the simplices here...
         Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {

            Simplex *s = it.Current();
            // count simplices to use...
            if (!m->Exterior(it.Current())) {
               psout << "newpath "
                     << (int) (scale*(C->P->V[s->v[0]].xyz[0] - xmin))+10 << " "
                     << (int) (scale*(C->P->V[s->v[0]].xyz[1] - ymin))+10 << " moveto "
                     << (int) (scale*(C->P->V[s->v[1]].xyz[0] - xmin))+10 << " "
                     << (int) (scale*(C->P->V[s->v[1]].xyz[1] - ymin))+10 << " lineto ";
               psout << "col0 stroke" << endl;
            }
            ++it;
         }
      }
   }

   //for (int i=0; i<(int)C->P0.size(); i++) {
   //  psout << (int) (scale*(C->P0[i].p.xyz[0] - xmin))+10 << " " << (int) (scale*(C->P0[i].p.xyz[1] - ymin))+10
   //      << " 5 0 360 arc closepath 0.0 setgray fill stroke" << endl;
   //}

   for (int i=0; i<(int)C->P->V.size(); i++) {

      if (C->P->type[i] == POINT_INPUT) {
         psout << (int) (scale*(C->P->V[i].xyz[0] - xmin))+10 << " " << (int) (scale*(C->P->V[i].xyz[1] - ymin))+10
               << " 5 0 360 arc closepath 0.0 setgray fill stroke" << endl;
      }
      if (C->P->type[i] == POINT_SEGMENT) {
         psout << (int) (scale*(C->P->V[i].xyz[0] - xmin))+10 << " " << (int) (scale*(C->P->V[i].xyz[1] - ymin))+10
               << " 3 0 360 arc closepath col1 fill stroke" << endl;
      }
      if (C->P->type[i] == POINT_FACE) {
         psout << (int) (scale*(C->P->V[i].xyz[0] - xmin))+10 << " " << (int) (scale*(C->P->V[i].xyz[1] - ymin))+10
               << " 3 0 360 arc closepath col2 fill stroke" << endl;
      }

   }

   psout << "showpage" << endl
         << "%%PageTrailer" << endl
         << "%%EndPage" << endl;



}




// FIXME: this is not complete...
void Mesh::PrintInfo()
{

   cout << "Data for Mesh: " << name << endl;
   cout << "Number of Points: " << np << ", Number of Simplices: " << ns
        << endl;

}


// FIXME: this needs to be updated...

void Mesh::RemovePoint(int i,set<Simplex *> & cavity) {

  set <Simplex *> oldcavity;
  RemovePoint(i, oldcavity, cavity);

}

void Mesh::RemovePoint(int i, set < Simplex * >&oldcavity,
		       set < Simplex * >&cavity) {
  cout << "Begin RemovePoint " << C->P->V[i] << " " << i << endl;
  
  set<int> nbs;
  set<Simplex *> star;
  vector<Face> bndry;
  
  Neighbors(i,nbs);
  Simplices(i,star);

  cout << "Neighbors " << nbs.size() << ", Tets " << star.size() << endl;
  foreach (Simplex * s, star) {
    Face f;
    
    int a,b,c,loc;
    a = s->v[0];
    b = s->v[1];
    c = s->v[2];
    loc = 3;
    
    if (s->v[0] == i) {
      a = s->v[3];
      loc = 0;
    }
    if (s->v[1] == i) {
      b = s->v[3];
            loc = 1;
    }
    if (s->v[2] == i) {
      c = s->v[3];
      loc = 2;
    }
    
    
    if (s->n[loc] != NULL) {
      f.s = s->n[loc];
      f.v = f.s->isNeighbor(s);
      if (f.v<0) {
	cout << "ERROR: problem in Remove point" << endl;
	exit(0);
      }
      bndry.push_back(f);
    } else {
      // make a dummy simplex...
      Simplex * dummy = C->SP->NewSimplex(a,b,c,-1);
      f.s = dummy;
      f.v = -1;
      bndry.push_back(f);
      //cout << "Dummy "; printMe(dummy);
    }
    //C->SP->DeleteSimplex(s);
  }
  
  cout << "Star Removed..." << endl;

  double maxd = 0.0;

  foreach(int iNb, nbs) {
    double myd = d(C->P->V[i],C->P->V[iNb]);
    
    if (myd > maxd) {
      maxd = myd;
    }
  }
  cout << maxd << endl;
   
  double x = C->P->V[i].xyz[0];
  double y = C->P->V[i].xyz[1];
  double z = C->P->V[i].xyz[2];

  Point p1(x,y,z+maxd*5.0);
  Point p2(x+5.0*maxd,y,z-maxd*5.0);
  Point p3(x-5.0*maxd,y+5.0*maxd,z-maxd*5.0);
  Point p4(x-5.0*maxd,y-5.0*maxd,z-maxd*5.0);

  int loc1 = C->P->AddPoint(p1,POINT_BBOX);
  int loc2 = C->P->AddPoint(p2,POINT_BBOX);
  int loc3 = C->P->AddPoint(p3,POINT_BBOX);
  int loc4 = C->P->AddPoint(p4,POINT_BBOX);

  Simplex * bigtet = C->SP->NewSimplex(loc1, loc2, loc3, loc4);
  SetRadiusAndCC(bigtet);
  Setp2s(bigtet);

  foreach (int nn, nbs) {
    InsertPoint(nn,loc1);
  }

  cout << "Points Triangulated" << endl;

  set<Simplex *> junk;

  // find tets to keep
  for(siterator sit(p2s[loc1]); sit!= send(); sit++) {
    if (C->G->inSphereStrict(i,*sit)) {
      cavity.insert(*sit);
    } else {
      junk.insert(*sit);
    }
  }

  // delete extra stuff
  foreach(Simplex * s, junk) {
    C->SP->DeleteSimplex(s);
  }

  int count = 0;

  // link up new tets...
  // be sure to leave p2s with correct data...
  foreach(Simplex * s1, cavity) {
    count = 0;
    foreach( Simplex * s2, cavity ) {
      // are s1 and s2 neighbors...
      int num = s1->isNeighbor(s2);
      if (num >= 0) {
	s1->n[num] = s2;
	count++;
      }
    }

    foreach (Face f, bndry) {
      // do I hook to the bndry...
      int num = s1->isNeighbor(f.s);
      
      if (num >= 0) {
	
	if (f.v >= 0) {
	  if ( f.s->isNeighbor(s1) == f.v) {
	    s1->n[num] = f.s;
	    f.s->n[f.v] = s1;
	    count++;
	  } else {

	  }
	} else {
	  s1->n[num] = NULL;
	  count++;
	}
      }
    }
    //cout << endl;
    if (count != 4) {
      cout << "ERROR: Remove did not stitch a tet in correctly..." << endl;
      cout << count << endl;
      exit(0);
    }
    
    Setp2s(s1);

  }
    
  // i is no longer in the mesh...
  p2s.erase(i);
  /*

    // paste in the new tets...
    Simplex * s = tmpc.begin();
    Simplex * sn;
    while (s != NULL) {
        sn = s->next;
        //cout << "--"; P->print(s);

        tmpc.RemoveSimplex(s);

        // this could be a problem?
        if (G->inSphereLoose(i,s)) {
            //P->print(s);
            cavity.insert(s);
            SC.AddSimplex(s);

        } else {
            // delete this simplex...
            delete s;
        }


        s = sn;
    }

    // link everything...

    int count =0;

    // be sure to leave p2s with correct data...
    foreach(Simplex * s1, cavity) {

        //cout << "---" << endl;
        //printMe(s1);
        //P->print(s1);
        //cout << "---" << endl;
        count = 0;
        foreach( Simplex * s2, cavity ) {
            // are s1 and s2 neighbors...
            int num = s1->isNeighbor(s2);
            if (num >= 0) {
                s1->n[num] = s2;
                //cout << "---" << endl;
                //cout << num << endl;
                //printMe(s1);
                //printMe(s2);
                //cout << "---" << endl;
                count++;
                //cout << num << " ";
            }
        }
        //cout << ": ";
        foreach (Face f, bndry) {
            // do I hook to the bndry...
            int num = s1->isNeighbor(f.s);

            if (num >= 0) {

                if (f.v >= 0) {
                    if ( f.s->isNeighbor(s1) == f.v) {
                        s1->n[num] = f.s;
                        f.s->n[f.v] = s1;
                        //cout << num << " " ;
                        //cout << "--a--" << endl;
                        //cout << num << " " << f.v << endl;
                        //printMe(s1);
                        //printMe(f.s);
                        //cout << "--a--" << endl;
                        count++;
                    } else {
                        //cout << "HMMMMM" << endl;
                    }
                } else {
                    s1->n[num] = NULL;
                    //cout << "NULL ";
                    //printMe(s1);
                    //cout << num << " " ;
                    //cout << "Do I ever get here..." << endl;
                    count++;
                }
            }
        }
        //cout << endl;
        if (count != 4) {
            cout << "ERROR: Remove did not stitch a tet in correctly..." << endl;
            cout << count << endl;
            exit(0);
        }
    }

    SC.p2s.erase(i);

    //cout << star.size() << " deleted, " << cavity.size() << " formed." << endl;
    */

   }





int Mesh::SegmentFirstPoint()
{
   if (dim != 1) {
      cout << "ERROR: SegmentFirstPoint() called on non-segment." <<
      endl;
      exit(0);
   }

   set < int >::iterator edgeIt = exteriorPoints.begin();
   return *edgeIt;
}

int Mesh::SegmentNextPoint(int cur, int prev)
{

   if (dim != 1) {
      cout << "ERROR: SegmentNextPoint() called on non-segment." << endl;
      exit(0);
   }

   set < int >nbs;
   Neighbors(cur, nbs);
   if (nbs.size() == 1) {
      return *(nbs.begin());
   }

   set < int >::iterator sIt = nbs.begin();
   if (*sIt != prev) {
      return *sIt;
   }
   sIt++;
   return *sIt;
}



Mesh::Iterator::Iterator()
{
   reset();
}

void Mesh::Iterator::reset()
{
   visited.clear();
   while (Q.size() > 0) {
      Q.pop();
   }
}

void Mesh::Iterator::Start(Simplex * s)
{
   Q.push(s);
   visited.insert(s);
}

Simplex *Mesh::Iterator::Current()
{
   if (Q.size() > 0) {
      return Q.front();
   } else {
      return NULL;
   }
}

void Mesh::Iterator::operator++()
{
   Simplex *s = Q.front();
   Q.pop();
   for (int i = 0; i < 4; i++) {
      if (s->n[i] != NULL) {
         if (visited.find(s->n[i]) == visited.end()) {
            Q.push(s->n[i]);
            visited.insert(s->n[i]);
         }
      }
   }
}


//----
// Point Iterator...

Mesh::PIterator::PIterator(Mesh * mm)
{
   m = mm;
   reset();
}

void Mesh::PIterator::reset()
{

   mIt = m->p2s.begin();

   /*
      visited.clear();
      while(Q.size() > 0) {
      Q.pop();
      }
      int i = m->GetPoint();
      Q.push(i);
      visited.insert(i);
    */

}

// FIXME: this is out of date... remove soon...
void Mesh::PIterator::Start(Simplex * s)
{
}

int Mesh::PIterator::Current()
{


   if (mIt == m->p2s.end()) {
      return -1;
   }
   return mIt->first;

   /*
      if (Q.size() > 0) {
      return Q.front();
      }

      return -1;
    */
}

void Mesh::PIterator::operator++()
{


   mIt++;
   /*
      int myi = Current();

      Q.pop();
      set<int> nbs;
      m->Neighbors(myi,nbs);

      foreach(int iN, nbs) {
      if (visited.find(iN) == visited.end()) {
      Q.push(iN);
      visited.insert(iN);
      }
      }
    */
}


bool Mesh::IsChild(Mesh * m1)
{

   foreach(Mesh * m, children) {
      if (m == m1) {
         return true;
      }
   }
   return false;
}


void Mesh::SetRadiusAndCC(Simplex * s)
{
   s->cc = C->G->circumcenter(s);
   s->rsq = d2(C->P->V[s->v[0]], s->cc);
}

void Mesh::DeleteMesh()
{

   // FIXME: implement this... the code below doesn't work since
   //        DeleteSimplex messes up the mesh and the iterator
   //        gets confused...

   /*
      //iterate through and delete all child simplices...
      Iterator it;
      it.Start(GetSimplex());
      while (it.Current() != NULL) {

      C->SP->DeleteSimplex(it.Current());
    ++it;
      }

      // clear p2s
      p2s.clear();


      // clear other stuff?
    */

}

// use this later?
/*
   bool operator< (const Mesh * & x, const Mesh * & y) {

    if (x->dim < y->dim) {
        return true;
    } else if (x->dim == y->dim && x->gid < y->gid) {
        return true;
    }
    return false;
   }
 */


/*
   class siterator {
public:
      siterator();
      void operator++();
      void operator++(int);
      siterator& operator=(const siterator & it);
      bool operator==(const siterator & it);
      bool operator!=(const siterator & it);
      Simplex * operator*();
      //Simplex * operator->() const { return &(operator*()); }

      set < Simplex * >V;
      queue < Simplex * >Q;
private:

   };
*/

Mesh::siterator::siterator() {

}

Mesh::siterator::siterator(Simplex * s) {
  V.insert(s);
  Q.push_front(s);
}

void Mesh::siterator::operator++()
{
   Simplex *s = Q.front();
   Q.pop_front();
   for (int i = 0; i < 4; i++) {
      if (s->n[i] != NULL) {
         if (V.find(s->n[i]) == V.end()) {
            Q.push_back(s->n[i]);
            V.insert(s->n[i]);
         }
      }
   }
}

void Mesh::siterator::operator++(int a) {
  ++(*this);
}

Simplex * Mesh::siterator::operator*()
{
   if (Q.size() > 0) {
      return Q.front();
   } else {
      return NULL;
   }
}

Mesh::siterator& Mesh::siterator::operator=(const siterator & it) {
  Q.clear();
  V.clear();
  
  foreach(Simplex * s, it.Q) {
    Q.push_back(s);
  }
  
  foreach (Simplex * s, it.V) {
    V.insert(s);
  }

  return *this;
}


bool Mesh::siterator::operator==(const siterator & it) {

  if (Q.size() == 0 && it.Q.size() == 0) {
    return true;
  }
  if (Q.size() == it.Q.size() && V.size() == it.V.size() ) {
    return true;
  }
  return false;

}

bool Mesh::siterator::operator!=(const siterator & it) {
  return !(*this == it);
}

Mesh::siterator Mesh::sbegin() {
  return   siterator(GetSimplex());
}

Mesh::siterator Mesh::send() {
  siterator it;
  return it;
}


/*
      piterator();
      piterator(map< int, Simplex *>::iterator m);
      void operator++();
      void operator++(int);
      piterator& operator=(const piterator & it);
      bool operator==(const piterator & it);
      bool operator!=(const piterator & it);
      int operator*();
*/


Mesh::piterator::piterator() { }

Mesh::piterator::piterator(map< int, Simplex *>::iterator m) {
  mit = m;
}
      
void Mesh::piterator::operator++() {
  mit++;
}

void Mesh::piterator::operator++(int a) {
  ++(*this);
}

Mesh::piterator& Mesh::piterator::operator=(const Mesh::piterator & it) {
  mit = it.mit;

  return *this;
}

bool Mesh::piterator::operator==(const Mesh::piterator & it) {
  return (mit == it.mit);
}

bool Mesh::piterator::operator!=(const Mesh::piterator & it) {
  return (mit != it.mit);
}

int Mesh::piterator::operator*() {
  return mit->first;
}

Mesh::piterator Mesh::pbegin() {
  return piterator(p2s.begin());
}

Mesh::piterator Mesh::pend() {
  return piterator(p2s.end());
}

