#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <queue>
#include <string>
#include <list>
#include <map>
#include <set>
#include <queue>
#include <algorithm>
#include <cmath>

#include "allh.h"
#include "common.h"

PLC::PLC()
{
  P = new Points;
  G = new Geometry(this);
  SP = new SimplexPool;
  M = NULL;
  config = new Configuration(this);
  stats = new Stats;
  counter = 0;
  logfileisopen = false;
  LOGFILE = NULL;
}

PLC::PLC(PLC * C)
{
  
  //cout << "Begin Copy Constructor." << endl;
  
  // duplicate the points class...
  
  //cout << "Copy Points." << endl;
  P = new Points;
  //P = C->P;
  
  G = new Geometry(this);
  
  config = new Configuration(this);
  stats = new Stats;
  
  //G = C->G;
  SP = C->SP;
  CopyPLC(C);
  
}

void PLC::CopyPLC(PLC * C)
{
  
  P->CopyPoints(C->P);
  
  delete config;
  delete stats;
  config = C->config;
  stats = C->stats;
  name = C->name;
  
  LOGFILE = C->LOGFILE;
  logfileisopen = C->logfileisopen;
  
  // duplicate all meshes
  
  set < Mesh * >old;
  
  M = new Mesh(this);
  old.insert(C->M);
  
  old2new[C->M] = M;
  foreach(Mesh * m, C->M->children) {
    Mesh *m1 = new Mesh(this);
    old2new[m] = m1;
    old.insert(m);
  }
  //copy over the mesh details...
  
  //cout << "Set l2m." << endl;
  
  for (int i = 0; i < (int)P->V.size(); i++) {
    foreach(Mesh * m, C->P->l2m[i]) {
      P->l2m[i].insert(old2new[m]);
    }
  }
  //cout << "Copy Meshes." << endl;
  
  // copy the meshes...
  foreach(Mesh * mold, old) {
    Mesh *mnew = old2new[mold];
    // fill mnew...
    mnew->dim = mold->dim;
    mnew->id = mold->id;
    mnew->gid = mold->gid;
    mnew->name = mold->name;
    
    mnew->bbox = mold->bbox;
    for (int i=0; i<3; i++) {
      mnew->upperbound[i] = mold->upperbound[i];
      mnew->lowerbound[i] = mold->lowerbound[i];
    }
    
    mnew->np = mold->np;
    mnew->ns = mold->ns;
    mnew->dir = mold->dir;
    mnew->base = mold->base;
    mnew->convex = mold->convex;
    mnew->protect = mold->protect;
    
    foreach(Mesh * child, mold->children) {
      mnew->children.push_back(old2new[child]);
    }
    foreach(Mesh * parent, mold->parents) {
      mnew->parents.push_back(old2new[parent]);
    }
    foreach(int ex, mold->exteriorPoints) {
      mnew->exteriorPoints.insert(ex);
    }
    // copy over the list of boundary segments...
    foreach(Mesh * m1, mold->bndseg) {
      mnew->bndseg.insert(old2new[m1]);
    }
    
    // copy the simplices...
    // not really efficient but it works for now...
    map < Simplex *, Simplex * >sold2new;
    vector < Simplex * >snew;
    
    Mesh::Iterator sIt;
    sIt.Start(mold->GetSimplex());
    while (sIt.Current() != NULL) {
      
      //Simplex * s = new Simplex(sIt.Current());
      Simplex *s = SP->NewSimplex(sIt.Current());
      sold2new[sIt.Current()] = s;
      snew.push_back(s);
      ++sIt;
    }
    
    // fix the pointers...
    foreach(Simplex * s, snew) {
      for (int i = 0; i < 4; i++) {
	if (s->n[i] != NULL) {
	  s->n[i] = sold2new[s->n[i]];
	}
      }
      mnew->Setp2s(s);
    }
   }
  
  // copy the bnd structures...
  foreach(Mesh * mold, old) {
    if (mold->bnd != NULL) {
      Mesh *mnew = old2new[mold];
      
      mnew->bnd = new Boundary(mnew);
      foreach(Mesh * segment, mold->bnd->seglist) {
	mnew->bnd->seglist.push_back(old2new[segment]);
	
      }
      
      
    }
  }
  
  // copy planestoprint
  foreach(Mesh * mold, C->planestoprint) {
    planestoprint.push_back(old2new[mold]);
  }
  
  
  //pointsToProtect = C->pointsToProtect;
  
  //foreach (Mesh * mm, C->segmentsToProtect) {
  //  segmentsToProtect.insert(old2new[mm]);
  //}
  
}


void PLC::OpenLogFile(string fname)
{
  LOGFILE = new ofstream(fname.c_str());
  logfileisopen = true;
}

void PLC::Step0()
{

   // add the points in a random order...
   vector < int >permut;
   permut.resize(P0.size());
   for (int i = 0; i < (int)P0.size(); i++) {
      permut[i] = i;
   }
   random_shuffle(permut.begin(), permut.end());

   PrintBegin("Step 0: Inserting All Input Points");
   //foreach(InputPoint p0, P0) {
   for (int j = 0; j < (int)P0.size(); j++) {
      InputPoint p0 = P0[permut[j]];

      // add the point to the points class...
      int pnum = P->AddPoint(p0.id, p0.p, POINT_INPUT);

      if (logfileisopen && config->verbose) {
         (*LOGFILE) << "Inserting " << pnum << " " << P->
         V[pnum] << endl;
      }
      //cout << "Inserting " << pnum << " " << P->V[pnum] << endl;

      // insert into each mesh...
      foreach(Mesh * m, p0.l2m) {
         //cout << "Inserting point " << p0.id << " into " <<
         //  m->name << " as point number " << pnum << endl;
         //cout << p0.p << m->base << m->dir << m->bbox << endl;

         set < Simplex * >oldcavity;
         set < Simplex * >newcavity;
         m->InsertPoint(pnum, oldcavity, newcavity);

         // remove the old simplex...
         foreach(Simplex * sim, oldcavity) {
            SP->DeleteSimplex(sim);
         }
      }
   }

   FormFaceBoundaries();
   if (logfileisopen) {
      (*LOGFILE) << "Face boundaries formed successfully." << endl;
   }

   MakeProtectionList();
   if (logfileisopen) {
      (*LOGFILE) << "Acute angle protection list formed." << endl;
      (*LOGFILE) << "Protecting " << segmentsToProtect.
      size() << " segments and " << pointsToProtect.
      size() << " input points." << endl;
      (*LOGFILE) << "Lazy Protection List Size - " << lazyProtectionList.size() << endl;
      (*LOGFILE) << "Protected Input Points - " << pointsToProtect.size() << endl;
      (*LOGFILE) << "Input Points - " << P0.size() << endl;
   }

   PrintEnd("Step 0: Insert All Input Points");
}

void PLC::Step1a()
{
   PrintBegin("Step 1a: Split Edges at 1/3 Length");

   int pointsinserted = 0;
   // compute the distance to the nearest neighbor for each point...
   map < int, double >p2d;

   Mesh::PIterator pIt(M);
   pIt.Start(M->GetSimplex());
   while (pIt.Current() != -1) {

      int iP = pIt.Current();

      double mindist = M->bbox * 100;

      set < int >nbrs;
      M->Neighbors(iP, nbrs);
      foreach(int iN, nbrs) {
         double newdist = d(P->V[iN], P->V[iP]);
         if (newdist < mindist) {
            mindist = newdist;
         }
      }
      p2d[iP] = mindist;

      ++pIt;
   }

   // split edges at the appropriate distance...

   // loop through the edges...
   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {
         // loop through all simplices...


         Mesh::Iterator it;
         it.Start(m->GetSimplex());
         while (it.Current() != NULL) {
            if (!m->Exterior(it.Current())) {
               int a, b;
               a = it.Current()->v[0];
               b = it.Current()->v[1];

               Point p1, p2;
               Vector v;
               // compute the two points to insert
               v = P->V[b] - P->V[a];
               v.normalize();
               p1 = P->V[a] + p2d[a] / 3.0 * v;
               p2 = P->V[b] - p2d[b] / 3.0 * v;

               // insert the two points in this mesh and all parent meshes...
               int i1 = P->AddPoint(p1, a, POINT_SEGMENT);
               int i2 = P->AddPoint(p2, b, POINT_SEGMENT);

               InsertInMeshAndParents(i1, m);
               InsertInMeshAndParents(i2, m);

               pointsinserted += 2;

               // there is only one segment per line... break out...
               break;
            }
            ++it;

         }
      }
   }
   if (logfileisopen) {
      (*LOGFILE) << "Number of Points Inserted: " << pointsinserted <<
      endl;
   }

   PrintEnd("Step 1a: Split Edges at 1/3 Length");
}

void PLC::Step1b()
{
   PrintBegin("Step 1b: Approximate fs_1");

   int segsplits = Step1bGroomEdges();

   if (logfileisopen) {
      (*LOGFILE) << "Number of Points Inserted: " << segsplits << endl;
   }

   PrintEnd("Step 1b: Approximate fs_1");

}

int PLC::Step1bGroomEdges()
{
   set < PointMesh > PQ;
   set < SimplexMesh > SQ;


   int segsplits = 0;

   // set up a priority queue of segments to split...
   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {

         Mesh::Iterator sIt;
         sIt.Start(m->GetSimplex());
         while (sIt.Current() != NULL) {
            if (!m->Exterior(sIt.Current())) {
               if (Step1bIsEncroached(sIt.Current())) {
                  // queue this simplex...
                  Step1bQueueSegment(sIt.Current(), SQ);
               }
            }

            ++sIt;
         }
      }
   }


   set < PointMesh >::iterator pIt;
   set < SimplexMesh >::iterator sIt;

   clock_t mytime = clock();
   int mycount = 0;

   // process the queue...
   while (SQ.size() > 0 || PQ.size() > 0) {


     if ((double) (clock() - mytime) > ((double) CLOCKS_PER_SEC) / 5.0) {
	cout << "\r Queue Length: " << setw(8) << SQ.size()
              << setw(10) << ",       Items Processed: " << setw(10) <<
         mycount;
         cout.flush();
         mytime = clock();
     }
      mycount++;

      if (PQ.size() > 0) {
         // insert a point into a mesh...

         pIt = PQ.begin();
         int p = pIt->p;
         Mesh *m = pIt->m;
         // cout << m->name << " : " << p << endl;

         // insert the point...
         //InsertInMesh(p, m);
         InsertInMeshAndDequeue(p, m, SQ);

         // requeue stuff...
         if (m->dim == 3) {
            Mesh *mSeg = NULL;

            // in the 3D mesh, see if anything needs requeueing...
            foreach(Mesh * m1, P->l2m[p]) {
               if (m1->dim == 1) {

                  mSeg = m1;

                  // look at the star and consider any segments...
                  set < Simplex * >star;
                  m1->Simplices(p, star);
                  foreach(Simplex * s, star) {
                     // ask if this simplex is encroached...
                     if (s->f == SIM_NOT_QUEUED
                         && Step1bIsEncroached(s)) {

                        Step1bQueueSegment(s, SQ);
                     }
                  }
               }
            }

            if (mSeg == NULL) {
               cout << "ERROR: mSeg == NULL" << endl;
               exit(0);
            }

            set < int >nbs;
            // get the new neighbors...
            m->Neighbors(p, nbs);

            foreach(int iN, nbs) {
               // does this neighbor belong to a segment that
               // p encroaches on...

               foreach(Mesh * m1, P->l2m[iN]) {
                  if (m1->dim == 1) {
                     set < Simplex * >star;
                     m1->Simplices(iN, star);
                     foreach(Simplex * s, star) {
                        if (s->f == SIM_NOT_QUEUED
                            && !m1->Exterior(s)
                            && Step1bDoesEncroach(p, s)) {
                           Step1bQueueSegment(s, SQ);
                        }
                     }
                  }
               }
            }
         }

         PQ.erase(pIt);

      } else {
         segsplits++;
         // split a simplex...
         sIt = SQ.begin();
         Simplex *s = sIt->s;
         Mesh *m = sIt->m;

         s->f = SIM_CURRENT;

         if (m->IsEndSegment(s)) {
            // queue my spindle...
         }
         // find the midpoint...
         // queue midpoint for insertion into all applicable meshes...
         Point p = s->cc;


         int i = P->AddPoint(p, s->v[0], POINT_SEGMENT);

         PointMesh pm;

         pm.p = i;
         pm.m = m;

         PQ.insert(pm);

         foreach(Mesh * mm, m->parents) {
            pm.m = mm;
            PQ.insert(pm);
         }


         SQ.erase(sIt);
      }
   }
   cout <<
   "\r                                                               ";
   cout.flush();
   cout << "\r Finished Step 1b.";
   cout.flush();
   cout << "\r";


   return segsplits;
}


bool PLC::Step1bIsEncroached(Simplex * s)
{
   //if (s->size () != 2) {
   //  cout << "ERROR: Expecting simplex of size 2 in Step1bIsEncroached" <<
   //    endl;
   //cout << "Size is " << s->size () << endl;
   //exit (0);
   //}


  //  commenting this allows to test the intestine..  
   if (config->hsq > 0.0) {
      if (s->rsq >
          ((double)(config->step2asplitfactor * config->step2asplitfactor))*config->hsq) {
	return true;
      }
   }
  

   int pa, pb;
   if (P->type[s->v[1]] == POINT_INPUT) {
      // pa is always the (potential) input point...
      pa = s->v[1];
      pb = s->v[0];
   } else {
      pa = s->v[0];
      pb = s->v[1];
   }

   double r2 = s->rsq;

   set < int >nbsa;

   // get all the neighbors...
   M->Neighbors(pa, nbsa);

   foreach(int iNb, nbsa) {
      // loop through the neighbors to look for encroachment...
      if (iNb != pb) {
         double dist = d2(P->V[pa], P->V[iNb]);
         if (dist < 2 * r2 - config->epsilon) {
            // the point is too close...
            // make sure it is a feature size witness, though...

            if (P->type[iNb] == POINT_INPUT ||
                !includes(P->l2m[pa].begin(), P->l2m[pa].end(),
                          P->l2m[iNb].begin(), P->l2m[iNb].end())) {
               // cout << "pa Encroached ";
               // cout << s;
               // cout << P->V[pa] << P->V[pb] << P->V[iNb] << endl;
               return true;
            }
         }
      }
   }

   set < int >nbsb;
   M->Neighbors(pb, nbsb);

   foreach(int iNb, nbsb) {
      // loop through the neighbors to look for encroachment...
      if (iNb != pa) {
         double dist = d2(P->V[pb], P->V[iNb]);
         if (dist < 2 * r2 - config->epsilon) {
            // the point is too close...
            // make sure it is a feature size witness, though...
            // IMPORTANT: pa here is correct since pa is the potential input point...
            if (P->type[iNb] == POINT_INPUT ||
                !includes(P->l2m[pa].begin(), P->l2m[pa].end(),
                          P->l2m[iNb].begin(), P->l2m[iNb].end())) {
               // cout << "pb Encroached ";
               // cout << s;
               // cout << P->V[pa] << P->V[pb] << P->V[iNb] << endl;
               return true;
            }
         }
      }
   }
   // cout << "Not Encroached ";
   // cout << s;
   // cout << P->V[pa] << P->V[pb] << endl;

   if (P->type[pa] == POINT_INPUT) {
      // I am "encroached" if there is an element of my spindle that is
      // queued and my same length...
      foreach(Mesh * m, P->l2m[pa]) {
         if (m->dim == 1) {
            set < Simplex * >star;
            m->Simplices(pa, star);

            foreach(Simplex * mys, star) {
               if (mys->f == SIM_QUEUED && !m->Exterior(mys)) {
                  if (mys->rsq < 2.0 * s->rsq) {
                     return true;
                  }
               }
            }

         }
      }




   }



   return false;
}


void PLC::Step1bQueueSegment(Simplex * s, set < SimplexMesh > &SQ)
{

   if (s->size() != 2) {
      cout << "ERROR: wrong size simplex in step 1b." << endl;
   }

   int pa, pb;

   pa = s->v[0];
   pb = s->v[1];

   if (P->type[pa] == POINT_INPUT) {

   } else if (P->type[pb] == POINT_INPUT) {
      int tmp = pa;
      pa = pb;
      pb = tmp;
   } else {

      // find the mesh that this belongs to...
      foreach(Mesh * m, P->l2m[pa]) {
         if (m->dim == 1) {
            SimplexMesh sm;
            sm.s = s;
            sm.m = m;
            s->f = SIM_QUEUED;
            SQ.insert(sm);
            return;
         }
      }

      cout << "ERROR: Step1bQueueSeg should not get here..." << endl;
      return;
   }

   //cout << "-----" << endl;
   // pa is an input point... queue the spindle...
   foreach(Mesh * m, P->l2m[pa]) {
      if (m->dim == 1) {
         set < Simplex * >star;
         m->Simplices(pa, star);


         foreach(Simplex * mys, star) {
            if (!m->Exterior(mys)) {
               SimplexMesh sm;
               sm.s = mys;
               sm.m = m;
               mys->f = SIM_QUEUED;
               SQ.insert(sm);
               //cout << mys << " " << m->name << " " << d(P->V[mys->v[0]],P->V[mys->v[1]]) << endl;
            }
         }

      }
   }
   //cout << "-----" << endl;


}

bool PLC::Step1bDoesEncroach(int p, Simplex * s)
{

   if (s->size() != 2) {
      cout << "ERROR: Expecting simplex of size 2 in Step1bDoesEncroach"
           << endl;
      cout << "Size is " << s->size() << endl;
      exit(0);
   }


   int pa, pb;
   if (P->type[s->v[1]] == POINT_INPUT) {
      // pa is always the (potential) input point...
      pa = s->v[1];
      pb = s->v[0];
   } else {
      pa = s->v[0];
      pb = s->v[1];
   }

   // if I am not a feature size witness, fail...
   if (includes
                    (P->l2m[pa].begin(), P->l2m[pa].end(), P->l2m[p].begin(),
                    P->l2m[p].end())) {
      return false;
   }

   double da, db;
   da = d2(P->V[p], P->V[pa]);
   db = d2(P->V[p], P->V[pb]);

   if (da < 2.0 * s->rsq - config->epsilon
       || db < 2.0 * s->rsq - config->epsilon) {
      return true;
   }

   return false;
}

void PLC::Step2a()
{
   PrintBegin("Step 2a: Split All Segments");
   int count=0;

   StoreInputNeighbors();


   if (config->step2asplitfactor > 0) {
      count = HalveAllSegments();
   }

   if (config->step2asplitfactor > 2) {
      count += HalveAllSegments();
   }

   if (logfileisopen) {
      (*LOGFILE) << "Step2a Split Factor Set to " << config->
      step2asplitfactor << endl;
      (*LOGFILE) << "Number of Points Inserted: " << count << endl;
   }
   PrintEnd("Step 2a: Split All Segments");
}


int PLC::HalveAllSegments()
{

   int count = 0;

   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {
         vector < Point > toInsert;
         vector < int >parentPoints;

         Mesh::Iterator sIt;
         sIt.Start(m->GetSimplex());

         Simplex *s = sIt.Current();

         while (s != NULL) {
            if (!m->Exterior(sIt.Current())) {
               Point p = s->cc;
               toInsert.push_back(p);
               parentPoints.push_back(s->v[0]);
            }
            ++sIt;
            s = sIt.Current();
         }
         for (int i = 0; i < (int)toInsert.size(); i++) {
            int loc =
               P->AddPoint(toInsert[i], parentPoints[i],
                           POINT_SEGMENT);
            InsertInMeshAndParents(loc, m);
            count++;
         }
      }
   }
   return count;
}

int PLC::HalveNonEndSegments()
{


   int count = 0;

   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {
         vector < Point > toInsert;
         vector < int >parentPoints;

         Mesh::Iterator sIt;
         sIt.Start(m->GetSimplex());

         Simplex *s = sIt.Current();

         while (s != NULL) {
            if (!m->Exterior(sIt.Current())
                && !m->IsEndSegment(sIt.Current())) {
               Point p = s->cc;
               toInsert.push_back(p);
               parentPoints.push_back(s->v[0]);
            }
            ++sIt;
            s = sIt.Current();
         }
         for (int i = 0; i < (int)toInsert.size(); i++) {
            int loc =
               P->AddPoint(toInsert[i], parentPoints[i],
                           POINT_SEGMENT);
            InsertInMeshAndParents(loc, m);
            count++;
         }
      }
   }
   return count;
}


int PLC::HalveNewNonEndSegments()
{

   // this function halves all segments between each input point ip
   // and its old neighbor ip2nb[ip]


   int count = 0;

   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {

         map < int, int >::iterator mit;
         for (mit = m->ip2nb.begin(); mit != m->ip2nb.end(); mit++) {


            //cout << mit->first << " " << mit->second << " " << m->name << endl;

            Vector dir = P->V[mit->second] - P->V[mit->first];

            int cp = mit->first;
            while (cp != mit->second) {
               set < Simplex * >star;

               m->Simplices(cp, star);
               if (star.size() != 2) {
                  cout <<
                  "ERROR: problem walking down an edge in HalveNewNonEnd"
                       << endl;
               }

               int loc = -1;
               foreach(Simplex * s, star) {
                  int newp = s->v[0];
                  if (newp == cp) {
                     newp = s->v[1];
                  }
                  Vector walkdir = P->V[newp] - P->V[cp];
                  if (walkdir * dir > 0) {

                     // split this segment if it is non-end...
                     // split the end segment too...

                     if (!m->IsEndSegment(s)) {
                        Point p = s->cc;
                        loc = P->AddPoint(p, cp, POINT_SEGMENT);
                     }

                     cp = newp;
                  }

               }

               if (loc >= 0) {
                  InsertInMeshAndParents(loc, m);
                  count++;
               }

            }
         }


      }
   }

   return count;
}


void PLC::Step2b()
{

   PrintBegin("Step 2b: Approximate lfs");

   int counts = 0;
   int countf = 0;

   set < PointMesh > PQ;
   set < SimplexMesh > SQ;

   int MaxSQSize = 0;

   // set up a priority queue of segments to split...
   foreach(Mesh * m, M->children) {
      Mesh::Iterator sIt;
      sIt.Start(m->GetSimplex());
      while (sIt.Current() != NULL) {
         if (!m->Exterior(sIt.Current())) {
            if (Step2bIsEncroached(sIt.Current(), m)) {
               // queue this simplex...
               Step2bQueueSimplex(sIt.Current(), m, SQ);
            }
         }

         ++sIt;
      }

   }

   set < PointMesh >::iterator pIt;
   set < SimplexMesh >::iterator sIt;

   clock_t mytime = clock();
   int mycount = 0;

   // process the queue...
   while (SQ.size() > 0 || PQ.size() > 0) {

      if ((int)SQ.size() > MaxSQSize) {
         MaxSQSize = SQ.size();
      }

      if ((double) (clock() - mytime) > ((double) CLOCKS_PER_SEC) / 5.0) {
         cout << "\r Queue Length: " << setw(8) << SQ.size()
              << setw(10) << ",       Items Processed: " << setw(10) <<
         mycount;
         cout.flush();
         mytime = clock();
      }
      mycount++;

      if (PQ.size() > 0) {
         // insert a point into a mesh...

         pIt = PQ.begin();
         int p = pIt->p;
         Mesh *m = pIt->m;

         // pop from queue...
         PQ.erase(pIt);

         // insert the point...
         // arand replacing this...
         //InsertInMesh(p, m);
         InsertInMeshAndDequeue(p, m, SQ);

         // requeue stuff based on the 3d mesh...
         if (m->dim == 3) {
            // fix this so that stuff gets requeued appropriately ...
            Mesh *mP = NULL;

            // in the 3D mesh, see if anything needs requeueing...
            foreach(Mesh * m1, P->l2m[p]) {

               if (m1->dim == 1 || m1->dim == 2) {

                  if (m1->dim == 1 || mP == NULL) {
                     // remember the lowest dimensional mesh that I belong too...
                     mP = m1;
                  }
                  // look at the star and consider any segments...
                  set < Simplex * >star;
                  m1->Simplices(p, star);
                  foreach(Simplex * s, star) {
                     // ask if this simplex is encroached...
                     if (s->f == SIM_NOT_QUEUED
                         && Step2bIsEncroached(s, m1)
                         && !m1->Exterior(s)) {

                        Step2bQueueSimplex(s, m1, SQ);
                     }
                  }
               }
            }


            if (mP == NULL) {
               cout << "ERROR: mP == NULL in Step2b" << endl;
               exit(0);
            }

            set < int >nbs;
            // get the new neighbors...
            m->Neighbors(p, nbs);

            foreach(int iN, nbs) {
               // does this neighbor belong to a segment or face that
               // p encroaches on...

               foreach(Mesh * m1, P->l2m[iN]) {
                  if ((m1->dim == 1 || m1->dim == 2)
                      && !mP->Contains(m1)) {
                     set < Simplex * >star;
                     m1->Simplices(iN, star);
                     foreach(Simplex * s, star) {
                        if (s->f == SIM_NOT_QUEUED
                            && !m1->Exterior(s)
                            && Step2bDoesEncroach(p, s, m1)) {
                           // requeue it...
                           Step2bQueueSimplex(s, m1, SQ);
                        }
                     }
                  }
               }
            }

         }

      } else {
         // split a simplex...
         sIt = SQ.begin();
         Simplex *s = sIt->s;
         Mesh *m = sIt->m;

         // pop from queue...
         SQ.erase(sIt);

         if (s->f == SIM_QUEUED) {

            // unqueue this...
            s->f = SIM_NOT_QUEUED;

            // find the circumcenter...

            // queue midpoint for insertion into all applicable meshes...
            //Point p = G->circumcenter(s);
            Point p = s->cc;

            int reject = 0;

            // add circumcenter rejection rule...
            // potentially reject this circumcenter...
            if (Step2bReject(p, s, m)) {
               reject = 1;
               //cout << "rejecting " << p << s << s->rsq << m->name << endl;
            }
            if (reject == 1) {
               // reject this point and un-queue the simplex
               s->f = SIM_NOT_QUEUED;

            } else {
               int i = -1;
               if (m->dim == 1) {
                  i = P->AddPoint(p, s->v[0], POINT_SEGMENT);
                  counts++;
               } else if (m->dim == 2) {
                  i = P->AddPoint(p, s->v[0], POINT_FACE);
                  countf++;
               }

               PointMesh pm;

               pm.p = i;
               pm.m = m;

               PQ.insert(pm);

               foreach(Mesh * mm, m->parents) {
                  pm.m = mm;
                  PQ.insert(pm);
               }
            }

         } else {
            cout << s << endl;
            if (s->f == SIM_NOT_QUEUED) {
               cout <<
               "Something not queued is on the simplex queue..."
                    << endl;
            }
            cout << "ERROR: Something bad on the simplex queue." <<
            endl;
            exit(0);
         }
      }
   }

   counts += HalveNewNonEndSegments();


   cout <<
   "\r                                                               ";
   cout.flush();
   cout << "\r Finished Step 2b.";
   cout.flush();
   cout << "\r";
   if (logfileisopen) {
      (*LOGFILE) << endl;
      (*LOGFILE) << "Max Queue Size: " << MaxSQSize << endl;
      (*LOGFILE) << "Number of Points Inserted: " << counts <<
      " midpoints, " << countf << " circumcenters" << endl;
   }
   PrintEnd("Step 2b: Approximate lfs");

}

bool PLC::Step2bDoesEncroach(int p, Simplex * s, Mesh * m)
{

   if (s->size() == 2) {
      return Step1bDoesEncroach(p, s);
   }

   if (s->size() != 3) {
      cout << "ERROR: size of simplex is " << s->size() << endl;
      exit(0);
   }

   int pa, pb, pc;
   pa = s->v[0];
   pb = s->v[1];
   pc = s->v[2];


   if (P->IsIn(p, m)) {
      // not a feature size witness...
      return false;
   }

   double da, db, dc;
   da = d2(P->V[p], P->V[pa]);
   db = d2(P->V[p], P->V[pb]);
   dc = d2(P->V[p], P->V[pc]);


   if (da < 2.0 * s->rsq - config->epsilon
       || db < 2.0 * s->rsq - config->epsilon
       || dc < 2.0 * s->rsq - config->epsilon) {
      return true;
   }

   return false;
}


bool PLC::Step2bIsEncroached(Simplex * s, Mesh * m)
{

   if (s->size() == 2) {
      return Step1bIsEncroached(s);
   } else if (s->size() == 3) {

      // unneeded complication???
      int pa, pb, pc;
      // find the point with the largest l2m set...
      int len0, len1, len2, lenm;
      len0 = P->l2m[s->v[0]].size();
      len1 = P->l2m[s->v[1]].size();
      len2 = P->l2m[s->v[2]].size();

      lenm = max(len0, len1);
      lenm = max(lenm, len2);

      if (len0 == lenm) {
         pa = s->v[0];
         pb = s->v[1];
         pc = s->v[2];
      } else if (len1 == lenm) {
         pa = s->v[1];
         pb = s->v[2];
         pc = s->v[0];
      } else {
         pa = s->v[2];
         pb = s->v[0];
         pc = s->v[1];
      }

      //double r = s->Radius();
      double r2 = s->rsq;

      set < int >nbsa;
      // get all the neighbors...
      M->Neighbors(pa, nbsa);

      foreach(int iNb, nbsa) {
         // loop through the neighbors to look for encroachment...
         if (iNb != pb && iNb != pc) {
            double dist = d2(P->V[pa], P->V[iNb]);
            if (dist < 2 * r2 - config->epsilon) {
               // the point is too close...
               // make sure it is a feature size witness... ie it is not in the face...
               if (P->l2m[iNb].find(m) == P->l2m[iNb].end()) {
                  // cout << s << " "; P->print(s);
                  // cout << P->V[iNb] << endl;
                  return true;
               }
            }
         }
      }

      set < int >nbsb;
      M->Neighbors(pb, nbsb);
      foreach(int iNb, nbsb) {
         // loop through the neighbors to look for encroachment...
         if (iNb != pa && iNb != pc) {
            double dist = d2(P->V[pb], P->V[iNb]);
            if (dist < 2 * r2 - config->epsilon) {
               // the point is too close...
               // make sure it is a feature size witness... ie it is not in the face...
               if (P->l2m[iNb].find(m) == P->l2m[iNb].end()) {
                  // cout << s << " "; P->print(s);
                  // cout << P->V[iNb] << endl;
                  return true;
               }
            }
         }
      }

      set < int >nbsc;
      M->Neighbors(pc, nbsc);
      foreach(int iNb, nbsc) {
         // loop through the neighbors to look for encroachment...
         if (iNb != pa && iNb != pb) {
            double dist = d2(P->V[pc], P->V[iNb]);
            if (dist < 2 * r2 - config->epsilon) {
               // the point is too close...
               // make sure it is a feature size witness... ie it is not in the face...
               if (P->l2m[iNb].find(m) == P->l2m[iNb].end()) {
                  // cout << s << " "; P->print(s);
                  // cout << P->V[iNb] << endl;
                  return true;
               }
            }
         }
      }

      // not encroached, return false below...
   } else {
      cout << "ERROR: Bad Simplex Size in Step2bIsEncroached: " << s->
      size() << endl;
      exit(0);
   }


   return false;
}



bool PLC::Step2bReject(Point p, Simplex * s, Mesh * m)
{

   if (m->dim != 2) {
      return false;
   }
   // look at neighbors in the 3D mesh...
   set < int >nbs;
   //M->Cavity(p,s,nbs);
   //M->Cavity(p,nbs);

   // look only in the plane?
   m->Cavity(p, s, nbs);

   foreach(int iN, nbs) {
      // loop through all the possible new neighbors...
      // only consider segments in the face m....
      if (P->l2m[iN].find(m) != P->l2m[iN].end()) {

         double dist = d2(p, P->V[iN]);

         foreach(Mesh * m2, P->l2m[iN]) {

            // consider meshes that are on the boundary of the face...
            if (m2->dim == 1) { // only want segments

               foreach(Mesh * mparent, m2->parents) {
                  if (mparent == m) { // want segments that are on the boundary of the face...

                     set < Simplex * >star;
                     m2->Simplices(iN, star);
                     // these are the segments that iN belongs too..

                     foreach(Simplex * s2, star) {

                        // s2 must be interior to the segment...
                        // otherwise the large artificial bounding segments mess up
                        // the protecting regions at the ends...
                        if (!m2->Exterior(s2)
                            && dist <
                            2 * s2->rsq - config->epsilon) {
                           //cout << "Reject " << dist << " " << 2*s2->rsq << endl;
                           return true;
                        }
                     }
                     break;
                  }
               }
            }
         }
      }
   }
   return false;
}


void PLC::Step2bQueueSimplex(Simplex * s, Mesh * m,
                             set < SimplexMesh > &SQ)
{
   if (s->size() == 2) {
      // use step1bQueueSimplex...
      Step1bQueueSegment(s, SQ);
   } else {
      // queue the triangle...
      SimplexMesh sm;
      sm.s = s;
      sm.m = m;
      s->f = SIM_QUEUED;
      SQ.insert(sm);
   }
}


void PLC::Step2c()
{
  if (config->useintestine) {
    PrintBegin("Step 2c: Insert Intestine");
    
    M->intestine = new Intestine();

    M->intestine->BuildIntestine(this);

    HalveNonEndSegments();
  
    PrintEnd("Step 2c: Insert Intestine");
    
  } else {
    
    PrintBegin("Step 2c: Insert Collar");
    
    // insert the collar...
    
    bool insphereval = config->exactinsphere;
    //config->exactinsphere = false;
    
    int midpoints = 0;
    int collarpoints = 0;
    
    // insert the collar...
    foreach(Mesh * m, M->children) {
      if (m->dim == 2) {
	m->col = new Collar(m);
	collarpoints += m->col->BuildCollar();
      }
    }
    
    if (logfileisopen) {
      (*LOGFILE) << "Number of Points Inserted: " << midpoints <<
	" midpoints, " << collarpoints << " collar points" << endl;
    }
    
    config->exactinsphere = insphereval;
    
    PrintEnd("Step 2c: Insert Collar");
  }
}  
void PLC::Step3Intestine()
{

  PrintBegin("Step 3: Create Conforming Quality Mesh");

  // priority queues...
   set < PointMesh > PQ;
   set < IntestineItem > IQ;
   set < SimplexMesh > SQ;

   // queue up initial simplices to split
   // first conformity...
   // set up a priority queue of simplices to
   foreach(Mesh * m, M->children) {
      //cout << m->name << endl;
      Mesh::Iterator sIt;
      sIt.Start(m->GetSimplex());
      while (sIt.Current() != NULL) {
         // do not queue exterior or collar elements: the collar starts ok...
         if (!m->Exterior(sIt.Current())
             && !m->IsIntestine(sIt.Current())) {
            if (Step3IsEncroached(sIt.Current())) {


               // queue this simplex...
               SimplexMesh sm;
               sm.s = sIt.Current();
               sm.m = m;

               // Queue the simplex...
               sm.s->f = SIM_QUEUED;
               SQ.insert(sm);

               //cout << "Inserting " << sm.s << " " << m->name << endl;
            }
         }

         ++sIt;
      }
   }

   Mesh::Iterator sIt1;
   sIt1.Start(M->GetSimplex());
   while (sIt1.Current() != NULL) {
      // do not queue exterior or collar elements: the collar starts ok...
     if (G->IsPoorQuality(sIt1.Current()) && 
	 !M->IsIntestine(sIt1.Current())) {
         // queue this simplex...
         SimplexMesh sm;
         sm.s = sIt1.Current();
         sm.m = M;

         // Queue the simplex...
         sm.s->f = SIM_QUEUED;
         SQ.insert(sm);
      }
     vector<Simplex> faces;
     sIt1.Current()->Faces(faces);
     foreach (Simplex f, faces) {
       SetRadiusAndCC(&f);
       Simplex stmp(f.v[0],f.v[1],f.v[2]);
       if (M->intestine->IsIntestineFace(f.v[0],f.v[1], f.v[2]) &&
	   M->IsIntestine(&stmp)) {
	 //cout << &f << endl;
	 set<int> nbrs;
	 M->Neighbors(f.v[0],nbrs);
	 M->Neighbors(f.v[1],nbrs);
	 M->Neighbors(f.v[2],nbrs);
	 
	 foreach (int iii, nbrs) {

	   if (G->inSphereReallyStrict(iii,&f)) {
	     //cout << "There " << &f << endl;
	     IntestineItem im;
	     im.a = f.v[0];
	     im.b = f.v[1];
	     im.c = f.v[2];
	     im.sort();
	     IQ.insert(im);
	   }
	 }
       }
     }

      ++sIt1;
   }

   int mycount = 0;
   clock_t mytime = 0;;


   set < PointMesh >::iterator pIt;
   set < IntestineItem >::iterator iIt;
   set < SimplexMesh >::iterator sIt;

   while (SQ.size() > 0 || IQ.size() > 0 || PQ.size() > 0) {

      mycount++;

      //cout << "Queues " << SQ.size() << " " << IQ.size() << " " << PQ.size() << endl;

      if (config->maxiterations >= 0 && mycount > config->maxiterations) {
         cout << endl << "WARNING: algortihm terminated based on number of operations"
              << endl;
         break;
      }

      if ((double) (clock() - mytime) > ((double) CLOCKS_PER_SEC) / 5.0) {

	cout << "\r Queue Length: " << setw(8) << SQ.size()
              << setw(10) << ",       Items Processed: " << setw(10) <<
         mycount;
         cout.flush();
         mytime = clock();
      }

      if (PQ.size() > 0) {
         pIt = PQ.begin();
         int p = pIt->p;
         Mesh *m = pIt->m;
         //cout << p << " " << P->V.size() << endl;
         //cout << P->V[p] << " " << endl;

         //cout << "Forcing " << p << " into " << m->name << endl;

         // pop from queue...
         PQ.erase(pIt);

         //cout << "Inserting " << p << "," << P->V[p] << ","<< P->type[p] << " in " << m->name << endl;

         // insert the point...
         InsertInMeshAndDequeue(p, m, SQ);

	 if (m->dim == 3) {
	   // requeue stuff
            set < SimplexMesh > encroached;
	    set <IntestineItem> encroachedintestine;
            Step3EncroachedInt(p, encroached, encroachedintestine);
	    // intestine case...

            foreach(SimplexMesh sm, encroached) {
	      if (!sm.m->IsIntestine(sm.s) && sm.s->f != SIM_QUEUED) {
		sm.s->f = SIM_QUEUED;
		SQ.insert(sm);
	      }
            }

            foreach(IntestineItem ii, encroachedintestine) {
	      IQ.insert(ii);

	      //cout << "II " << encroachedintestine.size() << endl;
	      //exit(0);
	    }


            // queue any new skinnies...

            set < Simplex * >star;
            M->Simplices(p, star);
            foreach(Simplex * s, star) {
	      if (G->IsPoorQuality(s) && !M->IsIntestine(s)) {
                  SimplexMesh sm;
                  sm.s = s;
                  sm.m = M;
                  sm.s->f = SIM_QUEUED;
                  SQ.insert(sm);
               }
            }

            // queue any new simplices which are encroached...

            foreach(Mesh * mm, P->l2m[p]) {
               if (mm->dim < 3) {
                  set < Simplex * >star;
                  mm->Simplices(p, star);

                  foreach(Simplex * news, star) {

                     if (!mm->Exterior(news)
                         && Step3IsEncroached(news)) {
                        if (mm->IsIntestine(news)) {
			  // don't worry about these guys...
                        } else if (news->f != SIM_QUEUED) {
                           SimplexMesh sm;
                           sm.s = news;
                           sm.m = mm;
                           // this simplex should be unqueued since it is "new"
                           sm.s->f = SIM_QUEUED;
                           SQ.insert(sm);
                        }
                     }
                  }
               }
            }
	 }
      } else if (IQ.size() > 0) {

	iIt = IQ.begin();

	int a = iIt->a;
	int b = iIt->b;
	int c = iIt->c;

	IQ.erase(iIt);

	//cout << endl << IQ.size() << "  Intestine Split " <<  a << " " << b <<" " << c << endl;
	//cout << P->type[a] << " " << P->type[b] <<" " << P->type[c] << endl;
	//cout << P->V[a] << " " << P->V[b] <<" " << P->V[c] << endl;

	Simplex stmp(a,b,c);
	bool stillexists = false;
	
	if (c != -1) {
	    
	    if (M->intestine->IsIntestineFace(a,b,c)
		&& M->IsIntestine(&stmp)) {
		stillexists = true;
	    }
	    /*
	    set<Simplex *> star;
	    M->Simplices(a,star);
	    foreach (Simplex * sstar, star) {
		if (!M->IsIntestine(sstar)) {
		    //if (!P->IsIntestine(sstar->v[0]) ||
		    //  !P->IsIntestine(sstar->v[1]) ||
		    //  !P->IsIntestine(sstar->v[2]) ||
		    //  !P->IsIntestine(sstar->v[3])) {
		    
		    //	      if ( ((c == -1) &&  sstar->contains(b) ) || 
		    //     (sstar->contains(b) && sstar->contains(c)) ) {
		    if (sstar->contains(b) && sstar->contains(c)) {
			
			stillexists = true;
			
		    }
		}
		
	    }
	    */
	} else {
	    if (M->intestine->IsIntestineProtected(a, b)) {
		stillexists = true;
	    }
	}
	
	
	//cout << "SE "<< stillexists << endl;
	if (stillexists) {
	  Point p;
	  
	  bool yield = false;
	  if (c != -1) {
	    // should I yield?
	    p = M->intestine->IntestineFaceSplitPoint(a,b,c);
	    set < int >nbs;
	    M->Cavity(p, M->GetSimplex(a), nbs);

	    foreach(int iN, nbs) {
	      
	      set<int> intNbs;
	      M->intestine->IntestineArcsSegsRings(iN,intNbs);

	      foreach (int jN, intNbs) {
		Simplex stmp(iN,jN);
		SetRadiusAndCC(&stmp);

		if (iN < jN && G->inSphereReallyStrict(p,&stmp)) {
		  yield = true;

		  IntestineItem im;
		  im.a = iN;
		  im.b = jN;
		  im.sort();
		  IQ.insert(im);

		  //cout << iN << " " << jN << endl;
		}
	      }
	      
	    }
	  }
	  //cout << "Yield " << yield << endl;
	  
	  if (!yield) {
	    
	    Mesh * minsert = NULL;
	    int loc = M->intestine->SplitIntestine(a,b,c,minsert);
	    //cout << "Loc " << loc << " " << minsert->name << endl;


	    if (loc != -1) {
	      //cout << P->V[loc] << " " << P->type[loc] << endl;
	      PointMesh pm;
	      
	      //cout <<"Collar Insertion" << endl;
	      //cout << i << " " << P->V[i] << endl;
	      //cout << "Splitting " << P->V[a] << " " << P->V[b] << endl;
	      pm.p = loc;
	      pm.m = minsert;
	      
	      PQ.insert(pm);
	      
	      foreach(Mesh * mm, minsert->parents) {
               pm.m = mm;
               PQ.insert(pm);
	      }
	    }
	  }
	}
      } else if (SQ.size() > 0) {
         sIt = SQ.begin();
         Simplex *s = sIt->s;
         Mesh *m = sIt->m;

         sIt++;
         
	 
	 //cout << "Current: " << s << "," << m->name << endl;
         //if (sIt != SQ.end()) {
         //  cout << "Next: " << sIt->s << "," << sIt->m->name << endl;
         //}
	 

         // pop from queue...
         SQ.erase(SQ.begin());

         if (s->f == SIM_REMOVED) {
            cout << "This should not occur" << endl;
            //cout << "Removing " << s << endl;
            //delete s;
            SP->DeleteSimplex(s);
         } else if (s->f == SIM_QUEUED) {

            // redundant check for debugging
            if (m->Exterior(s)) {
               cout << "ERROR: splitting exterior simplex " << endl;
               cout << m->name << " " << s << endl;
               exit(0);
            }
            if (m->IsIntestine(s)) {
               cout << "ERROR: Splitting Intestine Stuff" << endl;
            }

            // de-queue this...
            s->f = SIM_NOT_QUEUED;

            // find the circumcenter...

            // queue midpoint for insertion into all applicable meshes...
            if (s->rsq == -1.0) {
               cout << endl << "ERROR: radius not set in simplex." <<
               endl;
               cout << s << " " << s->rsq << endl;
               cout << m->name << endl;
               exit(0);
            }

            Point p = s->cc;
            //int reject = 0;

            set < SimplexMesh > encroached;
            set < IntestineItem > encroachedintestine;
            Step3EncroachedInt(p, s, m, encroached, encroachedintestine);

	    // yield...
	    if (encroached.size() > 0 || encroachedintestine.size() > 0) {
	      // queue the encroached stuff instead...
	      
	      foreach(SimplexMesh mysm, encroached) {
		
		if (mysm.s->f == SIM_REMOVED) {
		  cout << endl <<
		    "ERROR: removed simplex seems to be encroached"
		       << endl;
                        cout << mysm.s << " " << mysm.m->name << endl;
                        exit(0);
		}
		if (mysm.s->f != SIM_QUEUED) {
		  mysm.s->f = SIM_QUEUED;
		  SQ.insert(mysm);
		}
	      }
	      
	      foreach(IntestineItem im, encroachedintestine) {
		im.sort();
		IQ.insert(im);
	      }
	      
	      SimplexMesh sm;
	      sm.s = s;
	      sm.m = m;
	      s->f = SIM_QUEUED;
	      // requeue myself...
	      SQ.insert(sm);
	      
	    } else {
	      
	      // insert the circumcenter...
	      int i = -1;
	      
	      if (m->dim == 1) {
		i = P->AddPoint(p, s->v[0], POINT_SEGMENT);
		//counts++;
	      } else if (m->dim == 2) {
		i = P->AddPoint(p, s->v[0], POINT_FACE);
		//countf++;
	      } else if (m->dim == 3) {
		i = P->AddPoint(p, s->v[0], POINT_3D);
		//countt++;
	      }
	      
	      //cout << "Adding point " << p << " as CC of " << s << endl;
	      //cout << "Point ID " << i << endl;
	      //P->print(s);
	      
	      PointMesh pm;
	      
	      pm.p = i;
	      pm.m = m;
	      
	      PQ.insert(pm);
	      
	      foreach(Mesh * mm, m->parents) {
		pm.m = mm;
		PQ.insert(pm);
	      }
	    }
	 }
	 
      }

   }

  PrintEnd("Step 3: Create Conforming Quality Mesh");

}  


void PLC::Step3()
{
  if (M->intestine) {
    Step3Intestine();
    return;
  }

   PrintBegin("Step 3: Create Conforming Quality Mesh");

   int counts = 0;      // segment splits
   int countc = 0;      // collar splits
   int countf = 0;      // face splits
   int countt = 0;      // tet splits

   set < PointMesh > PQ;
   set < CollarMesh > CQ;
   set < SimplexMesh > SQ;

   // queue up initial simplices to split
   // first conformity...
   // set up a priority queue of simplices to
   foreach(Mesh * m, M->children) {
      //cout << m->name << endl;
      Mesh::Iterator sIt;
      sIt.Start(m->GetSimplex());
      while (sIt.Current() != NULL) {
         // do not queue exterior or collar elements: the collar starts ok...
         if (!m->Exterior(sIt.Current())
             && !m->IsCollar(sIt.Current())) {
            if (Step3IsEncroached(sIt.Current())) {


               // queue this simplex...
               SimplexMesh sm;
               sm.s = sIt.Current();
               sm.m = m;

               // Queue the simplex...
               sm.s->f = SIM_QUEUED;
               SQ.insert(sm);

               //cout << "Inserting " << sm.s << " " << m->name << endl;
            }
         }

         ++sIt;
      }
   }


   foreach(Mesh * m, M->children) {
      if (m->dim == 2) {
         // look at the collar... is it encroached?
         Simplex s;
         s.v[2] = -1;
         s.v[3] = -1;

         map < int, int >::iterator it;
         for (it = m->col->col2next.begin();
              it != m->col->col2next.end(); it++) {
            if (s.v[1] != -1) {
               s.v[0] = it->first;
               s.v[1] = it->second;
               if (Step3IsEncroached(&s)) {
                  // queue this collar segment...
                  CollarMesh cm;
                  cm.a = it->first;
                  cm.b = it->second;
                  cm.m = m;
                  CQ.insert(cm);

               }
            }
         }
      }
   }

   // queue poor quality elements also...

   Mesh::Iterator sIt1;
   sIt1.Start(M->GetSimplex());
   while (sIt1.Current() != NULL) {
      // do not queue exterior or collar elements: the collar starts ok...
      if (G->IsPoorQuality(sIt1.Current())) {
         // queue this simplex...
         SimplexMesh sm;
         sm.s = sIt1.Current();
         sm.m = M;

         // Queue the simplex...
         sm.s->f = SIM_QUEUED;
         SQ.insert(sm);
      }
      ++sIt1;
   }

   if (logfileisopen) {
      (*LOGFILE) << "Initial Queue Size: " << SQ.
      size() << " simplices and " << CQ.
      size() << " collar segments." << endl;
   }

   set < PointMesh >::iterator pIt;
   set < CollarMesh >::iterator cIt;
   set < SimplexMesh >::iterator sIt;

   int mycount = 0;

   clock_t mytime = clock();
   while (SQ.size() > 0 || CQ.size() > 0 || PQ.size() > 0) {

      mycount++;

      if (config->maxiterations >= 0 && mycount > config->maxiterations) {
         cout << endl <<
         "WARNING: algortihm terminated based on number of operations"
              << endl;
         break;
      }

      if ((double) (clock() - mytime) > ((double) CLOCKS_PER_SEC) / 5.0) {

         cout << "\r Queue Length: " << setw(8) << SQ.size() + CQ.size()
              << setw(10) << ",       Items Processed: " << setw(10) <<
         mycount;
         cout.flush();
         mytime = clock();
      }

      //cout << "Queues " << SQ.size() << " " << CQ.size()
      //     << " " << PQ.size() << endl;


      if (PQ.size() > 0) {
         pIt = PQ.begin();
         int p = pIt->p;
         Mesh *m = pIt->m;
                  
         // pop from queue...
         PQ.erase(pIt);

         //cout << "Inserting " << p << "," << P->V[p] << ","<< P->type[p] << " in " << m->name << endl;
	 //if (P->type[p] == POINT_COLLAR && m->dim == 2) {
	 //  if (m->col->col2p.find(p) != m->col->col2p.end()) {
	 //    cout << "Input Point " << m->col->col2p[p] << endl;
	 //  }
	 // }
	 

         // insert the point...
         InsertInMeshAndDequeue(p, m, SQ);

         // requeue here...
         if (m->dim == 3) {

            set < SimplexMesh > encroached;
            Step3Encroached(p, m, encroached);

            foreach(SimplexMesh sm, encroached) {

               // do I encroach a collar element
               // if so, split the collar boundary...
               if (sm.m->IsCollar(sm.s)) {

                  if (sm.m->dim == 2) {
                     //if (P->type[p] != POINT_COLLAR) {
                     //  cout << "I don't think this should happen..." << endl;
                     //}
                     //cout << p << " " << sm.s << " " << sm.m->name << endl;

                     Step3QueueCollarEdge(p, sm.s, sm.m, CQ);
                  } else {

		    /*
                     // do nothing... this is probably bad...
                     cout << endl <<
                     "ERROR: I don't think this should happen"
                          << endl;
                     cout << "Dimension " << sm.m->dim << endl;
                     cout << p << " " << P->V[p] << " " << P->
                     type[p] << " " << m->name << endl;
                     cout << sm.m->name << " " << sm.s << endl;
                     P->print(sm.s);
                     cout << "---" << endl;
                     foreach(Mesh * mmm, P->l2m[p]) {
                        cout << mmm->name << endl;
                        if (mmm->dim == 2) {
                           mmm->PrintOff("../error.off");
                        }
                     }

                     //return;
                     exit(0);
		    */
                  }

               } else {
                  // queue this stuff....
                  if (sm.s->f != SIM_QUEUED) {
                     sm.s->f = SIM_QUEUED;
                     SQ.insert(sm);
                  }
               }
            }

            // queue any new skinnies...

            set < Simplex * >star;
            M->Simplices(p, star);
            foreach(Simplex * s, star) {
               if (G->IsPoorQuality(s)) {
                  SimplexMesh sm;
                  sm.s = s;
                  sm.m = M;
                  sm.s->f = SIM_QUEUED;
                  SQ.insert(sm);
               }
            }

            // queue any new simplices which are encroached...

            foreach(Mesh * mm, P->l2m[p]) {
               if (mm->dim < 3) {
                  set < Simplex * >star;
                  mm->Simplices(p, star);

                  foreach(Simplex * news, star) {

                     if (!mm->Exterior(news)
                         && Step3IsEncroached(news)) {
                        if (mm->IsCollar(news)) {
                           // the collar should get split again to correct this...

                           set < int >enc;

                           Step3EncroachingPoint(news, enc);


                           //int stmp = CQ.size();
                           foreach(int ie, enc) {
                              Step3QueueCollarEdge(ie, news, mm,
                                                   CQ);
                           }

                        } else if (news->f != SIM_QUEUED) {
                           SimplexMesh sm;
                           sm.s = news;
                           sm.m = mm;
                           // this simplex should be unqueued since it is "new"
                           sm.s->f = SIM_QUEUED;
                           SQ.insert(sm);
                        }
                     }
                  }
               }
            }

            if (P->type[p] == POINT_COLLAR) {

               Mesh *myface = NULL;
               foreach(Mesh * m4, P->l2m[p]) {
                  if (m4->dim == 2) {
                     myface = m4;
                     break;
                  }
               }

               // ask if the new collar segments are encroached...
               //Simplex * s1 = new Simplex(p,myface->col->col2next[p]);
               Simplex *s1 =
                  SP->NewSimplex(p, myface->col->col2next[p]);
               SetRadiusAndCC(s1);
               //Simplex * s2 = new Simplex(p,myface->col->col2prev[p]);
               Simplex *s2 =
                  SP->NewSimplex(p, myface->col->col2prev[p]);
               SetRadiusAndCC(s2);

               if (Step3IsEncroached(s1)) {
                  //queue s1...
                  CollarMesh cm;
                  cm.a = s1->v[0];
                  cm.b = s1->v[1];
                  cm.m = myface;
                  CQ.insert(cm);
               }
               if (Step3IsEncroached(s2)) {
                  CollarMesh cm;
                  cm.a = s2->v[0];
                  cm.b = s2->v[1];
                  cm.m = myface;
                  CQ.insert(cm);
               }
               //delete s1;
               SP->DeleteSimplex(s1);
               //delete s2;
               SP->DeleteSimplex(s2);

            }

         }
      } else if (CQ.size() > 0) {

         cIt = CQ.begin();
         int a = cIt->a;
         int b = cIt->b;
         Mesh *m = cIt->m;

         // pop from queue...
         CQ.erase(cIt);

         if (m->col->Exists(a, b)) {


            int i = m->col->SplitCollar(a, b);
            countc++;
            PointMesh pm;

            //cout <<"Collar Insertion" << endl;
            //cout << i << " " << P->V[i] << endl;
            //cout << "Splitting " << P->V[a] << " " << P->V[b] << endl;
            pm.p = i;
            pm.m = m;

            PQ.insert(pm);

            foreach(Mesh * mm, m->parents) {
               pm.m = mm;
               PQ.insert(pm);
            }

         } else {
            // do nothing, this collar segment must have already been split...
         }
      } else if (SQ.size() > 0) {
         sIt = SQ.begin();
         Simplex *s = sIt->s;
         Mesh *m = sIt->m;

         sIt++;
	 
         //cout << "Current: " << s << "," << m->name << endl;
	 //P->print(s);
	 //for (int i = 0; i < s->size(); i++) {
	 //  cout << P->type[s->v[i]] << " "; 
	 //  if ( P->type[s->v[i]] != POINT_INPUT &&  P->type[s->v[i]] != POINT_BBOX) {
	 //    cout<< P->LowestDimensionalContainingMesh(s->v[i])->name;
	 //  }
	 //  cout << endl;
	 // }

         //if (sIt != SQ.end()) {
         //  cout << "Next: " << sIt->s << "," << sIt->m->name << endl;
         //}
	 

         // pop from queue...
         SQ.erase(SQ.begin());

         if (s->f == SIM_REMOVED) {

            cout << "This should not occur" << endl;
            //cout << "Removing " << s << endl;
            //delete s;
            SP->DeleteSimplex(s);
         } else if (s->f == SIM_QUEUED) {

	   // redundant check for debugging
	   if (m->Exterior(s)) {
	     cout << "ERROR: splitting exterior simplex " << endl;
	     cout << m->name << " " << s << endl;
	     exit(0);
	   }
	   if (m->IsCollar(s)) {
	     cout << "ERROR: Splitting Collar Stuff" << endl;
	     cout << s << endl;
	   }

	   // de-queue this...
	   s->f = SIM_NOT_QUEUED;
	   
	   // find the circumcenter...
	   
	   // queue midpoint for insertion into all applicable meshes...
	   //Point p = G->circumcenter(s);
	   if (s->rsq == -1.0) {
	     cout << endl << "ERROR: radius not set in simplex." <<
	       endl;
	     cout << s << " " << s->rsq << endl;
	     cout << m->name << endl;
	     exit(0);
	   }
	   
	   
	   Point p = s->cc;
	   //int reject = 0;
	   
	   //cout << "Proposed CC " << p << " of simplex " << s << endl;
            //P->print(s);
	   
	   bool lazySkip = false;
	   

	   if (config->lazyvertexprotection) {
	     // handle lazy protection case...
	     if (m->dim <= 2) {
	       set<int> nbs;
	       m->Cavity(p, s, nbs);
	       foreach(int a, nbs) {
		 
		 if (P->type[a] == POINT_INPUT && 
		     lazyProtectionList.find(P->ids[a]) != lazyProtectionList.end()) {
		   double mydist = d(P->V[a],p);
		   
		   if (mydist < 2.0*lazyProtectionDistance[a]) {
		     
		     lazySkip = true;
		     
		     // add the collar around this point.
		     
		     foreach(int nb, lazyProtectionNeighbors[P->ids[a]]) {
		       
		       PointMesh pm;
		       
		       Mesh * m  = lazyNb2Segment[nb];		    
		       
		       pm.p = nb;
		       pm.m = m;
		       
		       PQ.insert(pm);
		       
		       foreach(Mesh * mm, m->parents) {
			 pm.m = mm;
			 PQ.insert(pm);
		       }		    
		     }
		     
		     // remove this vertex from the lazy protection list...
		     lazyProtectionList.erase(lazyProtectionList.find(P->ids[a]));
		     
		   }
		 }
	       }
	     }
	   }

	   
	   if (!lazySkip) {
	      
	     // potentially reject this circumcenter...
	     // decide to reject or yield...
	     
	     set < SimplexMesh > encroached;
	     set < CollarMesh > encroachedcollar;
	     if (Step3Encroached(p, s, m, encroached, encroachedcollar) == 1) {
	       // reject this point and dequeue the simplex
	       //cout << "Reject "<< endl;
	       if (m->dim < 3) {
		 cout << "ERROR: bad reject " << endl;
		 //exit(0);
	       }
	       
	       s->f = SIM_NOT_QUEUED;
	       
	     } else {
	       // yield...
	       if (encroached.size() > 0 || encroachedcollar.size() > 0) {
		 // queue the encroached stuff instead...
		 
		 foreach(SimplexMesh mysm, encroached) {
		   
		   if (mysm.s->f == SIM_REMOVED) {
		     cout << endl <<
		       "ERROR: removed simplex seems to be encroached"
			  << endl;
		     cout << mysm.s << " " << mysm.m->name << endl;
		     exit(0);
		   }
		   //if (mysm.s->f != SIM_QUEUED || 1) { // this was a hack used at one point... seems unneeded...
		   if (mysm.s->f != SIM_QUEUED) {
		     mysm.s->f = SIM_QUEUED;
		     SQ.insert(mysm);
		   }
		   //else {
		   //  cout << endl << "Not Queueing " << mysm.s  << " in " << mysm.m->name << endl;
		   //}
		   
		 }
		 
		 foreach(CollarMesh mycm, encroachedcollar) {
		   CQ.insert(mycm);
		 }
		 
		 SimplexMesh sm;
		 sm.s = s;
		 sm.m = m;
		 s->f = SIM_QUEUED;
		 // requeue myself...
		 SQ.insert(sm);
		 
	       } else {
		 
		 // insert the circumcenter...
		 int i = -1;
		 
		 if (m->dim == 1) {
		   i = P->AddPoint(p, s->v[0], POINT_SEGMENT);
		   counts++;
		 } else if (m->dim == 2) {
		   i = P->AddPoint(p, s->v[0], POINT_FACE);
		   countf++;
		 } else if (m->dim == 3) {
		   i = P->AddPoint(p, s->v[0], POINT_3D);
		   countt++;
		 }
		 
		 //cout << "Adding point " << p << " as CC of " << s << endl;
		 //cout << "Point ID " << i << endl;
		 //P->print(s);
		 
		 PointMesh pm;
		 
		 pm.p = i;
		 pm.m = m;
		 
		 PQ.insert(pm);
		 
		 foreach(Mesh * mm, m->parents) {
		   pm.m = mm;
		   PQ.insert(pm);
		 }
	       }
	     }
	   }
	 } else {
	   cout << endl << "ERROR: queue is messed up." << endl;
	   cout << m->name << " " << s << endl;
	   exit(0);
	 }      
      }
   }

   cout << "\r                                                               ";
   cout.flush();
   cout << "\r Finished Step 3.";
   cout.flush();
   cout << "\r";

   //cout << "Remaining Queue " << endl;
   //foreach (SimplexMesh sm, SQ) {
   //  cout << sm.s << " in " << sm.m->name << endl;
   //}

   cout << "Number of Points Inserted: " << counts + countt + countf +
   countc << endl;

   PrintEnd("Step 3: Create Conforming Quality Mesh");
}


bool PLC::Step3IsEncroached(Simplex * s)
{
   set < int >nbs;


   for (int i = 0; i < s->size(); i++) {
      M->Neighbors(s->v[i], nbs);
   }

   foreach(int i, nbs) {
      // does this neighbor encroach the simplex?
      if (!s->contains(i)) {
         if (G->inSphereReallyStrict(i, s)) {
            return true;
         }
      }
   }

   return false;
}


void PLC::Step3EncroachingPoint(Simplex * s, set < int >&enc)
{
   set < int >nbs;

   for (int i = 0; i < s->size(); i++) {
      M->Neighbors(s->v[i], nbs);
   }

   int count = 0;

   foreach(int i, nbs) {
      // does this neighbor encroach the simplex?
      if (!s->contains(i)) {
         if (G->inSphereReallyStrict(i, s)) {
            enc.insert(i);
            count++;
         }
      }
   }

}


// if i encroach collar simplex s, queue the
// correct edge of s on the collar queue...
void PLC::Step3QueueCollarEdge(int p, Simplex * s, Mesh * m,
                               set < CollarMesh > &CQ)
{

   //cout << "QCE " << p << " " << s << " " << m->name << endl;

   int cp1 = -1, cp2 = -1, cp3 = -1;
   if (m->col->Exists(s->v[0]) &&
       m->col->Exists(s->v[1]) && m->col->Exists(s->v[2])) {
      cp1 = s->v[0];
      cp2 = s->v[1];
      cp3 = s->v[2];
   } else if (m->col->Exists(s->v[0]) && m->col->Exists(s->v[1])) {
      cp1 = s->v[0];
      cp2 = s->v[1];
   } else if (m->col->Exists(s->v[0]) && m->col->Exists(s->v[2])) {
      cp1 = s->v[0];
      cp2 = s->v[2];
   } else if (m->col->Exists(s->v[1]) && m->col->Exists(s->v[2])) {
      cp1 = s->v[2];
      cp2 = s->v[1];
   }
   // FIXME: this may be buggy...
   // need to handle the case where cp3 != -1
   if (cp1 != -1) {
      if (cp3 != -1) {

         int inserted = 0;
         Vector v1, v2;
         v1 = P->V[p] - P->V[cp1];
         v2 = P->V[cp2] - P->V[cp1];
         double d1 = v1 * v2 / v2.len2();

         if (m->col->Exists(cp1, cp2) && 0 <= d1 && d1 <= 1) {
            // queue this collar segment...
            CollarMesh cm;
            cm.a = cp1;
            cm.b = cp2;
            cm.m = m;
            CQ.insert(cm);
            inserted++;
         }


         v1 = P->V[p] - P->V[cp1];
         v2 = P->V[cp3] - P->V[cp1];
         d1 = v1 * v2 / v2.len2();
         if (m->col->Exists(cp1, cp3) && 0 <= d1 && d1 <= 1) {
            // queue this collar segment...
            CollarMesh cm;
            cm.a = cp1;
            cm.b = cp3;
            cm.m = m;
            CQ.insert(cm);
            inserted++;
         }

         v1 = P->V[p] - P->V[cp3];
         v2 = P->V[cp2] - P->V[cp3];
         d1 = v1 * v2 / v2.len2();
         if (m->col->Exists(cp3, cp2) && 0 <= d1 && d1 <= 1) {
            // queue this collar segment...
            CollarMesh cm;
            cm.a = cp3;
            cm.b = cp2;
            cm.m = m;
            CQ.insert(cm);
            inserted++;
         }

         if (inserted == 0) {
	   cout << "This shouldn't fail..." << endl;
         } else {
	   //cout << "Inserted=" << inserted << endl;
         }


      } else {
         Vector v1, v2;
         v1 = P->V[p] - P->V[cp1];
         v2 = P->V[cp2] - P->V[cp1];
         double d1 = v1 * v2 / v2.len2();

         if (m->col->Exists(cp1, cp2) && 0 <= d1 && d1 <= 1) {
            //if (m->col->Exists(cp1,cp2)) {
            // queue this collar segment...
            CollarMesh cm;
            cm.a = cp1;
            cm.b = cp2;
            cm.m = m;
            CQ.insert(cm);
         } else {

            if (!m->col->Exists(cp1, cp2)) {
	      //cout << "Seems ok to be here" << endl;
            } else {
               cout << "Queue Still failing" << endl;
               cout << P->V[p] << " " << P->V[cp1] << " " << P->
               V[cp2] << endl;
               cout << P->type[p] << " " << P->type[cp1] << " " << P->
               type[cp2] << endl;
               cout << v1 << " " << v2 << endl;
               cout << m->name << endl;
            }
         }

      }
   }
}


//
//  given point p, and Mesh * m, make a list of simplices in mesh m
//   which are encroached by the point p...
//
int PLC::Step3Encroached(Point p, Simplex * s, Mesh * m,
                         set < SimplexMesh > &encroached,
                         set < CollarMesh > &encroachedcollar)
{

   int reject = 0;

   //first, get the cavity...
   set < int >nbs;
   m->Cavity(p, s, nbs);
   //m->Cavity(p, nbs); // slower version...

   //cout << endl << nbs.size() << endl;

   // then look at the low dimensional meshes that each of these points belong to
   //  and see if edge or collar elements are encroached
   foreach(int iN, nbs) {
      foreach(Mesh * m1, P->l2m[iN]) {
         if (m->IsChild(m1)) {

            if (m1->dim == 2 || m1->dim == 1) {
               // find the encroached cavity here...
               // get the star of the point...

               set < Simplex * >star;

               m1->Simplices(iN, star);

               foreach(Simplex * s1, star) {
                  if (!m1->Exterior(s1)
                      && G->inSphereReallyStrict(p, s1)) {

                     if (m1->IsCollar(s1)) {
                        if (m->dim != 3) {

                        }
                        if (m->dim == 3) {
                           reject = 1;
                           if (m1->dim == 1) {
                              //cout << "Rejecting from " << m1->name << endl;
                           }
                           //cout << "Reject = 1" << endl;
                           // can I return immediately here?
                        }
                     } else {
                        SimplexMesh sm;
                        sm.s = s1;
                        sm.m = m1;
                        encroached.insert(sm);
                     }
                  }
               }

            }
         }

      }

      // maybe I encroach the boundary of the collar...
      if (m->dim == 2 && m->col->Exists(iN)) {
         int a, b;
         a = m->col->col2next[iN];
         b = m->col->col2prev[iN];

         Simplex *s1 = SP->NewSimplex(iN, a);
         SetRadiusAndCC(s1);
         Simplex *s2 = SP->NewSimplex(iN, b);
         SetRadiusAndCC(s2);

         if (a >= 0 && G->inSphereReallyStrict(p, s1)) {
            CollarMesh cm;
            cm.a = iN;
            cm.b = a;
            cm.m = m;
            encroachedcollar.insert(cm);
         }
         if (b >= 0 && G->inSphereReallyStrict(p, s2)) {
            CollarMesh cm;
            cm.a = iN;
            cm.b = b;
            cm.m = m;
            encroachedcollar.insert(cm);
         }
         //delete s1;
         SP->DeleteSimplex(s1);
         //delete s2;
         SP->DeleteSimplex(s2);

      }

   }

   return reject;
}

//
//  given point p, and Mesh * m, make a list of simplices in mesh m
//   which are encroached by the point p...
//
int PLC::Step3EncroachedInt(Point p, Simplex * s, Mesh * m,
                         set < SimplexMesh > &encroached,
                         set < IntestineItem > &encroachedintestine)
{

   int reject = 0;

   //first, get the cavity...
   set < int >nbs;
   m->Cavity(p, s, nbs);

   // then look at the low dimensional meshes that each of these points belong to
   foreach(int iN, nbs) {
      foreach(Mesh * m1, P->l2m[iN]) {
         if (m->IsChild(m1)) {

            if (m1->dim == 2 || m1->dim == 1) {
               // find the encroached cavity here...
               // get the star of the point...

               set < Simplex * >star;

               m1->Simplices(iN, star);

               foreach(Simplex * s1, star) {
                  if (!m1->Exterior(s1)
                      && G->inSphereReallyStrict(p, s1)) {

                     if (M->IsIntestine(s1)) {
		       // I don't think I should ever get here
		       cout << "WARNING: something wrong is going on..." << endl;

                     } else {
                        SimplexMesh sm;
                        sm.s = s1;
                        sm.m = m1;
                        encroached.insert(sm);
                     }
                  }
               }

            }
         }

      }
      // deal with the intestine...

      set <int> intNbs;
      
      M->intestine->IntestineArcsSegsRings(iN,intNbs);
      
      foreach (int jN, intNbs) {
	Simplex stmp(iN,jN);
	SetRadiusAndCC(&stmp);

	if (iN < jN && G->inSphereReallyStrict(p,&stmp)) {
	  
	  IntestineItem im;
	  im.a = iN;
	  im.b = jN;
	  encroachedintestine.insert(im);
	  
	  //cout << iN << " " << jN << endl;
	}
      }
      
      // also look for encroached intestine faces...
      if (m->dim > 2) {
	set<Simplex *> star;
	m->Simplices(iN,star);
	
	foreach(Simplex * s, star) {
	  if (!M->IsIntestine(s)) {
	    vector<Simplex> faces;
	    s->Faces(faces);
	    foreach (Simplex f, faces) {
	      SetRadiusAndCC(&f);
	      Simplex stmp(f.v[0],f.v[1],f.v[2]);
	      if (M->intestine->IsIntestineFace(f.v[0],f.v[1], f.v[2]) && 
		  M->IsIntestine(&stmp)) {
		if (G->inSphereReallyStrict(p,&f)) {
		  //cout << "There " << s << " " << &f << endl;
		  IntestineItem im;
		  im.a = f.v[0];
		  im.b = f.v[1];
		  im.c = f.v[2];
		  im.sort();
		  encroachedintestine.insert(im);
		}
	      }
	    }
	  }
	}
	
      }

   }

   return reject;
}





int PLC::Step3Encroached(int i, Mesh * m, set < SimplexMesh > &encroached)
{

   int reject = 0;

   //first, get the (3D) cavity...
   set < int >nbs;
   m->Neighbors(i, nbs);

   //cout << "N " << nbs.size() << endl;
   // then look at the low dimensional meshes that each of these points belong to
   foreach(int iN, nbs) {
      foreach(Mesh * m1, P->l2m[iN]) {
         if (m1->dim == 2 || m1->dim == 1) {
            // find the encroached cavity here...
            // get the star of the point...

            set < Simplex * >star;

            m1->Simplices(iN, star);

            foreach(Simplex * s1, star) {
               if (!m1->Exterior(s1) && !s1->contains(i)
                   && G->inSphereReallyStrict(i, s1)) {

                  SimplexMesh sm;
                  sm.s = s1;
                  sm.m = m1;
                  encroached.insert(sm);

                  if (m1->IsCollar(s1)) {
                     reject = 1;
                     if (P->type[i] == POINT_3D) {
                        cout << endl <<
                        "ERROR: circumcenter encroaches collar."
                             << endl;
                        cout << i << " " << P->V[i] << endl;
                        cout << s1 << " " << s1->rsq << " " << s1->
                        cc << " " << m1->name << endl;
                        cout << P->V[s1->v[0]] << " " << P->V[s1->
                                                              v[1]]
                             << endl;
                        cout << m->name << endl;

                        cout << s1->rsq << " " << d2(P->V[i],
                                                     s1->
                                                     cc) << endl;
                     }

                  }
               }
            }
         }
      }
   }

   return reject;
}

int PLC::Step3EncroachedInt(int i, set < SimplexMesh > &encroached,
			    set<IntestineItem> &encroachedintestine)
{

   int reject = 0;

   //first, get the (3D) cavity...
   set < int >nbs;
   M->Neighbors(i, nbs);

   //cout << "N " << nbs.size() << endl;

   // then look at the low dimensional meshes that each of these points belong to
   foreach(int iN, nbs) {
      foreach(Mesh * m1, P->l2m[iN]) {
         if (m1->dim == 2 || m1->dim == 1) {
            // find the encroached cavity here...
            // get the star of the point...

            set < Simplex * >star;

            m1->Simplices(iN, star);

            foreach(Simplex * s1, star) {
               if (!m1->Exterior(s1) && !s1->contains(i)
                   && G->inSphereReallyStrict(i, s1)) {

                  SimplexMesh sm;
                  sm.s = s1;
                  sm.m = m1;
                  encroached.insert(sm);

               }
            }
         }
      }
      set <int> intNbs;
      
      M->intestine->IntestineArcsSegsRings(iN,intNbs);
      
      foreach (int jN, intNbs) {
	Simplex stmp(iN,jN);
	
	if (iN < jN && G->inSphereReallyStrict(i,&stmp)) {
	  
	  IntestineItem im;
	  im.a = iN;
	  im.b = jN;
	  encroachedintestine.insert(im);
	  
	  //cout << iN << " " << jN << endl;
	}
      }


      set<Simplex *> star;
      M->Simplices(iN,star);
      
      foreach(Simplex * s, star) {
	vector<Simplex> faces;
	s->Faces(faces);
	foreach (Simplex f, faces) {
	  SetRadiusAndCC(&f);
	  Simplex stmp(f.v[0],f.v[1],f.v[2]);
	  if (M->intestine->IsIntestineFace(f.v[0],f.v[1], f.v[2]) && 
	      M->IsIntestine(&stmp)) {
	    if (i != f.v[0] && i != f.v[1] && i != f.v[2] && G->inSphereReallyStrict(i,&f)) {
	      
	      IntestineItem im;
	      im.a = f.v[0];
	      im.b = f.v[1];
	      im.c = f.v[2];
	      im.sort();
	      encroachedintestine.insert(im);
	    }
	  }
	}
      }



   }

   return reject;
}



// if the face of s opposite to point i part
// of a face in the plc?
bool PLC::IsFace(Simplex * s, int i)
{

   int a, b, c;

   if (0 == i) {
      a = s->v[1];
      b = s->v[2];
      c = s->v[3];
   } else if (1 == i) {
      a = s->v[0];
      b = s->v[2];
      c = s->v[3];
   } else if (2 == i) {
      a = s->v[0];
      b = s->v[1];
      c = s->v[3];
   } else if (3 == i) {
      a = s->v[0];
      b = s->v[1];
      c = s->v[2];
   } else {
      cout << "ERROR: Bad simplex in IsFace." << endl;
      exit(0);
   }
   /*
   // Doesn't work...
   Point p[3];
   p[0] = P->V[a];
   p[1] = P->V[b];
   p[2] = P->V[c];

   for (int i=0; i<3; i++) {
     if (d(p[i],M->base) < M->bbox/24.0) {
       cout << p[i] << M->base << " " << M->bbox << endl;
       return true;
     }
   }
   */

   
   /*
   // WARNING: this is a hack! FIXME
   PointType a1,b1,c1;
   a1 = P->type[a];
   b1 = P->type[b];
   c1 = P->type[c];

   if (a1 == POINT_BBOX || a1 == POINT_3D ||
       b1 == POINT_BBOX || b1 == POINT_3D ||
       c1 == POINT_BBOX || c1 == POINT_3D) {
     return false;
   }
   return true;
   // HACK ends here
   */

   foreach(Mesh * m, P->l2m[a]) {
      if (m->dim == 2) {
	  //set < int >nbors;
	  //m->Neighbors(a, nbors);

	  //if (nbors.find(b) != nbors.end()
          //   && nbors.find(c) != nbors.end()) {
          //  return true;
	  //}

	  if (P->IsIn(b,m) && P->IsIn(c,m)) {
	      return true;
	  }
      }
   }
   return false;

}

bool PLC::ContainsBadSliver(set<Simplex * > & star) {

  //cout << "CBS " << star.size() << endl;
  foreach (Simplex * s, star) {
    vector < double >angles;
    G->DihedralAngles(s, angles);
    
    if (angles[0] < .0001) {     
      /*
      cout << s << " --- " << angles[0] << endl;
      cout << P->type[s->v[0]] << " " << P->type[s->v[1]] << " "
		<< P->type[s->v[2]] << " " << P->type[s->v[3]] <<endl;
      cout << P->LowestDimensionalContainingMesh(s->v[0])->name << " ";
      cout << P->LowestDimensionalContainingMesh(s->v[1])->name << " ";
      cout << P->LowestDimensionalContainingMesh(s->v[2])->name << " ";
      cout << P->LowestDimensionalContainingMesh(s->v[3])->name << endl;;
      */
      return true;
    }
  }
  return false;

}

void PLC::FixCollarSlivers() {

  srand(clock());
  
  // queue collar points contained in bad slivers...

  set<int> badcollar;

  for (Mesh::piterator pit = M->pbegin(); pit != M->pend(); pit++) {
    if(P->type[*pit] == POINT_COLLAR) {


      set<Simplex *> star;
      
      M->Simplices(*pit,star);
      
      if (ContainsBadSliver(star)) {
	badcollar.insert(*pit);
      }


    }
  }

  cout << "# Bad Collar Vertices: " << badcollar.size() << endl;

  foreach(int i, badcollar) {

    set<Simplex *> star;
    
    M->Simplices(i,star);

    Vector last(0.0,0.0,0.0);

    while (ContainsBadSliver(star)) {
 
      // get neighbors

      set<int> nbs;
      M->Neighbors(i,nbs);
      double mind = d(P->V[i],P->V[*(nbs.begin())]);

      foreach(int iNb, nbs) {
	double myd = d(P->V[i],P->V[iNb]);
	if (myd < mind) {
	  mind = myd;
	}
      }

      Mesh * myFace = P->LowestDimensionalContainingMesh(i);
      
      // remove the point

      set<Simplex *> oldcavity;
      set<Simplex *> cavity;
      M->RemovePoint(i,oldcavity,cavity);
      foreach(Simplex * s, oldcavity) {
	SP->DeleteSimplex(s);
      }


      // perturb
      double epx = rand()/((double)RAND_MAX + 1);
      double epy = rand()/((double)RAND_MAX + 1);
      double epz = rand()/((double)RAND_MAX + 1);

      cout << epx << " " << epy << " " << epz << endl;

      Vector v(epx,epy,epz);
      //v = v - (v*myFace->dir)*myFace->dir;

      cout << myFace->dir << " " << v << " " << myFace->dir*v << endl;

      cout << "Loc " << M->Locate(i) << endl;

      P->V[i] = P->V[i] + (-1.0)*last + mind*.01*v;
      last = mind*.01*v;

      cout << "Ins " << P->V[i] << " " << mind << endl;

      //P->V[i].xyz[0] += epx*mind*.01;
      //P->V[i].xyz[1] += epy*mind*.01;
      //P->V[i].xyz[2] += epz*mind*.01;
      //P->V[i].xyz[0] += v.xyz[0]*mind*.0001;
      //P->V[i].xyz[1] += v.xyz[1]*mind*.0001;
      //P->V[i].xyz[2] += v.xyz[2]*mind*.0001;
      
      M->InsertPoint(i);
    }

    //set<Simplex *> cavity;

    
    
    

  }

  // for each point
  // 1.  check if it is still in a sliver
  // 2.  remove it from the meshes
  // 3.  perturb it until it does not create more slivers
  // 4.  re-insert into the mesh


}


// use a BFS to remove exterior tets
// if the input PLC is not "water tight"
//   this will remove everything
void PLC::StripBBoxTets()
{

   PrintBegin("Removing Bounding Box.");

   // point 0 is a bbox point for the main bounding box.
   Simplex *s = M->GetSimplex(0);

   set <Simplex *> sQ;
   

   tetsToRemove.insert(s);
   sQ.insert(s);
   s->f = SIM_BBOX_STRIP;

   while (sQ.size() > 0) {
      s = *(sQ.begin());
      //cout << s << " " << sQ.size() << endl;

      sQ.erase(sQ.begin());


      for (int j = 0; j < s->size(); j++) {
	  if (P->type[s->v[j]] == POINT_BBOX ||
	      P->type[s->v[j]] == POINT_3D) {
	    pointsToRemove.insert(s->v[j]);
	  } else if (M->p2s[s->v[j]] == s) {
	      // find a safe tet to use for p2s...
	      set < Simplex * >star;
	      M->Simplices(s->v[j], star);
	      foreach(Simplex * iStar, star) {
                           if (iStar->f != SIM_BBOX_STRIP) {
			       M->p2s[s->v[j]] = iStar;
                           }
	      }
	      
	  }
      }
      

      for (int i = 0; i < s->size(); i++) {

         // look at each neighbor of the simplex
         if (s->n[i] != NULL) {

            if (!IsFace(s, i)) {
               // check if the neighbor is through a boundary face
               if (s->n[i]->f != SIM_BBOX_STRIP) {
                  // if not, check if the neighbor is a bbox tet
                  s->n[i]->f = SIM_BBOX_STRIP;

		  sQ.insert(s->n[i]);
                  tetsToRemove.insert(s->n[i]);


               }
            }

         }
      }
   }

   //cout << pointsToRemove.size() << " " << tetsToRemove.size() << endl;
   
   // remove the bbox simplices and points...

   foreach(Simplex * s, tetsToRemove) {
      for (int i = 0; i < s->size(); i++) {
         if (s->n[i] != NULL) {
            for (int j = 0; j < s->n[i]->size(); j++) {
               if (s->n[i]->n[j] == s) {
                  s->n[i]->n[j] = NULL;
               }
            }
         }
      }
      M->ns--;
      SP->DeleteSimplex(s);
   }

   foreach(int ip, pointsToRemove) {
      M->p2s.erase(ip);
      P->l2m[ip].clear();
      M->np--;
   }

   int flipcount = 0;
   // point normals of exterior faces outwards...
   foreach (Mesh * mFace, M->children) {
       if (mFace->dim == 2) {
	   // find a non bbox point...
	   int p = -1;
	   Mesh::PIterator pIt(mFace);
	   pIt.Start(mFace->GetSimplex());
	   while (pIt.Current() != -1) {
	       if (P->type[pIt.Current()] != POINT_BBOX) {
		   p = pIt.Current();
		   break;
	       }
	       ++pIt;
	   }

	   // find a simplex is the 3D mesh touching the face
	   Simplex * spq = NULL;
	   int q = -1;
	   set<Simplex * > star;
	   M->Simplices(p,star);
	   
	   foreach(Simplex * s, star) {
	       if (P->IsIn(s->v[0],mFace)) {
		   if (P->IsIn(s->v[1],mFace)) { 
		       if (P->IsIn(s->v[2],mFace)) {
			   spq = s;
			   q = s->v[3];
			   break;
		       } else {
			   if (P->IsIn(s->v[3],mFace)) {
			       spq = s;
			       q = s->v[2];
			       break;
			   }
		       } 
		   } else {
			   if (P->IsIn(s->v[2],mFace) 
			       && P->IsIn(s->v[3],mFace)) {
			       spq = s;
			       q = s->v[1];
			       break;
			   }
		   }
	       } else {
		   if (P->IsIn(s->v[1],mFace) &&
		       P->IsIn(s->v[2],mFace) &&
		       P->IsIn(s->v[3],mFace) ) {
		       spq = s;
		       q = s->v[0];
		       break;
		   }
	       }

	   }

	   // check orientation...
	   Vector v = P->V[q] - P->V[p];
	   
	   if (v*mFace->dir > 0) {
	       flipcount++;
	       //cout << "Flipping " << mFace->name << endl;
	       mFace->dir = -mFace->dir;
	   }


       }
   }

   

   (*LOGFILE) << endl << "Orientation of " << flipcount 
	      << " faces was flipped." << endl;
   
   PrintEnd("Removing Bounding Box.");

}

//
// Gives all tets and triangles
// consistent orientation
//
void PLC::OrientPLC()
{
  foreach(Mesh * m, M->children) {
    if (m->dim == 2) {
      
      Mesh::Iterator sIt1;
      sIt1.Start(m->GetSimplex());
      while (sIt1.Current() != NULL) {
	Simplex *ss = sIt1.Current();
	
	Vector v0 = P->V[ss->v[0]] - P->V[ss->v[2]];
	
	Vector v1 = P->V[ss->v[1]] - P->V[ss->v[2]];
		
	if ((v0^v1)*m->dir >= 0.0) {
	  // correct orientation
	} else {
	  int tmpi;
	  Simplex * tmps;
	  tmpi = ss->v[1];
	  tmps = ss->n[1];
	  ss->v[1] = ss->v[2];
	  ss->n[1] = ss->n[2];
	  ss->v[2] = tmpi;
	  ss->n[2] = tmps;
	}
	++sIt1;
      }
    }
  }

  // tets...
   Mesh::Iterator sIt;

   sIt.Start(M->GetSimplex());
   while (sIt.Current() != NULL) {
     Simplex * ss = sIt.Current();
      if (G->Orient(ss) >= 0.0) {
	// correct orientation
      } else {
	  int tmpi;
	  Simplex * tmps;
	  tmpi = ss->v[2];
	  tmps = ss->n[2];
	  ss->v[2] = ss->v[3];
	  ss->n[2] = ss->n[3];
	  ss->v[3] = tmpi;
	  ss->n[3] = tmps;
      }
     

     ++sIt;
   }



}


void PLC::PrintTetView(string fname) {

  // loop through simplices
  // print some exterior simplices


  cout << M->base << M->bbox << endl;

  set<Simplex *> exteriorPrintList;

  //double factor = .5/config->boundingboxratio;
  double factor = 0.0;

  for (Mesh::siterator it = M->sbegin(); it != M->send(); ++it) {
    Simplex * s = *it;
    if (s->f == SIM_BBOX_STRIP) {

      if (d(P->V[s->v[0]],M->base) < M->bbox*factor && 
	  d(P->V[s->v[1]],M->base) < M->bbox*factor && 
	  d(P->V[s->v[2]],M->base) < M->bbox*factor && 
	  d(P->V[s->v[3]],M->base) < M->bbox*factor) {
	  //for (int i=0; i<4; i++) {
	  //if (P->V[s->v[i]].xyz[0] < M->base.xyz[0]) {
	
	  tetsToRemove.erase(tetsToRemove.find(s));
	  for(int j=0; j<4; j++) {
	    if (pointsToRemove.find(s->v[j]) != pointsToRemove.end()) {
	      pointsToRemove.erase(pointsToRemove.find(s->v[j]));
	    }
	  }
	  //break;
	}
	//}
    }
  }



   ofstream eleout,nodeout;
   eleout.open((fname+".ele").c_str());
   nodeout.open((fname+".node").c_str());

   nodeout << setprecision(20)
        << setiosflags(ios::scientific) ;


   nodeout << M->np - pointsToRemove.size() << " 3 0 0" << endl;

   eleout << M->ns - tetsToRemove.size() << " 4 1" << endl;

   cout << "PrintTetView " << pointsToRemove.size() << " " << tetsToRemove.size() << endl;


   map < int, int >pointmap;

   int nextpoint = 1;
   for (int i = 0; i < (int)P->V.size(); i++) {
     if (P->IsIn(i, M) && pointsToRemove.find(i) == pointsToRemove.end()) {
	nodeout << nextpoint << " " << P->V[i].xyz[0] << " "
              << P->V[i].xyz[1] << " " << P->V[i].xyz[2] << endl;
         pointmap[i] = nextpoint;
         nextpoint++;
      }
   }

   
   Mesh::Iterator sIt;

   int count = 1;

   sIt.Start(M->GetSimplex());
   while (sIt.Current() != NULL) {

      Simplex *ss = sIt.Current();

      //if (ss->f != SIM_BBOX_STRIP) {
      if (tetsToRemove.find(ss) == tetsToRemove.end()) {

	if (G->Orient(ss) >= 0.0) {
	  eleout << count << " " << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]] << " "
		 << pointmap[ss->v[2]] << " " << pointmap[ss->v[3]];
	} else {
	  eleout << count << " " << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]] << " "
		 << pointmap[ss->v[3]] << " " << pointmap[ss->v[2]];
	  cout << "ERROR: Orientation problem.  OrientPLC() must have failed." << endl;
	}

	if (ss->f == SIM_BBOX_STRIP) {
	  eleout << " 10" << endl;
	} else {
	  eleout << " 11" << endl;
	}

      }
      ++sIt;
      count++;
   }
   

}

void PLC::PrintRaw(string fname) {

   ofstream fout;
   fout.open((fname).c_str());

   //fout << setprecision(20)
   //     << setiosflags(ios::scientific) ;

   fout << M->np << " " << M->ns << endl;

   fout << M->np - pointsToRemove.size() << " " 
	<< M->ns - tetsToRemove.size() << endl;

   map < int, int >pointmap;

   int nextpoint = 0;
   for (int i = 0; i < (int)P->V.size(); i++) {
      if (P->IsIn(i, M) && pointsToRemove.find(i) == pointsToRemove.end()) { 
      //if (P->IsIn(i, M)) {
         fout << P->V[i].xyz[0] << " "
              << P->V[i].xyz[1] << " " << P->V[i].xyz[2] << " 1" << endl;
         pointmap[i] = nextpoint;
         nextpoint++;
      }
   }

   Mesh::Iterator sIt;

   int count = 0;

   sIt.Start(M->GetSimplex());
   while (sIt.Current() != NULL) {

      Simplex *ss = sIt.Current();

      if (tetsToRemove.find(ss) == tetsToRemove.end()) {
	if (G->Orient(ss) >= 0.0) {
	  fout << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]] << " "
	       << pointmap[ss->v[2]] << " " << pointmap[ss->v[3]] << endl;
	} else {
	  fout << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]] << " "
              << pointmap[ss->v[3]] << " " << pointmap[ss->v[2]] << endl;
	  cout << "ERROR: Orientation problem.  OrientPLC() must have failed." << endl;
	}
      }
      ++sIt;
      count++;
   }

}

void PLC::PrintTPLC(string fname)
{
   ofstream fout;
   fout.open((fname).c_str());

   fout << setprecision(20)
        << setiosflags(ios::scientific) ;

   fout << "TPLC" << endl;
   fout << "DIMENSION 3" << endl;

   fout << "POINTS " << M->np << endl;

   map < int, int >pointmap;

   int nextpoint = 0;
   for (int i = 0; i < (int)P->V.size(); i++) {
      if (P->IsIn(i, M)) {
         fout << nextpoint << " " << P->V[i].xyz[0] << " "
              << P->V[i].xyz[1] << " " << P->V[i].xyz[2] << endl;
         pointmap[i] = nextpoint;
         nextpoint++;
      }
   }
   int ns, nf;
   ns = 0;
   nf = 0;
   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {
         ns++;
      } else if (m->dim == 2) {
         nf++;
      }
   }

   fout << "SEGMENTS " << ns << endl;
   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {
         // print info for this segment...
         fout << m->id << " " << m->np << endl;
         int cur, prev, next;
         prev = -1;
         cur = m->SegmentFirstPoint();
         for (int i = 0; i < m->np; i++) {
            next = m->SegmentNextPoint(cur, prev);
            fout << pointmap[next] << " ";
            prev = cur;
            cur = next;
         }
         fout << endl;
      }
   }

   fout << "FACES " << nf << endl;
   foreach(Mesh * m, M->children) {
      if (m->dim == 2) {
         Mesh::Iterator sIt;

         int ntri = 0;
         sIt.Start(m->GetSimplex());
         while (sIt.Current() != NULL) {
            if (!m->Exterior(sIt.Current())) {
               ntri++;
            }
            ++sIt;
         }
         fout << m->id << " " << ntri << endl;

         Mesh::Iterator sIt1;
         sIt1.Start(m->GetSimplex());
         while (sIt1.Current() != NULL) {
            if (!m->Exterior(sIt1.Current())) {
               Simplex *ss = sIt1.Current();

               Vector v0 = P->V[ss->v[0]] - P->V[ss->v[2]];

               Vector v1 = P->V[ss->v[1]] - P->V[ss->v[2]];


               if ((v0^v1)*m->dir >= 0.0) {
                  fout << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]]
                       << " " << pointmap[ss->v[2]] << endl;
               } else {
                  fout << pointmap[ss->v[0]] << " " << pointmap[ss->v[2]]
                       << " " << pointmap[ss->v[1]] << endl;
		  cout << "Error: OrientPLC should give correct orientation." << endl;
               }
            }
            ++sIt1;
         }
      }
   }

   Mesh::Iterator sIt;

   fout << "TETS " << M->ns << endl;


   //int ntri = 0;
   int count = 0;

   sIt.Start(M->GetSimplex());
   while (sIt.Current() != NULL) {

      Simplex *ss = sIt.Current();
       if (pointmap[ss->v[0]] == pointmap[ss->v[1]] ||
	   pointmap[ss->v[0]] == pointmap[ss->v[2]] ||
	   pointmap[ss->v[0]] == pointmap[ss->v[3]] ||
	   pointmap[ss->v[1]] == pointmap[ss->v[2]] ||
	   pointmap[ss->v[1]] == pointmap[ss->v[3]] ||
	   pointmap[ss->v[2]] == pointmap[ss->v[3]]) {

	   cout << ss << endl;

	   cout << pointmap[ss->v[0]] << " " 
		<< pointmap[ss->v[1]] << " " 
		<< pointmap[ss->v[2]] << " " 
		<< pointmap[ss->v[3]] << endl;

	   cout << P->type[ss->v[0]] << " " 
		<< P->type[ss->v[1]] << " " 
		<< P->type[ss->v[2]] << " " 
		<< P->type[ss->v[3]] << endl; 

	   cout << P->V[ss->v[0]] << " " 
		<< P->V[ss->v[1]] << " " 
		<< P->V[ss->v[2]] << " " 
		<< P->V[ss->v[3]] << endl; 
	   
	   cout << P->IsIn(ss->v[0], M) << " " 
		<< P->IsIn(ss->v[1], M) << " " 
		<< P->IsIn(ss->v[2], M) << " " 
		<< P->IsIn(ss->v[3], M) << endl;

	   for (int i=0; i<4; i++) {
	       cout << i << endl;
	       foreach(Mesh * mmm, P->l2m[ss->v[i]]) {
		   cout << mmm->name << endl;
	       }
	       cout << endl;
	       if (ss->n[i] != NULL) {
		   cout << ss->n[i] << endl;
	       }
	       
	   }

       }


      if (G->Orient(ss) >= 0.0) {
         fout << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]] << " "
              << pointmap[ss->v[2]] << " " << pointmap[ss->v[3]] << endl;
      } else {
         fout << pointmap[ss->v[0]] << " " << pointmap[ss->v[1]] << " "
              << pointmap[ss->v[3]] << " " << pointmap[ss->v[2]] << endl;
	 cout << "ERROR: Orientation problem.  OrientPLC() must have failed." << endl;
      }
      ++sIt;
      count++;
   }
}


void PLC::PrintFreeFEM(string fname)
{

  cout << pointsToRemove.size() << endl;
   ofstream fout;
   fout.open((fname).c_str());

   fout << setprecision(20)
        << setiosflags(ios::scientific) ;

   fout << "MeshVersionFormatted 1" << endl;
   fout << "Dimension 3" << endl;

   fout << endl << "Vertices" << endl << M->np << endl;

   map < int, int >pointmap;

   int nextpoint = 1;
   for (int i = 0; i < (int)P->V.size(); i++) {
      if (P->IsIn(i, M)) {

         fout << P->V[i].xyz[0] << " "
              << P->V[i].xyz[1] << " " << P->V[i].xyz[2] 
	      << " 0" << endl;
         pointmap[i] = nextpoint;
         nextpoint++;
      }
   }
   int ntri=0;
   foreach(Mesh * m, M->children) {
      if (m->dim == 2) {
         Mesh::Iterator sIt;

         sIt.Start(m->GetSimplex());
         while (sIt.Current() != NULL) {
            if (!m->Exterior(sIt.Current())) {
               ntri++;
            }
            ++sIt;
         }
      }
   }


   fout << endl << "Triangles" << endl << ntri << endl;

   int fId = 1;
   foreach(Mesh * m, M->children) {
     fId = m->id;
     //fId = fId%2;
      if (m->dim == 2) {
         Mesh::Iterator sIt1;
         sIt1.Start(m->GetSimplex());
         while (sIt1.Current() != NULL) {
            if (!m->Exterior(sIt1.Current())) {
               Simplex *ss = sIt1.Current();

               Vector v0 = P->V[ss->v[0]] - P->V[ss->v[2]];

               Vector v1 = P->V[ss->v[1]] - P->V[ss->v[2]];


               if ((v0^v1)*m->dir >= 0.0) {
		  fout << pointmap[ss->v[0]] << " " <<  pointmap[ss->v[1]] << " "
		       <<  pointmap[ss->v[2]] << " " << fId << endl;

               } else {
		  fout <<  pointmap[ss->v[2]] << " " <<  pointmap[ss->v[1]] << " "
		       <<  pointmap[ss->v[0]] << " " << fId << endl;
               }
            }
            ++sIt1;
         }
      }
   }

   Mesh::Iterator sIt;

   fout << endl << "Tetrahedra" << endl << M->ns << endl;

   sIt.Start(M->GetSimplex());
   while (sIt.Current() != NULL) {

      Simplex *ss = sIt.Current();

      if (G->Orient(ss) <= 0.0) {
	fout <<  pointmap[ss->v[0]] << " " <<  pointmap[ss->v[1]] << " "
              <<  pointmap[ss->v[2]] << " " <<  pointmap[ss->v[3]]  << " 0" << endl;
      } else {
	fout <<  pointmap[ss->v[1]] << " " <<  pointmap[ss->v[0]] << " "
              <<  pointmap[ss->v[2]] << " " <<  pointmap[ss->v[3]]  << " 0" << endl;
      }
      ++sIt;
   }


   fout << endl << "End" << endl;

}



void PLC::PrintOff(string prefix, string suffix)
{

   //M->PrintSurfaceOff(prefix + M->name + suffix);
   //M->PrintSurfaceCOff(prefix + M->name + "-color" + suffix);

   //cout << "Printing " << prefix + M->name + suffix << endl;
   M->PrintSurfaceCOff(prefix + M->name + suffix);
   //M->PrintSurfaceGMV(prefix + M->name + suffix + ".gmv");
   foreach(Mesh * child, planestoprint) {
      child->PrintOff(prefix + child->name + suffix);
   }
}


void PLC::PrintCollarOff(string prefix, string suffix)
{
   foreach(Mesh * child, planestoprint) {
      child->PrintOff(prefix + child->name + suffix);
   }
}



void PLC::PrintBegin(string s)
{

   cout << s << endl;
   if (logfileisopen) {
      (*LOGFILE) << endl <<
      "-------------------------------------------------" << endl;
      (*LOGFILE) << "BEGIN: " << s << endl;
      (*LOGFILE) << endl;
   }
   start = clock();
}

void PLC::PrintEnd(string s)
{
   if (logfileisopen) {
      (*LOGFILE) << endl;
      if (M != NULL) {
         (*LOGFILE) << "Verices in the Mesh: " << M->np << endl;
         (*LOGFILE) << "Tets in the Mesh: " << M->ns << endl;
      }
      (*LOGFILE) << "Seconds Elapsed: " << ((double) (clock() - start)) /
      ((double) CLOCKS_PER_SEC)
                 << endl;
      (*LOGFILE) << endl << "END: " << s << endl;
      (*LOGFILE) << "-------------------------------------------------"
                 << endl;
   }
}

void PLC::InsertInMesh(int i, Mesh * m)
{

   set < Simplex * >oldcavity;
   set < Simplex * >newcavity;
   m->InsertPoint(i, oldcavity, newcavity);

   // delete old simplices...
   foreach(Simplex * sim, oldcavity) {
      SP->DeleteSimplex(sim);
   }
}

void PLC::InsertInMesh(int i, int hint, Mesh * m)
{

   set < Simplex * >oldcavity;
   set < Simplex * >newcavity;
   m->InsertPoint(i, hint, oldcavity, newcavity);

   // delete old simplices...
   foreach(Simplex * sim, oldcavity) {
      SP->DeleteSimplex(sim);
   }
}

void PLC::InsertInMesh(int i, Simplex * s, Mesh * m)
{

   set < Simplex * >oldcavity;
   set < Simplex * >newcavity;
   m->InsertPoint(i, s, oldcavity, newcavity);

   // delete old simplices...
   foreach(Simplex * sim, oldcavity) {
      SP->DeleteSimplex(sim);
   }
}


void PLC::InsertInMeshAndDequeue(int i, Mesh * m, set < SimplexMesh > &SQ)
{

   set < Simplex * >oldcavity;
   set < Simplex * >newcavity;
   m->InsertPoint(i, oldcavity, newcavity);

   foreach(Simplex * sim, oldcavity) {

      if (sim->f == SIM_QUEUED) {
         SimplexMesh sm;
         sm.s = sim;
         sm.m = m;

         //SQ.erase(sm);
         set < SimplexMesh >::iterator it;
         it = SQ.find(sm);
         if (it != SQ.end()) {
            SQ.erase(it);
         } else {
            cout << "Problem finding SM pair" << endl;
         }
      }

      SP->DeleteSimplex(sim);
   }
}

void PLC::InsertInMeshAndDequeue(int i, int hint, Mesh * m,
                                 set < SimplexMesh > &SQ)
{

   set < Simplex * >oldcavity;
   set < Simplex * >newcavity;
   m->InsertPoint(i, hint, oldcavity, newcavity);

   foreach(Simplex * sim, oldcavity) {

      if (sim->f == SIM_QUEUED) {
         SimplexMesh sm;
         sm.s = sim;
         sm.m = m;

         //SQ.erase(sm);
         set < SimplexMesh >::iterator it;
         it = SQ.find(sm);
         if (it != SQ.end()) {
            SQ.erase(it);
         } else {
            cout << "Problem finding SM pair" << endl;
         }
      }
      SP->DeleteSimplex(sim);
   }
}



void PLC::InsertInMeshAndParents(int i, Mesh * m)
{
   InsertInMesh(i, m);

   foreach(Mesh * m1, m->parents) {
      InsertInMesh(i, m1);
   }

}

void PLC::InsertInMeshAndParents(int i, int hint, Mesh * m)
{


   InsertInMesh(i, m);

   foreach(Mesh * m1, m->parents) {
      InsertInMesh(i, hint, m1);
   }

}



int PLC::GetNextGID()
{
   counter++;
   return counter;
}


// WARNING: calling this when the face and segment are not
//          synchronized could cause a problem...
Vector PLC::NormalInFace(Mesh * segment, Mesh * face)
{
   // get the direction of the normal into the face...
   Vector NIF;

   if (!face->convex) {
      int bboxpoint = segment->SegmentFirstPoint();
      int p1 = segment->SegmentNextPoint(bboxpoint, -1);
      int p2 = segment->SegmentNextPoint(p1, bboxpoint);
      if (face->bnd->Exists(p1)) {
         Vector vv;
         if (face->bnd->next[p1] == p2) {
            vv = P->V[p1] - P->V[p2];
         } else {
            vv = P->V[p2] - P->V[p1];
         }
         vv.normalize();
         NIF = vv ^ face->dir;
      } else {
         NIF = segment->dir ^ face->dir; // cross the direction of the edge
         if (NIF * (face->base - segment->base) < 0) {
            NIF = -NIF;
         }
      }
   } else {
      NIF = segment->dir ^ face->dir; // cross the direction of the edge
      if (NIF * (face->base - segment->base) < 0) {
         NIF = -NIF;
      }
   }

   return NIF;
}

Vector PLC::DirTowardsSegment(Point p, Mesh * segment)
{

   set < int >::iterator sit = segment->exteriorPoints.begin();

   int p1, p2;

   p1 = *sit;

   sit++;

   p2 = *sit;

   Vector mydir;
   if (d2(p, P->V[p1]) < d2(p, P->V[p2])) {
      mydir = p - P->V[p1];
   } else {
      mydir = p - P->V[p2];
   }

   mydir.normalize();
   return mydir;

}


// 10-30: seems to work now...
// FIXME: make protection angles a configuration option
void PLC::MakeProtectionList()
{

  double minangle = 100.0;
  double minangle2 = 100.0;

   foreach(Mesh * m, M->children) {
      if (m->dim == 1) {

         vector < Vector > faceDirs;

         //cout << m->name << " : ";
         foreach(Mesh * m1, m->parents) {
            if (m1->dim == 2) {
               faceDirs.push_back(NormalInFace(m, m1));

               //cout << m->name << " " << m1->name << " " << NormalInFace(m,m1) << endl;

               // for interior segments, insert both directions...
               if (m1->bndseg.find(m) == m1->bndseg.end()) {
                  faceDirs.push_back(-NormalInFace(m, m1));
               }
               //cout << NormalInFace(m,m1) << " ";
            }
         }          //cout << endl;

         // protect segments between sharply adjacent faces...
         int protectme = 0;

         if (faceDirs.size() > 5 || config->protectallsegments) {
	   segmentsToProtect.insert(m);
            protectme = 1;
            m->protect = true;
         } else {
	   
	   for (int i = 0; i < (int)faceDirs.size() && protectme == 0; i++) {
	     for (int j = i + 1; j < (int)faceDirs.size() && protectme == 0; j++) {
	       // check the difference...
	       
	       //70 degrees based on Hang Si paper...
	       // if (faceDirs[i] *faceDirs[j] > .333333333) {
	       if (faceDirs[i] *faceDirs[j] > 0) {
		 segmentsToProtect.insert(m);
		 protectme = 1;
		 m->protect = true;
		     
		 double angle = acos(faceDirs[i] *faceDirs[j]);
		 if (angle < minangle) {
		   minangle = angle;
		 }
		 
		 
	       }
	     }
	   }
         }
         if (protectme == 0) {
	   if (logfileisopen && config->verbose) {
	     (*(LOGFILE)) << m->name << " not protected." << endl;
	   }
         } else {
	   if (logfileisopen && config->verbose) {
	     (*(LOGFILE)) << m->name << " protected." << endl;
	   }
         }
      }
   }
   
   foreach(InputPoint ip, P0) {
     
     int protectme = 0;
     
     vector < Vector > segDirs;
     
     foreach(Mesh * m, ip.l2m) {
       if (m->dim == 1) {
	 
	 if (segmentsToProtect.find(m) != segmentsToProtect.end()) {
	   protectme = 1;
	   break;
	 }
	 segDirs.push_back(DirTowardsSegment(ip.p, m));
       }
     }
     
     // check minimum angle?
     if (protectme == 1 || segDirs.size() > 6) {
       pointsToProtect.insert(ip.id);
       protectme = 1;
     } else {
       foreach(Mesh * m, ip.l2m) {
	 foreach(Mesh * m1, ip.l2m) {
	   
	   if (m != m1 && m->dim == 1) {
	     if (m1->dim == 1) {
	       
	       //if (DirTowardsSegment(ip.p, m) * DirTowardsSegment(ip.p,m1) > config->epsilon) {
	       
	       // 60 degrees
	       if (DirTowardsSegment(ip.p, m) *
		   DirTowardsSegment(ip.p,m1) > .5) {
		 pointsToProtect.insert(ip.id);
		 protectme = 1;
		 
		 double angle = acos(DirTowardsSegment(ip.p, m) * DirTowardsSegment(ip.p,m1));
		 if (angle < minangle2) {
		   minangle2 = angle;
		 }
		 
	       }
	     } else if (m1->dim == 2 && !m1->Contains(m)) {
	       
	       // fill in this case...
	       Vector segDir = DirTowardsSegment(ip.p, m);
	       Vector faceNorm = m1->dir;
	       
	       // is this an interior point or a boundary point...
	       bool interior = true;
	       
	       vector < Mesh * >mybnd;
	       
	       foreach(Mesh * m2, ip.l2m) {
		 if (m1->bndseg.find(m2) != m1->bndseg.end()) {
		   interior = false;
		   mybnd.push_back(m2);
		 }
	       }
	       
	       //cout << m1->name << " " << m->name << " " << interior << " " 
	       //	    << segDir << " " << faceNorm << " " << segDir * faceNorm <<  endl;
	       
	       //if (segDir * faceNorm < 1 - EPSILON && segDir * faceNorm > -1 + config->epsilon) {
	       if (segDir * faceNorm < sqrt(3)/2 && segDir * faceNorm > -1.0*sqrt(3)/2) {

		 if (interior) {
		   pointsToProtect.insert(ip.id);
		   protectme = 1;
		 
		 } else {

		   //cout << "Boundary case" << endl;
		   // FIXME: implementation missing here!
		   // this should allow a 60 degree angle...
		 
		   if (m1->convex) {
		     
		     Vector v1, v2, v3;
		     v1 = DirTowardsSegment(ip.p, m) - (DirTowardsSegment(ip.p, m)*m1->dir)*m1->dir;
		     v2 = DirTowardsSegment(ip.p, mybnd[0]);
		     v3 = DirTowardsSegment(ip.p, mybnd[1]);


		     v1.normalize();
		     v2.normalize();
		     v3.normalize();
		     
		     //cout << v1 << v2 << v3 << endl;
		     //cout << (v3^v2)*(v1^v2) << " " << (v1^v2)*(v3^v1) << endl;

		     
		     if ((v3^v2)*(v1^v2) >= 0 &&
			 (v1^v2)*(v3^v1) >= 0 ) {
		       //cout << "Protect " << ip.id << endl;
		       pointsToProtect.insert(ip.id);
		       protectme = 1;
		     }		     
		     
		   } else {
		     
		     // FIXME: just need to implement the nonconvex case...

		     Vector v1, v2, v3;
		     v1 = DirTowardsSegment(ip.p, m) - (DirTowardsSegment(ip.p, m)*m1->dir)*m1->dir;

		     //cout << "TTT " << P->id2loc[ip.id] << " " <<m1->bnd->prev[P->id2loc[ip.id]] << "  " << m1->bnd->next[P->id2loc[ip.id]] << endl;

		     //cout <<P->V[m1->bnd->prev[P->id2loc[ip.id]]] << " " << ip.p << " " << P->V[m1->bnd->next[P->id2loc[ip.id]]] << endl;

		     v2 = P->V[m1->bnd->prev[P->id2loc[ip.id]]] - ip.p;
		     v3 = P->V[m1->bnd->next[P->id2loc[ip.id]]] - ip.p;
		     v1.normalize();
		     v2.normalize();
		     v3.normalize();
		     
		     //cout << v1 << v2 << v3 << endl;

		     double angle1, angle2;
		     
		     // decide if I need to protect based on these vectors...
		     double dotprod = v1 * v2;
		     if (dotprod > 1.0) {
		       dotprod = 1.0;
		     }
		     if (dotprod < -1.0) {
		       dotprod = -1.0;
		     }
		     
		     if ((v1 ^ v2) * m1->dir >= 0.0) {
		       angle1 =
			 acos(dotprod) * 180.0 / PI;
		     } else {
		       angle1 =
			 360.0 - acos(dotprod) * 180.0 / PI;
		     }

		     dotprod = v3 * v2;
		     if (dotprod > 1.0) {
		       dotprod = 1.0;
		     }
		     if (dotprod < -1.0) {
		       dotprod = -1.0;
		     }
		     
		     if ((v3 ^ v2) * m1->dir >= 0.0) {
		       angle2 =
			 acos(dotprod) * 180.0 / PI;
		     } else {
		       angle2 =
			 360.0 - acos(dotprod) * 180.0 / PI;
		     }
		     

		     // FIXME 10-29-2009
		     // remember to work here...
		     //cout << "A " << angle1 << " " << angle2 << endl;

		     if (angle1 < angle2) {
		       pointsToProtect.insert(ip.id);
		       protectme = 1;
		     }
		     
		   }
		 }
	       }	     
	     }

	   }
	 }
       }
     }
      if (protectme == 0) {
         if (logfileisopen && config->verbose) {
            (*(LOGFILE)) << ip.id << " not protected." << endl;
         }
      } else {
         if (logfileisopen && config->verbose) {
            (*(LOGFILE)) << ip.id << " protected." << endl;
         }
      }

   }

   //cout << "Min face-face input angle " << minangle*180/PI << endl;
   //cout << "Min segment-segment input angle " << minangle2*180/PI << endl;


   // fill the lazy protection list...
   foreach (int i, pointsToProtect) {

     bool lazyprotect = true;

     // does this point belong to a protected segment?
     foreach(Mesh * m, P0[i].l2m) {
       if (segmentsToProtect.find(m) != segmentsToProtect.end()) {
	 lazyprotect = false;
       }
     }

     if (config->lazyvertexprotection == true && lazyprotect) {
       lazyProtectionList.insert(i); 
     }

   }

}


void PLC::StoreInputNeighbors()
{
   // not really efficient...
   for (int i = 0; i < (int)P->V.size(); i++) {
      if (P->type[i] == POINT_INPUT) {

         foreach(Mesh * m, P->l2m[i]) {
            if (m->dim == 1) {
               set < int >nbs;
               m->Neighbors(i, nbs);

               if (nbs.size() != 2) {
                  cout << "ERROR: confusing segment..." << endl;
                  cout << nbs.size() << endl;
                  cout << i << P->V[i] << endl;
                  exit(0);
               } else {
                  foreach(int in, nbs) {
                     if (P->type[in] == POINT_SEGMENT) {
                        m->ip2nb[i] = in;
                        //cout << i << " " << in << " " << m->name << endl;
                     }
                  }
               }
            }
         }
      }
   }
}


bool PLC::IsProtected(int ip)
{

   if (P->type[ip] == POINT_INPUT) {
      if (pointsToProtect.find(P->ids[ip]) != pointsToProtect.end()) {
         return true;
      }
      foreach(Mesh * m1, P->l2m[ip]) {
         if (m1->dim == 1) {
            if (segmentsToProtect.find(m1) != segmentsToProtect.end()) {
               return true;
            }
         }
      }
   }
   return false;

}

bool PLC::IsProtected(Mesh * m)
{
  if (segmentsToProtect.find(m) != segmentsToProtect.end() ) {
    return true;
  }
  return false;
}


void PLC::PrintInfo()
{

  /*
   for (Mesh::siterator it = M->sbegin(); it != M->send(); ++it) {
     cout << *it << endl;
  }

   for (Mesh::piterator pit = M->pbegin(); pit != M->pend(); pit++) {
     cout << *pit << endl;
   }
  */


   if (config->printoff) {
      PrintOff("", ".off");
   }

   if (config->printtplc) {
      PrintTPLC(name + ".tplc");
   }

   if (config->printraw) {
      PrintRaw(name + "_tet.raw");
   }

   if (config->printtetview) {
      PrintTetView(name);
   }


   if (config->printfreefem) {
      PrintFreeFEM(name + ".mesh");
   }

   if (logfileisopen) {

      PrintBegin("Printing Mesh Summary.");


      (*LOGFILE) << "Info for mesh " << name << endl;

      (*LOGFILE) << endl;
      (*LOGFILE) << "Mesh reported numbers of vertices and tets" << endl;
      (*LOGFILE) << "  Number of Vertices: " << M->np << endl;
      (*LOGFILE) << "      Number of Tets: " << M->ns << endl;

      (*LOGFILE) << endl << "Counting the number of vertices and tets."
                 << endl;

      int nVert, nTet;
      nVert = 0;
      nTet = 0;

      //enum PointType { POINT_BBOX, POINT_INPUT, POINT_SEGMENT, POINT_COLLAR, POINT_FACE, POINT_3D };

      int pointhist[6];
      for (int i = 0; i < 6; i++) {
         pointhist[i] = 0;
      }


      Mesh::PIterator pIt(M);
      pIt.Start(M->GetSimplex());
      while (pIt.Current() != -1) {

         if (P->type[pIt.Current()] == POINT_BBOX) {
            pointhist[0]++;
         }
         if (P->type[pIt.Current()] == POINT_INPUT) {
            pointhist[1]++;
         }
         if (P->type[pIt.Current()] == POINT_SEGMENT) {
            pointhist[2]++;
         }
         if (P->type[pIt.Current()] == POINT_COLLAR || 
	     P->type[pIt.Current()] == POINT_SPHERE || 
	     P->type[pIt.Current()] == POINT_RING || 
	     P->type[pIt.Current()] == POINT_CYLINDER ) {
            pointhist[3]++;
         }
         if (P->type[pIt.Current()] == POINT_FACE) {
            pointhist[4]++;
         }
         if (P->type[pIt.Current()] == POINT_3D) {
            pointhist[5]++;
         }

         nVert++;

         ++pIt;
      }

      double total_volume = 0.0;
      double this_volume;
      double max_volume = 0.0;
      double min_volume = 1000000000000.0;
      double max_dist = 0.0;
      double min_dist = 1000000000000.0;
      double ave_dist = 0.0;

      Mesh::Iterator sIt;
      sIt.Start(M->GetSimplex());
      while (sIt.Current() != NULL) {
	Simplex * sa = sIt.Current();;	

	if (sa->f != SIM_BBOX_STRIP) {

	  this_volume = G->Volume(sIt.Current());
	  total_volume += this_volume;
	  if (this_volume > max_volume) {
	    max_volume = this_volume;
	  }
	  if (this_volume < min_volume) {
	    min_volume = this_volume;
	  }
	  
	  
	  int a,b,c,e;
	  
	  a = sa->v[0];
	  b = sa->v[1];
	  c = sa->v[2];
	  e = sa->v[3];
	  
	  double da[6];
	  da[0] = d(P->V[a],P->V[b]);
	  da[1] = d(P->V[a],P->V[c]);
	  da[2] = d(P->V[a],P->V[e]);
	  da[3] = d(P->V[b],P->V[c]);
	  da[4] = d(P->V[b],P->V[e]);
	  da[5] = d(P->V[c],P->V[e]);
	  
	  for (int i=0; i<6; i++) {
	    ave_dist += da[i];
	    
	    if (da[i] > max_dist) {
	      max_dist = da[i];
	    }
	    if (da[i] < min_dist) {
	      min_dist = da[i];
	    }
	  }
	  
	  nTet++;
	}
	++sIt;
      }

      ave_dist = ave_dist/nTet/6.0;
	
      (*LOGFILE) << "  Number of Vertices: " << nVert << endl;
      (*LOGFILE) << "      Number of Tets: " << nTet << endl;
      (*LOGFILE) << "Total Volume of Tets: "
                 << setprecision(10)
                 << setiosflags(ios::scientific)
                 << total_volume << endl;
      (*LOGFILE) << "  Minimum Tet Volume: " << setprecision(10)
                 << min_volume << endl;
      (*LOGFILE) << "  Maximum Tet Volume: " << setprecision(10)
                 << max_volume << endl;
      (*LOGFILE) << "  Minimum Vertex Distance: " << setprecision(10)
                 << min_dist << endl;
      (*LOGFILE) << "  Maximum Vertex Distance: " << setprecision(10)
                 << max_dist << endl;
      (*LOGFILE) << "  Average Vertex Distance: " << setprecision(10)
                 << ave_dist << endl;
      
      (*LOGFILE) << setprecision(6) << resetiosflags(ios::scientific);
      
      
      (*LOGFILE) << endl << "Number of points by type." << endl;
      (*LOGFILE) << " POINT_BBOX: " << pointhist[0] << endl;
      (*LOGFILE) << "  POINT_INPUT: " << pointhist[1] << endl;
      (*LOGFILE) << "POINT_SEGMENT: " << pointhist[2] << endl;
      (*LOGFILE) << " POINT_COLLAR: " << pointhist[3] << endl;
      (*LOGFILE) << " POINT_FACE: " << pointhist[4] << endl;
      (*LOGFILE) << "   POINT_3D: " << pointhist[5] << endl;

      if (config->lazyvertexprotection) {
	(*LOGFILE) << endl;      
	(*LOGFILE) << lazyProtectionList.size() << " vertices unprotected due to lazy protection" << endl;;
      }
      
      (*LOGFILE) << endl <<
	"Counting the number of encroached facets in each mesh..." <<
	endl;
      
      int facenum = 0;
      int foundEnc = 0;
      foreach(Mesh * m, M->children) {
	
	int numEncroached = 0;
	int numSimp = 0;
         int numVert = 0;

         Mesh::PIterator pIt(m);
         pIt.Start(m->GetSimplex());
         while (pIt.Current() != -1) {

	   /*
	   if (P->type[pIt.Current()] == POINT_INPUT) {
	     cout << pIt.Current() << endl;
	     set<int> nbrs;
	     M->Neighbors(pIt.Current(), nbrs);
	     int i = 0;
	     foreach(int nbr, nbrs) {
	       cout << i <<" " << P->type[nbr] << endl;
	       i++;
	     }
	   }
	   */

            if (P->type[pIt.Current()] != POINT_BBOX) {
               numVert++;
            }

            ++pIt;
         }

         Mesh::Iterator sIt;
         sIt.Start(m->GetSimplex());
         while (sIt.Current() != NULL) {
            if (!m->Exterior(sIt.Current())
                && Step3IsEncroached(sIt.Current())) {
               numEncroached++;
            }

            if (!m->Exterior(sIt.Current())) {
               numSimp++;
            }

            ++sIt;
         }

         facenum++;
         if (numEncroached > 0) {

            cout << m->name << " : " << numVert << " vertices, "
                 << numSimp << " simplices, " << numEncroached
                 << " encroached simplices." << endl;
            foundEnc = 1;

         }

      }

      if (foundEnc == 0) {
         (*LOGFILE) <<
         "No facets are encroached in any lower dimensional feature. "
                    << endl;
      }


      (*LOGFILE) << endl << "Total Number of Simplices Allocated: " <<
      SP->size << endl;
      (*LOGFILE) << endl << "Total Number of Location Steps Taken: " <<
      stats->nlocsteps << endl;


      // angle histogram...
      double minangle = 60.0, maxangle = 0.0;

      Simplex * worsttet = NULL;

      int anglehist[18];

      for (int ii = 0; ii < 18; ii++) {
         anglehist[ii] = 0;
      }
      Mesh::Iterator sIt1;
      sIt1.Start(M->GetSimplex());
      while (sIt1.Current() != NULL) {
         vector < double >angles;
         G->DihedralAngles(sIt1.Current(), angles);

         //if (angles[0] <= 0.0) {
         //cout << angles[0] << endl;
         //cout << sIt1.Current() << endl;
         //P->print(sIt1.Current());
         //Simplex * sss = sIt1.Current();
         //cout << P->type[sss->v[0]] <<"  "
         //      << P->type[sss->v[1]] <<"  "
         //             <<  P->type[sss->v[2]] <<"  "
         //                     <<   P->type[sss->v[3]] << endl;
         //
         //}
         if (angles[0] < minangle) {
            minangle = angles[0];
         }
         if (angles[5] > maxangle) {
            maxangle = angles[5];
	    worsttet = sIt1.Current();
         }
         int loc = (int) (angles[0] / 5.0);


         if (loc < 0 || loc > 17) {
            cout << "loc = " << loc << endl;
            Simplex *ss = sIt1.Current();
            cout << ss << endl;
            cout << P->V[ss->v[0]] << " "
                 << P->V[ss->v[1]] << " "
                 << P->V[ss->v[2]] << " " << P->V[ss->v[3]] << endl;

            cout << angles[0] << " "
                 << angles[1] << " "
                 << angles[2] << " "
                 << angles[3] << " "
                 << angles[4] << " " << angles[5] << endl;
         } else {
            anglehist[loc]++;
         }

         ++sIt1;
      }
      (*LOGFILE) << endl << "Number of simplicies by minimum angle." <<
      endl;
      for (int i = 0; i < 18; i++) {
         (*LOGFILE) << " " << i * 5 << "-" << i * 5 +
         5 << " degrees: " << anglehist[i] << endl;
      }

      (*LOGFILE) << endl << "Min Dihedral Angle: " << minangle << endl;
      (*LOGFILE) << "Max Dihedral Angle: " << maxangle << endl << endl;

      (*LOGFILE) << "Worst Tet " << worsttet << endl;
      //P->print(worsttet);

      //cout << P->type[worsttet->v[0]] << " "
      //   << P->type[worsttet->v[1]] << " "
      //   << P->type[worsttet->v[2]] << " "
      //   << P->type[worsttet->v[3]] << endl;

      /*
         cout << "Boundary Check" << endl;

         foreach (Mesh * m, M->children) {

         if (m->dim == 2 && !m->convex) {
         cout << m->name << endl;
         m->bnd->Print();
         }

         }
         cout << "End Boundary Check" << endl;
       */

      PrintEnd("Printing Mesh Summary.");
   }
}


void PLC::SetRadiusAndCC(Simplex * s)
{
   s->cc = G->circumcenter(s);
   s->rsq = d2(P->V[s->v[0]], s->cc);
}



//
// read PLC is an update of ReadFile
// this implements the new plc file type including nonconvex faces
// and interior points are correctly handled...
//
void PLC::ReadPLC(string fname, vector < Simplex * >&initialSimplices)
{

   PrintBegin("Input: Reading Data From " + fname);

   // keep this for later?
   map < int, Mesh * >id2mesh;
   map <int, int> e2p1;
   map <int, int> e2p2;
   map < Mesh *, set < int > >mesh2p;

   M = new Mesh(this);

   // read an input file...
   ifstream in((fname).c_str());
   if (!in) {
      cerr << "ERROR: can't open file:" << fname << endl;
      exit(0);
   }
   M->name = fname.substr(0, fname.size() - 4) + "-out";

   string junk;

   int nVert, dim;

   in >> junk;

   if (junk.compare("PLC") != 0) {
      if (logfileisopen) {
         (*LOGFILE) << "Specified file is not in correct format." <<
         endl;
      }
   } else {
      if (logfileisopen) {
         (*LOGFILE) << "PLC file " << fname << " opened successfully."
                    << endl;
      }
   }

   in >> junk;
   if (junk.compare("DIMENSION") != 0) {
      cout << "Error reading dimension section of data file." << endl;
      exit(0);
   }

   in >> dim;

   M->dim = dim;

   if (dim != 3 && dim != 2) {
      cout << "ERROR: mesh dimension is " << dim << "." <<
      endl;
      exit(0);
   }
   //cout << "readFile: " << fname << endl;
   //cout << "Dim " << dim << endl;    // someday handle 2D or general dim...

   in >> junk >> nVert;

   if (junk.compare("POINTS") != 0) {
      cout << "Error reading points section of data file." << endl;
      exit(0);
   }

   if (logfileisopen) {
      (*LOGFILE) << "Reading " << nVert << " Points" << endl;
   }
   // read the nodes...
   int id;

   for (int ii=0; ii<3; ii++) {
      M->upperbound[ii] = -1.0*config->infinity;
      M->lowerbound[ii] = config->infinity;
   }
   double *p = new double[3];
   p[0] = 0.0;
   p[1] = 0.0;
   p[2] = 0.0;

   vector < int >myids;
   vector < Point > mypoints;
   map < int, int >id2point;

   for (int i = 0; i < nVert; i++) {
      in >> id;
      for (int j = 0; j < dim; j++) {
         double tmp;
         in >> tmp;
         p[j] = tmp;
         if (tmp > M->upperbound[j]) {
            M->upperbound[j] = tmp;
         }
         if (tmp < M->lowerbound[j]) {
            M->lowerbound[j] = tmp;
         }
      }
      Point p1(p[0], p[1], p[2]);

      mypoints.push_back(p1);
      myids.push_back(id);
      id2point[id] = i;

      InputPoint ptmp;
      ptmp.p = p1;
      ptmp.id = id;
      ptmp.l2m.insert(M);
      P0.push_back(ptmp);
   }

   delete[] p;

   Point center((M->upperbound[0]+M->lowerbound[0])/2.0,
                (M->upperbound[1]+M->lowerbound[1])/2.0,
                (M->upperbound[2]+M->lowerbound[2])/2.0);

   M->base = center;

   if (dim == 2) {
      M->upperbound[2] = 0.0;
      M->lowerbound[2] = 0.0;
   }

   double bboxsize = max(max(M->upperbound[0]-M->lowerbound[0],
                             M->upperbound[1]-M->lowerbound[1]),
                         M->upperbound[2]-M->lowerbound[2])/2.0;

   M->bbox = bboxsize * config->boundingboxratio;

   M->Initialize(initialSimplices);     // set up the initial cube

   // this eventually should be able to be forgotten...
   M->bbox = M->bbox + .00001;  // FIXME: create a little wiggle room... robust math fixes this (?)...

   if (logfileisopen) {
      (*LOGFILE) << "Bounding box center: " << M->base << endl;
      (*LOGFILE) << "Bounding box size:   " << M->bbox << endl;
   }

   // add lines...
   int nLines;

   in >> junk >> nLines;

   if (junk.compare("SEGMENTS") != 0) {
      cout << "Error reading lines section of data file." << endl;
      exit(0);
   }


   if (logfileisopen) {
      (*LOGFILE) << "Reading " << nLines << " Segments" << endl;
   }

   for (int i = 0; i < nLines; i++) {
      int npoints, edgeid;

      in >> edgeid;

      npoints = 2;

      //cout << "Edge " << edgeid << endl;

      Mesh *newEdge = new Mesh(this);

      newEdge->id = edgeid;

      vector < int >edgepoints;
      int tmp;
      for (int j = 0; j < npoints; j++) {
         in >> tmp;
         edgepoints.push_back(id2point[tmp]);
         P0[id2point[tmp]].l2m.insert(newEdge);
      }


      std::ostringstream oss1;
      oss1 << "Line-" << edgeid;
      newEdge->name = oss1.str();

      newEdge->parents.push_back(M);
      M->children.push_back(newEdge);
      newEdge->dim = 1;

      // create the 1-d bounding box...
      int ll = edgepoints[0];
      int rr = edgepoints[1];
      Point pa, pb;

      pa = P0[ll].p + (P0[ll].p - P0[rr].p);
      pb = P0[rr].p + (P0[rr].p - P0[ll].p);


      int ia = P->AddPoint(pa, POINT_BBOX);
      int ib = P->AddPoint(pb, POINT_BBOX);

      //Simplex * edges = new Simplex(ia, ib);
      Simplex *edges = SP->NewSimplex(ia, ib);
      SetRadiusAndCC(edges);
      newEdge->Setp2s(edges);

      // identify the points inserted as the bounding box...
      newEdge->exteriorPoints.insert(ia);
      newEdge->exteriorPoints.insert(ib);

      newEdge->dir = pb - pa;
      newEdge->dir.normalize();
      newEdge->base = (pa + pb) / 2;

      // cout << newEdge->dir << newEdge->base << endl;
      // cout << pa << pb << P0[ll].p << P0[rr].p << endl;

      id2mesh[edgeid] = newEdge;
      e2p1[edgeid] = ll;
      e2p2[edgeid] = rr;


      for (int i = 0; i < (int)edgepoints.size(); i++) {
         mesh2p[newEdge].insert(edgepoints[i]);
      }
   }

   if (dim == 3) {
      // add planes
      int nPlanes;

      in >> junk >> nPlanes;

      if (junk.compare("FACES") != 0) {
         cout << "Error reading planes section of data file." << endl;
         cout << junk << endl;
         exit(0);
      }


      if (logfileisopen) {
         (*LOGFILE) << "Reading " << nPlanes << " Faces" << endl;
      }

      for (int iPlane = 0; iPlane < nPlanes; iPlane++) {
         int faceid;
         char facetype;

         in >> faceid;

         int nbdlines, nintlines, nintpoints;

         in >> nbdlines >> nintlines >> nintpoints;

         in >> facetype;

         // create a Mesh for the face...
         Mesh *newFace = new Mesh(this);
         newFace->id = faceid;

         std::ostringstream oss1;
         oss1 << "Plane-" << faceid;
         newFace->name = oss1.str();

         newFace->dim = 2;

         if (config->planestoprint.find(faceid) !=
             config->planestoprint.end()) {
            planestoprint.push_back(newFace);
         }

         if (facetype == 'C') {
            newFace->convex = true;
         } else if (facetype == 'N') {
            newFace->convex = false;
         } else {
            cout << "ERROR: ReadPLC failing" << endl;
         }

         vector < int >bdlines;
         for (int i = 0; i < nbdlines; i++) {
            int tmp;
            in >> tmp;
            bdlines.push_back(tmp);
         }

         vector < int >intlines;
         for (int i = 0; i < nintlines; i++) {
            int tmp;
            in >> tmp;
            intlines.push_back(tmp);
         }

         vector < int >intpoints;
         for (int i = 0; i < nintpoints; i++) {
            int tmp;
            in >> tmp;
            intpoints.push_back(tmp);
         }

         // make a set of all vertices in the mesh...
         set < int >myvert;


         // attempt to find the normal...
         Vector correctNormal;;
         bool normalFound = false;

         int prev1 =  -1;
         int prev2 = -1;

         for (int iEdge = 0; iEdge < nbdlines; iEdge++) {

            newFace->bndseg.insert(id2mesh[bdlines[iEdge]]);
            foreach(int thisvert, mesh2p[id2mesh[bdlines[iEdge]]]) {
               myvert.insert(thisvert);
            }

            if (!normalFound) {
               int a,b;
               set<int>::iterator tmpIt = id2mesh[bdlines[iEdge]]->exteriorPoints.begin();
               a = *tmpIt;
               tmpIt++;
               b = *tmpIt;

               a = e2p1[bdlines[iEdge]];
               b = e2p2[bdlines[iEdge]];

               Vector v1,v2;

               if (prev1 == -1) {
                  prev1 = a;
                  prev2 = b;
               } else {
                  if (a == prev1) {
                     v1 = P0[b].p - P0[a].p;
                     v2 = P0[prev1].p - P0[prev2].p;
                  } else if (b == prev1) {
                     v1 = P0[a].p - P0[b].p;
                     v2 = P0[prev1].p - P0[prev2].p;
                  } else if (a == prev2) {
                     v1 = P0[b].p - P0[a].p;
                     v2 = P0[prev2].p - P0[prev1].p;
                  } else if (b == prev2) {
                     v1 = P0[a].p - P0[b].p;
                     v2 = P0[prev2].p - P0[prev1].p;
                  }
                  correctNormal = v1^v2;
                  if (correctNormal.len2() > config->epsilon) {
                     correctNormal.normalize();
                     normalFound = true;
                  }

               }



            }

         }
         for (int iEdge = 0; iEdge < nintlines; iEdge++) {

            foreach(int thisvert, mesh2p[id2mesh[intlines[iEdge]]]) {
               myvert.insert(thisvert);
            }
         }

         for (int iInt = 0; iInt < nintpoints; iInt++) {
            // add this vertex to the plane also...
            myvert.insert(id2point[intpoints[iInt]]);
         }


         // set each point to include this mesh...
         foreach(int vv, myvert) {
            P0[vv].l2m.insert(newFace);
         }

         Point center;
         Vector nn, b1, b2;
         double mysize;
         GetCenterNormalAndBBox(myvert, center, nn, b1, b2, mysize);

         newFace->base = center;
         if (normalFound) {
            newFace->dir = correctNormal;;
         } else {
            newFace->dir = nn;
         }

         //cout << nn << endl;

         CreateBoundingSquare(mysize, center, b1, b2, newFace);

         id2mesh[faceid] = newFace;

         // set parent/children pointers...
         newFace->parents.push_back(M);
         M->children.push_back(newFace);

         // set up pointers to and from children edges...
         for (int iEdge = 0; iEdge < nbdlines; iEdge++) {
            newFace->children.push_back(id2mesh[bdlines[iEdge]]);
            id2mesh[bdlines[iEdge]]->parents.push_back(newFace);
         }

         for (int iEdge = 0; iEdge < nintlines; iEdge++) {
            newFace->children.push_back(id2mesh[intlines[iEdge]]);
            id2mesh[intlines[iEdge]]->parents.push_back(newFace);
         }

         if (newFace->convex == false) {
            newFace->bnd = new Boundary(newFace);
            // add segments to the boundary...
            // the boundary must be set up later...
            for (int iEdge = 0; iEdge < nbdlines; iEdge++) {
               newFace->bnd->seglist.push_back(id2mesh[bdlines[iEdge]]);
            }
         }

      }


   }
   //M->GetSimplex()->C = this;

   foreach(Mesh * m, M->children) {
     if (m->dim ==1) {
       int count = 0;
       foreach (Mesh * m1, m->parents) {
	 if (m1->dim == 2) {
	   count++;
	 }
       }

       if (count > 2) {
	 //cout << m->name << " belongs to " << count << " faces: ";
	 //foreach (Mesh * m1, m->parents) {
	 //  if (m1->dim == 2) {
	 //    cout << m1->name << " ";
	   
	     /*
	     // eliminate this face...
	     foreach(Mesh * m2, m1->children) {
	       for (int i=0; i<(int)m2->parents.size(); i++) {
		 if (m2->parents[i] == m1) {
		   m2->parents.erase(m2->parents.begin()+i);
		   break;
		 }
	       }
	     }
	     
	     foreach(Mesh * m2, m1->parents) {
	       for (int i=0; i<(int)m2->children.size(); i++) {
		 if (m2->children[i] == m1) {
		   m2->children.erase(m2->children.begin()+i);
		   break;
	       }
	       }
	     }
	     
	     // fix l2m...
	     foreach(InputPoint ip, P0) {
	       if (ip.l2m.find(m1) != ip.l2m.end()) {
		 ip.l2m.erase(ip.l2m.find(m1));
	       }
	     }
	     */
	 //}

	 //}
	 //cout << endl;
       }
       
     }
   }




   PrintEnd("Input: Reading Data");
}


void PLC::GetCenterNormalAndBBox(set < int >&myvert, Point & center,
                                 Vector & normal, Vector & b1, Vector & b2,
                                 double &bbox)
{


   if (myvert.size() < 3) {
      cout << "ERROR: Face contains less than 3 points." << endl;
      cout << myvert.size() << endl;
      exit(0);
   }

   set < int >::iterator sIt = myvert.begin();
   Point a, b, c;
   a = P0[*sIt].p;
   sIt++;
   b = P0[*sIt].p;
   sIt++;
   c = P0[*sIt].p;

   b1 = b - a;
   b1.normalize();

   b2 = c - a;
   b2.normalize();

   while (1 - b1 * b2 < config->epsilon || 1 + b1 * b2 < config->epsilon) {
      sIt++;
      if (sIt == myvert.end()) {
         cout << "ERROR: all points in face seem to be colinear." <<
         endl;
         cout << myvert.size() << endl;
	 Point d(0,0,0);
	 b2 = d - a;
	 b2.normalize();

         cout << b1 << " " << b2 << " " << a << "  " << b << " " << c <<
         endl;

	 //break;
         exit(0);
      }

      a = b;
      b = c;
      c = P0[*sIt].p;

      b1 = b - a;
      b1.normalize();

      b2 = c - a;
      b2.normalize();


   }

   b2 = b2 - (b2 * b1) * b1;
   b2.normalize();

   normal = b1 ^ b2;

   // b1 and b2 are a normalized basis...
   //cout <<"dir:b1:b2 " << " " << normal << " : " << b1 << " : " << b2 << endl;

   // ptmp will be the anchor point for the bounding box...
   Point ptmp(0, 0, 0);
   //ptmp.x =0;
   //ptmp.y =0;
   //ptmp.z =0;

   double xmin, xmax, ymin, ymax, zmin, zmax;
   xmin = M->bbox;
   ymin = M->bbox;
   zmin = M->bbox;
   xmax = -M->bbox;
   ymax = -M->bbox;
   zmax = -M->bbox;

   foreach(int iV, myvert) {
      ptmp = ptmp + P0[iV].p;
      if (P0[iV].p.xyz[0] < xmin) {
         xmin = P0[iV].p.xyz[0];
      }
      if (P0[iV].p.xyz[1] < ymin) {
         ymin = P0[iV].p.xyz[1];
      }
      if (P0[iV].p.xyz[2] < zmin) {
         zmin = P0[iV].p.xyz[2];
      }
      if (P0[iV].p.xyz[0] > xmax) {
         xmax = P0[iV].p.xyz[0];
      }
      if (P0[iV].p.xyz[1] > ymax) {
         ymax = P0[iV].p.xyz[1];
      }
      if (P0[iV].p.xyz[2] > zmax) {
         zmax = P0[iV].p.xyz[2];
      }
   }
   center = ptmp / ((double) myvert.size());

   bbox = xmax - xmin + ymax - ymin + zmax - zmin;
}

void PLC::CreateBoundingSquare(double mysize, Point & center, Vector & b1,
                               Vector & b2, Mesh * m)
{
   // add bounding box points...
   Point box1 =
      center + config->boundingboxratio * mysize * b1 +
      config->boundingboxratio * mysize * b2;
   Point box2 =
      center - config->boundingboxratio * mysize * b1 +
      config->boundingboxratio * mysize * b2;
   Point box3 =
      center - config->boundingboxratio * mysize * b1 -
      config->boundingboxratio * mysize * b2;
   Point box4 =
      center + config->boundingboxratio * mysize * b1 -
      config->boundingboxratio * mysize * b2;

   // set up initial triangulation of the bounding box

   int iA = P->AddPoint(box1, POINT_BBOX);
   int iB = P->AddPoint(box2, POINT_BBOX);
   int iC = P->AddPoint(box3, POINT_BBOX);
   int iD = P->AddPoint(box4, POINT_BBOX);

   //Simplex * t1 = new Simplex(iA, iB, iC);
   Simplex *t1 = SP->NewSimplex(iA, iB, iC);
   SetRadiusAndCC(t1);
   //Simplex * t2 = new Simplex(iC, iD, iA);
   Simplex *t2 = SP->NewSimplex(iC, iD, iA);
   SetRadiusAndCC(t2);

   m->Setp2s(t1);
   m->Setp2s(t2);

   // link the two triangles together...
   t1->SetNeighbor(1, t2);
   t2->SetNeighbor(1, t1);

   // eventually this will be unnecessary...
   m->exteriorPoints.insert(iA);
   m->exteriorPoints.insert(iB);
   m->exteriorPoints.insert(iC);
   m->exteriorPoints.insert(iD);

}


void PLC::FormFaceBoundaries()
{

   foreach(Mesh * m, M->children) {

      if (!m->convex) {
         if (m->dim != 2) {
            cout << "ERROR: nonconvex mesh of dimension " << m->dim << endl;
         }

         m->bnd->Build();
      }
   }
}

// DEBUG routine...
void PLC::VerifySimplex(Simplex * s)
{
   bool nullnbr = false;
   bool bboxpoint = false;

   if (s->size() == 4) {
      for (int i = 0; i < s->size(); i++) {
         if (s->n[i] == NULL) {
            nullnbr = true;
         }
         if (P->type[s->v[i]] == POINT_BBOX) {
            bboxpoint = true;
         }
      }
   }

   if (nullnbr && !bboxpoint) {
      cout << "VerifySimplex failed." << endl;
   }

}

void PLC::CopyMeshOver(PLC * Cbbox)
{

   int mystart = Cbbox->P->V.size();

   // copy the point set...
   // this avoids renumbering and fixing associated info in the collar...
   // FIXME: eventually I will avoid copying all the points...
   for (int i = mystart; i < (int)P->V.size(); i++) {
      int loc = Cbbox->P->AddPoint(P->V[i], P->type[i]);      
      if (loc != i) {
         cout << "Warning: indexing issue..." << endl;
      }
   }
   // insert the input points.
   for (int i = mystart; i < (int)Cbbox->P->V.size(); i++) {
      if (Cbbox->P->type[i] == POINT_INPUT) {
         foreach(Mesh * m, P->l2m[i]) {
            Mesh *mnew = Cbbox->old2new[m];

            mnew->InsertPoint(i);
         }
      }
   }

   // set up the boundary.
   Cbbox->FormFaceBoundaries();

   // copy over the rest of the points to keep
   for (int i = mystart; i < (int)Cbbox->P->V.size(); i++) {
      if (keepPoint(i)) {

	//cout << "keepPoint " << i << " " << P->LowestDimensionalContainingMesh(i)->name << endl;

         // need to insert the point in meshes starting from lowest dimension to the highest...
         // FIXME: this is a hack but I can live with it for now...
         foreach(Mesh * m, P->l2m[i]) {
            if (m->dim == 1) {
               Mesh *mnew = Cbbox->old2new[m];
               mnew->InsertPoint(i);
	       //cout << " *";
            }
         }
         foreach(Mesh * m, P->l2m[i]) {
            if (m->dim == 2) {
               Mesh *mnew = Cbbox->old2new[m];
               mnew->InsertPoint(i);
	       //cout << " -";
            }
         }
         foreach(Mesh * m, P->l2m[i]) {
            if (m->dim == 3) {
               Mesh *mnew = Cbbox->old2new[m];
               mnew->InsertPoint(i);
	       //cout << " =";
            }
         }
	 //cout << endl;
      }

   }

   //cout << endl;


   // set up lazy protection neighbors...
   Cbbox->SetLazyProtectionNeighbors(this);


   map < int, int >::iterator mIt;
   for (mIt = P->id2loc.begin(); mIt != P->id2loc.end(); mIt++) {
      Cbbox->P->id2loc[mIt->first] = mIt->second;
      //cout << mIt->first << " " << mIt->second << endl;
   }


   // copy the list of ids for input points...
   for (mIt = P->ids.begin(); mIt != P->ids.end(); mIt++) {
      Cbbox->P->ids[mIt->first] = mIt->second;
      //cout << mIt->first << " " << mIt->second << endl;
   }
   //Cbbox->P->ids = C->P->ids;

   // copy the protection lists...
   foreach(int i, pointsToProtect) {
      Cbbox->pointsToProtect.insert(i);
   }
   foreach(Mesh * m, segmentsToProtect) {
      Cbbox->segmentsToProtect.insert(Cbbox->old2new[m]);
   }
   foreach(InputPoint ip, P0) {
      InputPoint ip1;
      ip1.p = ip.p;
      ip1.id = ip.id;
      foreach(Mesh * mm, ip.l2m) {
         ip1.l2m.insert(Cbbox->old2new[mm]);
      }

      Cbbox->P0.push_back(ip1);
   }

   map < Mesh *, Mesh * >::iterator mapIt;
   for (mapIt = Cbbox->old2new.begin(); mapIt != Cbbox->old2new.end();
        mapIt++) {
      mapIt->second->ip2nb = mapIt->first->ip2nb;
      mapIt->second->protect = mapIt->first->protect;
   }

   foreach(Mesh * mm, M->children) {
      mm->DeleteMesh();
   }
   M->DeleteMesh();
}


void PLC::SetLazyProtectionNeighbors(PLC * COld) {

   foreach (int id, COld->lazyProtectionList) {
     
     lazyProtectionList.insert(id);

     int i = COld->P->id2loc[id];

     int j = -1;

     foreach (Mesh * m, COld->P->l2m[i]) {
       if (m->dim == 1) {
	 // get neighbors and add to the list...
	 set<int> nbs;
	 m->Neighbors(i,nbs);
	 foreach(int iNb, nbs) {
	   if (COld->P->type[iNb] == POINT_SEGMENT) {
	     lazyProtectionNeighbors[id].insert(iNb);
	     j = iNb;
	     lazyNb2Segment[iNb] = old2new[m];

	   }
	 }
       }
     }

     if (j == -1) {
       cout << "ERROR: No Neighbor found in PLC::SetLazyProtectionNeighbors()" <<endl;
     } else {
       lazyProtectionDistance[i] = d(COld->P->V[i],COld->P->V[j]);
     }
   }

}


void PLC::ReadAndRefine(string fn)
{

   // strip .plc from the file name
   if (fn.size() > 4
       && fn.compare(fn.size() - 4, fn.size() - 1, ".plc") == 0) {
      fn = fn.substr(0, fn.size() - 4);
   }


   cout << "-------------------------------------------------" << endl;
   cout << "Begin processing file: " << fn << ".plc" << endl << endl;;

   name = fn;

   clock_t start = clock();

   //PLC * Cbbox;

   PLC *Ctmp = new PLC;
   Ctmp->name = fn;

   vector < Simplex * >s;

   // open the log file
   Ctmp->OpenLogFile(fn + ".log");

   // read the configuration file
   Ctmp->config->ReadConfiguration(fn+".cfg");

   // read in the plc file
   Ctmp->ReadPLC(fn + ".plc", s);

   if (Ctmp->M->dim == 3) {

      // build the new PLC  for use in step 2c and 3...
      CopyPLC(Ctmp);
     
      Ctmp->Step0();
      if (Ctmp->config->printintermediate) {
         Ctmp->PrintOff("", "-0.off");
      }

      Ctmp->Step1a();
      if (Ctmp->config->printintermediate) {
         Ctmp->PrintOff("", "-1a.off");
      }

      Ctmp->Step1b();
      if (Ctmp->config->printintermediate) {
         Ctmp->PrintOff("", "-1b.off");
      }

      Ctmp->Step2a();
      if (Ctmp->config->printintermediate) {
         Ctmp->PrintOff("", "-2a.off");
      }

      Ctmp->Step2b();
      if (Ctmp->config->printintermediate) {
         Ctmp->PrintOff("", "-2b.off");
      }

      Ctmp->CopyMeshOver(this);
      if (config->printintermediate) {
         PrintOff("", "-2copy.off");
      }

      Step2c();
      if (config->printintermediate) {
         PrintOff("", "-2c.off");
      }


      Step3();
      if (config->printintermediate) {
         PrintOff("", "-3.off");
      }

      // Not working yet...
      //FixCollarSlivers();

      if (config->removeboundingbox) {
         StripBBoxTets();
      }

      OrientPLC();
      
      PrintInfo();

   } else {

     // unsupported, not working stuff...

      Ctmp->Step0();

      Ctmp->M->PrintSurfaceEPS("input.eps");
      Ctmp->M->PrintEPS("step0.eps");

      Ctmp->TwoDStep1a();
      Ctmp->M->PrintEPS("step1a.eps");

      Ctmp->TwoDStep1b();
      Ctmp->M->PrintEPS("step1b.eps");

      Ctmp->TwoDStep2();
      Ctmp->M->PrintEPS("step2.eps");


      Ctmp->M->PrintSurfaceEPS("outputsurf.eps");

   }

   cout << endl << "Finished processing file: " << fn << endl;
   cout << "Total Time: " << ((double) (clock() - start)) /
   ((double) CLOCKS_PER_SEC) << endl;
   cout << "-------------------------------------------------" << endl;

}

bool PLC::keepPoint(int i)
{

   if (P->type[i] == POINT_INPUT) {
      return false;
      // input points copied over in the first pass...
   }

   /*
   //copy over happens before collar insertion
   // so this should never happen...
   if (P->type[i] == POINT_COLLAR) {
      return true;
   }
   */

   if (P->type[i] == POINT_SEGMENT) {
      // find segment I belong to...
      Mesh *myseg = NULL;
      foreach(Mesh * m, P->l2m[i]) {
         if (m->dim == 1) {
            myseg = m;
            break;
         }
      }
      // if it is collared, keep me
      if (segmentsToProtect.find(myseg) != segmentsToProtect.end()) {
         return true;
      }
      // if not, check if I am neighboring a protected input point...
      set < int >myNeigh;
      myseg->Neighbors(i, myNeigh);
      foreach(int iN, myNeigh) {
         if (P->type[iN] == POINT_INPUT) {
            if (pointsToProtect.find(P->ids[iN]) !=
                pointsToProtect.end()) {
	      
	      if (lazyProtectionList.find(P->ids[iN]) == lazyProtectionList.end()) {
		//cout << "Here " << endl;
               return true;
	      } else {
		//cout << "lazy" << endl;
	      }
            }
            foreach(Mesh * m1, P->l2m[iN]) {
               if (m1->dim == 1) {
                  if (segmentsToProtect.find(m1) !=
                      segmentsToProtect.end()) {
                     return true;
                  }
               }
            }
         }
      }
   }
   return false;

}


bool PLC::TwoDStep1aIsEncroached(Simplex * s)
{
   int pa,pb;
   bool enda,endb;

   enda = false;
   endb = false;

   pa = s->v[0];
   pb = s->v[1];
   if (P->type[pa] == POINT_INPUT) {
      enda = true;
   }
   if (P->type[pb] == POINT_INPUT) {
      endb = true;
   }

   set < int >nbs;

   for (int i = 0; i < s->size(); i++) {
      M->Neighbors(s->v[i], nbs);
   }
   foreach(int i, nbs) {
      // does this neighbor encroach the simplex?
      if (!s->contains(i)) {
         if (G->inSphereReallyStrict(i, s)) {

            bool dosomething = true;
            if (P->type[i] != POINT_INPUT) {
               if (enda) {
                  Mesh *  m = P->LowestDimensionalContainingMesh(i);
                  if (P->l2m[pa].find(m) != P->l2m[pa].end()) {
                     if (SpindleLengthSq(pa) > .25*s->rsq) {
                        dosomething = false;
                     }
                  }
               }
               if (endb) {
                  Mesh *  m = P->LowestDimensionalContainingMesh(i);
                  if (P->l2m[pb].find(m) != P->l2m[pb].end()) {
                     if (SpindleLengthSq(pb) > .25*s->rsq) {
                        dosomething = false;
                     }
                  }
               }

            }
            if (dosomething)
               return true;
         }
      }
   }

   return false;
}

bool PLC::TwoDStep1aDoesEncroach(int p, Simplex * s)
{
   if (s->size() != 2) {
      cout << "ERROR: Expecting simplex of size 2 in TwoDStep1aDoesEncroach"
           << endl;
      cout << "Size is " << s->size() << endl;
      exit(0);
   }

   if (!G->inSphereReallyStrict(p, s)) {
      return false;
   }

   int pa,pb;
   bool enda,endb;

   enda = false;
   endb = false;

   pa = s->v[0];
   pb = s->v[1];
   if (P->type[pa] == POINT_INPUT) {
      enda = true;
   }
   if (P->type[pb] == POINT_INPUT) {
      endb = true;
   }

   if (enda) {
      Mesh *  m = P->LowestDimensionalContainingMesh(p);
      if (P->type[p] == POINT_INPUT || P->l2m[pa].find(m) != P->l2m[pa].end()) {
         if (SpindleLengthSq(pa) > .25*s->rsq) {
            return false;
         }
      }
   }

   if (endb) {
      Mesh *  m = P->LowestDimensionalContainingMesh(p);
      if (P->type[p] == POINT_INPUT || P->l2m[pb].find(m) != P->l2m[pb].end()) {
         if (SpindleLengthSq(pb) > .25*s->rsq) {
            return false;
         }
      }
   }

   return true;
}




void PLC::TwoDStep1a()
{
   PrintBegin("Step 1a: Estimate lfs-1");

   int count = 0;       // segment splits

   set < PointMesh > PQ;
   set < SimplexMesh > SQ;

   // queue up initial simplices to split
   // first conformity...
   // set up a priority queue of simplices to
   foreach(Mesh * m, M->children) {
      //cout << m->name << endl;
      Mesh::Iterator sIt;
      sIt.Start(m->GetSimplex());
      while (sIt.Current() != NULL) {
         // do not queue exterior or collar elements: the collar starts ok...
         if (!m->Exterior(sIt.Current())) {
            if (TwoDStep1aIsEncroached(sIt.Current())) {
               // queue this simplex...
               SimplexMesh sm;
               sm.s = sIt.Current();
               sm.m = m;

               // Queue the simplex...
               sm.s->f = SIM_QUEUED;
               SQ.insert(sm);
            }
         }

         ++sIt;
      }
   }

   set < PointMesh >::iterator pIt;
   set < SimplexMesh >::iterator sIt;

   clock_t mytime = clock();
   int mycount = 0;

   // process the queue...
   while (SQ.size() > 0 || PQ.size() > 0) {


      if ((double) (clock() - mytime) > ((double) CLOCKS_PER_SEC) / 5.0) {
         cout << "\r Queue Length: " << setw(8) << SQ.size()
              << setw(10) << ",       Items Processed: " << setw(10) <<
         mycount;
         cout.flush();
         mytime = clock();
      }
      mycount++;

      if (PQ.size() > 0) {
         // insert a point into a mesh...

         pIt = PQ.begin();
         int p = pIt->p;
         Mesh *m = pIt->m;
         // cout << m->name << " : " << p << endl;

         // insert the point...
         //InsertInMesh(p, m);
         InsertInMeshAndDequeue(p, m, SQ);

         // requeue stuff...
         if (m->dim == 2) {

            Mesh * mSeg = NULL;

            // in the 3D mesh, see if anything needs requeueing...
            foreach(Mesh * m1, P->l2m[p]) {
               if (m1->dim == 1) {

                  mSeg = m1;

                  // look at the star and consider any segments...
                  set < Simplex * >star;
                  m1->Simplices(p, star);
                  foreach(Simplex * s, star) {
                     // ask if this simplex is encroached...
                     if (s->f == SIM_NOT_QUEUED && TwoDStep1aIsEncroached(s)) {
                        SimplexMesh sm;
                        sm.s = s;
                        sm.m = m1;

                        // Queue the simplex...
                        sm.s->f = SIM_QUEUED;
                        SQ.insert(sm);
                     }
                  }
               }
            }

            set < int >nbs;
            // get the new neighbors...
            m->Neighbors(p, nbs);

            foreach(int iN, nbs) {
               // does this neighbor belong to a segment that
               // p encroaches on...

               foreach(Mesh * m1, P->l2m[iN]) {
                  if (m1->dim == 1) {
                     set < Simplex * >star;
                     m1->Simplices(iN, star);
                     foreach(Simplex * s, star) {
                        if (s->f == SIM_NOT_QUEUED&& !m1->Exterior(s)
                            && TwoDStep1aDoesEncroach(p, s)) {
                           SimplexMesh sm;
                           sm.s = s;
                           sm.m = m1;

                           // Queue the simplex...
                           sm.s->f = SIM_QUEUED;
                           SQ.insert(sm);
                        }
                     }
                  }
               }
            }

         }
         PQ.erase(pIt);

      } else {

         // split a simplex...
         sIt = SQ.begin();
         Simplex *s = sIt->s;
         Mesh *m = sIt->m;

         s->f = SIM_CURRENT;

         // find the midpoint...
         // queue midpoint for insertion into all applicable meshes...
         Point p = s->cc;
         //cout << "Split " << s << " at " << p << endl;
         count++;

         int i = P->AddPoint(p, s->v[0], POINT_SEGMENT);

         PointMesh pm;

         pm.p = i;
         pm.m = m;

         PQ.insert(pm);

         foreach(Mesh * mm, m->parents) {
            pm.m = mm;
            PQ.insert(pm);
         }


         SQ.erase(sIt);
      }
   }

   cout << "Inserted " << count << " points in the mesh." << endl;

   PrintEnd("Step 1a: Estimate lfs-1");
}

double PLC::SpindleLengthSq(int p) {

   double len = 10000000000.0;

   if (P->type[p] != POINT_INPUT) {
      cout << "WARNING: computing spindle size of a non-input point." << endl;
   }
   foreach (Mesh * m, P->l2m[p]) {
      if (m->dim == 1) {
         set<Simplex *> star;
         m->Simplices(p,star);
         foreach(Simplex * s, star) {
            if (!m->Exterior(s) && s->rsq < len) {
               len = s->rsq;
            }
         }
      }
   }
   return len;

}

void PLC::TwoDStep1b()
{
   PrintBegin("Step 1b: Insert Collar");

   int count = 0;


   cout <<"Number of Points to Protect: " << pointsToProtect.size() << endl;
   foreach(int i, pointsToProtect) {

      cout << "Protecting point " << i << ":" << P->V[P->id2loc[i]] << endl;

      int p = P->id2loc[i];

      double len = config->infinity;
      if (config->maxoutputradius > 0.0) {
         len = config->maxoutputradius;
      }

      set<int> nbors;
      M->Neighbors(p,nbors);
      foreach(int iN, nbors) {
         if (P->type[iN] == POINT_INPUT) {
            if (len > .333* d(P->V[p],P->V[iN])) {
               len = .333* d(P->V[p],P->V[iN]);
            }
         }
         if (P->type[iN] == POINT_SEGMENT) {
            if (len > .5* d(P->V[p],P->V[iN])) {
               len = .5* d(P->V[p],P->V[iN]);
            }
         }
      }

      cout << "Protection radius: " << len << endl;

      foreach(Mesh * m, P->l2m[p]) {
         if (m->dim == 1) {
            set<int>nbrs;
            m->Neighbors(p,nbrs);
            foreach(int iN, nbrs) {
               if (P->type[iN] != POINT_BBOX) {
                  Vector dir = P->V[iN] - P->V[p];
                  dir.normalize();
                  Point q = P->V[p] + len*dir;

                  int i = P->AddPoint(q, p, POINT_SEGMENT);

                  InsertInMeshAndParents(i, p, m);
                  count++;
               }
            }
         }
      }
   }



   cout << "Inserted " << count << " points in the mesh." << endl;

   PrintEnd("Step 1b: Insert Collar");
}


void PLC::TwoDStep2()
{
   PrintBegin("Step 2: Quality Refinement");

   int counts = 0;      // segment splits
   int countt = 0;      // triangle splits

   set < PointMesh > PQ;
   set < SimplexMesh > SQ;

   // queue up initial simplices to split
   // first conformity...
   // set up a priority queue of simplices to
   foreach(Mesh * m, M->children) {
      //cout << m->name << endl;
      Mesh::Iterator sIt;
      sIt.Start(m->GetSimplex());
      while (sIt.Current() != NULL) {
         // do not queue exterior or collar elements: the collar starts ok...
         if (!m->Exterior(sIt.Current())) {
            if (TwoDStep2IsEncroached(sIt.Current())) {
               // queue this simplex...
               SimplexMesh sm;
               sm.s = sIt.Current();
               sm.m = m;

               // Queue the simplex...
               sm.s->f = SIM_QUEUED;
               SQ.insert(sm);
            }
         }

         ++sIt;
      }
   }
   // queue for quality...
   Mesh::Iterator sIt1;
   sIt1.Start(M->GetSimplex());
   while (sIt1.Current() != NULL) {
      // do not queue exterior or collar elements: the collar starts ok...
      if (G->TwoDIsPoorQuality(sIt1.Current())) {
         // queue this simplex...
         SimplexMesh sm;
         sm.s = sIt1.Current();
         sm.m = M;

         // Queue the simplex...
         sm.s->f = SIM_QUEUED;
         SQ.insert(sm);
      }
      ++sIt1;
   }

   if (logfileisopen) {
      (*LOGFILE) << "Initial Queue Size: "
                 << SQ.size() << " simplices." << endl;
   }

   set < PointMesh >::iterator pIt;
   set < SimplexMesh >::iterator sIt;

   clock_t mytime = clock();
   int mycount = 0;

   // process the queue...
   while (SQ.size() > 0 || PQ.size() > 0) {


      if ((double) (clock() - mytime) > ((double) CLOCKS_PER_SEC) / 5.0) {
         cout << "\r Queue Length: " << setw(8) << SQ.size()
              << setw(10) << ",       Items Processed: " << setw(10) <<
         mycount;
         cout.flush();
         mytime = clock();
      }
      mycount++;

      if (PQ.size() > 0) {
         // insert a point into a mesh...

         pIt = PQ.begin();
         int p = pIt->p;
         Mesh *m = pIt->m;
         // cout << m->name << " : " << p << endl;

         // insert the point...
         //InsertInMesh(p, m);
         InsertInMeshAndDequeue(p, m, SQ);

         // requeue stuff...
         if (m->dim == 2) {

            Mesh * mSeg = NULL;

            // in the 3D mesh, see if anything needs requeueing...
            foreach(Mesh * m1, P->l2m[p]) {
               if (m1->dim == 1) {

                  mSeg = m1;

                  // look at the star and consider any segments...
                  set < Simplex * >star;
                  m1->Simplices(p, star);
                  foreach(Simplex * s, star) {
                     // ask if this simplex is encroached...
                     if (s->f == SIM_NOT_QUEUED && TwoDStep2IsEncroached(s)) {
                        SimplexMesh sm;
                        sm.s = s;
                        sm.m = m1;

                        // Queue the simplex...
                        sm.s->f = SIM_QUEUED;
                        SQ.insert(sm);
                     }
                  }
               }
            }

            set < int >nbs;
            // get the new neighbors...
            m->Neighbors(p, nbs);

            foreach(int iN, nbs) {
               // does this neighbor belong to a segment that
               // p encroaches on...

               foreach(Mesh * m1, P->l2m[iN]) {
                  if (m1->dim == 1) {
                     set < Simplex * >star;
                     m1->Simplices(iN, star);
                     foreach(Simplex * s, star) {
                        if (s->f == SIM_NOT_QUEUED&& !m1->Exterior(s)
                            && TwoDStep2DoesEncroach(p, s)) {
                           SimplexMesh sm;
                           sm.s = s;
                           sm.m = m1;

                           // Queue the simplex...
                           sm.s->f = SIM_QUEUED;
                           SQ.insert(sm);
                        }
                     }
                  }
               }
            }

            // requeue for quality...
            set < Simplex * >star;
            M->Simplices(p, star);
            foreach(Simplex * s, star) {
               if (G->TwoDIsPoorQuality(s)) {
                  SimplexMesh sm;
                  sm.s = s;
                  sm.m = M;
                  sm.s->f = SIM_QUEUED;
                  SQ.insert(sm);
               }
            }
         }
         PQ.erase(pIt);

      } else {

         // split a simplex...
         sIt = SQ.begin();
         Simplex *s = sIt->s;
         Mesh *m = sIt->m;

         s->f = SIM_CURRENT;

         Point p = s->cc;

         vector<SimplexMesh> encroached;
         bool reject = false;
         if (s->size() == 3) {
            // does this triangle encroach upon segments
            // if so, queue the segments
            // unless they are protected end segments...


            set <int> nbs;
            m->Cavity(p, s, nbs);
            foreach(int iN, nbs) {

               foreach(Mesh * m1, P->l2m[iN]) {

                  if (m1->dim == 1) {

                     set<Simplex *> star;
                     m1->Simplices(iN,star);
                     foreach (Simplex * iS, star) {
                        if (!m1->Exterior(iS) &&
                            G->inSphereReallyStrict(p,iS)) {
                           // is iS protected?
                           int pa = iS->v[0];
                           int pb = iS->v[1];
                           if ((P->type[pa] == POINT_INPUT && pointsToProtect.find(P->ids[pa]) != pointsToProtect.end()) ||
                               (P->type[pb] == POINT_INPUT && pointsToProtect.find(P->ids[pb]) != pointsToProtect.end()) )  {
                              reject = true;
                           } else {

                              SimplexMesh sm;
                              sm.s = iS;
                              sm.m = m1;

                              encroached.push_back(sm);
                           }
                        }
                     }
                  }
               }

            }


         }

         if (!reject) {

            if (encroached.size() > 0) {
               foreach (SimplexMesh sm, encroached) {
                  SQ.insert(sm);
               }
               s->f = SIM_QUEUED;
            } else {

               //cout << "Split " << s << " at " << p << endl;

               int i;
               if (m->dim == 1) {
                  i = P->AddPoint(p, s->v[0], POINT_SEGMENT);
                  counts++;
               }
               else {
                  i = P->AddPoint(p, s->v[0], POINT_FACE);
                  countt++;
               }


               PointMesh pm;

               pm.p = i;
               pm.m = m;

               PQ.insert(pm);

               foreach(Mesh * mm, m->parents) {
                  pm.m = mm;
                  PQ.insert(pm);
               }
               SQ.erase(sIt);
            }
         } else {
            cout << "Reject " << endl;
            SQ.erase(sIt);
         }
      }
   }

   cout << "Inserted " << counts << " midpoints and " << countt
        << " circumcenters into the mesh." << endl;

   // queue up initial simplices to split
   // first conformity...
   // set up a priority queue of simplices to
   foreach(Mesh * m, M->children) {
      //cout << m->name << endl;
      Mesh::Iterator sIt;
      sIt.Start(m->GetSimplex());
      while (sIt.Current() != NULL) {
         // do not queue exterior or collar elements: the collar starts ok...
         if (!m->Exterior(sIt.Current())) {
            if (TwoDStep2IsEncroached(sIt.Current())) {
               cout << "Encroachment: " << m->name << " " << sIt.Current() << endl;

            }
         }

         ++sIt;
      }
   }



   PrintEnd("Step 2: Quality Refinement");
}

bool PLC::TwoDStep2IsEncroached(Simplex * s)
{
   int pa,pb;

   pa = s->v[0];
   pb = s->v[1];

   set < int >nbs;

   for (int i = 0; i < s->size(); i++) {
      M->Neighbors(s->v[i], nbs);
   }
   foreach(int i, nbs) {
      // does this neighbor encroach the simplex?
      if (!s->contains(i)) {
         if (G->inSphereReallyStrict(i, s)) {
            return true;
         }
      }
   }

   return false;
}

bool PLC::TwoDStep2DoesEncroach(int p, Simplex * s)
{
   if (s->size() != 2) {
      cout << "ERROR: Expecting simplex of size 2 in TwoDStep2DoesEncroach"
           << endl;
      cout << "Size is " << s->size() << endl;
      exit(0);
   }

   if (!G->inSphereReallyStrict(p, s)) {
      return false;
   }

   return true;
}

