#include <iostream>
#include <fstream>
#include <math.h>
#include <iomanip>

#include "point.h"


istream & operator>>(istream & input, Point & P)
{
   input >> P.xyz[0];
   input >> P.xyz[1];
   input >> P.xyz[2];
   return input;
}

ostream & operator<<(ostream & output, Point P)
{

   output << "(" << P.xyz[0] << ", " << P.xyz[1] << ", " 
	  << P.xyz[2] << ")";
   return output;
}

ostream & operator<<(ostream & output, Vector P)
{

   output << "<" << P.xyz[0] << ", " << P.xyz[1] << ", " 
	  << P.xyz[2] << ">";
   return output;
}


Point & Point::operator=(const Point & p)
{
   xyz[0] = p.xyz[0];
   xyz[1] = p.xyz[1];
   xyz[2] = p.xyz[2];
   return *this;
}

int Point::operator==(Point Q)
{
   return (xyz[0] == Q.xyz[0] && xyz[1] == Q.xyz[1]
           && xyz[2] == Q.xyz[2]);
}

int Point::operator!=(Point Q)
{
   return (xyz[0] != Q.xyz[0] || xyz[1] != Q.xyz[1]
           || xyz[2] != Q.xyz[2]);
}


Vector Point::operator-(Point Q)
{
   Vector v;
   v.xyz[0] = xyz[0] - Q.xyz[0];
   v.xyz[1] = xyz[1] - Q.xyz[1];
   v.xyz[2] = xyz[2] - Q.xyz[2];
   return v;
}

Point Point::operator+(Vector v)
{
   Point P;
   P.xyz[0] = xyz[0] + v.xyz[0];
   P.xyz[1] = xyz[1] + v.xyz[1];
   P.xyz[2] = xyz[2] + v.xyz[2];
   return P;
}

Point Point::operator-(Vector v)
{
   Point P;
   P.xyz[0] = xyz[0] - v.xyz[0];
   P.xyz[1] = xyz[1] - v.xyz[1];
   P.xyz[2] = xyz[2] - v.xyz[2];
   return P;
}

Point & Point::operator+=(Vector v)
{
   xyz[0] += v.xyz[0];
   xyz[1] += v.xyz[1];
   xyz[2] += v.xyz[2];
   return *this;
}

Point & Point::operator-=(Vector v)
{
   xyz[0] -= v.xyz[0];
   xyz[1] -= v.xyz[1];
   xyz[2] -= v.xyz[2];
   return *this;
}


Point operator*(int c, Point Q)
{
   Point P;
   P.xyz[0] = c * Q.xyz[0];
   P.xyz[1] = c * Q.xyz[1];
   P.xyz[2] = c * Q.xyz[2];
   return P;
}

Point operator*(double c, Point Q)
{
   Point P;
   P.xyz[0] = c * Q.xyz[0];
   P.xyz[1] = c * Q.xyz[1];
   P.xyz[2] = c * Q.xyz[2];
   return P;
}

Point operator*(Point Q, int c)
{
   Point P;
   P.xyz[0] = c * Q.xyz[0];
   P.xyz[1] = c * Q.xyz[1];
   P.xyz[2] = c * Q.xyz[2];
   return P;
}

Point operator*(Point Q, double c)
{
   Point P;
   P.xyz[0] = c * Q.xyz[0];
   P.xyz[1] = c * Q.xyz[1];
   P.xyz[2] = c * Q.xyz[2];
   return P;
}

Point operator/(Point Q, int c)
{
   Point P;
   P.xyz[0] = Q.xyz[0] / c;
   P.xyz[1] = Q.xyz[1] / c;
   P.xyz[2] = Q.xyz[2] / c;
   return P;
}

Point operator/(Point Q, double c)
{
   Point P;
   P.xyz[0] = Q.xyz[0] / c;
   P.xyz[1] = Q.xyz[1] / c;
   P.xyz[2] = Q.xyz[2] / c;
   return P;
}

Point operator+(Point Q, Point R)
{
   Point P;
   P.xyz[0] = Q.xyz[0] + R.xyz[0];
   P.xyz[1] = Q.xyz[1] + R.xyz[1];
   P.xyz[2] = Q.xyz[2] + R.xyz[2];
   return P;
}


double d(Point P, Point Q)
{                               // Euclidean distance
   double dx = P.xyz[0] - Q.xyz[0];
   double dy = P.xyz[1] - Q.xyz[1];
   double dz = P.xyz[2] - Q.xyz[2];
   return sqrt(dx * dx + dy * dy + dz * dz);
}

double d2(Point P, Point Q)
{                               // squared distance
   double dx = P.xyz[0] - Q.xyz[0];
   double dy = P.xyz[1] - Q.xyz[1];
   double dz = P.xyz[2] - Q.xyz[2];
   return (dx * dx + dy * dy + dz * dz);
}



Vector Vector::operator-()
{
   Vector v;
   v.xyz[0] = -xyz[0];
   v.xyz[1] = -xyz[1];
   v.xyz[2] = -xyz[2];
   return v;
}

//------------------------------------------------------------------
//  Scalar Ops
//------------------------------------------------------------------

// Scalar multiplication
Vector operator*(int c, Vector w)
{
   Vector v;
   v.xyz[0] = c * w.xyz[0];
   v.xyz[1] = c * w.xyz[1];
   v.xyz[2] = c * w.xyz[2];
   return v;
}

Vector operator*(double c, Vector w)
{
   Vector v;
   v.xyz[0] = c * w.xyz[0];
   v.xyz[1] = c * w.xyz[1];
   v.xyz[2] = c * w.xyz[2];
   return v;
}

Vector operator*(Vector w, int c)
{
   Vector v;
   v.xyz[0] = c * w.xyz[0];
   v.xyz[1] = c * w.xyz[1];
   v.xyz[2] = c * w.xyz[2];
   return v;
}

Vector operator*(Vector w, double c)
{
   Vector v;
   v.xyz[0] = c * w.xyz[0];
   v.xyz[1] = c * w.xyz[1];
   v.xyz[2] = c * w.xyz[2];
   return v;
}

Vector operator/(Vector w, int c)
{
   Vector v;
   v.xyz[0] = w.xyz[0] / c;
   v.xyz[1] = w.xyz[1] / c;
   v.xyz[2] = w.xyz[2] / c;
   return v;
}

Vector operator/(Vector w, double c)
{
   Vector v;
   v.xyz[0] = w.xyz[0] / c;
   v.xyz[1] = w.xyz[1] / c;
   v.xyz[2] = w.xyz[2] / c;
   return v;
}


Vector Vector::operator+(Vector w)
{
   Vector v;
   v.xyz[0] = xyz[0] + w.xyz[0];
   v.xyz[1] = xyz[1] + w.xyz[1];
   v.xyz[2] = xyz[2] + w.xyz[2];
   return v;
}

Vector Vector::operator-(Vector w)
{
   Vector v;
   v.xyz[0] = xyz[0] - w.xyz[0];
   v.xyz[1] = xyz[1] - w.xyz[1];
   v.xyz[2] = xyz[2] - w.xyz[2];
   return v;
}

double Vector::operator*(Vector w)
{
   return (xyz[0] *w.xyz[0] + xyz[1] *w.xyz[1] + xyz[2] *w.xyz[2]);
}

double Vector::operator*(Point w)
{
   return (xyz[0] *w.xyz[0] + xyz[1] *w.xyz[1] + xyz[2] *w.xyz[2]);
}

Vector Vector::operator^(Vector w)
{
   Vector v;
   v.xyz[0] = xyz[1] *w.xyz[2] - xyz[2] *w.xyz[1];
   v.xyz[1] = xyz[2] *w.xyz[0] - xyz[0] *w.xyz[2];
   v.xyz[2] = xyz[0] *w.xyz[1] - xyz[1] *w.xyz[0];
   return v;
}

Vector & Vector::operator*=(double c)
{
   xyz[0] *= c;
   xyz[1] *= c;
   xyz[2] *= c;
   return *this;
}

Vector & Vector::operator/=(double c)
{
   xyz[0] /= c;
   xyz[1] /= c;
   xyz[2] /= c;
   return *this;
}

Vector & Vector::operator+=(Vector w)
{
   xyz[0] += w.xyz[0];
   xyz[1] += w.xyz[1];
   xyz[2] += w.xyz[2];
   return *this;
}

Vector & Vector::operator-=(Vector w)
{
   xyz[0] -= w.xyz[0];
   xyz[1] -= w.xyz[1];
   xyz[2] -= w.xyz[2];
   return *this;
}

Vector & Vector::operator^=(Vector w)
{
   double ox = xyz[0], oy = xyz[1], oz = xyz[2];
   xyz[0] = oy * w.xyz[2] - oz * w.xyz[1];
   xyz[1] = oz * w.xyz[0] - ox * w.xyz[2];
   xyz[2] = ox * w.xyz[1] - oy * w.xyz[0];
   return *this;
}


void Vector::normalize()
{
   double ln = sqrt(xyz[0] *xyz[0] + xyz[1] *xyz[1] + xyz[2] *xyz[2]);
   if (ln == 0) {
      return;                   // do nothing to zero vector...
   }

   xyz[0] /= ln;
   xyz[1] /= ln;
   xyz[2] /= ln;
}
