#include "simplex.h"
#include "geometry.h"
#include "queuepair.h"
#include "plc.h"

Simplex::Simplex(Simplex * s)
{
   v[0] = s->v[0];
   v[1] = s->v[1];
   v[2] = s->v[2];
   v[3] = s->v[3];
   n[0] = s->n[0];
   n[1] = s->n[1];
   n[2] = s->n[2];
   n[3] = s->n[3];
   f = s->f;
   rsq = s->rsq;
   cc = s->cc;
}

Simplex::Simplex()
{
   v[0] = -1;
   v[1] = -1;
   v[2] = -1;
   v[3] = -1;
   n[0] = NULL;
   n[1] = NULL;
   n[2] = NULL;
   n[3] = NULL;
   f = SIM_NOT_QUEUED;
   rsq = -1.0;
}

Simplex::Simplex(int a)
{
   v[0] = a;
   v[1] = -1;
   v[2] = -1;
   v[3] = -1;
   n[0] = NULL;
   n[1] = NULL;
   n[2] = NULL;
   n[3] = NULL;
   f = SIM_NOT_QUEUED;
   rsq = -1.0;

}

Simplex::Simplex(int a, int b)
{
   v[0] = a;
   v[1] = b;
   v[2] = -1;
   v[3] = -1;
   n[0] = NULL;
   n[1] = NULL;
   n[2] = NULL;
   n[3] = NULL;
   f = SIM_NOT_QUEUED;
   rsq = -1.0;
}

Simplex::Simplex(int a, int b, int c)
{
   v[0] = a;
   v[1] = b;
   v[2] = c;
   v[3] = -1;
   n[0] = NULL;
   n[1] = NULL;
   n[2] = NULL;
   n[3] = NULL;
   f = SIM_NOT_QUEUED;
   rsq = -1.0;
}
Simplex::Simplex(int a, int b, int c, int d)
{
   v[0] = a;
   v[1] = b;
   v[2] = c;
   v[3] = d;
   n[0] = NULL;
   n[1] = NULL;
   n[2] = NULL;
   n[3] = NULL;
   f = SIM_NOT_QUEUED;
   rsq = -1.0;
}

void Simplex::link(Simplex * s1, Simplex * s2, Simplex * s3, Simplex * s4)
{
   n[0] = s1;
   n[1] = s2;
   n[2] = s3;
   n[3] = s4;
}

void Simplex::SetVertex(int i, int myv)
{
   v[i] = myv;
}

void Simplex::SetNeighbor(int i, Simplex * myn)
{
   n[i] = myn;
}


bool Simplex::contains(int i)
{
   if (v[0] == i || v[1] == i || v[2] == i || v[3] == i) {
      return true;
   }
   return false;
}

bool Simplex::ncontains(Simplex * i)
{
   if (n[0] == i || n[1] == i || n[2] == i || n[3] == i) {
      return true;
   }
   return false;
}

int Simplex::isNeighbor(Simplex * s)
{
   int a, b, c, d;
   a = 0;
   b = 0;
   c = 0;
   d = 0;
   if (s->contains(v[0])) {
      a = 1;
   }
   if (s->contains(v[1])) {
      b = 1;
   }
   if (s->contains(v[2])) {
      c = 1;
   }
   if (s->contains(v[3])) {
      d = 1;
   }

   if (a + b + c + d < 3) {
      return -1;
   } else {
      if (a == 0) {
         return 0;
      }
      if (b == 0) {
         return 1;
      }
      if (c == 0) {
         return 2;
      }
      if (d == 0) {
         return 3;
      }
   }
   return -2;
}

int Simplex::size()
{
   if (v[3] != -1) {
      return 4;
   } else if (v[2] != -1) {
      return 3;
   } else if (v[1] != -1) {
      return 2;
   } else if (v[0] != -1) {
      return 1;
   }
   return 0;
}

void Simplex::Faces(vector<Simplex> & faces) {
  if (size() == 4) {
    Simplex s1(v[0],v[1],v[2]);
    Simplex s2(v[0],v[1],v[3]);
    Simplex s3(v[0],v[2],v[3]);
    Simplex s4(v[1],v[2],v[3]);

    faces.push_back(s1);
    faces.push_back(s2);
    faces.push_back(s3);
    faces.push_back(s4);
  }
  
}


// debugging routine...
bool Simplex::Verify()
{
   cout << "Calling Verify.  This should only occur with debug code." <<
   endl;

   int success = 0;
   for (int i = 0; i < size(); i++) {
      if (n[i] != NULL && !n[i]->ncontains(this)) {
         success = 1;
      }
   }

   if (success == 1) {
      cout << endl << "Simplex " << this << " failed verification.  " <<
      f << endl;
      for (int i = 0; i < size(); i++) {
         if (n[i] == NULL) {
            cout << "NULL ";
         } else {
            cout << n[i] << " : ";
            for (int j = 0; j < n[i]->size(); j++) {
               if (n[i]->n[j] == NULL) {
                  cout << "NULL ";
               } else {
                  cout << n[i]->n[j] << " ";
               }
            }
         }
         cout << endl;
      }
      return false;
   }
   return true;
}

bool operator<(Simplex & x, Simplex & y)
{

   if (x.size() < y.size()) {
      return true;
   } else if (x.size() > y.size()) {
      return false;
   } else if (x.size() == y.size()) {
      for (int i = 0; i < x.size(); i++) {
         if (x.v[i] < y.v[i]) {
            return true;
         }
         if (x.v[i] > y.v[i]) {
            return false;
         }
      }
   }

   return false;
}

bool operator<(const Face & x, const Face & y)
{

   if (x.s < y.s) {
      return true;
   } else if (x.s == y.s && x.v < y.v) {
      return true;
   }
   return false;
}

ostream & operator<<(ostream & output, Simplex * s)
{
   if (s == NULL) {
      output << "[]" << endl;
      return output;
   }

   int a = s->size();
   if (a == 4) {
      output << "[" << s->v[0] << ", " << s->v[1] << ", " << s->
      v[2] << ", " << s->v[3] << "]";
   } else if (a == 3) {
      output << "[" << s->v[0] << ", " << s->v[1] << ", " << s->
      v[2] << "]";
   } else if (a == 2) {
      output << "[" << s->v[0] << ", " << s->v[1] << "]";
   } else if (a == 1) {
      output << "[" << s->v[0] << "]";
   }

   return output;
}

ostream & operator<<(ostream & output, SimplexFlag f)
{

   if (f == SIM_NOT_QUEUED) {
      output << "SIM_NOT_QUEUED";
   }
   if (f == SIM_QUEUED) {
      output << "SIM_QUEUED";
   }
   if (f == SIM_CURRENT) {
      output << "SIM_CURRENT";
   }
   if (f == SIM_REMOVED) {
      output << "SIM_REMOVED";
   }
   if (f == SIM_DELETED) {
      output << "SIM_DELETED";
   }
   return output;
}
